Package: Cyclops
Type: Package
Title: Cyclic Coordinate Descent for Logistic, Poisson and Survival
        Analysis
Version: 3.7.1
Authors@R: c(
    person("Marc A.", "Suchard", email = "msuchard@ucla.edu", role = c("aut","cre")),
    person("Martijn J.", "Schuemie", role = "aut"),
    person("Trevor R.", "Shaddox", role = "aut"),
    person("Yuxi", "Tian", role = "aut"),
    person("Jianxiao", "Yang", role = "aut"),
    person("Eric", "Kawaguchi", role = "aut"),
    person("Sushil", "Mittal", role ="ctb"),
    person("Observational Health Data Sciences and Informatics", role = "cph"),
    person("Marcus", "Geelnard", role = c("cph", "ctb"),
      comment = "provided the TinyThread library"),
    person("Rutgers University", role = c("cph", "ctb"),
      comment = "provided the HParSearch routine"),
    person("R Development Core Team", role = c("cph", "ctb"),
      comment = "provided the ZeroIn routine")
    )
Description: This model fitting tool incorporates cyclic coordinate descent and
    majorization-minimization approaches to fit a variety of regression models
    found in large-scale observational healthcare data.  Implementations focus
    on computational optimization and fine-scale parallelization to yield
    efficient inference in massive datasets.  Please see:
    Suchard, Simpson, Zorych, Ryan and Madigan (2013) <doi:10.1145/2414416.2414791>.
License: Apache License 2.0
LazyData: Yes
Biarch: true
URL: https://github.com/ohdsi/cyclops
BugReports: https://github.com/ohdsi/cyclops/issues
Depends: R (>= 3.5.0)
Imports: rlang, Matrix, Rcpp (>= 0.12.12), Andromeda, dplyr, methods,
        survival, tidyr, bit, bit64
LinkingTo: Rcpp, RcppEigen (>= 0.3.2)
Suggests: testthat, readr, MASS, gnm, ggplot2, microbenchmark, cmprsk,
        boot, rsample
NeedsCompilation: yes
RoxygenNote: 7.3.2
Encoding: UTF-8
Packaged: 2026-05-26 15:24:17 UTC; msuchard
Author: Marc A. Suchard [aut, cre],
  Martijn J. Schuemie [aut],
  Trevor R. Shaddox [aut],
  Yuxi Tian [aut],
  Jianxiao Yang [aut],
  Eric Kawaguchi [aut],
  Sushil Mittal [ctb],
  Observational Health Data Sciences and Informatics [cph],
  Marcus Geelnard [cph, ctb] (provided the TinyThread library),
  Rutgers University [cph, ctb] (provided the HParSearch routine),
  R Development Core Team [cph, ctb] (provided the ZeroIn routine)
Maintainer: Marc A. Suchard <msuchard@ucla.edu>
Repository: CRAN
Date/Publication: 2026-05-27 06:20:14 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-27 08:28:06 UTC; unix
Archs: Cyclops.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(coef,cyclopsFit)
S3method(confint,cyclopsFit)
S3method(convertToCyclopsData,data.frame)
S3method(convertToCyclopsData,tbl_dbi)
S3method(logLik,cyclopsFit)
S3method(predict,cyclopsFit)
S3method(print,cyclopsData)
S3method(print,cyclopsFit)
S3method(print,cyclopsZph)
S3method(residuals,cyclopsFit)
S3method(summary,cyclopsData)
S3method(survfit,cyclopsFit)
S3method(vcov,cyclopsFit)
export(Multitype)
export(aconfint)
export(cacheCyclopsModelForJava)
export(clearCyclopsModelCache)
export(convertToCyclopsData)
export(convertToTimeVaryingCoef)
export(coverage)
export(createAutoGridCrossValidationControl)
export(createControl)
export(createCyclopsData)
export(createNonSeparablePrior)
export(createParameterizedPrior)
export(createPrior)
export(createWeightBasedSearchControl)
export(finalizeSqlCyclopsData)
export(fitCyclopsModel)
export(fitCyclopsSimulation)
export(getCovariateIds)
export(getCovariateTypes)
export(getCyclopsProfileLogLikelihood)
export(getFineGrayWeights)
export(getFloatingPointSize)
export(getHyperParameter)
export(getNumberOfCovariates)
export(getNumberOfRows)
export(getNumberOfStrata)
export(getUnivariableCorrelation)
export(getUnivariableSeparability)
export(gradient)
export(isInitialized)
export(listGPUDevices)
export(meanLinearPredictor)
export(mse)
export(readCyclopsData)
export(runBootstrap)
export(setOpenCLDevice)
export(simulateCyclopsData)
export(splitTime)
export(testProportionality)
import(Matrix)
import(Rcpp)
import(dplyr)
importFrom(bit64,as.integer64)
importFrom(bit64,is.integer64)
importFrom(methods,as)
importFrom(rlang,.data)
importFrom(stats,aggregate)
importFrom(stats,approx)
importFrom(stats,as.formula)
importFrom(stats,coef)
importFrom(stats,coefficients)
importFrom(stats,confint)
importFrom(stats,contrasts)
importFrom(stats,deviance)
importFrom(stats,model.matrix)
importFrom(stats,model.offset)
importFrom(stats,model.response)
importFrom(stats,pchisq)
importFrom(stats,poisson)
importFrom(stats,predict)
importFrom(stats,printCoefmat)
importFrom(stats,qchisq)
importFrom(stats,qnorm)
importFrom(stats,rbinom)
importFrom(stats,residuals)
importFrom(stats,rexp)
importFrom(stats,rnorm)
importFrom(stats,rpois)
importFrom(stats,runif)
importFrom(stats,terms)
importFrom(stats,time)
importFrom(stats,vcov)
importFrom(survival,Surv)
importFrom(survival,survSplit)
importFrom(survival,survfit)
useDynLib(Cyclops, .registration = TRUE)

Cyclops
=======

[![Build Status](https://github.com/ohdsi/Cyclops/actions/workflows/R_CMD_check_Hades.yaml/badge.svg)](https://github.com/OHDSI/Cyclops/actions?query=workflow%3AR-CMD-check)
[![codecov.io](https://codecov.io/github/OHDSI/Cyclops/coverage.svg?branch=main)](https://app.codecov.io/github/OHDSI/Cyclops?branch=main)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/Cyclops)](https://CRAN.R-project.org/package=Cyclops)
[![CRAN_Status_Badge](https://cranlogs.r-pkg.org/badges/Cyclops)](https://cran.r-project.org/package=Cyclops)

Cyclops is part of the [HADES](https://ohdsi.github.io/Hades/).


Introduction
============

Cyclops (Cyclic coordinate descent for logistic, Poisson and survival analysis) is an R package for performing large scale regularized regressions.

Features
========
 - Regression of very large problems: up to millions of observations, millions of variables
 - Supports (conditional) logistic regression, (conditional) Poisson regression, as well as (conditional) Cox regression
 - Uses a sparse representation of the independent variables when appropriate
 - Supports using no prior, a normal prior or a Laplace prior
 - Supports automatic selection of hyperparameter through cross-validation
 - Efficient estimation of confidence intervals for a single variable using a profile-likelihood for that variable

Examples
========

```r
  library(Cyclops)
  cyclopsData <- createCyclopsDataFrame(formula)
  cyclopsFit <- fitCyclopsModel(cyclopsData)
```

Technology
============
Cyclops in an R package, with most functionality implemented in C++. Cyclops uses cyclic coordinate descent to optimize the likelihood function, which makes use of the sparse nature of the data.

System Requirements
===================
Requires R (version 3.1.0 or higher). Compilation on Windows requires [RTools >= 3.4]( https://CRAN.R-project.org/bin/windows/Rtools/).

Installation
============
In R, to install the latest stable version, install from CRAN:

```r
install.packages("Cyclops")
```

To install the latest development version, install from GitHub. Note that this will require RTools to be installed.

```r
install.packages("devtools")
devtools::install_github("OHDSI/Cyclops")
```


User Documentation
==================
Documentation can be found on the [package website](https://ohdsi.github.io/Cyclops/).

PDF versions of the documentation are also available:
* Package manual: [Cyclops manual](https://raw.githubusercontent.com/OHDSI/Cyclops/main/extras/Cyclops.pdf)

Support
=======
* Developer questions/comments/feedback: <a href="http://forums.ohdsi.org/c/developers">OHDSI Forum</a>
* We use the <a href="https://github.com/OHDSI/Cyclops/issues">GitHub issue tracker</a> for all bugs/issues/enhancements

Contributing
============
Read [here](https://ohdsi.github.io/Hades/contribute.html) how you can contribute to this package.

License
=======
Cyclops is licensed under Apache License 2.0.   Cyclops contains the TinyThread libray.

The TinyThread library is licensed under the [zlib/libpng](https://opensource.org/license/zlib) license as described [here](https://tinythreadpp.bitsnbites.eu/).

Development
===========
Cyclops is being developed in R Studio.

Acknowledgements
================
- This project is supported in part through the National Science Foundation grants IIS 1251151 and DMS 1264153.

#' @title Create a Cyclops control object that supports multiple hyperparameters
#'
#' @description
#' \code{createCrossValidationControl} creates a Cyclops control object for use with \code{\link{fitCyclopsModel}}
#' that supports multiple hyperparameters through an auto-search in one dimension and a grid-search over the remaining
#' dimensions
#'
#' @param outerGrid     List or data.frame of grid parameters to explore
#' @param autoPosition  Vector position for auto-search parameter (concatenated into outerGrid)
#' @param refitAtMaximum Logical: re-fit Cyclops object at maximal cross-validation parameters
#' @param cvType        Must equal "auto"
#' @param initialValue  Initial value for auto-search parameter
#' @param ...           Additional parameters passed through to \code{\link{createControl}}
#'
#' @return
#' A Cyclops prior object of class inheriting from \code{"cyclopsPrior"} and \code{"cyclopsFunctionalPrior"}
#' for use with \code{fitCyclopsModel}.
#'
#' @export
createAutoGridCrossValidationControl <- function(outerGrid,
                                         autoPosition = 1,
                                         refitAtMaximum = TRUE,
                                         cvType = "auto",
                                         initialValue = 1,
                                         ...) {
    if (cvType != "auto") {
        stop("Only auto cross-validation allowed in inner loop")
    }
    control <- createControl(cvType = cvType, ...)

    # Set-up grid
    outerGrid <- as.data.frame(outerGrid)
    numRows <- nrow(outerGrid)
    numCols <- ncol(outerGrid)
    innerAuto <- rep(initialValue, numRows)

    if (autoPosition <= 0 || autoPosition > (numCols + 1)) {
        stop("Auto-position is invalid")
    }

    if (autoPosition == 1) {
        outerGrid <- cbind(innerAuto, outerGrid)
    } else if (autoPosition == (numCols + 1)) {
        outerGrid <- cbind(outerGrid, innerAuto)
    } else {
        outerGrid <- cbind(outerGrid[,1:(autoPosition - 1)],
                           innerAuto,
                           outerGrid[,autoPosition:numCols])
    }

    control$setHook <- function(cyclopsData, prior, control, ...) {
        # closure to capture arguments

        if (!inherits(prior, "cyclopsParameterizedPrior")) {
            stop("Auto-grid cross-validation is only implemented for parameterized priors")
        }

        if (!prior$useCrossValidation) {
            stop("Prior specification must require cross-validation")
        }

        control$setHook <- NULL # Do not re-enter call-back
        # parametrize <- prior$parameterize

        # Delegate parameterized prior functions
        priors <- apply(outerGrid, MARGIN = 1,
                         function(point) {
                             newPrior <- createParameterizedPrior(
                                 priorType = prior$priorType,
                                 parameterize =  function(x) {
                                     point[autoPosition] <- x
                                     prior$parameterize(point)
                                 },
                                 values = prior$values[autoPosition],
                                 useCrossValidation = TRUE,
                                 forceIntercept = prior$forceIntercept)
                             newPrior$point <- point
                             newPrior$point[autoPosition] <- NA
                             return(newPrior)
                         })

        # Run grid search
        searchResults <- lapply(priors, # TODO Could run in parallel
                                function(prior) {
                                    fit <- fitCyclopsModel(cyclopsData,
                                                           prior = prior,
                                                           control = control, ...)
                                    cvInfo <- getCrossValidationInfo(fit)
                                    pt <- cvInfo$point
                                    cvInfo$point <- prior$point
                                    cvInfo$point[is.na(cvInfo$point)] <- pt
                                    list(fit = fit,
                                         cvInfo = cvInfo)
                                })

        gridEvals <- sapply(searchResults,
                            function(fit) {
                                fit$cvInfo$ordinate
                            })

        whichMax <- which(gridEvals == max(gridEvals))
        cvInfo <- searchResults[[whichMax]]$cvInfo

        if (refitAtMaximum) {
            maxFit <- fitCyclopsModel(cyclopsData,
                                      prior = priors[[whichMax]],
                                      control = control, ...)
        } else {
            maxFit <- gridEvals[[whichMax]]
        }

        maxFit$searchResults <- searchResults
        maxFit$cross_validation <- paste0(c(cvInfo$ordinate,
                                            cvInfo$point), collapse = " ")
        return(maxFit)
    }

    return(control)
}

#' @method survfit cyclopsFit
#' @title Calculate baseline hazard function
#'
#' @description
#' \code{survfit.cyclopsFit} computes baseline hazard function
#'
#' @param formula A Cyclops survival model fit object
#' @param type type of baseline survival, choices are: "aalen" (Breslow)
#' @param ... for future methods
#'
#' @return Baseline survival function for mean covariates
#'
#' @importFrom survival survfit
#'
#' @export
survfit.cyclopsFit <- function(formula, type="aalen", ...) {
    cyclopsFit <- formula
    delta = meanLinearPredictor(cyclopsFit)

    times = getTimeVector(cyclopsFit$cyclopsData)
    events = getYVector(cyclopsFit$cyclopsData)
    predictors = exp(predict(cyclopsFit))

    if (type == "aalen") {
        accDenom = Reduce("+",predictors,accumulate=TRUE)
        newAccDenom = accDenom[length(times)+1-match(unique(times),times[length(times):1])]
        newTimes = unique(times)
        n = length(newTimes)

        t = table(times*events)[-1]
        g = match(as.numeric(names(t)), newTimes)
        newY = rep(0,n)
        newY[g] = t

        l = newY/newAccDenom
        L = rep(0,n)
        for (i in 1:n) {
            L[n+1-i] = sum(l[i:n])
        }
        return (list(time = newTimes[order(newTimes)],
                     surv = 1/exp(L)^exp(delta)))
    }
}

# @file Residuals.R
#
# Copyright 2024 Observational Health Data Sciences and Informatics
#
# This file is part of cyclops
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#' @method residuals cyclopsFit
#' @title Model residuals
#'
#' @description
#' \code{residuals.cyclopsFit} computes model residuals for Cox model-based Cyclops objects
#'
#' @param object    A Cyclops model fit object
#' @param parm      A specification of which parameters require residuals,
#'                  either a vector of numbers or covariateId names
#' @param type      Character string indicating the type of residual desires. Possible
#'                  values are "schoenfeld".
#' @param ...		Additional parameters for compatibility with S3 parent function
#'
#' @importFrom stats residuals
#'
#' @export
residuals.cyclopsFit <- function(object, parm = NULL, type = "schoenfeld", ...) {

    .checkInterface(object$cyclopsData, testOnly = TRUE)

    if (object$cyclopsData$modelType != "cox") {
        stop("Residuals for only Cox models are currently implemented")
    }
    if (type != "schoenfeld") {
        stop("Only Schoenfeld residuals are currently implemented")
    }

    if (getNumberOfCovariates(object$cyclopsData) != 1) {
        stop("Only single-covariate models are currently implemented")
    }

    res <- .cyclopsGetSchoenfeldResiduals(object$interface, NULL)

    res <- res[order(res$strata, res$times),]

    result <- res$residuals
    names(result) <- res$times

    tbl <- table(res$strata)
    if (dim(tbl) > 1) {
        names(tbl) <- paste0("stratum=", names(tbl))
        attr(result, "strata") <- tbl
    }

    return(result)
}

#' @title Test hazard ratio proportionality assumption
#'
#' @description
#' \code{testProportionality} tests the hazard ratio proportionality assumption
#' of a Cyclops model fit object using a score test
#'
#' @param object    A Cyclops model fit object
#' @param parm      A specification of which parameters require a proportionality test,
#'                  either a vector of numbers or covariateId names
#' @param transform Same as `transform` in `cox.zph`
#'
#' @export
testProportionality <- function(object, parm = NULL, transform = "km") {

    .checkInterface(object$cyclopsData, testOnly = TRUE)

    if (object$cyclopsData$modelType != "cox") {
        stop("Proportionality test for only Cox models are currently implemented")
    }

    nCovariates <- getNumberOfCovariates(object$cyclopsData)
    if (nCovariates != 1) {
        stop("Only single-covariate models are currently implemented")
    }

    times <- getTimeVector(object$cyclopsData)
    y <- getYVector(object$cyclopsData)
    survY <- survival::Surv(time = times, event = y)
    if (is.character(transform)) {
        tname <- transform
        ttimes <- switch(transform, identity = times, rank = rank(times),
                         log = log(times),
                         km = {
                             temp <- survival::survfitKM(
                                 factor(rep(1L, nrow(survY))), survY,
                                 se.fit = FALSE)
                             indx <- findInterval(times, temp$time, left.open = TRUE)
                             1 - c(1, temp$surv)[indx + 1]
                             },
                         stop("Unrecognized transform"))
    } else {
        tname <- deparse(substitute(transform))
        if (length(tname) > 1) {
            tname <- "user"
        }
        ttimes <- transform(times)
    }
    transformedTimes <- ttimes - mean(ttimes[y == 1])

    res <- .cyclopsTestProportionality(object$interface, NULL, transformedTimes)
    nCovariates <- 1 # TODO Remove
    res$hessian <- matrix(res$hessian, nrow = (nCovariates + 1))

    if (any(abs(res$gradient[1:nCovariates]) > 1E-5)) {
        stop("Internal state of Cyclops 'object' is not at its mode: ", res$gradient)
    }

    u <- c(rep(0, nCovariates), res$gradient[nCovariates + 1])
    test <- drop(solve(res$hessian, u) %*% u)
    df <- 1

    tbl <- cbind(test, df, pchisq(test, df, lower.tail = FALSE))

    names <- as.character(getCovariateIds(object$cyclopsData)[1])
    if (!is.null(object$cyclopsData$coefficientNames)) {
        names <- object$coefficientNames[1]
    }

    dimnames(tbl) <- list(names, c("chisq", "df", "p"))
    res$table <- tbl

    class(res) <- "cyclopsZph"

    return(res)
}

#' @method print cyclopsZph
#' @importFrom stats printCoefmat
#'
#' @export
print.cyclopsZph <- function(x, digits = max(options()$digits - 4, 3),
                             signif.stars = FALSE, ...) {
    invisible(printCoefmat(x$table, digits=digits, signif.stars=signif.stars,
                           P.values=TRUE, has.Pvalue=TRUE, ...))
}

library("testthat")
library("survival")

suppressWarnings(RNGversion("3.5.0"))

test_that("Check Schoenfeld residuals and PH test, no strata", {
    gfit <- coxph(Surv(futime, fustat) ~ age,
                  data=ovarian, method = "breslow")
    gres <- residuals(gfit, "schoenfeld")

    ovarian$mage <- ovarian$age - mean(ovarian$age)

    data <- createCyclopsData(Surv(futime, fustat) ~ age,
                              data=ovarian, modelType = "cox")
    cfit <- fitCyclopsModel(data)
    cres <- residuals(cfit, "schoenfeld")

    expect_equal(cres, gres)

    gtest <- cox.zph(gfit, transform = "identity", global = FALSE)
    ctest <- testProportionality(cfit, parm = NULL, transform = "identity")

    expect_equal(ctest$table, gtest$table)
})

test_that("Check Schoenfeld residuals and PH test, with strata", {
    gfit <- coxph(Surv(futime, fustat) ~ age + strata(ecog.ps),
                  data=ovarian, method = "breslow")
    gres <- residuals(gfit, "schoenfeld")

    data <- createCyclopsData(Surv(futime, fustat) ~ age + strata(ecog.ps),
                              data=ovarian, modelType = "cox")
    cfit <- fitCyclopsModel(data)
    cres <- residuals(cfit, "schoenfeld")

    expect_equivalent(cres, gres)

    gtest <- cox.zph(gfit, transform = "identity", global = FALSE)
    ctest <- testProportionality(cfit, parm = NULL, transform = "identity")

    expect_equal(ctest$table, gtest$table)
})

test_that("Check Schoenfeld residuals and PH test, with sparse covariates", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0.2,0
0, 3,  0,0,1
0, 3,  1,2,0 # was x1 = 2
0, 2,  1,0.5,1
0, 2,  1,0.1,1
0, 2,  1,1.2,1
0, 1,  0,1,0
0, 1,  1,1,0
")

    gfit <- coxph(Surv(length, event) ~ x1, test, ties = "breslow")
    gres <- residuals(gfit, "schoenfeld")
    gtest <- cox.zph(gfit, transform = "km", global = FALSE)

    data <- createCyclopsData(Surv(length, event) ~ 1,
                              sparseFormula = ~ x1,
                              data = test, modelType = "cox")

    cfit <- fitCyclopsModel(data)
    cres <- residuals(cfit, "schoenfeld")

    expect_equal(gres, cres)

    ctest <- testProportionality(cfit, parm = NULL, transform = "km")

    expect_equal(gtest$table, ctest$table)

    ctest2 <- testProportionality(cfit, parm = NULL)

    expect_equal(ctest, ctest2)

#     #  check in Stata
#
#     clear
#     input start length event x1 x2
#     0 4 1   0   0
#     0 3 0   0   1
#     0 3 1   2   0   // was x1 = 2
#     0 2 1   0.1 1
#     0 2 1   1.2 1
#     0 1 0   1   0   // was 0
#     0 1 1   1   0   // was 1
#     end
#
#     stset length, failure(event)
#     stcox x1
#     predict sch_x1, schoenfeld
#     show sch_x1
})

test_that("Check residuals on large Cox regression with weighting",{
    tolerance <- 1E-4
    set.seed(123)
    sim <- simulateCyclopsData(nstrata=1000, #1
                               ncovars=1,
                               nrows=10000, #30
                               effectSizeSd=0.5,
                               eCovarsPerRow=2,
                               model="survival")
    sim$outcomes$weights <- 1/sim$outcomes$rr
    # sim$outcomes$weights <- rep(1, length(sim$outcomes$weights))
    sim$outcomes$weights <- runif(length(sim$outcomes$weights))

    # Gold standard
    covariates <- sim$covariates
    ncovars <- max(covariates$covariateId)
    nrows <- nrow(sim$outcomes)
    m <- matrix(0,nrows,ncovars)
    for (i in 1:nrow(covariates)){
        m[covariates$rowId[i],covariates$covariateId[i]] <- 1
    }
    data <- as.data.frame(m)

    data$rowId <- 1:nrow(data)
    data <- merge(data,sim$outcomes)
    data <- data[order(data$stratumId,data$rowId),]
    formula <- as.formula(paste(c("Surv(time,y) ~ strata(stratumId)",paste("V",1:ncovars,sep="")),collapse=" + "))
    fitCoxph <- survival::coxph(formula, data = data, weights = data$weights, ties = "breslow")

    # Cyclops
    cyclopsData <- convertToCyclopsData(outcomes = sim$outcomes,
                                        covariates = sim$covariates,
                                        modelType = "cox")
    fitCyclops <- fitCyclopsModel(cyclopsData = cyclopsData)

    expect_equivalent(coef(fitCyclops), coef(fitCoxph), tolerance = tolerance)

    resCoxph <- residuals(fitCoxph, type = "schoenfeld")
    resCyclops <- residuals(fitCyclops, type = "schoenfeld")

    expect_equivalent(resCoxph, resCyclops)

    gtest1 <- cox.zph(fitCoxph, transform = "km", global = FALSE)
    ctest1 <- testProportionality(fitCyclops, transform = "km")
    expect_equivalent(gtest1$table, ctest1$table)

    gtest2 <- cox.zph(fitCoxph, transform = "identity", global = FALSE)
    ctest2 <- testProportionality(fitCyclops, transform = "identity")
    expect_equivalent(gtest2$table, ctest2$table)
})

test_that("Check Schoenfeld residuals and PH test, with sparse covariates", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3,  1,2,0
0, 3,  0,0,1
0, 2,  1,0,1
0, 2,  1,1,1
0, 1,  0,1,0
0, 1,  1,1,0
")

    gfit <- coxph(Surv(length, event) ~ x1 + strata(x2), test, ties = "breslow")
    gres <- residuals(gfit, "schoenfeld")

    data <- createCyclopsData(Surv(length, event) ~ strata(x2),
                                    sparseFormula = ~ x1,
                                    data = test, modelType = "cox")

    cfit <- fitCyclopsModel(data)
    cres <- residuals(cfit, "schoenfeld")

    expect_equivalent(cres, gres)
})

# @file PackageMaintenance
#
# Copyright 2014 Observational Health Data Sciences and Informatics
#
# This file is part of Cyclops
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


.createManualAndVignettes <- function(){
    if (.Platform$OS.type == "unix") {
        system("rm extras/Cyclops.pdf")
        system("R CMD Rd2pdf ./ --output=extras/Cyclops.pdf")
    } else {
        shell("rm extras/Cyclops.pdf")
        shell("R CMD Rd2pdf ./ --output=extras/Cyclops.pdf")
    }

    #No vignettes yet
    #require(rmarkdown)
    #render("vignettes/SingleStudies.Rmd", pdf_document(latex_engine = "pdflatex",toc = TRUE,number_sections = TRUE))
}

# @file Predict.R
#
# Copyright 2016 Observational Health Data Sciences and Informatics
#
# This file is part of cyclops
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#' @method predict cyclopsFit
#' @title Model predictions
#'
#' @description
#' \code{predict.cyclopsFit} computes model response-scale predictive values for all data rows
#'
#' @param object    A Cyclops model fit object
#' @param newOutcomes  An optional data frame or Andromeda table object, similar to the object used in \code{\link{convertToCyclopsData}}.
#' @param newCovariates  An optional data frame or Andromeda table object, similar to the object used in \code{\link{convertToCyclopsData}}.
#' @param ...   Additional arguments
#'
#' @importFrom stats predict
#'
#' @export
predict.cyclopsFit <- function(object, newOutcomes, newCovariates, ...) {
    if (!missing(newOutcomes) && (missing(newCovariates) || is.null(newCovariates)))
        stop("Need to specify both newOutcomes and newCovariates")
    if (!missing(newCovariates) && (missing(newOutcomes) || is.null(newOutcomes)))
        stop("Need to specify both newOutcomes and newCovariates")
    if (missing(newOutcomes)) {
        # Predict for same data that was used to fit the model:
        .checkInterface(object$cyclopsData, testOnly = TRUE)
        pred <- .cyclopsPredictModel(object$cyclopsData$cyclopsInterfacePtr)
        values <- pred$prediction
        if (is.null(names(values))) {
            names(values) <- object$rowNames
        }
        return(values)
    } else {
        # Predict for new data:
        modelType <- object$cyclopsData$modelType
        if (modelType == "cox")
            stop("Prediction for Cox models not implemented")
        if (modelType == "cpr" || modelType == "clr")
            stop("Prediction for conditional models not implemented")

        if (any(class(newOutcomes) != class(newCovariates))) {
            stop("`newCovariates` and `newOutcomes` must be of the same type")
        }

        coefficients <- coef(object)
        intercept <- coefficients[1]
        coefficients <- coefficients[2:length(coefficients)]
        coefficients <- data.frame(beta = as.numeric(coefficients),
                                   covariateId = as.numeric(names(coefficients)))
        coefficients <- coefficients[coefficients$beta != 0, ]

        if (inherits(newCovariates, "tbl_dbi")) {

            # Optimized for Andromeda
            if (nrow(coefficients) == 0) {
                if ("time" %in% colnames(newOutcomes)) {
                    prediction <- newOutcomes %>%
                        select("rowId", "time") %>%
                        collect()
                } else {
                    prediction <- newOutcomes %>%
                        select("rowId") %>%
                        collect()
                }
                prediction <- prediction %>%
                    mutate(value = intercept)
            } else {
                prediction <- inner_join(newCovariates,
                                         coefficients, by = "covariateId", copy = TRUE)

                prediction <- prediction %>%
                    mutate(value = .data$covariateValue * .data$beta) %>%
                    group_by(.data$rowId) %>%
                    summarize(value = sum(.data$value, na.rm = TRUE))

                prediction <- left_join(newOutcomes,
                                         prediction, by = "rowId") %>%
                    collect()

                prediction$value[is.na(prediction$value)] <- 0
                prediction$value <- prediction$value + intercept
            }
            prediction <- prediction %>%
                arrange(.data$rowId)
        } else {
            # Not using Andromeda
            if (nrow(coefficients) == 0) {
                prediction <- newOutcomes
                prediction$value <- intercept
            } else {
                prediction <- merge(newCovariates, coefficients, by = "covariateId")
                prediction$value <- prediction$covariateValue * prediction$beta
                prediction <- aggregate(value ~ rowId, data = prediction, sum)
                prediction <- merge(newOutcomes, prediction, by = "rowId", all.x = TRUE)
                prediction$value[is.na(prediction$value)] <- 0
                prediction$value <- prediction$value + intercept
            }
        }

        if (modelType == "lr") {
            link <- function(x) {
                return(1/(1 + exp(0 - x)))
            }
            prediction$value <- link(prediction$value)
        } else if (modelType == "pr") {
            prediction$value <- exp(prediction$value) * prediction$time
        }

        result <- prediction$value
        names(result) <- prediction$rowId
        return(result)
    }

}

#' @title Calculates xbar*beta
#' @description
#' \code{meanLinearPredictor} computes xbar*beta for model fit
#'
#' @param cyclopsFit A Cyclops model fit object
#'
#' @export
meanLinearPredictor <- function(cyclopsFit) {
    cyclopsData = cyclopsFit$cyclopsData
    dataSummary = summary(cyclopsData)
    dataSummary$xbar = dataSummary$nzCount*dataSummary$nzMean/getNumberOfRows(cyclopsData)
    dataSummary$beta = coef(cyclopsFit)[match(rownames(dataSummary),names(coef(cyclopsFit)))]
    dataSummary$xbarBeta = dataSummary$xbar * dataSummary$beta
    delta = sum(dataSummary$xbarBeta)
    return (delta)
}

#' @title Create a Cyclops parameterized prior object
#'
#' @description
#' \code{createParameterizedPrior} creates a Cyclops prior object for use with \code{\link{fitCyclopsModel}}
#' in which arbitrary \code{R} functions parameterize the prior location and variance.
#'
#' @param priorType     Character vector: specifies prior distribution.  See below for options
#' @param parameterize  Function list: parameterizes location and variance
#' @param values        Numeric vector: initial parameter values
#' @param useCrossValidation    Logical: Perform cross-validation to determine \code{parameters}.
#' @param forceIntercept  Logical: Force intercept coefficient into prior
#'
#' @return
#' A Cyclops prior object of class inheriting from \code{"cyclopsPrior"} and \code{"cyclopsFunctionalPrior"}
#' for use with \code{fitCyclopsModel}.
#'
#' @export
createParameterizedPrior <- function(priorType,
                                  parameterize,
                                  values,
                                  useCrossValidation = FALSE,
                                  forceIntercept = FALSE) {

    validNames = c("none", "laplace","normal")
    stopifnot(priorType %in% validNames)

    if (missing(values)) {
        stop("Must provide parameter values")
    }

    parameterizationEval <- parameterize(values)
    if (!inherits(parameterizationEval, "list")) {
        stop("Parameterization function should return a list")
    }

    if (length(priorType) != length(parameterizationEval)) {
        stop("Prior types and functions have a dimensionality mismatch")
    }

    if (all(priorType == "none") && useCrossValidation) {
        stop("Cannot perform cross validation with a flat prior")
    }

    setHook <- function(cyclopsData, warnings) {
        # closure to capture arguments
        if (length(priorType) > 1) {
            if (length(priorType) != getNumberOfCovariates(cyclopsData)) {
                stop("Length of priors must equal the number of covariates")
            }
        }

        if (priorType[1] != "none" && .cyclopsGetHasIntercept(cyclopsData) && !forceIntercept) {
            priorType[1] <- "none"
            if (warnings) {
                warning("Excluding intercept from regularization")
            }
        }

        .cyclopsSetParameterizedPrior(cyclopsData$cyclopsInterfacePtr,
                                      priorType,
                                      parameterize,
                                      values,
                                      excludeNumeric = NULL)
    }

    structure(list(priorType = priorType,
                   useCrossValidation = useCrossValidation,
                   forceIntercept = forceIntercept,
                   setHook = setHook,
                   parameterize = parameterize,
                   values = values),
              class = c("cyclopsPrior", "cyclopsParameterizedPrior"))
}

# @file NewDataConversion.R
#
# Copyright 2020 Observational Health Data Sciences and Informatics
#
# This file is part of cyclops
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

isSorted <- function(data, columnNames, ascending = rep(TRUE, length(columnNames))){
    return(.isSorted(data, columnNames, ascending))
}

#' Convert data from two data frames or ffdf objects into a CyclopsData object
#'
#' @description
#' \code{convertToCyclopsData} loads data from two data frames or ffdf objects, and inserts it into a Cyclops data object.
#'
#' @param outcomes      A data frame or ffdf object containing the outcomes with predefined columns (see below).
#' @param covariates    A data frame or ffdf object containing the covariates with predefined columns (see below).
#' @param modelType     Cyclops model type. Current supported types are "pr", "cpr", lr", "clr", or "cox"
#' @param addIntercept  Add an intercept to the model?
#' @param checkSorting  (DEPRECATED) Check if the data are sorted appropriately, and if not, sort.
#' @param checkRowIds   Check if all rowIds in the covariates appear in the outcomes.
#' @param normalize     String: Name of normalization for all non-indicator covariates (possible values: stdev, max, median)
#' @param quiet         If true, (warning) messages are suppressed.
#' @param floatingPoint Specified floating-point representation size (32 or 64)
#' @param timeEffectMap A data frame or ffdf object containing the convariates that have time-varying effects on the outcome
#'
#' @details
#' These columns are expected in the outcome object:
#' \tabular{lll}{
#'   \verb{stratumId}    \tab(integer) \tab (optional) Stratum ID for conditional regression models \cr
#'   \verb{rowId}  	\tab(integer) \tab Row ID is used to link multiple covariates (x) to a single outcome (y) \cr
#'   \verb{y}    \tab(real) \tab The outcome variable \cr
#'   \verb{time}    \tab(real) \tab For models that use time (e.g. Poisson or Cox regression) this contains time \cr
#'                  \tab        \tab(e.g. number of days) \cr
#'   \verb{weights} \tab(real) \tab (optional) Non-negative weights to apply to outcome \cr
#'   \verb{censorWeights} \tab(real) \tab (optional) Non-negative censoring weights for competing risk model; will be computed if not provided.
#' }
#'
#' These columns are expected in the covariates object:
#' \tabular{lll}{
#'   \verb{stratumId}    \tab(integer) \tab (optional) Stratum ID for conditional regression models \cr
#'   \verb{rowId}  	\tab(integer) \tab Row ID is used to link multiple covariates (x) to a single outcome (y) \cr
#'   \verb{covariateId}    \tab(integer) \tab A numeric identifier of a covariate  \cr
#'   \verb{covariateValue}    \tab(real) \tab The value of the specified covariate \cr
#' }
#'
#' These columns are expected in the timeEffectMap object:
#' \tabular{lll}{
#'   \verb{covariateId}    \tab(integer) \tab A numeric identifier of the covariates that have time-varying effects on the outcome \cr
#' }
#'
#' @return
#' An object of type cyclopsData
#'
#' @examples
#' #Convert infert dataset to Cyclops format:
#' covariates <- data.frame(stratumId = rep(infert$stratum, 2),
#'                          rowId = rep(1:nrow(infert), 2),
#'                          covariateId = rep(1:2, each = nrow(infert)),
#'                          covariateValue = c(infert$spontaneous, infert$induced))
#' outcomes <- data.frame(stratumId = infert$stratum,
#'                        rowId = 1:nrow(infert),
#'                        y = infert$case)
#' #Make sparse:
#' covariates <- covariates[covariates$covariateValue != 0, ]
#'
#' #Create Cyclops data object:
#' cyclopsData <- convertToCyclopsData(outcomes, covariates, modelType = "clr",
#'                                     addIntercept = FALSE)
#'
#' #Fit model:
#' fit <- fitCyclopsModel(cyclopsData, prior = createPrior("none"))
#'
#' @export
convertToCyclopsData <- function(outcomes,
                                 covariates,
                                 modelType = "lr",
				 timeEffectMap = NULL,
                                 addIntercept = TRUE,
                                 checkSorting = NULL,
                                 checkRowIds = TRUE,
                                 normalize = NULL,
                                 quiet = FALSE,
                                 floatingPoint = 64) {
    UseMethod("convertToCyclopsData")
}

#' @describeIn convertToCyclopsData Convert data from two \code{data.frame}
#' @export
# <<<<<<< HEAD
# convertToCyclopsData.ffdf <- function(outcomes,
#                                       covariates,
#                                       modelType = "lr",
#                                       addIntercept = TRUE,
#                                       checkSorting = TRUE,
#                                       checkRowIds = TRUE,
#                                       normalize = NULL,
#                                       quiet = FALSE,
#                                       floatingPoint = 64){
#     if ((modelType == "clr" | modelType == "cpr" | modelType == "clr_exact" | modelType == "clr_efron") & addIntercept){
#         if(!quiet) {
#             warning("Intercepts are not allowed in conditional models, removing intercept",call.=FALSE)
#         }
# =======
convertToCyclopsData.data.frame <- function(outcomes,
                                            covariates,
                                            modelType = "lr",
					    timeEffectMap = NULL,
                                            addIntercept = TRUE,
                                            checkSorting = NULL,
                                            checkRowIds = TRUE,
                                            normalize = NULL,
                                            quiet = FALSE,
                                            floatingPoint = 64) {
    if (!is.null(checkSorting))
        warning("The 'checkSorting' argument has been deprecated. Sorting is now always checked")

    if ((modelType == "clr" | modelType == "cpr") & addIntercept) {
        if (!quiet)
            warning("Intercepts are not allowed in conditional models, removing intercept", call. = FALSE)
# >>>>>>> master
        addIntercept = FALSE
    }
    if (modelType == "pr" | modelType == "cpr")
        if (any(outcomes$time <= 0))
            stop("time cannot be non-positive", call. = FALSE)

    if (modelType == "lr" | modelType == "pr") {
        outcomes$stratumId <- NULL
        covariates$stratumId <- NULL
    }
    if ((modelType == "cox" | modelType == "cox_time" | modelType == "fgr") & !"stratumId" %in% colnames(outcomes)) {
        outcomes$stratumId <- 0
        covariates$stratumId <- 0
    }

    if (modelType == "lr" | modelType == "pr") {
        if (!isSorted(outcomes, c("rowId"))) {
            if (!quiet)
                writeLines("Sorting outcomes by rowId")
            outcomes <- outcomes[order(outcomes$rowId),]
        }
        if (!isSorted(covariates, c("covariateId", "rowId"))) {
            if (!quiet)
                writeLines("Sorting covariates by covariateId and rowId")
            covariates <- covariates[order(covariates$covariateId, covariates$rowId),]
        }
    }

    if (modelType == "clr" | modelType == "cpr") {
        if (!isSorted(outcomes, c("stratumId","rowId"))) {
            if (!quiet)
                writeLines("Sorting outcomes by stratumId and rowId")
            outcomes <- outcomes[order(outcomes$stratumId,outcomes$rowId),]
        }
        if (!isSorted(covariates, c("covariateId", "stratumId","rowId"))) {
            if (!quiet)
                writeLines("Sorting covariates by covariateId, stratumId, and rowId")
            covariates <- covariates[order(covariates$covariateId, covariates$stratumId, covariates$rowId),]
        }
    }

    if (modelType == "cox" | modelType == "cox_time" | modelType == "fgr") {

        if ((modelType == "cox" | modelType == "cox_time") & length(unique(outcomes$y)) > 2) {
            stop("Cox model only accepts one outcome type")
        }
# <<<<<<< HEAD
#         if (modelType == "clr" | modelType == "cpr" | modelType == "clr_exact" | modelType == "clr_efron"){
#             if (!isSorted(outcomes,c("stratumId","rowId"))){
#                 if(!quiet) {
#                     writeLines("Sorting outcomes by stratumId and rowId")
#                 }
#                 rownames(outcomes) <- NULL #Needs to be null or the ordering of ffdf will fail
#                 outcomes <- outcomes[ff::ffdforder(outcomes[c("stratumId","rowId")]),]
#             }
#             if (!isSorted(covariates,c("covariateId", "stratumId","rowId"))){
#                 if(!quiet) {
#                     writeLines("Sorting covariates by covariateId, stratumId and rowId")
#                 }
#                 rownames(covariates) <- NULL #Needs to be null or the ordering of ffdf will fail
#                 covariates <- covariates[ff::ffdforder(covariates[c("covariateId", "stratumId","rowId")]),]
#             }
# =======
        if ("subjectId" %in% colnames(outcomes)) {
            if (!isSorted(outcomes,
                          c("stratumId", "time", "y", "subjectId", "rowId"),
                          c(TRUE, FALSE, TRUE, TRUE, TRUE))) {
                if (!quiet)
                    writeLines("Sorting outcomes by stratumId, time (descending), y, subjectId and rowId")
                outcomes <- outcomes[order(outcomes$stratumId, -outcomes$time, outcomes$y, outcomes$subjectId, outcomes$rowId),]
            }
        } else {
            if (!isSorted(outcomes,
                          c("stratumId", "time", "y", "rowId"),
                          c(TRUE, FALSE, TRUE, TRUE))) {
                if (!quiet)
                    writeLines("Sorting outcomes by stratumId, time (descending), y and rowId")
                    outcomes <- outcomes[order(outcomes$stratumId, -outcomes$time, outcomes$y, outcomes$rowId),]
            }
        }
        if (!"time" %in% colnames(covariates)) {
            covariates$time <- NULL
            covariates$y <- NULL
            covariates$stratumId <- NULL
            covariates <- merge(covariates, outcomes, by = c("rowId"))
        }
        if (!isSorted(covariates,
                      c("covariateId", "stratumId", "time", "y", "rowId"),
                      c(TRUE, TRUE, FALSE, TRUE, TRUE))) {
            if (!quiet)
                writeLines("Sorting covariates by covariateId, stratumId, time (descending), y, and rowId")
            covariates <- covariates[order(covariates$covariateId, covariates$stratumId, -covariates$time, covariates$y, covariates$rowId),]
        }
    }

    if (checkRowIds) {
        mapping <- match(covariates$rowId,outcomes$rowId)
        if (any(is.na(mapping))) {
            if (!quiet)
                writeLines("Removing covariate values with rowIds that are not in outcomes")
            covariateRowsWithMapping <- which(!is.na(mapping))
            covariates <- covariates[covariateRowsWithMapping,]
        }
    }

    dataPtr <- createSqlCyclopsData(modelType = modelType, floatingPoint = floatingPoint)

    loadNewSqlCyclopsDataY(object = dataPtr,
                           stratumId = if ("stratumId" %in% colnames(outcomes)) outcomes$stratumId else NULL,
                           rowId = outcomes$rowId,
                           y = outcomes$y,
                           time = if ("time" %in% colnames(outcomes)) outcomes$time else NULL)

    if (addIntercept & (modelType != "cox" & modelType != "cox_time" & modelType != "fgr")) {
        loadNewSqlCyclopsDataX(dataPtr, 0, NULL, NULL, name = "(Intercept)")
    }

    covarNames <- unique(covariates$covariateId)
    loadNewSqlCyclopsDataMultipleX(object = dataPtr,
                                   covariateId = covariates$covariateId,
                                   rowId = covariates$rowId,
                                   covariateValue = covariates$covariateValue,
                                   name = covarNames)

    if (modelType == "cox_time" && !is.null(timeEffectMap)) {
        if (!all(timeEffectMap$covariateId %in% covarNames)) stop("Invalid covariateId for time effects.")
        loadNewSqlCyclopsDataStratTimeEffects(object = dataPtr,
					      stratumId = outcomes$stratumId,
					      rowId = outcomes$rowId,
					      subjectId = outcomes$subjectId,
					      timeEffectCovariateId = sort(timeEffectMap$covariateId))
    }

    if (modelType == "pr" || modelType == "cpr")
        finalizeSqlCyclopsData(dataPtr, useOffsetCovariate = -1)

    if (!is.null(normalize)) {
        .normalizeCovariates(dataPtr, normalize)
    }

    if ("weights" %in% colnames(outcomes)) {
        dataPtr$weights <- outcomes$weights
    } else {
        dataPtr$censorWeights <- NULL
    }

    if ("censorWeights" %in% colnames(outcomes)) {
        dataPtr$censorWeights <- outcomes$censorWeights
    } else {
        if (modelType == "fgr") {
            dataPtr$censorWeights <- getFineGrayWeights(outcomes$time, outcomes$y)$weights
            writeLines("Generating censoring weights")
        } else {
            dataPtr$censorWeights <- NULL
        }
    }

    return(dataPtr)
}

#' @describeIn convertToCyclopsData Convert data from two \code{Andromeda} tables
#' @export
# <<<<<<< HEAD
# convertToCyclopsData.data.frame <- function(outcomes,
#                                             covariates,
#                                             modelType = "lr",
#                                             addIntercept = TRUE,
#                                             checkSorting = TRUE,
#                                             checkRowIds = TRUE,
#                                             normalize = NULL,
#                                             quiet = FALSE,
#                                             floatingPoint = 64){
#     if ((modelType == "clr" | modelType == "cpr" | modelType == "clr_exact" | modelType == "clr_efron") & addIntercept){
#         if(!quiet)
#             warning("Intercepts are not allowed in conditional models, removing intercept",call.=FALSE)
# =======
convertToCyclopsData.tbl_dbi <- function(outcomes,
                                         covariates,
                                         modelType = "lr",
					 timeEffectMap = NULL,
                                         addIntercept = TRUE,
                                         checkSorting = NULL,
                                         checkRowIds = TRUE,
                                         normalize = NULL,
                                         quiet = FALSE,
                                         floatingPoint = 64) {
    if (!is.null(checkSorting))
        warning("The 'checkSorting' argument has been deprecated. Sorting is now always checked")

    if ((modelType == "clr" | modelType == "cpr") & addIntercept) {
        if (!quiet) {
            warning("Intercepts are not allowed in conditional models, removing intercept", call. = FALSE)
        }
# >>>>>>> master
        addIntercept = FALSE
    }

    if (modelType == "pr" | modelType == "cpr") {
        if (any(pull(outcomes, time) <= 0)) {
            stop("time cannot be non-positive", call. = FALSE)
        }
    }

    providedNoStrata <- !"stratumId" %in% colnames(outcomes)

# <<<<<<< HEAD
#         if (modelType == "clr" | modelType == "cpr" | modelType == "clr_exact" | modelType == "clr_efron"){
#             if (!isSorted(outcomes,c("stratumId","rowId"))){
#                 if(!quiet)
#                     writeLines("Sorting outcomes by stratumId and rowId")
#                 outcomes <- outcomes[order(outcomes$stratumId,outcomes$rowId),]
#             }
#             if (!isSorted(covariates,c("covariateId", "stratumId","rowId"))){
#                 if(!quiet)
#                     writeLines("Sorting covariates by covariateId, stratumId, and rowId")
#                 covariates <- covariates[order(covariates$covariateId, covariates$stratumId,covariates$rowId),]
#             }
# =======
    if (modelType == "cox" | modelType == "cox_time" | modelType == "fgr") {
        if (providedNoStrata) {
            outcomes <- outcomes %>%
                mutate(stratumId = 0)
            covariates <- covariates %>%
                mutate(stratumId = 0)
# >>>>>>> master
        }
    }

    if (checkRowIds) {
        covariateRowIds <- covariates %>%
            distinct(.data$rowId) %>%
            pull()
        outcomeRowIds <- select(outcomes, "rowId") %>%
            pull()
        mapping <- match(covariateRowIds, outcomeRowIds)
        if (any(is.na(mapping))) {
            if (!quiet) {
                writeLines("Removing covariate values with rowIds that are not in outcomes")
            }
            covariates <- covariates %>%
                filter(.data$rowId %in% outcomeRowIds)
        }
    }

    # Sorting should be last, as other operations may change ordering.
    # Also, should always explicitly define sorting, else not guaranteed.
    if (modelType == "lr" | modelType == "pr") {
        outcomes <- outcomes %>%
            arrange(.data$rowId)

        covariates <- covariates %>%
            arrange(.data$covariateId, .data$rowId)
    }

    if (modelType == "clr" | modelType == "cpr") {
        outcomes <- outcomes %>%
            arrange(.data$stratumId, .data$rowId)

        covariates <- covariates %>%
            arrange(.data$covariateId, .data$stratumId, .data$rowId)
    }

    if (modelType == "cox" | modelType == "cox_time" | modelType == "fgr") {

        if ((modelType == "cox" | modelType == "cox_time") &
            (select(outcomes, "y") %>% distinct() %>% count() %>% collect() > 2)) {
            stop("Cox model only accepts one outcome type")
        }
        if (!"time" %in% colnames(covariates)) {
            covariates <- covariates %>%
                inner_join(select(outcomes, "rowId", "time", "y"), by = "rowId")
        }
        if ("subjectId" %in% colnames(outcomes)) {
            outcomes <- outcomes %>%
                arrange(.data$stratumId, desc(.data$time), .data$y, .data$subjectId, .data$rowId)
        } else {
            outcomes <- outcomes %>%
                arrange(.data$stratumId, desc(.data$time), .data$y, .data$rowId)
	}
        covariates <- covariates %>%
            arrange(.data$covariateId, .data$stratumId, desc(.data$time), .data$y, .data$rowId)
    }

    dataPtr <- createSqlCyclopsData(modelType = modelType, floatingPoint = floatingPoint)

    outcomes <- collect(outcomes)
    if (modelType == "lr" | modelType == "pr") {
        outcomes$stratumId <- NULL
    }

    loadNewSqlCyclopsDataY(object = dataPtr,
                           stratumId = if ("stratumId" %in% colnames(outcomes)) outcomes$stratumId else NULL,
                           rowId = outcomes$rowId,
                           y = outcomes$y,
                           time = if ("time" %in% colnames(outcomes)) (select(outcomes, time) %>% pull()) else NULL)

    if (addIntercept & (modelType != "cox" & modelType != "cox_time" & modelType != "fgr")) {
        loadNewSqlCyclopsDataX(dataPtr, 0, NULL, NULL, name = "(Intercept)")
    }

    loadCovariates <- function(batch) {
        covarNames <- unique(batch$covariateId)
        loadNewSqlCyclopsDataMultipleX(object = dataPtr,
                                        covariateId = batch$covariateId,
                                        rowId = batch$rowId,
                                        covariateValue = batch$covariateValue,
                                        name = covarNames,
                                        append = TRUE)
    }

    Andromeda::batchApply(covariates,
                          loadCovariates,
                          batchSize = 100000) # TODO Pick magic number

    if (modelType == "cox_time" && !is.null(timeEffectMap)) {
        covarNames <- unique(pull(covariates, .data$covariateId))
        if (!all(timeEffectMap$covariateId %in% covarNames)) stop("Invalid covariateId for time effects.")
        loadNewSqlCyclopsDataStratTimeEffects(object = dataPtr,
					      stratumId = outcomes$stratumId,
					      rowId = outcomes$rowId,
					      subjectId = outcomes$subjectId,
					      timeEffectCovariateId = sort(timeEffectMap$covariateId))
    }

    if (modelType == "pr" || modelType == "cpr")
        finalizeSqlCyclopsData(dataPtr, useOffsetCovariate = -1)

    if (!is.null(normalize)) {
        .normalizeCovariates(dataPtr, normalize)
    }

    if ("weights" %in% colnames(outcomes)) {
        dataPtr$weights <- outcomes %>% pull(.data$weights)
    } else {
        dataPtr$weights <- NULL
    }

    if ("censorWeights" %in% colnames(outcomes)) {
        dataPtr$censorWeights <- outcomes %>% pull(.data$censorWeights)
    } else {
        if (modelType == "fgr") {
            dataPtr$censorWeights <- getFineGrayWeights(
                outcomes %>% pull("time"),
                outcomes %>% pull("y")
            )$weights
            writeLines("Generating censoring weights")
        } else {
            dataPtr$censorWeights <- NULL
        }
    }

    return(dataPtr)
}

library("testthat")
library("survival")
library("cmprsk")

suppressWarnings(RNGversion("3.5.0"))

test_that("Check that Cox fits only have 2 outcome identifies", {

    test <- read.table(header=T, sep = ",", text = "
                       start, length, event, x1, x2
                       0, 4,  1,0,0
                       0, 3.5,2,2,0
                       0, 3,  0,0,1
                       0, 2.5,2,0,1
                       0, 2,  1,1,1
                       0, 1.5,0,1,0
                       0, 1,  1,1,0")
    outcomes <- data.frame(rowId = 1:7,
                           time = c(4, 3.5, 3, 2.5, 2, 1.5, 1),
                           y = c(1, 2, 0, 0, 1, 1, 1))

    covariates <- data.frame(rowId = c(2, 3, 4, 5, 5, 6, 7),
                             covariateId = c(1, 2, 2, 1, 2, 1, 1),
                             covariateValue = c(2, 1, 1, 1, 1, 1, 1))

    expect_error(ptr <- convertToCyclopsData(outcomes, covariates, modelType = "cox"))

    andro <- Andromeda::andromeda(outcomes = outcomes, covariates = covariates)
    expect_error(ptr <- convertToCyclopsData(andro$outcomes, andro$covariates, modelType = "cox"))
})

test_that("Check very small Fine-Gray example with no ties", {
    test <- read.table(header=T, sep = ",", text = "
                       start, length, event, x1, x2
                       0, 4,  1,0,0
                       0, 3.5,2,2,0
                       0, 3,  0,0,1
                       0, 2.5,2,0,1
                       0, 2,  1,1,1
                       0, 1.5,0,1,0
                       0, 1,  1,1,0")

    fgDat <- Cyclops:::getFineGrayWeights(test$length, test$event)
    dataPtr <- Cyclops:::createCyclopsData(fgDat$surv ~ test$x1 + test$x2, modelType = "fgr", censorWeights = fgDat$weights)
    cyclopsFit <- Cyclops:::fitCyclopsModel(dataPtr)
    goldFit <- crr(test$length, test$event, cbind(test$x1, test$x2), variance = FALSE)

    tolerance <- 1E-4
    expect_equivalent(coef(cyclopsFit), goldFit$coef, tolerance = tolerance)
})

test_that("Check very small Fine-Gray example with time ties, but no failure ties", {
    test <- read.table(header=T, sep = ",", text = "
                       start, length, event, x1, x2
                       0, 4,1,0,0
                       0, 3,2,2,0
                       0, 3,2,0,1
                       0, 3,1,0,1
                       0, 2,0,1,0
                       0, 2,0,1,0
                       0, 2,1,1,1")

    fgDat <- Cyclops:::getFineGrayWeights(test$length, test$event)
    dataPtr <- Cyclops:::createCyclopsData(fgDat$surv ~ test$x1 + test$x2, modelType = "fgr", censorWeights = fgDat$weights)
    cyclopsFit <- Cyclops:::fitCyclopsModel(dataPtr)
    goldFit <- crr(test$length, test$event, cbind(test$x1, test$x2), variance = FALSE)

    tolerance <- 1E-4
    expect_equivalent(coef(cyclopsFit), goldFit$coef, tolerance = tolerance)
})

test_that("Check very small Fine-Gray example with time ties and failure ties", {
    test <- read.table(header=T, sep = ",", text = "
                       start, length, event, x1, x2
                       0, 4,  1,0,0
                       0, 3, 1,2,0
                       0, 3,  1,0,1
                       0, 3, 2,0,1
                       0, 2,  0,1,1
                       0, 1, 2,1,0
                       0, 1, 0,1,1,
                       0, 1, 1,1,0,
                       0, 1, 1,1,1")

    fgDat <- Cyclops:::getFineGrayWeights(test$length, test$event)
    dataPtr <- Cyclops:::createCyclopsData(fgDat$surv ~ test$x1 + test$x2, modelType = "fgr", censorWeights = fgDat$weights)
    cyclopsFit <- Cyclops:::fitCyclopsModel(dataPtr)
    goldFit <- crr(test$length, test$event, cbind(test$x1, test$x2), variance = FALSE)

    tolerance <- 1E-4
    expect_equivalent(coef(cyclopsFit), goldFit$coef, tolerance = tolerance)
})

test_that("Check very small Fine-Gray example with no ties (sparse vs. dense)", {
    test <- read.table(header=T, sep = ",", text = "
                       start, length, event, x1, x2
                       0, 4,  1,0,0
                       0, 3.5,2,2,0
                       0, 3,  0,0,1
                       0, 2.5,2,0,1
                       0, 2,  1,1,1
                       0, 1.5,0,1,0
                       0, 1,  1,1,0")

    fgDat <- Cyclops:::getFineGrayWeights(test$length, test$event)
    dataPtr <- Cyclops:::createCyclopsData(fgDat$surv ~ test$x1 + test$x2, modelType = "fgr", censorWeights = fgDat$weights)
    denseFit <- Cyclops:::fitCyclopsModel(dataPtr)

    outcomes <- data.frame(rowId = 1:7, time = test$length, y = test$event, censorWeights = fgDat$weights)
    covariates <- data.frame(rowId = c(2, 3, 4, 5, 5, 6, 7),
                             covariateId = c(1, 2, 2, 1, 2, 1, 1),
                             covariateValue = c(2, 1, 1, 1, 1, 1, 1))

    dataPtr <- convertToCyclopsData(outcomes, covariates, modelType = "fgr")
    sparseFit <- fitCyclopsModel(dataPtr)

    tolerance <- 1E-8
    expect_equivalent(coef(denseFit), coef(sparseFit), tolerance = tolerance)

    woWeights <- data.frame(rowId = 1:7, time = test$length, y = test$event)
    woPtr <- convertToCyclopsData(woWeights, covariates, modelType = "fgr")
    woFit <- fitCyclopsModel(woPtr)
    expect_equal(coef(woFit), coef(sparseFit))
})

test_that("Check very small Fine-Gray example with no ties using Andromeda", {

    test <- read.table(header=T, sep = ",", text = "
                       start, length, event, x1, x2
                       0, 4,  1,0,0
                       0, 3.5,2,2,0
                       0, 3,  0,0,1
                       0, 2.5,2,0,1
                       0, 2,  1,1,1
                       0, 1.5,0,1,0
                       0, 1,  1,1,0")

    fgDat <- Cyclops:::getFineGrayWeights(test$length, test$event)
    dataPtr <- Cyclops:::createCyclopsData(fgDat$surv ~ test$x1 + test$x2, modelType = "fgr", censorWeights = fgDat$weights)
    denseFit <- Cyclops:::fitCyclopsModel(dataPtr)

    outcomes <- data.frame(rowId = 1:7, time = test$length, y = test$event, censorWeights = fgDat$weights)
    covariates <- data.frame(rowId = c(2, 3, 4, 5, 5, 6, 7),
                             covariateId = c(1, 2, 2, 1, 2, 1, 1),
                             covariateValue = c(2, 1, 1, 1, 1, 1, 1))

    andro <- andromeda(outcomes = outcomes, covariates = covariates)

    dataPtr <- convertToCyclopsData(andro$outcomes, andro$covariates, modelType = "fgr")
    sparseFit <- fitCyclopsModel(dataPtr)

    tolerance <- 1E-8
    expect_equivalent(coef(denseFit), coef(sparseFit), tolerance = tolerance)

    woWeights <- data.frame(rowId = 1:7, time = test$length, y = test$event)

    woAndro <- andromeda(outcomes = woWeights, covariates = covariates)

    woPtr <- convertToCyclopsData(woAndro$outcomes, woAndro$covariates, modelType = "fgr")
    woFit <- fitCyclopsModel(woPtr)
    expect_equal(coef(woFit), coef(sparseFit))
})


test_that("Check very small Fine-Gray example with time ties, but no failure ties (sparse vs. dense)", {
    test <- read.table(header=T, sep = ",", text = "
                       start, length, event, x1, x2
                       0, 4,1,0,0
                       0, 3,2,2,0
                       0, 3,2,0,1
                       0, 3,1,0,1
                       0, 2,0,1,0
                       0, 2,0,1,0
                       0, 2,1,1,1")

    fgDat <- Cyclops:::getFineGrayWeights(test$length, test$event)
    dataPtr <- Cyclops:::createCyclopsData(fgDat$surv ~ test$x1 + test$x2, modelType = "fgr", censorWeights = fgDat$weights)
    denseFit <- Cyclops:::fitCyclopsModel(dataPtr)

    outcomes <- data.frame(rowId = 1:7, time = test$length, y = test$event, censorWeights = fgDat$weights)
    covariates <- data.frame(rowId = c(2, 3, 4, 5, 6, 7, 7),
                             covariateId = c(1, 2, 2, 1, 1, 1, 2),
                             covariateValue = c(2, 1, 1, 1, 1, 1, 1))

    dataPtr <- convertToCyclopsData(outcomes, covariates, modelType = "fgr")
    sparseFit <- fitCyclopsModel(dataPtr)

    tolerance <- 1E-8
    expect_equivalent(coef(denseFit), coef(sparseFit), tolerance = tolerance)
})

test_that("Check very small Fine-Gray example with time ties and failure ties (sparse vs. dense)", {
    test <- read.table(header=T, sep = ",", text = "
                       start, length, event, x1, x2
                       0, 4,  1,0,0
                       0, 3, 1,2,0
                       0, 3,  1,0,1
                       0, 3, 2,0,1
                       0, 2,  0,1,1
                       0, 1, 2,1,0
                       0, 1, 0,1,1,
                       0, 1, 1,1,0,
                       0, 1, 1,1,1")

    fgDat <- Cyclops:::getFineGrayWeights(test$length, test$event)
    dataPtr <- Cyclops:::createCyclopsData(fgDat$surv ~ test$x1 + test$x2, modelType = "fgr", censorWeights = fgDat$weights)
    denseFit <- Cyclops:::fitCyclopsModel(dataPtr)

    outcomes <- data.frame(rowId = 1:9, time = test$length, y = test$event, censorWeights = fgDat$weights)
    covariates <- data.frame(rowId = c(2, 3, 4, 5, 5, 6, 7, 7, 8, 9, 9),
                             covariateId = c(1, 2, 2, 1, 2, 1, 1, 2, 1, 1, 2),
                             covariateValue = c(2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1))

    dataPtr <- convertToCyclopsData(outcomes, covariates, modelType = "fgr")
    sparseFit <- fitCyclopsModel(dataPtr)

    tolerance <- 1E-8
    expect_equivalent(coef(denseFit), coef(sparseFit), tolerance = tolerance)
})

test_that("Check very small Fine-Gray versus Cox for 0/1 events (censorWeights undefined for Cox model).", {
    test <- read.table(header=T, sep = ",", text = "
                       start, length, event, x1, x2
                       0, 4,  1,0,0
                       0, 3.5,0,2,0
                       0, 3,  0,0,1
                       0, 2.5,0,0,1
                       0, 2,  1,1,1
                       0, 1.5,0,1,0
                       0, 1,  1,1,0")

    fgDat <- Cyclops:::getFineGrayWeights(test$length, test$event)
    dataPtr <- Cyclops:::createCyclopsData(fgDat$surv ~ test$x1 + test$x2, modelType = "fgr", censorWeights = fgDat$weights)
    cyclopsFit1 <- Cyclops:::fitCyclopsModel(dataPtr)
    dataPtr <- Cyclops:::createCyclopsData(fgDat$surv ~ test$x1 + test$x2, modelType = "cox")
    cyclopsFit2 <- Cyclops:::fitCyclopsModel(dataPtr)

    tolerance <- 1E-4
    expect_equivalent(coef(cyclopsFit1), coef(cyclopsFit2), tolerance = tolerance)
})

test_that("Check very small Fine-Gray versus Cox for 0/1 events (censorWeights defined, but not used, for Cox model).", {
    test <- read.table(header=T, sep = ",", text = "
                       start, length, event, x1, x2
                       0, 4,  1,0,0
                       0, 3.5,0,2,0
                       0, 3,  0,0,1
                       0, 2.5,0,0,1
                       0, 2,  1,1,1
                       0, 1.5,0,1,0
                       0, 1,  1,1,0")

    fgDat <- Cyclops:::getFineGrayWeights(test$length, test$event)
    dataPtr <- Cyclops:::createCyclopsData(fgDat$surv ~ test$x1 + test$x2, modelType = "fgr", censorWeights = fgDat$weights)
    cyclopsFit1 <- Cyclops:::fitCyclopsModel(dataPtr)
    dataPtr <- Cyclops:::createCyclopsData(fgDat$surv ~ test$x1 + test$x2, modelType = "cox", censorWeights = fgDat$weights)
    cyclopsFit2 <- Cyclops:::fitCyclopsModel(dataPtr, warnings = FALSE)

    tolerance <- 1E-4
    expect_equivalent(coef(cyclopsFit1), coef(cyclopsFit2), tolerance = tolerance)
})

test_that("Check very small Cox example with weights and censor weights defined (the latter is unused for Cox models).", {
    test <- read.table(header=T, sep = ",", text = "
                       start, length, event, x1, x2
                       0, 4,  1,0,0
                       0, 3.5,1,2,0
                       0, 3,  0,0,1
                       0, 2.5,1,0,1
                       0, 2,  1,1,1
                       0, 1.5,0,1,0
                       0, 1,  1,1,0")

    fgDat <- Cyclops:::getFineGrayWeights(test$length, test$event)
    dataPtr <- Cyclops:::createCyclopsData(fgDat$surv ~ test$x1 + test$x2, modelType = "cox",
                                           weights = c(1, 1, 1, 0, 0, 1, 1), censorWeights = fgDat$weights)
    cyclopsFit1 <- Cyclops:::fitCyclopsModel(dataPtr, warnings = FALSE)
    dataPtr <- Cyclops:::createCyclopsData(fgDat$surv ~ test$x1 + test$x2, modelType = "cox",
                                           weights = c(1, 1, 1, 0, 0, 1, 1))
    cyclopsFit2 <- Cyclops:::fitCyclopsModel(dataPtr)

    tolerance <- 1E-4
    expect_equivalent(coef(cyclopsFit1), coef(cyclopsFit2), tolerance = tolerance)
})

test_that("Check very small Fine-Gray example with weights and censor weights defined.", {
    test <- read.table(header=T, sep = ",", text = "
                       start, length, event, x1, x2
                       0, 6,  1,0,1
                       0, 5,  0,1,1
                       0, 4,  1,0,0
                       0, 3.5,2,2,0
                       0, 3,  0,0,1
                       0, 2.5,2,0,1
                       0, 2,  1,1,1
                       0, 2,  1,2,0
                       0, 1.5,0,1,0
                       0, 1,  1,1,0")
    weights <- c(0,0,1,1,1,1,1,0,1,1)
    sub <- c(FALSE, FALSE, TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, TRUE, TRUE)

    fgDat <- Cyclops:::getFineGrayWeights(test$length, test$event, weights)
    dataPtr <- Cyclops:::createCyclopsData(fgDat$surv ~ test$x1 + test$x2, modelType = "fgr", censorWeights = fgDat$weights)
    dataFit <- Cyclops:::fitCyclopsModel(dataPtr, weights = weights)

    goldFit <- crr(test$length, test$event, cbind(test$x1, test$x2), subset = sub, variance = FALSE)

    tolerance <- 1E-6
    expect_equivalent(coef(dataFit), goldFit$coef, tolerance = tolerance)


    test_sub <- read.table(header=T, sep = ",", text = "
                       start, length, event, x1, x2
#                        0, 6,  1,0,1
#                        0, 5,  0,1,1
                        0, 4,  1,0,0
                        0, 3.5,2,2,0
                        0, 3,  0,0,1
                        0, 2.5,2,0,1
                        0, 2,  1,1,1
#                        0, 2,  1,2,0
                        0, 1.5,0,1,0
                        0, 1,  1,1,0")
    fgDat_sub <- Cyclops:::getFineGrayWeights(test_sub$length, test_sub$event)

    dataPtr_sub <- Cyclops:::createCyclopsData(fgDat_sub$surv ~ test_sub$x1 + test_sub$x2, modelType = "fgr", censorWeights = fgDat_sub$weights)
    dataFit_sub <- Cyclops:::fitCyclopsModel(dataPtr_sub)

    goldFit_sub <- crr(test_sub$length, test_sub$event, cbind(test_sub$x1, test_sub$x2), variance = FALSE)

    expect_equivalent(coef(dataFit_sub), goldFit_sub$coef, tolerance = tolerance)
    expect_equivalent(coef(dataFit_sub), coef(dataFit), tolerance = tolerance)
})

# @file ExampleData.R
#
# Copyright 2014 Observational Health Data Sciences and Informatics
#
# This file is part of Cyclops
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#' Oxford self-controlled case series data
#'
#' A dataset containing the MMR vaccination / meningitis in Oxford example from
#' Farrington and Whitaker.  There are 10 patients comprising 38 unique exposure intervals.
#'
#' @name oxford
#'
#' @docType data
#'
#' @usage data(oxford)
#'
#' @format A data frame with 38 rows and 6 variables:
#' \describe{
#'   \item{indiv}{patient identifier}
#'   \item{event}{number of events in interval}
#'   \item{interval}{interval length in days}
#'   \item{agegr}{age group}
#'   \item{exgr}{exposure group}
#'   \item{loginterval}{log interval length}
#'   ...
#' }
NULL

# @file DataManagement.R
#
# Copyright 2014 Observational Health Data Sciences and Informatics
#
# This file is part of cyclops
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#' @title Create a Cyclops data object
#'
#' @description
#' \code{createCyclopsData} creates a Cyclops data object from an R formula or data matrices.
#'
#' @details
#' This function creates a Cyclops model data object from R \code{"\link{formula}"} or directly from
#' numeric vectors and matrices to define the model response and covariates.
#' If specifying a model using a \code{"\link{formula}"}, then the left-hand side define the model response and the
#' right-hand side defines dense covariate terms.
#' Objects provided with \code{"sparseFormula"} and \code{"indicatorFormula"} must be include left-hand side responses and terms are
#' coersed into sparse and indicator representations for computational efficiency.
#'
#' Items to discuss:
#' * Only use formula or (y,dx,...)
#' * stratum() in formula
#' * offset() in formula
#' * when \code{"stratum"} (renamed from pid) are necessary
#' * when \code{"time"} are necessary
#'
#' @template types
#'
#' @param formula
#' An object of class \code{"\link{formula}"} that provides a symbolic description of the numerically dense model response and terms.
#' @param sparseFormula
#' An object of class \code{"\link{formula}"} that provides a symbolic description of numerically sparse model terms.
#' @param indicatorFormula
#' An object of class \code{"\link{formula}"} that provides a symbolic description of \{0,1\} model terms.
#' @param data
#' An optional data frame, list or environment containing the variables in the model.
#' @param subset
#' Currently unused
#' @param weights
#' Currently unused
#' @param censorWeights
#' Vector of subject-specific censoring weights (between 0 and 1). Currently only supported in \code{modelType = "fgr"}.
#' @param offset
#' Currently unused
#' @param pid
#' Optional vector of integer stratum identifiers. If supplied, all rows must be sorted by increasing identifiers
#' @param y
#' Currently undocumented
#' @param type
#' Currently undocumented
#' @param time
#' Currently undocumented
#' @param dx
#' Optional dense \code{"\link[Matrix]{Matrix}"} of covariates
#' @param sx
#' Optional sparse \code{"\link[Matrix]{Matrix}"} of covariates
#' @param ix
#' Optional \{0,1\} \code{"\link[Matrix]{Matrix}"} of covariates
#' @param model
#' Currently undocumented
#' @param method
#' Currently undocumented
#' @param normalize
#' String: Name of normalization for all non-indicator covariates (possible values: stdev, max, median)
#' @param floatingPoint
#' Integer: Floating-point representation size (32 or 64)
#'
#' @return
#' A list that contains a Cyclops model data object pointer and an operation duration
#'
#' @examples
#' ## Dobson (1990) Page 93: Randomized Controlled Trial :
#' counts <- c(18, 17, 15, 20, 10, 20, 25, 13, 12)
#' outcome <- gl(3, 1, 9)
#' treatment <- gl(3, 3)
#' cyclopsData <- createCyclopsData(
#'      counts ~ outcome + treatment,
#'      modelType = "pr")
#' cyclopsFit <- fitCyclopsModel(cyclopsData)
#'
#' cyclopsData2 <- createCyclopsData(
#'      counts ~ outcome,
#'      indicatorFormula = ~ treatment,
#'      modelType = "pr")
#' summary(cyclopsData2)
#' cyclopsFit2 <- fitCyclopsModel(cyclopsData2)
#'
#' @export
createCyclopsData <- function(formula, sparseFormula, indicatorFormula, modelType,
                              data, subset = NULL, weights = NULL, censorWeights = NULL, offset = NULL, time = NULL, pid = NULL, y = NULL, type = NULL, dx = NULL,
                              sx = NULL, ix = NULL, model = FALSE, normalize = NULL,
                              floatingPoint = 64,
                              method = "cyclops.fit") {
    cl <- match.call() # save to return
    mf.all <- match.call(expand.dots = FALSE)

    if (!.isValidModelType(modelType)) stop("Invalid model type.")

    .checkFloatingPoint(floatingPoint)

    hasIntercept <- FALSE
    colnames <- NULL

    contrasts <- NULL
    sortOrder <- NULL

    if (!missing(formula)) { # Use formula to construct Cyclops matrices
        if (missing(data)) {
            data <- environment(formula)
        }
        mf.all <- match.call(expand.dots = FALSE)
        m.d <- match(c("formula", "data", "subset", "weights", "censorWeights",
                       "offset"), names(mf.all), 0L)
        mf.d <- mf.all[c(1L, m.d)]
        mf.d$drop.unused.levels <- TRUE
        mf.d[[1L]] <- quote(stats::model.frame)
        mf.d <- eval(mf.d, parent.frame())

        outcome <- model.response(mf.d)
        if (inherits(outcome, "Surv")) {
            if (!.isSurvivalModelType(modelType)) {
                stop("Censored outcomes are currently only supported for Cox regression.")
            }
            if (dim(outcome)[2] == 3) {
                time <- as.numeric(outcome[,2] - outcome[,1])
                y <- as.numeric(outcome[,3])
            } else {
                time <- as.numeric(outcome[,1])
                y <- as.numeric(outcome[,2])
            }
        } else if (inherits(outcome, "Multitype")) {
            contrasts <- attr(outcome, "contrasts")
            type <- outcome[,2]
            y <- outcome[,1]
        } else {
            y <- outcome
        }

        mt.d <- attr(mf.d, "terms")

        # Handle strata
        specialTerms <- terms(formula, "strata")
        special <- attr(specialTerms, "special")
        hasStrata <- !is.null(special$strata)
        strata <- NULL

        if (hasStrata) {
            pid <- as.numeric(strata(mf.d[ , special$strata], shortlabel = TRUE))
            nterm <- survival::untangle.specials(specialTerms, "strata")$terms
            mt.d <- mt.d[-nterm]

            ## Must sort outcomes
            if (.isSurvivalModelType(modelType)) {
                sortOrder <- order(pid, -time, y)
            } else {
                sortOrder <- order(pid)
            }
        } else {
            if (.isSurvivalModelType(modelType)) {
                sortOrder <- order(-time, y)
                if (missing(pid)) {
                    pid <- rep(1, length(y))
                }
            } else {
                if (missing(pid)) {
                    pid <- c(1:length(y))
                }
            }
        }

        #         if (.removeIntercept(modelType)) { # This does not work with constrasts
        #             attr(mt.d, "intercept") <- FALSE
        #         }

        dtmp <- model.matrix(mt.d, mf.d)
        dlabels <- labels(dtmp)[[2]]
        if (attr(mt.d, "intercept") && .removeIntercept(modelType)) {
            dx <- Matrix(as.matrix(dtmp[,-1]), sparse = FALSE)
            dlabels <- dlabels[-1]
        } else {
            dx <- Matrix(dtmp, sparse = FALSE)
        }

        off.d <- model.offset(mf.d)
        if (!is.null(time) && !is.null(off.d)) {
            stop("Supplied both 'time' and 'offset' quantities")
        }
        if (!is.null(off.d)) {
            time <- as.vector(off.d)
        }

        #colnames <- c(colnames, dx@Dimnames[[2]])
        colnames <- c(colnames, dlabels)

        if (attr(mt.d, "intercept") && !.removeIntercept(modelType)) { # Has intercept
            hasIntercept <- TRUE
        }

        if (!missing(sparseFormula)) {
            if (missing(data)) {
                data <- environment(sparseFormula)
            }
            m.s <- match(c("sparseFormula", "data", "subset", "weights", "censorWeights",
                           "offset"), names(mf.all), 0L)
            mf.s <- mf.all[c(1L, m.s)]
            mf.s$drop.unused.levels <- TRUE
            mf.s[[1L]] <- quote(stats::model.frame)
            names(mf.s)[2] = "formula"
            mf.s <- eval(mf.s, parent.frame())

            if (!is.null(model.response(mf.s))) {
                stop("Must only provide outcome variable in dense formula.")
            }
            mt.s <- attr(mf.s, "terms")
            stmp <- model.matrix(mt.s, mf.s)
            #stmp <- model.matrix(mf.s, data)
            slabels <- labels(stmp)[[2]]
            if (attr(attr(mf.s, "terms"), "intercept") == 1) { # Remove intercept
                sx <- Matrix(as.matrix(stmp[,-1]), sparse=TRUE)
                slabels <- slabels[-1]
            } else {
                sx <- Matrix(as.matrix(stmp), sparse=TRUE)
            }
            colnames <- c(colnames, slabels)
        }

        if (!missing(indicatorFormula)) {
            if (missing(data)) {
                data <- environment(indicatorFormula)
            }
            m.i <- match(c("indicatorFormula", "data", "subset", "weights", "censorWeights",
                           "offset"), names(mf.all), 0L)
            mf.i <- mf.all[c(1L, m.i)]
            mf.i$drop.unused.levels <- TRUE
            mf.i[[1L]] <- quote(stats::model.frame)
            names(mf.i)[2] = "formula"
            mf.i <- eval(mf.i, parent.frame())

            if (!is.null(model.response(mf.i))) {
                stop("Must only provide outcome variable in dense formula.")
            }

            # TODO Check that all values in mf.i are 0/1
            mt.i <- attr(mf.i, "terms")
            itmp <- model.matrix(mt.i, mf.i)
            #itmp <- model.matrix(mf.i, data)
            ilabels <- labels(itmp)[[2]]
            if (attr(attr(mf.i, "terms"), "intercept") == 1) { # Remove intercept
                ix <- Matrix(as.matrix(itmp[,-1]), sparse=TRUE)
                ilabels <- ilabels[-1]
            } else {
                ix <- Matrix(as.matrix(itmp), sparse=TRUE)
            }
            colnames <- c(colnames, ilabels)
        }

        if (!is.null(sortOrder)) {
            pid <- pid[sortOrder]
            y <- y[sortOrder]
            if (!missing(type)) {
                type <- type[sortOrder]
            }
            if (!missing(weights)) {
                weights <- weights[sortOrder]
            }
            if (!missing(censorWeights)) {
                censorWeights <- censorWeights[sortOrder]
            }
            time <- time[sortOrder]
            dx <- dx[sortOrder, ]
            if (inherits(dx,"numeric")) {
                dx = as(dx,"dgeMatrix")
            }
            sx <- sx[sortOrder, ]
            if (inherits(sx, "numeric")) {
                sx = Matrix(sx, ncol = 1, sparse = TRUE)
            }
            ix <- ix[sortOrder, ]
            if (inherits(ix, "numeric")) {
                ix = Matrix(ix, ncol = 1, sparse = TRUE)
            }
        }

        if (identical(method, "model.frame")) {
            result <- list()
            if (exists("mf.d")) {
                result$dense <- mf.d
                result$dx <- dx
                result$y <- y
                result$time <- time
            }
            if (exists("mf.s")) {
                result$sparse <- mf.s
                result$sx <- sx
            }
            if (exists("mf.i")) {
                result$indicator <- mf.i
                result$ix <- ix
            }
            return(result)
        }
    } else  {
        if (!missing(sparseFormula) || !missing(indicatorFormula)) {
            stop("Must provide a dense formula when specifying sparse or indicator covariates.")
        }
    }

    # Handle multiple outcome types
    numTypes <- 1
    if (!missing(type)) {
        if (is.null(contrasts)) {
            contrasts <- contrasts(as.factor(type))
        }
        numTypes <- length(contrasts)
        type <- contrasts[type]
        if (!is.null(colnames)) {
            colnames <- as.vector(sapply(colnames, function(x, y) { paste(x,y,sep=":")},
                                         y = dimnames(contrasts)[[1]], USE.NAMES=F))
        }
    }

    # TODO Check types and dimensions

    useTimeAsOffset <- FALSE
    if (!is.null(time) && !.useOffsetModelType(modelType)) {
        useTimeAsOffset <- TRUE
    }

    if (is.null(pid)) {
        pid <- c(1:length(y)) # TODO Should not be necessary
    }

    md <- .cyclopsModelData(pid, y, type, time, dx, sx, ix, modelType, useTimeAsOffset, numTypes,
                            floatingPoint)
    result <- new.env(parent = emptyenv())
    result$cyclopsDataPtr <- md$data
    result$modelType <- modelType
    result$timeLoad <- md$timeLoad
    result$call <- cl
    #     if (exists("mf") && model == TRUE) {
    #         result$mf <- mf
    #     }
    result$cyclopsInterfacePtr <- NULL
    result$call <- cl
    result$coefficientNames <- colnames

    if (!is.null(dx)) {
        result$rowNames <- dx@Dimnames[[1]]
    } else if (!is.null(pid)) {
        result$rowNames <- pid
    } else {
        result$rowNames <- c(1:length(y))
    }

    result$sortOrder <- sortOrder
    result$weights <- weights
    result$censorWeights <- censorWeights

    if (identical(method, "debug")) {
        result$debug <- list()
        result$debug$dx <- dx
        result$debug$sx <- sx
        result$debug$ix <- ix
        result$debug$y <- y
        result$debug$pid <- pid
        result$debug$time <- time
    }
    class(result) <- "cyclopsData"

    if (hasIntercept == TRUE) {
        .cyclopsSetHasIntercept(result, hasIntercept = TRUE)
    }

    if (!is.null(normalize)) {
        .normalizeCovariates(result, normalize)
    }

    result
}

.checkFloatingPoint <- function(floatingPoint) {
      if (floatingPoint != 64 && floatingPoint != 32) {
          stop("Invalid floating point precision")
      }
}

.normalizeCovariates <- function(cyclopsData, type) {
    scale <- .cyclopsNormalizeCovariates(cyclopsData, type)

    if (is.null(cyclopsData$scale)) {
        cyclopsData$scale <- scale
    } else {
        cyclopsData$scale <- cyclopsData$scale * scale # Can call function multiple times without bad effects
    }

    cyclopsData$cyclopsInterfacePtr <- NULL # Force new engine for next fit

    invisible(scale)
}

# @title isValidModelType
#
# @description
# \code{isValidModelType} checks for a valid Cyclops model type
#
# @template types
#
# @return TRUE/FALSE
#
# @examples
# isValidModelType("pr")
# #TRUE
#
# isValidModelType("abc")
# #FALSE
#
# @export
.isValidModelType <- function(modelType) {
    types <- .cyclopsGetModelTypeNames()
    modelType %in% types
}

.removeIntercept <- function(modelType) {
    types <- .cyclopsGetRemoveInterceptNames()
    modelType %in% types
}

.isSurvivalModelType <- function(modelType) {
    types <- .cyclopsGetIsSurvivalNames()
    modelType %in% types
}

.useOffsetModelType <- function(modelType) {
    types <- .cyclopsGetUseOffsetNames()
    modelType %in% types
}

#' @title Read Cyclops data from file
#'
#' @description
#' \code{readCyclopsData} reads a Cyclops-formatted text file.
#'
#' @details
#' This function reads a Cyclops-formatted text file and returns a Cyclops data object. The first line of the
#' file may start with '\samp{#}', indicating that it contains header options.  Valid header options are:
#'
#'  \tabular{ll}{
#'  \verb{	row_label}		\tab (assume file contains a numeric column of unique row identifiers) \cr
#'  \verb{	stratum_label}\tab (assume file contains a numeric column of stratum identifiers) \cr
#'  \verb{	weight}				\tab (assume file contains a column of row-specific model weights, currently unused) \cr
#' 	\verb{	offset}				\tab (assume file contains a dense column of linear predictor offsets) \cr
#' 	\verb{	bbr_outcome}	\tab (assume logistic outcomes are encoded -1/+1 following BBR) \cr
#' 	\verb{	log_offset}		\tab (assume file contains a dense column of values x_i for which log(x_i) is the offset) \cr
#' 	\verb{	add_intercept}\tab (automatically include an intercept column of all 1s for each entry) \cr
#' 	\verb{	indicator_only}\tab(assume all covariates 0/1-valued and only covariate name is given) \cr
#' 	\verb{	sparse}				\tab (force all BBR formatted covariates to be represented as sparse, instead of \cr
#'                              \tab sparse-indicator, columns .. really only for debugging) \cr
#' 	\verb{	dense}				\tab (force all BBR formatted covariates to be represented as dense columns.. really\cr
#'                              \tab only for debugging) \cr
#' }
#'
#' Successive lines of the file are white-space delimited and follow the format:
#'
#' \preformatted{	[Row ID] {Stratum ID} [Weight] <Outcome> {Censored} {Offset} <BBR covariates>}
#'
#' \itemize{
#'   	\item \verb{[optional]}
#'   	\item \verb{<required>}
#'   	\item \verb{{required or optional depending on model}}
#'  }
#'
#' Bayesian binary regression (BBR) covariates are white-space delimited and generally in a sparse
#' \samp{<name>:<value>} format, where \samp{name} must (currently) be numeric and \samp{value} is non-zero.
#' If option \samp{indicator_only} is specified, then format is simply \samp{<name>}.
#' \samp{Row ID} and \samp{Stratum ID} must be numeric, and rows must be sorted such that equal \samp{Stratum ID}
#' are consecutive.
#' \samp{Stratum ID} is required for \samp{clr} and \samp{sccs} models.
#' \samp{Censored} is required for a \samp{cox} model.
#' \samp{Offset} is (currently) required for a \samp{sccs} model.
#'
#' @template types
#'
#' @param fileName          Name of text file to be read. If fileName does not contain an absolute path,
#'
#' @return
#' A list that contains a Cyclops model data object pointer and an operation duration
#'
#' @examples
#' \dontrun{
#' dataPtr = readCyclopsData(system.file("extdata/infert_ccd.txt", package="Cyclops"), "clr")
#' }
#' @export
readCyclopsData <- function(fileName, modelType) {
    cl <- match.call() # save to return

    if (!.isValidModelType(modelType)) stop("Invalid model type.")

    read <- .cyclopsReadData(fileName, modelType)
    result <- new.env(parent = emptyenv())
    result$cyclopsDataPtr <- read$cyclopsDataPtr
    result$modelType <- modelType
    result$timeLoad <- read$timeLoad
    result$cyclopsInterfacePtr <- NULL
    result$call <- cl

    class(result) <- "cyclopsData"
    result
}

#' @title Get univariable correlation
#'
#' @description \code{getUnivariableCorrelation} reports covariates that have high correlation with the outcome
#'
#' @param cyclopsData    A Cyclops data object
#' @param covariates Integer or string vector: list of covariates to report; default (NULL) implies all covariates
#' @param threshold Correlation threshold for reporting
#'
#' @return A list of covariates whose absolute correlation with the outcome is greater than or equal to the threshold
#'
#' @export
getUnivariableCorrelation <- function(cyclopsData, covariates = NULL, threshold = 0.0) {
    # Check for valid arguments
    .checkData(cyclopsData)
    if (.isSurvivalModelType(cyclopsData$modelType)) {
        stop("Univariable correlation for time-to-events model is not yet implemented.")
    }

    labels <- covariates
    covariates <- .checkCovariates(cyclopsData, covariates)
    if (is.null(covariates)) {
        covariates <- bit64::as.integer64(c()) # zero-length vector
        labels <- cyclopsData$coefficientNames
    }

    correlations <- .cyclopsUnivariableCorrelation(cyclopsData, covariates)
    if (threshold > 0.0) {
        select <- abs(correlations) >= threshold
        correlations <- correlations[select]
        labels <- labels[select]
    }

    names(correlations) <- labels
    return(correlations)
}

#' @title Get univariable linear separability
#'
#' @description \code{getUnivariableSeparability} reports covariates that are univariably separable with the outcome
#'
#' @param cyclopsData    A Cyclops data object
#' @param covariates Integer or string vector: list of covariates to report; default (NULL) implies all covariates
#'
#' @return A list of covariates that are univariably separable with the outcome
#'
#' @export
getUnivariableSeparability <- function(cyclopsData, covariates = NULL) {
    # Check for valid arguments
    .checkData(cyclopsData)

    labels <- covariates
    covariates <- .checkCovariates(cyclopsData, covariates)
    ids <- covariates
    if (is.null(covariates)) {
        covariates <- bit64::as.integer64(c()) # zero-length vector
        labels <- cyclopsData$coefficientNames
        ids <- getCovariateIds(cyclopsData)
    }

    types <- getCovariateTypes(cyclopsData, ids)
    if (any(types == "sparse") || any(types == "dense")) {
        warning("Sparse and dense types are not yet implemented")
    }

    separability <- .cyclopsUnivariableSeparability(cyclopsData, covariates)

    names(separability) <- labels
    return(separability == 1)
}

#' @title Apply simple data reductions
#'
#' @description \code{reduce} reports the count of non-zero elements, sum and sum-of-squares for specified covariates in a Cyclops data object.
#'
#' @param object    A Cyclops data object
#' @param covariates Integer or string vector: list of covariates to report
#' @param groupBy   Integer or string (optional): generates a segmented reduction stratified by this covariate.  Setting \code{groupBy = "stratum"} segments reduction for strataID
#' @param power Integer: 0 = non-zero count, 1 = sum, 2 = sum-of-squares
#'
#' @return Specified reduction as number or \code{data.frame} if segmented.
#'
#' @keywords internal
reduce <- function(object, covariates, groupBy, power = 1) {
    if (!isInitialized(object)) {
        stop("Object is no longer or improperly initialized.")
    }
    covariates <- .checkCovariates(object, covariates)

    if (!(power %in% c(0,1,2))) {
        stop("Only powers 0, 1 and 2 are allowed.")
    }

    if (missing(groupBy)) {
        .cyclopsSum(object, covariates, power)
    } else {
        if (length(groupBy) != 1L) {
            stop("Only single stratification is currently implemented")
        }
        if (groupBy == "stratum") {
            as.data.frame(.cyclopsSumByStratum(object, covariates, power),
                          row.names = c(1L:getNumberOfStrata(object)))
        } else {
            groupBy <- .checkCovariates(object, groupBy)
            as.data.frame(.cyclopsSumByGroup(object, covariates, groupBy, power),
                          row.names = c(0L,1L))
        }
    }
}

#' @title appendSqlCyclopsData
#'
#' @description
#' \code{appendSqlCyclopsData} appends data to an OHDSI data object.
#'
#' @details Append data using two tables.  The outcomes table is dense and contains ...  The covariates table is sparse and contains ...
#' All entries in the outcome table must be sorted in increasing order by (oStratumId, oRowId).  All entries in the covariate table
#' must be sorted in increasing order by (cRowId). Each cRowId value must match exactly one oRowId value.
#'
#' @param object    OHDSI Cyclops data object to append entries
#' @param oStratumId    Integer vector (optional): non-unique stratum identifier for each row in outcomes table
#' @param oRowId        Integer vector: unique row identifier for each row in outcomes table
#' @param oY            Numeric vector: model outcome variable for each row in outcomes table
#' @param oTime         Numeric vector (optional): exposure interval or censoring time for each row in outcomes table
#' @param cRowId        Integer vector: non-unique row identifier for each row in covariates table that matches a single outcomes table entry
#' @param cCovariateId  Integer vector: covariate identifier
#' @param cCovariateValue   Numeric vector: covariate value
#'
#' @keywords internal
appendSqlCyclopsData <- function(object,
                                 oStratumId,
                                 oRowId,
                                 oY,
                                 oTime,
                                 cRowId,
                                 cCovariateId,
                                 cCovariateValue) {
    if (!isInitialized(object)) {
        stop("Object is no longer or improperly initialized.")
    }

    if (is.unsorted(oStratumId)
        #|| is.unsorted(oRowId) || is.unsorted(cRowId)
    ) {
        stop("All columns must be sorted first by stratumId (if supplied) and then by rowId")
    }

    .appendSqlCyclopsData(object,
                          oStratumId,
                          oRowId,
                          oY,
                          oTime,
                          cRowId,
                          cCovariateId,
                          cCovariateValue)
}

#' @importFrom bit64 as.integer64 is.integer64
#' @keywords internal
loadNewSqlCyclopsDataMultipleX <- function(object,
                                           covariateId, # Vector
                                           rowId, # Vector
                                           covariateValue = NULL, # Vector
                                           name,
                                           append = FALSE,
                                           checkSorting = FALSE,
                                           checkCovariateIds = FALSE,
                                           checkCovariateBounds = FALSE,
                                           forceSparse = FALSE) {

    if (!isInitialized(object)) stop("Object is no longer or improperly initialized.")

    if (length(covariateId) != length(rowId)) stop("Vector length mismatch")

    if (is.null(covariateValue)) {
        if (forceSparse) stop("Must provide covariate values when forcing sparse")
        covariateValue <- as.numeric(c())
    } else {
        if (length(covariateId) != length(covariateValue)) stop ("Vector length mismatch")
    }

    if (checkSorting) {
        # TODO: Check sorted by (1) covariateId, (2) rowId
        stop("Not yet implemented")
    }

    if (!bit64::is.integer64(covariateId)) {
      covariateId <- bit64::as.integer64(covariateId)
    }

    if (!bit64::is.integer64(rowId)) {
      rowId <- bit64::as.integer64(rowId)
    }

    index <- .loadCyclopsDataMultipleX(object,
                                       covariateId,
                                       rowId,
                                       covariateValue,
                                       checkCovariateIds,
                                       checkCovariateBounds,
                                       append,
                                       forceSparse)

    if (!missing(name)) {
        if (is.null(object$coefficientNames)) {
            object$coefficientNames <- as.character(c())
        }
        start <- index + 1
        end <- index + length(name)
        object$coefficientNames[start:end] <- as.character(name)
    }
}

#' @keywords internal
loadNewSqlCyclopsDataX <- function(object,
                                   covariateId, # Scalar
                                   rowId = NULL, # Vector
                                   covariateValue = NULL, # Vector
                                   name,
                                   replace = FALSE,
                                   append = FALSE,
                                   forceSparse = FALSE) {
    if (!isInitialized(object)) {
        stop("Object is no longer or improperly initialized.")
    }

    if (replace == TRUE) {
        stop("Replacing an X column is not currently supported")
    }

    if (is.null(rowId)) rowId <- bit64::as.integer64(c())
    if (is.null(covariateValue)) {
        if (forceSparse) stop("Must provide covariate values when forcing sparse")
        covariateValue <- as.numeric(c())
    }

    if (!bit64::is.integer64(covariateId)) {
      covariateId <- bit64::as.integer64(covariateId)
    }

    if (!bit64::is.integer64(rowId)) {
      rowId <- bit64::as.integer64(rowId)
    }

    # throws error if covariateId already exists and append == FALSE
    index <- .loadCyclopsDataX(object,
                               covariateId,
                               rowId,
                               covariateValue,
                               replace,
                               append,
                               forceSparse)

    if (!missing(name)) {
        if(is.null(object$coefficientNames)) {
            object$coefficientNames <- as.character(c())
        }
        object$coefficientNames[index + 1] <- name
    }
}

#' @keywords internal
loadNewSqlCyclopsDataY <- function(object,
                                   stratumId = NULL,
                                   rowId = NULL,
                                   y,
                                   time = NULL) {
    if (!isInitialized(object)) {
        stop("Object is no longer or improperly initialized.")
    }

    if (is.null(stratumId)) stratumId <- bit64::as.integer64(c())
    if (is.null(rowId)) rowId <- bit64::as.integer64(c())

    if (is.unsorted(stratumId)) {
        stop("All columns must be sorted first by stratumId (if supplied) and then by rowId")
    }

    if (is.null(time)) {
        if (.isSurvivalModelType(object$modelType)) stop("Must provide time for survival model")
        time <- as.numeric(c())
    }

    if (!bit64::is.integer64(stratumId)) {
      stratumId <- bit64::as.integer64(stratumId)
    }

    if (!bit64::is.integer64(rowId)) {
      rowId <- bit64::as.integer64(rowId)
    }

    .loadCyclopsDataY(object,
                      stratumId,
                      rowId,
                      y,
                      time)
}

#' @keywords internal
loadNewSqlCyclopsDataStratTimeEffects <- function(object,
                                                  stratumId = NULL,
                                                  rowId = NULL,
                                                  subjectId = NULL,
                                                  timeEffectCovariateId) {

    if (!isInitialized(object)) stop("Object is no longer or improperly initialized.")

    if (is.unsorted(stratumId)) {
        stop("All columns must be sorted first by stratumId (if supplied) and then by rowId")
    }

    if (!bit64::is.integer64(stratumId)) {
        stratumId <- bit64::as.integer64(stratumId)
    }

    if (!bit64::is.integer64(rowId)) {
        rowId <- bit64::as.integer64(rowId)
    }

    if (!bit64::is.integer64(subjectId)) {
        subjectId <- bit64::as.integer64(subjectId)
    }

    if (!bit64::is.integer64(timeEffectCovariateId)) {
        timeEffectCovariateId <- bit64::as.integer64(timeEffectCovariateId)
    }

    timeEffectCovariatesName <- .loadCyclopsDataStratTimeEffects(object,
                                                                 stratumId,
                                                                 rowId,
                                                                 subjectId,
                                                                 timeEffectCovariateId)
    if (!is.null(object$coefficientNames)) {
        object$coefficientNames <- append(object$coefficientNames, timeEffectCovariatesName)
    }
}

#' @title finalizeSqlCyclopsData
#'
#' @description
#' \code{finalizeSqlCyclopsData} finalizes a Cyclops data object
#'
#' @param object							Cyclops data object
#' @param addIntercept				Add an intercept covariate if one was not imported through SQL
#' @param useOffsetCovariate	Specify is a covariate should be used as an offset (fixed coefficient = 1).
#' 														Set option to \code{-1} to specify the time-to-event column,
#' 														otherwise include a single numeric or character covariate name.
#' @param offsetAlreadyOnLogScale						Set to \code{TRUE} to indicate that offsets were log-transformed before importing into Cyclops data object.
#' @param sortCovariates			Sort covariates in numeric-order with intercept first if it exists.
#' @param makeCovariatesDense List of numeric or character covariates names to densely represent in Cyclops data object.
#' 														For efficiency, we suggest making at least the intercept dense.
##' @keywords internal
#' @export
finalizeSqlCyclopsData <- function(object,
                                   addIntercept = FALSE,
                                   useOffsetCovariate = NULL,
                                   offsetAlreadyOnLogScale = FALSE,
                                   sortCovariates = FALSE,
                                   makeCovariatesDense = NULL) {
    if (!isInitialized(object)) {
        stop("Object is no longer or improperly initialized.")
    }

    savedUseOffsetCovariate <- useOffsetCovariate
    useOffsetCovariate <- .checkCovariates(object, useOffsetCovariate)
    if (length(useOffsetCovariate) > 1) {
        stop("Can only supply one offset")
    }

    makeCovariatesDense <- .checkCovariates(object, makeCovariatesDense)

    .cyclopsFinalizeData(object, addIntercept, useOffsetCovariate,
                         offsetAlreadyOnLogScale, sortCovariates,
                         makeCovariatesDense)

    if (addIntercept == TRUE) {
        if (!is.null(object$coefficientNames)) {
            object$coefficientNames = c("(Intercept)",
                                        object$coefficientNames)
        }
    }
    if (!is.null(useOffsetCovariate) && useOffsetCovariate != -1) {
        if (!is.null(object$coefficientNames)) {
            object$coefficientNames = object$coefficientNames[-as.integer(useOffsetCovariate)]
        }
    }
}

#' @keywords internal
createSqlCyclopsData <- function(modelType, control,
                                 floatingPoint = 64) {
    cl <- match.call() # save to return

    if (!.isValidModelType(modelType)) stop("Invalid model type.")

    noiseLevel <- "silent"
    if (!missing(control)) { # Set up control
        stopifnot(inherits(control, "cyclopsControl"))
        noiseLevel <- control$noiseLevel
    }

    sql <- .cyclopsNewSqlData(modelType, noiseLevel, floatingPoint)
    result <- new.env(parent = emptyenv()) # TODO Remove code duplication with two functions above
    result$cyclopsDataPtr <- sql$cyclopsDataPtr
    result$modelType <- modelType
    result$timeLoad <- 0
    result$cyclopsInterfacePtr <- NULL
    result$call <- cl
    class(result) <- "cyclopsData"
    result
}

#' @title Check if a Cyclops data object is initialized
#'
#' @description
#' \code{isInitialized} determines if an Cyclops data object is properly
#' initialized and remains in memory.  Cyclops data objects do not
#' serialized/deserialize their back-end memory across R sessions.
#'
#' @param object    Cyclops data object to test
#'
#' @export
isInitialized <- function(object) {
    return(!is.null(object$cyclopsDataPtr) && !.isRcppPtrNull(object$cyclopsDataPtr))
}


#' @title Get covariate types
#'
#' @description
#' \code{getCovariateTypes} returns a vector covariate types in a Cyclops data object
#'
#' @param object    A Cyclops data object
#' @param covariateLabel Integer vector: covariate identifiers to return
#'
#' @export
getCovariateTypes <- function(object, covariateLabel) {
  if (!isInitialized(object)) {
    stop("Cyclops data object is no longer or improperly initialized")
  }

  if (!bit64::is.integer64(covariateLabel)) {
    covariateLabel <- bit64::as.integer64(covariateLabel)
  }

  return(.getCovariateTypes(object, covariateLabel))
}

#' @title Cyclops data object summary
#'
#' @method summary cyclopsData
#'
#' @description \code{summary.cyclopsData} summarizes the data held in an Cyclops data object.
#'
#' @param object    A Cyclops data object
#' @param ...       Additional arguments
#'
#' @return
#' Returns a \code{data.frame} that reports simply summarize statistics for each covariate in a Cyclops data object.
#'
#' @export
summary.cyclopsData <- function(object, ...) {
    if (!isInitialized(object)) {
        stop("Cyclops data object is no longer or improperly initialized")
    }
    covariates <- getCovariateIds(object)
    counts <- reduce(object, covariates, power = 0)
    sums <- reduce(object, covariates, power = 1)
    sumsSquared <- reduce(object, covariates, power = 2)
    types <- getCovariateTypes(object, covariates)

    tmean <- sums / counts;
    tvar <- (sumsSquared -  counts * tmean * tmean) / counts
    tvar[tvar < 0] <- 0

    if (is.null(object$scale)) {
        tdf <- data.frame(covariateId = covariates,
                          nzCount = counts,
                          nzMean = tmean,
                          nzVar = tvar,
                          type = types)
    } else {
        tdf <- data.frame(covariateId = covariates,
                          nzCount = counts,
                          nzMean = tmean,
                          nzVar = tvar,
                          type = types,
                          scale = object$scale)
    }

    if (!is.null(object$coefficientNames)) {
        row.names(tdf) <- object$coefficientNames
    }
    tdf
}


#' @method print cyclopsData
#' @title Print a Cyclops data object
#'
#' @description
#' \code{print.cyclopsData} displays information about a Cyclops data model object.
#'
#' @param x    A Cyclops data model object
#' @param show.call Logical: display last call to construct the Cyclops data model object
#' @param ...   Additional arguments
#'
#' @export
print.cyclopsData <- function(x, show.call=TRUE ,...) {
    cat("Cyclops Data Object\n\n")

    if (show.call && !is.null(x$call)) {
        cat("Call: ",paste(deparse(x$call),sep="\n",collapse="\n"),"\n\n",sep="")
    }
    cat("     Model: ", x$modelType, "\n", sep="")

    if (isInitialized(x)) {
        nRows <- getNumberOfRows(x)
        cat("      Rows: ", nRows, "\n", sep = "")
        cat("Covariates: ", getNumberOfCovariates(x), "\n", sep = "")
        nStrata <- getNumberOfStrata(x)
        if (nRows != nStrata) {
            cat("    Strata: ", nStrata, "\n", sep = "")
        }
        if (.cyclopsGetHasOffset(x)) {
            cat("    Offset: ", .cyclopsGetMeanOffset(x), " (mean)\n", sep = "")
        }
        cat("   FP size:", getFloatingPointSize(x))
    } else {
        cat("\nObject is no longer or improperly initialized.\n")
    }
    cat("\n")
    if (!is.null(x$cyclopsInterfacePtr) && !.isRcppPtrNull(x$cyclopsInterfacePtr)) {
        cat("Initialized interface (details coming soon).\n")
    } else {
        cat("Uninitialized interface.\n")
    }
    invisible(x)
}

library("testthat")

context("test-normalization.R")

test_that("Simple normalization", {
    counts <- c(18,17,15,20,10,20,25,13,12)
    outcome <- gl(3,1,9)
    treatment <- gl(3,3)
    tolerance <- 1E-4

    gold <- glm(counts ~ outcome + treatment, family = poisson())

    dataPtr <- createCyclopsData(counts ~ 1,
                                 sparseFormula = ~outcome,
                                 indicatorFormula = ~treatment,
                                 modelType = "pr")
    expect_null(dataPtr$scales)
    fit1 <- fitCyclopsModel(dataPtr, prior = createPrior("none"))

    # Normalize
    Cyclops:::.normalizeCovariates(dataPtr, type = "stdev")

    expect_error(confint(fit1, parm = "outcome2")) # Data are now invalid for fit1

    sum2 <- summary(dataPtr)
    expect_equal(sum(sum2$nzMean != 1.0), 2) # Only sparse covariates are normalized
    fit2 <- fitCyclopsModel(dataPtr, prior = createPrior("none"))

    # Normalizing again should not change values
    Cyclops:::.normalizeCovariates(dataPtr, type = "stdev")
    sum3 <- summary(dataPtr)
    expect_equal(sum2$scale, sum3$scale, tolerance = tolerance * tolerance)
    fit3 <- fitCyclopsModel(dataPtr, prior = createPrior("none"))

    # Check equality of first results at end
    expect_equal(coef(gold), coef(fit1), tolerance = tolerance)
    # And second after rescaling
    expect_equal(coef(gold), coef(fit2, rescale = TRUE), tolerance = tolerance)
    # And third after rescaling
    expect_equal(confint(gold, parm = "outcome2"),
                 confint(fit3, parm = "outcome2", rescale = TRUE)[,2:3], tolerance = sqrt(tolerance))

    # Repeat process with normalization during construction
    dataPtr2 <- createCyclopsData(counts ~ 1,
                                 sparseFormula = ~outcome,
                                 indicatorFormula = ~treatment,
                                 modelType = "pr", normalize = "stdev")

    fit4 <- fitCyclopsModel(dataPtr2, prior = createPrior("none"))
    expect_equal(coef(fit2), coef(fit4))

    # Normalize via max
    covariate <- c(1:length(counts))
    dataPtr3 <- createCyclopsData(counts ~ 1,
                                  sparseFormula = ~covariate,
                                  indicatorFormula = ~treatment,
                                  modelType = "pr", normalize = "max")
    sum4 <- summary(dataPtr3)
    expect_equal(sum4$scale[2], 1 / max(covariate))

    # Normalize via median
    dataPtr4 <- createCyclopsData(counts ~ 1,
                                  sparseFormula = ~covariate,
                                  indicatorFormula = ~treatment,
                                  modelType = "pr", normalize = "median")
    sum5 <- summary(dataPtr4)
    expect_equal(sum5$scale[2], 1 / median(covariate))
})

library("testthat")

context("test-sqlConstructor.R")

#
# Tests for the SQL constructor of ModelData
#

test_that("Test constructor and append", {
    counts <- c(18,17,15,20,10,20,25,13,12)
    outcome <- gl(3,1,9)
    treatment <- gl(3,3)

    oStratumId <- c(1:9)
    oRowId <- c(1:9)
    oY <- c(18,17,15,20,10,20,25,13,12)
    oTime <- rep(0,9)
    cRowId <- c(
        1,
        2,2,
        3,3,
        4,4,
        5,5,5,
        6,6,6,
        7,7,
        8,8,8,
        9,9,9)
    cCovariateId <- c(
        1,
        1,2,
        1,3,
        1,4,
        1,2,4,
        1,3,4,
        1,5,
        1,2,5,
        1,3,5)
    cCovariateValue <- rep(1, 21)

    dataPtr <- createSqlCyclopsData(modelType = "pr")

    # Should have no data
    expect_equal(getNumberOfRows(dataPtr), 0)
    expect_equal(getNumberOfStrata(dataPtr), 0)
    expect_equal(getNumberOfCovariates(dataPtr), 0)
    expect_error(fitCyclopsModel(dataPtr))

    count <- appendSqlCyclopsData(dataPtr,
                              oStratumId,
                              oRowId,
                              oY,
                              oTime,
                              cRowId,
                              cCovariateId,
                              cCovariateValue)
    finalizeSqlCyclopsData(dataPtr) # Not yet implemented

    expect_equal(count, 9)
    expect_equal(getNumberOfRows(dataPtr), 9)
    expect_equal(getNumberOfStrata(dataPtr), 9)
    expect_equal(getNumberOfCovariates(dataPtr), 5)

    cyclopsFit <- fitCyclopsModel(dataPtr, control = createControl(noiseLevel = "silent"),
                                  startingCoefficients = rep(0, 5))

    dataPtrF <- createCyclopsData(counts ~ outcome + treatment, modelType = "pr")
    cyclopsFitF <- fitCyclopsModel(dataPtrF, control = createControl(noiseLevel = "silent"),
                                   startingCoefficients = rep(0, 5))
    expect_equivalent(coef(cyclopsFit), coef(cyclopsFitF)) # Have different coefficient names

    # Test chucked append
    dataPtrC <- createSqlCyclopsData(modelType = "pr")
    count <- appendSqlCyclopsData(dataPtrC,
                              oStratumId[1:5],
                              oRowId[1:5],
                              oY[1:5],
                              oTime[1:5],
                              cRowId[1:10],
                              cCovariateId[1:10],
                              cCovariateValue[1:10])

    count <- appendSqlCyclopsData(dataPtrC,
                              oStratumId[6:9],
                              oRowId[6:9],
                              oY[6:9],
                              oTime[6:9],
                              cRowId[11:21],
                              cCovariateId[11:21],
                              cCovariateValue[11:21])
    finalizeSqlCyclopsData(dataPtrC) # Not yet implemented

    expect_equal(count, 4)
    expect_equal(getNumberOfRows(dataPtrC), 9)
    expect_equal(getNumberOfStrata(dataPtrC), 9)
    expect_equal(getNumberOfCovariates(dataPtrC), 5)

    cyclopsFitC <- fitCyclopsModel(dataPtrC, control = createControl(noiseLevel = "silent"))
    expect_equal(coef(cyclopsFitC), coef(cyclopsFit))
})

test_that("Test bad stratum IDs", {
   binomial_bid <- c(1,5,10,20,30,40,50,75,100,150,200)
   binomial_n <- c(31,29,27,25,23,21,19,17,15,15,15)
   binomial_y <- c(0,3,6,7,9,13,17,12,11,14,13)

   log_bid <- log(c(rep(rep(binomial_bid, binomial_n - binomial_y)), rep(binomial_bid, binomial_y)))
   y <- c(rep(0, sum(binomial_n - binomial_y)), rep(1, sum(binomial_y)))

   dataPtr <- createSqlCyclopsData(modelType = "lr")
   count <- appendSqlCyclopsData(dataPtr,
                             rep(0, length(y)),
                             1:length(y),
                             y,
                             rep(0,length(y)),
                             1:length(y),
                             rep(0,length(y)),
                             log_bid)
   finalizeSqlCyclopsData(dataPtr) # Not yet implemented

#     fitCyclopsModel(dataPtr, prior = createPrior("none")) #crashes R
})

library("testthat")

suppressWarnings(RNGversion("3.5.0"))

test_that("Check int64 handling", {
    set.seed(123)

    data <- simulateCyclopsData(nstrata = 1,
                                nrows = 1000,
                                ncovars = 5,
                                model="logistic")

    # using int32
    cyclopsData <- convertToCyclopsData(data$outcomes,data$covariates, modelType = "lr", addIntercept = TRUE)
    fit <- fitCyclopsModel(cyclopsData)
    int32Result <- coef(fit)

    # using int64
    data$outcomes$rowId <- bit64::as.integer64(data$outcomes$rowId)
    data$covariates$rowId <- bit64::as.integer64(data$covariates$rowId)
    data$covariates$covariateId <- bit64::as.integer64(data$covariates$covariateId)

    cyclopsData64 <- convertToCyclopsData(data$outcomes,data$covariates, modelType = "lr", addIntercept = TRUE)
    fit64 <- fitCyclopsModel(cyclopsData64)
    int64Result <- coef(fit64)

    expect_equal(int64Result, int32Result)

    # Change to int64 labels
    longName <- "1234567890"
    data$covariates$covariateId[which(data$covariates$covariateId == bit64::as.integer64(1))] <- bit64::as.integer64(longName)
    cyclopsData64b <- convertToCyclopsData(data$outcomes,data$covariates, modelType = "lr", addIntercept = TRUE)
    fit2 <- fitCyclopsModel(cyclopsData64b)
    result2 <- coef(fit2)
    expect_error(
        expect_equal(sort(result2), sort(int64Result)),
        "1 string mismatch")
    expect_equal(rownames(confint(fit2, parm = longName)), longName)
    expect_equal(rownames(confint(fit2, parm = bit64::as.integer64(longName))), longName)
})

test_that("Check int64 reductions", {
    set.seed(123)

    data <- simulateCyclopsData(nstrata = 2,
                                nrows = 1000,
                                ncovars = 5,
                                model="logistic")

    # using int64
    data$outcomes$rowId <- bit64::as.integer64(data$outcomes$rowId)
    data$covariates$rowId <- bit64::as.integer64(data$covariates$rowId)
    data$covariates$covariateId <- bit64::as.integer64(data$covariates$covariateId)
    data$covariates$covariateId[1] <- bit64::as.integer64("9223372036854775807")
    data$covariates$covariateId[2] <- bit64::as.integer64("9223372036854775806")

    cyclopsData64 <- convertToCyclopsData(data$outcomes,data$covariates, modelType = "lr", addIntercept = TRUE)
    covariates <- getCovariateIds(cyclopsData64)

    expect_equal(length(Cyclops:::reduce(cyclopsData64, covariates, power = 0)), 8)
    #expect_equal(ncol(Cyclops:::reduce(cyclopsData64, covariates, groupBy = "stratum", power = 0)), 8) # Causes crash
    expect_equal(nrow(summary(cyclopsData64)), 8)

    expect_equal(
        getUnivariableCorrelation(cyclopsData64)[7],
        getUnivariableCorrelation(cyclopsData64, covariates = bit64::as.integer64("9223372036854775807"))
    )

    expect_equal(
        getUnivariableSeparability(cyclopsData64)[7],
        getUnivariableSeparability(cyclopsData64, covariates = bit64::as.integer64("9223372036854775807"))
    )
})

library("testthat")

context("test-cprViaSql.R")

test_that("Try loading cohort via SQL", {
   load(system.file("extdata/cpr_test.RData",
                    package="Cyclops")) # loads "test" and "cov"

   test$TIME[which(test$TIME == 0)] <- 1

   cd <- createSqlCyclopsData(modelType="cpr")

   appendSqlCyclopsData(cd, test$STRATUM_ID, test$ROW_ID,
                    test$Y,
                    test$TIME,
                    #rep(1, length(test$Y)),
                    cov$ROW_ID, cov$COVARIATE_ID,
                    cov$COVARIATE_VALUE)

   finalizeSqlCyclopsData(cd,
                          useOffsetCovariate=-1,
                          addIntercept=FALSE)

   fit <- fitCyclopsModel(cd, prior = createPrior("laplace",0.1))
   coef(fit)[which(coef(fit) != 0)]
})

library("testthat")
library("survival")

context("test-xyConstructor.R")
suppressWarnings(RNGversion("3.5.0"))

#
# Tests for the XY constructor of ModelData
#

test_that("Test basic XY constructor", {
    counts <- c(18,17,15,20,10,20,25,13,12)
    outcome <- gl(3,1,9)
    treatment <- gl(3,3)
    rowId <- c(1:9)

    # gold standard
    tolerance <- 1E-4
    glmFit <- glm(counts ~ outcome + treatment, family = poisson()) # gold standard

    # Dense interface
    dataPtrD <- createSqlCyclopsData(modelType = "pr")
    loadNewSqlCyclopsDataY(dataPtrD, NULL, NULL, counts, NULL)
    loadNewSqlCyclopsDataX(dataPtrD, 0, NULL, rep(1,9), name = "(Intercept)")
    loadNewSqlCyclopsDataX(dataPtrD, 1, NULL, rep(c(0,1,0),3), name = "outcome2")
    loadNewSqlCyclopsDataX(dataPtrD, 2, NULL, rep(c(0,0,1),3), name = "outcome3")
    loadNewSqlCyclopsDataX(dataPtrD, 3, NULL, c(0,0,0,1,1,1,0,0,0), name = "treatment2")
    loadNewSqlCyclopsDataX(dataPtrD, 4, NULL, c(0,0,0,0,0,0,1,1,1), name = "treatment3")
    expect_equal(as.character(summary(dataPtrD)[,"type"]), rep("dense",5))
    expect_equal(coef(fitCyclopsModel(dataPtrD)), coef(glmFit), tolerance = tolerance)

    # Indicator interface
    dataPtrI <- createSqlCyclopsData(modelType = "pr")
    loadNewSqlCyclopsDataY(dataPtrI, NULL, rowId, counts, NULL)
    loadNewSqlCyclopsDataX(dataPtrI, 0, c(1:9), NULL, name = "(Intercept)")
    loadNewSqlCyclopsDataX(dataPtrI, 1, c(2,5,8), NULL, name = "outcome2")
    loadNewSqlCyclopsDataX(dataPtrI, 2, c(3,6,9), NULL, name = "outcome3")
    loadNewSqlCyclopsDataX(dataPtrI, 3, c(4:6), NULL, name = "treatment2")
    loadNewSqlCyclopsDataX(dataPtrI, 4, c(7:9), NULL, name = "treatment3")
    expect_equal(as.character(summary(dataPtrI)[,"type"]), rep("indicator",5))
    expect_equal(coef(fitCyclopsModel(dataPtrI)), coef(glmFit), tolerance = tolerance)

    # Sparse interface
    dataPtrS <- createSqlCyclopsData(modelType = "pr")
    loadNewSqlCyclopsDataY(dataPtrS, NULL, rowId, counts, NULL)
    loadNewSqlCyclopsDataX(dataPtrS, 0, c(1:9), rep(1,9), name = "(Intercept)", forceSparse = TRUE)
    loadNewSqlCyclopsDataX(dataPtrS, 1, c(2,5,8), rep(1,3), name = "outcome2", forceSparse = TRUE)
    loadNewSqlCyclopsDataX(dataPtrS, 2, c(3,6,9), rep(1,3), name = "outcome3", forceSparse = TRUE)
    loadNewSqlCyclopsDataX(dataPtrS, 3, c(4:6), rep(1,3), name = "treatment2", forceSparse = TRUE)
    loadNewSqlCyclopsDataX(dataPtrS, 4, c(7:9), rep(1,3), name = "treatment3", forceSparse = TRUE)
    expect_equal(as.character(summary(dataPtrS)[,"type"]), rep("sparse",5))
    expect_equal(coef(fitCyclopsModel(dataPtrS)), coef(glmFit), tolerance = tolerance)
})

test_that("Test strata XY constructor", {
    # Gold standard
    gold <- clogit(case ~ spontaneous + induced + strata(stratum), data=infert)
    tolerance <- 1E-4

    permute <- order(infert$stratum)
    pInfert <- infert[permute,]

    # Sparse interface
    dataPtrS <- createSqlCyclopsData(modelType = "clr")
    loadNewSqlCyclopsDataY(dataPtrS, pInfert$stratum, c(1:248), pInfert$case, NULL)
    loadNewSqlCyclopsDataX(dataPtrS, 0,
                           which(pInfert$spontaneous != 0),
                           pInfert$spontaneous[which(pInfert$spontaneous != 0)],
                           name = "spontaneous")
    loadNewSqlCyclopsDataX(dataPtrS, 1,
                           which(pInfert$induced != 0),
                           pInfert$induced[which(pInfert$induced != 0)],
                           name = "induced")
    expect_equal(coef(fitCyclopsModel(dataPtrS)), coef(gold), tolerance = tolerance)
})

test_that("Check stratified Cox XY constructor", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3,  1,2,0
0, 3,  0,0,1
0, 2,  1,0,1
0, 2,  1,1,1
0, 1,  0,1,0
0, 1,  1,1,0
")
    gold <- coxph(Surv(length, event) ~ x1 + strata(x2), test, ties = "breslow")
    tolerance <- 1E-4

    permute <- order(test$x2)
    pTest <- test[permute,]

    # Sparse interface
    dataPtrS <- createSqlCyclopsData(modelType = "cox")
    loadNewSqlCyclopsDataY(dataPtrS, pTest$x2, c(1:7), pTest$event, pTest$length)
    loadNewSqlCyclopsDataX(dataPtrS, 0,
                           which(pTest$x1 != 0),
                           pTest$x1[which(pTest$x1 != 0)],
                           name = "x1")
    cyclops <- fitCyclopsModel(dataPtrS)
    expect_equal(coef(cyclops), coef(gold), tolerance = tolerance)
    expect_equal(logLik(cyclops)[1], logLik(gold)[1], tolerance = tolerance)
})

# test_that("Replace covariate", {
#     counts <- c(18,17,15,20,10,20,25,13,12)
#     outcome <- gl(3,1,9)
#     treatment <- gl(3,3)
#     rowId <- c(1:9)
#
#     # gold standard
#     tolerance <- 1E-4
#     glmFit <- glm(counts ~ outcome + treatment, family = poisson()) # gold standard
#
#     # Sparse interface
#     dataPtrS <- createSqlCyclopsData(modelType = "pr")
#     loadNewSqlCyclopsDataY(dataPtrS, NULL, rowId, counts, NULL)
#     loadNewSqlCyclopsDataX(dataPtrS, 0, c(1:9), rep(1,9), name = "(Intercept)")
#     loadNewSqlCyclopsDataX(dataPtrS, 1, c(2,5,8), rep(1,3), name = "outcome2")
#     loadNewSqlCyclopsDataX(dataPtrS, 2, c(3,6,9), rep(2,3), name = "outcome3")
#     loadNewSqlCyclopsDataX(dataPtrS, 3, c(4:6), rep(1,3), name = "treatment2")
#     loadNewSqlCyclopsDataX(dataPtrS, 4, c(7:9), rep(1,3), name = "treatment3")
#     expect_equal(as.character(summary(dataPtrS)[,"type"]), rep("sparse",5))
#     expect_equal(coef(fitCyclopsModel(dataPtrS))[3] * 2, coef(glmFit)[3], tolerance = tolerance)
#
#     # Replace variable
#     loadNewSqlCyclopsDataX(dataPtrS, 2, c(3,6,9), NULL, name = "outcome3", replace = TRUE)
#     summary(dataPtrS)
#
#     expect_equal(coef(fitCyclopsModel(dataPtrS)), coef(glmFit), tolerance = tolerance)
# })

test_that("Intercept covariate", {
    counts <- c(18,17,15,20,10,20,25,13,12)
    outcome <- gl(3,1,9)
    treatment <- gl(3,3)

    # gold standard
    tolerance <- 1E-4
    glmFit <- glm(counts ~ outcome + treatment, family = poisson()) # gold standard

    # Dense interface
    dataPtrD <- createSqlCyclopsData(modelType = "pr")
    loadNewSqlCyclopsDataY(dataPtrD, NULL, NULL, counts, NULL)
    loadNewSqlCyclopsDataX(dataPtrD, 0, NULL, NULL, name = "(Intercept)")
    loadNewSqlCyclopsDataX(dataPtrD, 1, NULL, rep(c(0,1,0),3), name = "outcome2")
    loadNewSqlCyclopsDataX(dataPtrD, 2, NULL, rep(c(0,0,1),3), name = "outcome3")
    loadNewSqlCyclopsDataX(dataPtrD, 3, NULL, c(0,0,0,1,1,1,0,0,0), name = "treatment2")
    loadNewSqlCyclopsDataX(dataPtrD, 4, NULL, c(0,0,0,0,0,0,1,1,1), name = "treatment3")
    expect_equal(as.character(summary(dataPtrD)[1,"type"]), "intercept")
    expect_equal(coef(fitCyclopsModel(dataPtrD)), coef(glmFit), tolerance = tolerance)
})

test_that("Test COO-constructor", {
    counts <- c(18,17,15,20,10,20,25,13,12)
    outcome <- gl(3,1,9)
    treatment <- gl(3,3)

    # gold standard
    tolerance <- 1E-4
    glmFit <- glm(counts ~ outcome + treatment, family = poisson()) # gold standard

    rowId <- c(
        1,2,3,4,5,6,7,8,9,
        2,5,8,
        3,6,9,
        4,5,6,
        7,8,9)
    covariateId <- c(
        1,1,1,1,1,1,1,1,1,
        2,2,2,
        3,3,3,
        4,4,4,
        5,5,5)

    # Indicator interface
    dataPtrI <- createSqlCyclopsData(modelType = "pr")
    loadNewSqlCyclopsDataY(dataPtrI, NULL, c(1:9), counts, NULL)  # TODO Crashes without row IDs
    loadNewSqlCyclopsDataMultipleX(dataPtrI, covariateId, rowId, NULL,
                                    name = c("(Intercept)","outcome2","outcome3","treatment2","treatment3"))
    expect_equal(as.character(summary(dataPtrI)[,"type"]), rep("indicator",5))
    expect_equal(coef(fitCyclopsModel(dataPtrI)), coef(glmFit), tolerance = tolerance)

    # Sparse interface
    dataPtrS <- createSqlCyclopsData(modelType = "pr")
    loadNewSqlCyclopsDataY(dataPtrS, NULL, c(1:9), counts, NULL)  # TODO Crashes without row IDs
    loadNewSqlCyclopsDataMultipleX(dataPtrS, covariateId, rowId, rep(1,length(covariateId)),
                                    name = c("(Intercept)","outcome2","outcome3","treatment2","treatment3"),
                                    forceSparse = TRUE)
    expect_equal(as.character(summary(dataPtrS)[,"type"]), rep("sparse",5))
    expect_equal(coef(fitCyclopsModel(dataPtrS)), coef(glmFit), tolerance = tolerance)

    # Mixed interface (should use in practice)
    dataPtrM <- createSqlCyclopsData(modelType = "pr")
    loadNewSqlCyclopsDataY(dataPtrM, NULL, c(1:9), counts, NULL)  # TODO Crashes without row IDs
    loadNewSqlCyclopsDataX(dataPtrM, 0, NULL, NULL, name = "(Intercept)")
    loadNewSqlCyclopsDataMultipleX(dataPtrM, covariateId[10:21], rowId[10:21], NULL,
                                    name = c("outcome2","outcome3","treatment2","treatment3"))
    expect_equal(coef(fitCyclopsModel(dataPtrM)), coef(glmFit), tolerance = tolerance)

    # Multiple adds
    dataPtrA <- createSqlCyclopsData(modelType = "pr")
    loadNewSqlCyclopsDataY(dataPtrA, NULL, c(1:9), counts, NULL)  # TODO Crashes without row IDs
    loadNewSqlCyclopsDataMultipleX(dataPtrA, covariateId[1:11], rowId[1:11], NULL)
    loadNewSqlCyclopsDataMultipleX(dataPtrA, covariateId[12:21], rowId[12:21], NULL, append = TRUE)
    expect_equal(as.character(summary(dataPtrA)[,"type"]), rep("indicator",5))
    cf <- coef(fitCyclopsModel(dataPtrA))
    names(cf) <- c("(Intercept)","outcome2","outcome3","treatment2","treatment3")
    expect_equal(cf, coef(glmFit), tolerance = tolerance)
})

test_that("Data errors and casting in COO-constructor", {
    dataPtr <- createSqlCyclopsData(modelType = "lr")
    expect_error(loadNewSqlCyclopsDataMultipleX(dataPtr, c(1,1), c(1,1), NULL))

    loadNewSqlCyclopsDataMultipleX(dataPtr, c(2,2), c(1,2), c(1,1))
    expect_equal(as.character(summary(dataPtr)[2,"type"]), "indicator")

    loadNewSqlCyclopsDataMultipleX(dataPtr, c(3,3), c(1,2), c(1,2))
    expect_equal(as.character(summary(dataPtr)[3,"type"]), "sparse")

    loadNewSqlCyclopsDataMultipleX(dataPtr, c(4,4), c(1,2), c(0,1))
    expect_equal(as.character(summary(dataPtr)[4,"type"]), "indicator")

    expect_error(loadNewSqlCyclopsDataX(dataPtr, 5, c(1,1)))
})

test_that("Poisson xy-construction with offset", {
    set.seed(666)
    sim <- simulateCyclopsData(nstrata = 1, nrows = 1000, ncovars = 10,
                               model = "poisson")

    cyclopsData1 <- convertToCyclopsData(outcomes = sim$outcomes,
                                         covariates = sim$covariates,
                                         modelType = "pr",
                                         checkRowIds = FALSE)

    sim$covariates <- sim$covariates[order(sim$covariates$covariateId,
                                           sim$covariates$rowId),]
    cyclopsData2 <- createSqlCyclopsData(modelType = "pr")
    loadNewSqlCyclopsDataY(cyclopsData2, NULL, sim$outcomes$rowId,
                           sim$outcomes$y, sim$outcomes$time)
    loadNewSqlCyclopsDataX(cyclopsData2, 0, NULL, NULL,
                           name = "(Intercept)") # names are not necessary
    covarNames <- unique(sim$covariates$covariateId)
    loadNewSqlCyclopsDataMultipleX(cyclopsData2, sim$covariates$covariateId,
                                    sim$covariates$rowId,
                                    sim$covariates$covariateValue,
                                    #NULL, # pass NULL if you want indicators instead of sparse
                                    name = covarNames) # names are not necessary

    finalizeSqlCyclopsData(cyclopsData2, addIntercept = FALSE,
                           useOffsetCovariate = -1)

    fit1 <- fitCyclopsModel(cyclopsData1)
    fit2 <- fitCyclopsModel(cyclopsData2)

    tolerance <- 1E-6
    expect_equal(coef(fit2)[1], coef(fit1)[1], tolerance = tolerance)
})

library("testthat")
library("survival")

context("test-correlation.R")

#
# Small Poisson MLE regression
#

test_that("univariable correlation", {
    dobson <- data.frame(
        counts = c(18,17,15,20,10,20,25,13,12),
        outcome = gl(3,1,9),
        treatment = gl(3,3)
    )
    tolerance <- 1E-4

    dataPtrD <- createCyclopsData(counts ~ outcome + treatment, data = dobson,
                                  modelType = "pr")

    allCorrelations <- getUnivariableCorrelation(dataPtrD)
    expect_equal(length(allCorrelations), 5)

    gold <- c(NA, sapply(2:5, function(i) {
        with(dobson, cor(counts, model.matrix(~outcome + treatment)[,i]))
    }))
    expect_equivalent(gold, allCorrelations, tolerance)

    someCorrelations <- getUnivariableCorrelation(dataPtrD, c("outcome2","outcome3"))
    expect_equal(length(someCorrelations), 2)

    someCorrelations <- getUnivariableCorrelation(dataPtrD, c("outcome2","outcome3"),
                                               threshold = 0.5)
    expect_equal(length(someCorrelations), 1)

    # Try SQL data constructor
    covariates <- data.frame(stratumId = rep(infert$stratum, 2),
                             rowId = rep(1:nrow(infert), 2),
                             covariateId = rep(4:5, each = nrow(infert)),
                             covariateValue = c(infert$spontaneous, infert$induced))
    outcomes <- data.frame(stratumId = infert$stratum,
                           rowId = 1:nrow(infert),
                           y = infert$case)
    covariates <- covariates[covariates$covariateValue != 0, ]

    cyclopsData <- convertToCyclopsData(outcomes, covariates, modelType = "clr",
                                        addIntercept = FALSE)

    allCorrelations <- getUnivariableCorrelation(cyclopsData, threshold = 0.3)
    expect_equal(names(allCorrelations), c("4"))
})

library("testthat")

context("test-smallBernoulli.R")

#
# Small Bernoulli MLE regression
#

test_that("Small Bernoulli dense regression", {
	binomial_bid <- c(1,5,10,20,30,40,50,75,100,150,200)
	binomial_n <- c(31,29,27,25,23,21,19,17,15,15,15)
	binomial_y <- c(0,3,6,7,9,13,17,12,11,14,13)

	log_bid <- log(c(rep(rep(binomial_bid, binomial_n - binomial_y)), rep(binomial_bid, binomial_y)))
	y <- c(rep(0, sum(binomial_n - binomial_y)), rep(1, sum(binomial_y)))

	tolerance <- 1E-4

	glmFit <- glm(y ~ log_bid, family = binomial()) # gold standard

	dataPtrD <- createCyclopsData(y ~ log_bid, modelType = "lr")
	cyclopsFitD <- fitCyclopsModel(dataPtrD, prior = createPrior("none"),
	                       control = createControl(noiseLevel = "silent"))
	expect_equal(coef(cyclopsFitD), coef(glmFit), tolerance = tolerance)
	expect_equal(cyclopsFitD$log_likelihood, logLik(glmFit)[[1]], tolerance = tolerance)
	expect_equal(confint(cyclopsFitD, c(1:2))[,2:3], confint(glmFit, c(1:2)), tolerance = tolerance)
	expect_equal(predict(cyclopsFitD), predict(glmFit, type = "response"), tolerance = tolerance)

	# Try with Zhang-Oles criterion
	zoFit <- fitCyclopsModel(dataPtrD, prior = createPrior("none"), forceNewObject = TRUE,
	                         control = createControl(convergenceType = "zhang", noiseLevel = "noisy"))
	expect_equal(coef(cyclopsFitD), coef(zoFit), tolerance = tolerance)

	# Try with KKT swindle
	gold <- fitCyclopsModel(dataPtrD, prior = createPrior("laplace"), forceNewObject = TRUE, warnings = FALSE)

	kktFit <- fitCyclopsModel(dataPtrD, prior = createPrior("laplace"), forceNewObject = TRUE, warnings = FALSE,
	                          control = createControl(useKKTSwindle = TRUE, noiseLevel = "quiet"))
	expect_equal(coef(gold), coef(kktFit))

	# Test old logging routines
	file <- tempfile()
	Cyclops:::.cyclopsLogResults(cyclopsFitD$interface, file, TRUE)
	result <- read.csv(file)
	expect_equivalent(result$estimate, coef(cyclopsFitD), tolerance = tolerance)

	cap <- capture.output(print(dataPtrD))
	expect_false(is.null(cap))
})

test_that("Old BLR format", {
    data <- readCyclopsData(system.file("extdata/infert_ccd.txt", package = "Cyclops"), "clr")
    expect_equal(getNumberOfRows(data), 248)
})

test_that("Add intercept via finalize", {
    binomial_bid <- c(1,5,10,20,30,40,50,75,100,150,200)
    binomial_n <- c(31,29,27,25,23,21,19,17,15,15,15)
    binomial_y <- c(0,3,6,7,9,13,17,12,11,14,13)

    log_bid <- log(c(rep(rep(binomial_bid, binomial_n - binomial_y)), rep(binomial_bid, binomial_y)))
    y <- c(rep(0, sum(binomial_n - binomial_y)), rep(1, sum(binomial_y)))

    tolerance <- 1E-4

    glmFit <- glm(y ~ log_bid, family = binomial()) # gold standard

    dataPtrD <- createCyclopsData(y ~ log_bid - 1, modelType = "lr")
    finalizeSqlCyclopsData(dataPtrD, addIntercept = TRUE)
    cyclopsFitN <- fitCyclopsModel(dataPtrD, prior = createPrior("none"), forceNewObject = TRUE,
                           control = createControl(noiseLevel = "silent"))
    expect_equal(coef(cyclopsFitN), coef(glmFit), tolerance = tolerance)

    expect_error(finalizeSqlCyclopsData(dataPtrD, addIntercept = TRUE))

    dataPtrI <- createCyclopsData(y ~ log_bid, modelType = "lr")
    expect_error(finalizeSqlCyclopsData(dataPtrI, addIntercept = TRUE))
})

test_that("Small Bernoulli sparse regression", {
	binomial_bid <- c(1,5,10,20,30,40,50,75,100,150,200)
	binomial_n <- c(31,29,27,25,23,21,19,17,15,15,15)
	binomial_y <- c(0,3,6,7,9,13,17,12,11,14,13)

	log_bid <- log(c(rep(rep(binomial_bid, binomial_n - binomial_y)), rep(binomial_bid, binomial_y)))
	y <- c(rep(0, sum(binomial_n - binomial_y)), rep(1, sum(binomial_y)))

	tolerance <- 1E-4

	glmFit <- glm(y ~ log_bid, family = binomial()) # gold standard

	dataPtrS <- createCyclopsData(y ~ 1, sparseFormula = ~ log_bid, modelType = "lr")
	cyclopsFitS <- fitCyclopsModel(dataPtrS, prior = createPrior("none"),
												 control = createControl(noiseLevel = "silent"))
	expect_equal(coef(cyclopsFitS), coef(glmFit), tolerance = tolerance)
	expect_equal(cyclopsFitS$log_likelihood, logLik(glmFit)[[1]], tolerance = tolerance)
	expect_equal(confint(cyclopsFitS, c(1:2))[,2:3], confint(glmFit, c(1:2)), tolerance = tolerance)
	expect_equal(predict(cyclopsFitS), predict(glmFit, type = "response"), tolerance = tolerance)
})

library("testthat")

context("test-reductions.R")

test_that("Simple reductions", {
    counts <- c(18,17,15,20,10,20,25,13,12)
    outcome <- gl(3,1,9)
    treatment <- gl(3,3)
    tolerance <- 1E-4

    dataPtr <- createCyclopsData(counts ~ 1,
                                  indicatorFormula = ~ outcome + treatment,
                                  modelType = "pr")

    expect_error(reduce(dataPtr, 0))
    expect_error(reduce(dataPtr, "BAD"))
    expect_equal(reduce(dataPtr, c(1,2)),
                 c(9,3))

    expect_equivalent(reduce(dataPtr, 4, groupBy = 3),
                      as.data.frame(c(2,1)))
    expect_equivalent(reduce(dataPtr, 3, groupBy = "treatment2"),
                      as.data.frame(c(2,1)))
    expect_equal(dim(reduce(dataPtr, 3, groupBy = "stratum")),
                 c(9,1))
    expect_error(reduce(dataPtr, 4, groupBy = c(3,1)))

    #throw error? when # strata = # row
})

library(testthat)

context("test-dataManagement.R")

test_that("Make covariates dense", {
    counts <- c(18,17,15,20,10,20,25,13,12)
    outcome <- gl(3,1,9)
    treatment <- gl(3,3)
    tolerance <- 1E-4

    dataPtr <- createCyclopsData(counts ~ outcome, indicatorFormula =  ~ treatment,
                                  modelType = "pr")

    expect_equal(as.character(summary(dataPtr)["treatment2","type"]),
                 "indicator")

    finalizeSqlCyclopsData(dataPtr, makeCovariatesDense = "treatment2")

    expect_equal(as.character(summary(dataPtr)["treatment2","type"]),
                 "dense")
})

library("testthat")
library("survival")

context("test-survfit.R")

test_that("Check very small Cox example with no ties, breslow baseline", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3.5,1,2,0
0, 3,  0,0,1
0, 2.5,1,0,1
0, 2,  1,1,1
0, 1.5,0,1,0
0, 1,  1,1,0
")

    if (packageVersion("survival") >= '3.2.9') {
        goldFit <-  coxph(Surv(start, length, event) ~ x1 + x2, test, ties = "breslow", nocenter = NULL)
    } else {
        goldFit <-  coxph(Surv(start, length, event) ~ x1 + x2, test, ties = "breslow")
    }
    goldSurv <- survfit(goldFit)

    dataPtr <- createCyclopsData(Surv(length, event) ~ x1 + x2, data = test, modelType = "cox")
    cyclopsFit <- fitCyclopsModel(dataPtr)
    cyclopsSurv <- survfit(cyclopsFit, type="aalen")

    expect_equal(goldSurv$time, cyclopsSurv$time)
    tolerance <- 1E-4
    expect_equal(goldSurv$surv, cyclopsSurv$surv, tolerance = tolerance)
})

test_that("Check very small Cox example with time-ties, but no failure ties, breslow baseline", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3,  1,2,0
0, 3,  0,0,1
0, 2,  1,0,1
0, 2,  0,1,1
0, 1,  0,1,0
0, 1,  1,1,0
")

    if (packageVersion("survival") >= '3.2.9') {
        goldFit <-  coxph(Surv(start, length, event) ~ x1 + x2, test, ties = "breslow", nocenter = NULL)
    } else {
        goldFit <-  coxph(Surv(start, length, event) ~ x1 + x2, test, ties = "breslow")
    }
    goldSurv <- survfit(goldFit)

    dataPtr <- createCyclopsData(Surv(length, event) ~ x1 + x2, data = test, modelType = "cox")
    cyclopsFit <- fitCyclopsModel(dataPtr)
    cyclopsSurv <- survfit(cyclopsFit, type="aalen")

    expect_equal(goldSurv$time, cyclopsSurv$time)
    tolerance <- 1E-4
    expect_equal(goldSurv$surv, cyclopsSurv$surv, tolerance = tolerance)
})

test_that("Check very small Cox example with failure ties, no risk-set contribution after tie, breslow baseline", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3,  1,2,0
0, 3,  0,0,1
0, 2,  1,0,1
0, 2,  1,1,1
0, 2,  1,0,0
")
    if (packageVersion("survival") >= '3.2.9') {
        goldFit <-  coxph(Surv(start, length, event) ~ x1 + x2, test, ties = "breslow", nocenter = NULL)
    } else {
        goldFit <-  coxph(Surv(start, length, event) ~ x1 + x2, test, ties = "breslow")
    }
    goldSurv <- survfit(goldFit)

    dataPtr <- createCyclopsData(Surv(length, event) ~ x1 + x2, data = test, modelType = "cox")
    cyclopsFit <- fitCyclopsModel(dataPtr)
    cyclopsSurv <- survfit(cyclopsFit, type="aalen")

    expect_equal(goldSurv$time, cyclopsSurv$time)
    tolerance <- 1E-4
    expect_equal(goldSurv$surv, cyclopsSurv$surv, tolerance = tolerance)})

test_that("Check very small Cox example with failure ties, with risk-set contribution after tie, breslow baseline", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3,  1,2,0
0, 3,  0,0,1
0, 2,  1,0,1
0, 2,  1,1,1
0, 1,  0,1,0
0, 1,  1,1,0
")

    if (packageVersion("survival") >= '3.2.9') {
        goldFit <-  coxph(Surv(start, length, event) ~ x1 + x2, test, ties = "breslow", nocenter = NULL)
    } else {
        goldFit <-  coxph(Surv(start, length, event) ~ x1 + x2, test, ties = "breslow")
    }
    goldSurv <- survfit(goldFit)

    dataPtr <- createCyclopsData(Surv(length, event) ~ x1 + x2, data = test, modelType = "cox")
    cyclopsFit <- fitCyclopsModel(dataPtr)
    cyclopsSurv <- survfit(cyclopsFit, type="aalen")

    expect_equal(goldSurv$time, cyclopsSurv$time)
    tolerance <- 1E-4
    expect_equal(goldSurv$surv, cyclopsSurv$surv, tolerance = tolerance)
})

# @file Multitype.R
#
# Copyright 2014 Observational Health Data Sciences and Informatics
#
# This file is part of cyclops
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#' @title Create a multitype outcome object
#' 
#' @description
#' \code{Multitype} creates a multitype outcome object, usually used as a response variable in a 
#' hierarchical Cyclops model fit.
#' 
#' @param y     Numeric: Response count(s)
#' @param type Numeric or factor: Response type
#' 
#' @return An object of class \code{Multitype} with length equal to the length of \code{y} and \code{type}.
#' @examples
#' Multitype(c(0,1,0), as.factor(c("A","A","B")))
#' 
#' @export
Multitype <- function(y, type) {
    if (missing(y) || !is.numeric(y)) stop("Must have outcome counts")
    if (missing(type)) stop("Must have outcome types")
    if (length(y) != length(type)) stop("Must have equal lengths")
    
    #mt <- data.frame(y, type)
    
    #mt$dim <- c(length(y),2)
    #attr(mt, "dim")  <- c(length(y), 2)
    mt <- cbind(y = y, type = type)
    attr(mt, "contrasts") <- contrasts(type)
    class(mt) <- c("Multitype")
    
    mt
}

# @file Simulation.R
#
# Copyright 2014 Observational Health Data Sciences and Informatics
#
# This file is part of cyclops
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#' @title Simulation Cyclops dataset
#'
#' @description
#' \code{simulateCyclopsData} generates a simulated large, sparse data set for use by \code{fitCyclopsSimulation}.
#'
#' @param nstrata   Numeric: Number of strata
#' @param nrows Numeric: Number of observation rows
#' @param ncovars Numeric: Number of covariates
#' @param effectSizeSd Numeric: Standard derivation of the non-zero simulated regression coefficients
#' @param zeroEffectSizeProp Numeric: Expected proportion of zero effect size
#' @param eCovarsPerRow Number: Effective number of non-zero covariates per data row
#' @param model String: Simulation model. Choices are: \code{logistic}, \code{poisson} or \code{survival}
#'
#' @return A simulated data set
#'
#' @template elaborateExample
#'
#' @export
simulateCyclopsData <- function(nstrata = 200,
                                nrows = 10000,
                                ncovars = 20,
                                effectSizeSd = 1,
                                zeroEffectSizeProp = 0.9,
                                eCovarsPerRow = ncovars/100,
                                model="survival"){

    sd <- rep(effectSizeSd, ncovars) * rbinom(ncovars, 1, 1 - zeroEffectSizeProp)
    effectSizes <- data.frame(covariateId=1:ncovars,rr=exp(rnorm(ncovars,mean=0,sd=sd)))

    covarsPerRow <- rpois(nrows,eCovarsPerRow)
    covarsPerRow[covarsPerRow > ncovars] <- ncovars
    covarsPerRow <- data.frame(covarsPerRow = covarsPerRow)
    covarRows <- sum(covarsPerRow$covarsPerRow)
    covariates <- data.frame(rowId = rep(0,covarRows), covariateId = rep(0,covarRows), covariateValue = rep(1,covarRows))
    cursor <- 1
    for (i in 1:nrow(covarsPerRow)){
        n <- covarsPerRow$covarsPerRow[i]
        if (n != 0){
            covariates$rowId[cursor:(cursor+n-1)] <- i
            covariates$covariateId[cursor:(cursor+n-1)] <- sample.int(size=n,ncovars)
            cursor = cursor+n
        }
    }

    outcomes <- data.frame(rowId = 1:nrows, stratumId = round(runif(nrows,min=1,max=nstrata)), y=0)
    covariates <- merge(covariates,outcomes[,c("rowId","stratumId")])

    rowId_to_rr <- aggregate(rr ~ rowId, data=merge(covariates,effectSizes),prod)

    outcomes <- merge(outcomes,rowId_to_rr,all.x=TRUE)
    outcomes$rr[is.na(outcomes$rr)] <- 1

    if (model == "survival"){
        strataBackgroundProb <- runif(nstrata,min=0.01,max=0.03)
        outcomes$rate <-  strataBackgroundProb[outcomes$stratumId] * outcomes$rr
        outcomes$timeToOutcome <- 1+round(rexp(n=nrow(outcomes),outcomes$rate))
        outcomes$timeToCensor <- 1+round(runif(n=nrow(outcomes),min=0,max=499))
        outcomes$time <- outcomes$timeToOutcome
        outcomes$time[outcomes$timeToCensor < outcomes$timeToOutcome] <- outcomes$timeToCensor[outcomes$timeToCensor < outcomes$timeToOutcome]
        outcomes$y <- as.integer(outcomes$timeToCensor > outcomes$timeToOutcome)
    } else if (model == "logistic") {
        strataBackgroundProb <- runif(nstrata,min=0.1,max=0.3)
        outcomes$prob <-  strataBackgroundProb[outcomes$stratumId] * outcomes$rr
        outcomes$y <- as.integer(runif(nrows,min=0,max=1) < outcomes$prob)
    } else if (model == "poisson"){
        strataBackgroundProb <- runif(nstrata,min=0.01,max=0.03)
        outcomes$rate <-  strataBackgroundProb[outcomes$stratumId] * outcomes$rr
        outcomes$time <- 1+round(runif(n=nrow(outcomes),min=0,max=499))
        outcomes$y <- rpois(nrows,outcomes$rate * outcomes$time)
    } else
        stop(paste("Unknown model:",model))

    outcomes <- outcomes[order(outcomes$stratumId,outcomes$rowId),]
    covariates <- covariates[order(covariates$stratumId,covariates$rowId,covariates$covariateId),]
    sparseness <- 1-(nrow(covariates)/(nrows*ncovars))
    intercepts <- data.frame(stratumId = 1:nstrata, prob = strataBackgroundProb, logProb = log(strataBackgroundProb))
    writeLines(paste("Sparseness =",sparseness*100,"%"))
    list(outcomes = outcomes,
         covariates = covariates,
         effectSizes = effectSizes,
         sparseness = sparseness,
         intercepts = intercepts)
}

# .figureOutGlmnetComparison <- function() {
#
#     sim <-simulateCyclopsData(1, 100000, 1000, 0.5,
#                               zeroEffectSizeProp = 0.9, model = "logistic")
#
#     convertCyclopsSimulationToGlmnet <- function(sim) {
#         mat <- Matrix(data = 0,
#                       nrow = nrow(sim$outcomes),
#                       ncol = max(sim$covariates$covariateId),
#                       sparse = TRUE)
#         nnz <- length(sim$covariates$rowId)
#         for (i in 1:nnz) {
#             mat[sim$covariates$rowId[i],
#                 sim$covariates$covariateId[i]] <- sim$covariates$covariateValue[i]
#         }
#         mat
#     }
#
#     mat <- convertCyclopsSimulationToGlmnet(sim)
#
#     mat <- sparseX
#     y <- y
#
#     start <- Sys.time()
#     f <- glmnet(y = y, x = mat,
#                 family = "binomial",
#                 lambda = sqrt(2 / 0.1) / nrow(mat),
#                 intercept = FALSE, standardize = FALSE)
#     delta <- Sys.time() - start
#     writeLines(paste("Analysis took", signif(delta,3), attr(delta,"units")))
#
#     start <- Sys.time()
#     cd <- createCyclopsData(y = y, ix = mat, modelType = "lr")
#     ff <- fitCyclopsModel(cd, prior = createPrior("laplace", 0.1))
#     delta <- Sys.time() - start
#     writeLines(paste("Analysis took", signif(delta,3), attr(delta,"units")))
#
# }

.fitUsingClogit <- function(sim,coverage=TRUE){
    start <- Sys.time()
    covariates <- sim$covariates
    ncovars <- max(covariates$covariateId)
    nrows <- nrow(sim$outcomes)
    m <- matrix(0,nrows,ncovars)
    for (i in 1:nrow(covariates)){
        m[covariates$rowId[i],covariates$covariateId[i]] <- 1
    }
    data <- as.data.frame(m)

    data$rowId <- 1:nrow(data)
    data <- merge(data,sim$outcomes)
    data <- data[order(data$stratumId,data$rowId),]
    formula <- as.formula(paste(c("y ~ strata(stratumId)",paste("V",1:ncovars,sep="")),collapse=" + "))
    fit <- survival::clogit(formula,data=data)
    if (coverage) {
        ci <- confint(fit)
    } else {
        ci <- matrix(0,nrow=1,ncol=2)
    }

    delta <- Sys.time() - start
    writeLines(paste("Analysis took", signif(delta,3), attr(delta,"units")))

    data.frame(coef = coef(fit), lbCi95 = ci[,1], ubCi95 = ci[,2])
}

.fitUsingCoxph <- function(sim, coverage = TRUE){
    if (!requireNamespace("survival")) {
        stop("survival library required")
    }
    start <- Sys.time()
    covariates <- sim$covariates
    ncovars <- max(covariates$covariateId)
    nrows <- nrow(sim$outcomes)
    m <- matrix(0,nrows,ncovars)
    for (i in 1:nrow(covariates)){
        m[covariates$rowId[i],covariates$covariateId[i]] <- 1
    }
    data <- as.data.frame(m)

    data$rowId <- 1:nrow(data)
    data <- merge(data,sim$outcomes)
    data <- data[order(data$stratumId,data$rowId),]
    formula <- as.formula(paste(c("Surv(time,y) ~ strata(stratumId)",paste("V",1:ncovars,sep="")),collapse=" + "))
    fit <- survival::coxph(formula,data=data)
    if (coverage) {
        ci <- confint(fit)
    } else {
        ci <- matrix(0,nrow=1,ncol=2)
    }

    delta <- Sys.time() - start
    writeLines(paste("Analysis took", signif(delta,3), attr(delta,"units")))

    data.frame(coef = coef(fit), lbCi95 = ci[,1], ubCi95 = ci[,2])
}

# .fitUsingGnm <- function(sim,coverage=TRUE){
#     if (!requireNamespace("gnm")) {
#         stop("gnm library required")
#     }
#     start <- Sys.time()
#     covariates <- sim$covariates
#     ncovars <- max(covariates$covariateId)
#     nrows <- nrow(sim$outcomes)
#     m <- matrix(0,nrows,ncovars)
#     for (i in 1:nrow(covariates)){
#         m[covariates$rowId[i],covariates$covariateId[i]] <- 1
#     }
#     data <- as.data.frame(m)
#
#     data$rowId <- 1:nrow(data)
#     data <- merge(data,sim$outcomes)
#     data <- data[order(data$stratumId,data$rowId),]
#     formula <- as.formula(paste(c("y ~ v1",paste("v",2:ncovars,sep="")),collapse=" + "))
#
#     fit = gnm::gnm(formula, family=poisson, offset=log(time), eliminate=as.factor(data$stratumId), data = data)
#     #Todo: figure out how to do confidence intervals correctly
#     confint(fit)
#     fit0 = gnm::gnm(y ~ 1, family=poisson, offset=log(time), eliminate=as.factor(data$stratumId), data = data)
#     se <- abs(coef(fit)[[1]]/qnorm(1-pchisq(deviance(fit0)-deviance(fit),1)))
#
#
#     fit <- survival::coxph(formula,data=data)
#     if (coverage) {
#         ci <- confint(fit)
#     } else {
#         ci <- matrix(0,nrow=1,ncol=2)
#     }
#
#     delta <- Sys.time() - start
#     writeLines(paste("Analysis took", signif(delta,3), attr(delta,"units")))
#
#     data.frame(coef = coef(fit), lbCi95 = ci[,1], ubCi95 = ci[,2])
# }


#' @title Fit simulated data
#'
#' @description
#' \code{fitCyclopsSimulation} fits simulated Cyclops data using Cyclops or a standard routine.
#' This function is useful for simulation studies comparing the performance of Cyclops when considering
#' large, sparse datasets.
#'
#' @param sim    A simulated Cyclops dataset generated via \code{simulateCyclopsData}
#' @param useCyclops    Logical: use Cyclops or a standard routine
#' @param model  String: Fitted regression model type
#' @param coverage Logical: report coverage statistics
#' @param includePenalty   Logical: include regularized regression penalty in computing profile likelihood based confidence intervals
#' @param computeDevice String: Name of compute device to employ; defaults to \code{"native"} C++ on CPU
#'
#' @export
fitCyclopsSimulation <- function(sim,
                                 useCyclops = TRUE,
                                 model = "logistic",
                                 coverage = TRUE,
                                 includePenalty = FALSE,
				 computeDevice = "native") {
    if (useCyclops) {
        .fitCyclopsSimulationUsingCyclops(sim, model, coverage = coverage, includePenalty = includePenalty, computeDevice = computeDevice)
    } else {
        .fitCyclopsSimulationUsingOtherThanCyclops(sim, model, coverage)
    }
}

.fitCyclopsSimulationUsingOtherThanCyclops <- function(sim, model="logistic",coverage=TRUE){
    if (model == "logistic"){
        writeLines("Fitting model using clogit")
        .fitUsingClogit(sim,coverage)
    } else if (model == "poisson"){
        stop("Poisson not yet implemented")
    } else if (model == "survival"){
        writeLines("Fitting model using coxpht")
        .fitUsingCoxph(sim,coverage)
    } else {
        stop(paste("Unknown model:",model))
    }
}

.fitCyclopsSimulationUsingCyclops <- function(sim,
                                              model = "logistic",
                                              regularized = FALSE,
                                              coverage = TRUE,
                                              includePenalty = FALSE,
					      computeDevice = computeDevice){
    if (!regularized)
        includePenalty = FALSE
    start <- Sys.time()
    stratified <- max(sim$outcomes$stratumId) > 1
    if (stratified){
        if (model == "logistic") modelType = "clr"
        if (model == "poisson") modelType = "cpr"
        if (model == "survival") modelType = "cox"
    } else {
        if (model == "logistic") modelType = "lr"
        if (model == "poisson") modelType = "pr"
        if (model == "survival") modelType = "cox"
    }

    dataPtr <- convertToCyclopsData(sim$outcomes,sim$covariates,modelType = modelType,addIntercept = !stratified)

    # if (regularized){
        coefCyclops <- rep(0,length(sim$effectSizes$rr))
        lbCi95 <- rep(0,length(sim$effectSizes$rr))
        ubCi95 <- rep(0,length(sim$effectSizes$rr))
        for (i in 1:length(sim$effectSizes$rr)){
            if (model == "survival"){ # For some reason can't get confint twice on same dataPtr object, so recreate:
                dataPtr <- convertToCyclopsData(sim$outcomes,sim$covariates,modelType = modelType,addIntercept = !stratified)
            }
            if (regularized) {
                prior <- createPrior("laplace", 0.1, exclude = i)
            } else {
                prior <- createPrior("none")
            }
            cyclopsFit <- fitCyclopsModel(dataPtr,prior = prior)
            coefCyclops[i] <- coef(cyclopsFit)[names(coef(cyclopsFit)) == as.character(i)]
            if (coverage) {
                if (model == "survival"){
                    ci <- confint(cyclopsFit,parm=i,includePenalty = includePenalty)
                    lbCi95[i] <- ci[,2]
                    ubCi95[i] <- ci[,3]
                } else {
                    ci <- confint(cyclopsFit,parm=i,includePenalty = includePenalty)
                    lbCi95[i] <- ci[,2]
                    ubCi95[i] <- ci[,3]
                }
            }
        }
    # } else {
    #     cyclopsFit <- fitCyclopsModel(dataPtr, prior = createPrior("none"))
    #     coefCyclops <- data.frame(covariateId = as.integer(names(coef(cyclopsFit))),beta=coef(cyclopsFit))
    #     coefCyclops <- coefCyclops[order(coefCyclops$covariateId),]
    #     coefCyclops <- coefCyclops$beta
    #     if (coverage) {
    #         if (model == "survival"){
    #             ci <- confint(cyclopsFit,parm=i,includePenalty = includePenalty)
    #             lbCi95 <- ci[,2]
    #             ubCi95 <- ci[,3]
    #         } else {
    #             ci <- confint(cyclopsFit,parm=i,includePenalty = includePenalty)
    #             lbCi95 <- ci[,2]
    #             ubCi95 <- ci[,3]
    #         }
    #     }
    # }
    delta <- Sys.time() - start
    writeLines(paste("Analysis took", signif(delta,3), attr(delta,"units"),
                     "(",
                     signif(cyclopsFit$timeFit,3), attr(cyclopsFit$timeFit,"units"),
                     ")"))

    if (coverage) {
        df <- data.frame(coef = coefCyclops, lbCi95 = lbCi95, ubCi95 = ubCi95)
    } else {
        df <- data.frame(coef = coefCyclops)
    }
    attr(df, "dataPtr") <- dataPtr
    df
}

#' @title Mean squared error
#'
#' @description
#' \code{mse} computes the mean squared error between two numeric vectors
#'
#' @param goldStandard Numeric vector
#' @param estimates  Numeric vector
#'
#' @return MSE(\code{goldStandard}, \code{estimates})
#'
#' @export
mse <- function(goldStandard, estimates){
    mean((goldStandard - estimates)^2)
}

#' @title Coverage
#'
#' @description
#' \code{coverage} computes the coverage on confidence intervals
#'
#' @param goldStandard Numeric vector
#' @param lowerBounds  Numeric vector. Lower bound of the confidence intervals
#' @param upperBounds  Numeric vector. Upper bound of the confidence intervals
#'
#' @return The proportion of times \code{goldStandard} falls between \code{lowerBound} and \code{upperBound}
#'
#' @export
coverage <- function(goldStandard, lowerBounds, upperBounds){
    sum(goldStandard >= lowerBounds & goldStandard <= upperBounds) / length(goldStandard)
}

#' @title Plot Cyclops simulation model fit
#'
#' @description
#' \code{plotCyclopsSimulationFit} generates a plot that compares \code{goldStandard} coefficients to their
#' Cyclops model \code{fit}.
#'
#' @param fit   A Cyclops simulation fit generated by \code{fitCyclopsSimulation}
#' @param goldStandard Numeric vector.  True relative risks.
#' @param label String. Name of estimate type.
#'
#' @keywords internal
plotCyclopsSimulationFit <- function(fit,goldStandard,label){
    if (requireNamespace("ggplot2")) {
        ggplot2::ggplot(fit, ggplot2::aes(x= goldStandard , y=coef, ymin=fit$lbCi95, ymax=fit$ubCi95), environment=environment()) +
            ggplot2::geom_abline(intercept = 0, slope = 1) +
            ggplot2::geom_pointrange(alpha=0.2) +
            ggplot2::scale_y_continuous(label)
    } else {
        stop("gglot2 package required")
    }
}


# TODO Move to vignette
#
# runSimulation1 <- function(){
#     model = "survival"      # "logistic", "survival", or "poisson"
#     sim <- simulateCyclopsData(nstrata=2000,
#                         ncovars=100,
#                         nrows=10000,
#                         effectSizeSd=0.5,
#                         eCovarsPerRow=2,
#                         model=model)
#     coefGoldStandard <- log(sim$effectSizes$rr)
#
#     fitCyclops <- fitCyclopsSimulation(sim, useCyclops = TRUE, regularized=TRUE,model)
#     fit <- fitCyclopsSimulation(sim, useCyclops = FALSE, model)
#
#     writeLines(paste("MSE Cyclops:", mse(fitCyclops$coef,coefGoldStandard)))
#     writeLines(paste("MSE other:", mse(fit$coef,coefGoldStandard)))
#
#     plotCyclopsSimulationFit(fitCyclops,coefGoldStandard,"Cyclops")
#     plotCyclopsSimulationFit(fit,coefGoldStandard,"Other")
#
#     writeLines(paste("Coverage Cyclops:", coverage(coefGoldStandard,fitCyclops$lbCi95, fitCyclops$ubCi95)))
#     writeLines(paste("Coverage other:", coverage(coefGoldStandard,fit$lbCi95, fit$ubCi95)))
# }
#
# runSimulation2 <- function(){
#     model = "logistic"      # "logistic", "survival", or "poisson"
#     sim <- simulateCyclopsData(nstrata=2000,
#                         ncovars=100,
#                         nrows=10000,
#                         effectSizeSd=0.5,
#                         eCovarsPerRow=2,
#                         model=model)
#     coefGoldStandard <- log(sim$effectSizes$rr)
#
#     fit <- fitCyclopsSimulation(sim, useCyclops = FALSE, model,coverage=FALSE)
#     fitCyclops <- fitCyclopsSimulation(sim, useCyclops = TRUE, regularized=TRUE,model,coverage=FALSE)
#
#     writeLines(paste("MSE Cyclops:", mse(fitCyclops$coef,coefGoldStandard)))
#     writeLines(paste("MSE other:", mse(fit$coef,coefGoldStandard)))
# }

library("testthat")

context("test-simulation.R")

#
# Simulation
#

test_that("Check logistic output from other programs", {
    set.seed(123)
    tolerance <- 1E-2
    data <- simulateCyclopsData(nstrata = 1, nrows = 1000, ncovars = 2, zeroEffectSizeProp = 0.0, eCovarsPerRow = 0.5,
                                model = "logistic")
    suppressWarnings(cyclopsFit <- fitCyclopsSimulation(data, model = "logistic", useCyclops = TRUE))
    otherFit <- fitCyclopsSimulation(data, model = "logistic", useCyclops = FALSE)
    expect_equal(cyclopsFit$coef, otherFit$coef, tolerance = tolerance)
})

test_that("Check survival output from other programs", {
    set.seed(123)
    tolerance <- 1E-2
    data <- simulateCyclopsData(nstrata = 1, nrows = 1000, ncovars = 2, zeroEffectSizeProp = 0.0, eCovarsPerRow = 0.5,
                                model = "survival")
    suppressWarnings(cyclopsFit <- fitCyclopsSimulation(data, model = "survival", useCyclops = TRUE))
    otherFit <- fitCyclopsSimulation(data, model = "survival", useCyclops = FALSE)
    expect_equal(cyclopsFit$coef, otherFit$coef, tolerance = tolerance)

    plot <- Cyclops:::plotCyclopsSimulationFit(cyclopsFit, log(data$effectSizes$rr), label = "Cyclops")
    expect_false(is.null(plot))
})

test_that("Check simulation edge cases", {
    expect_error(simulateCyclopsData(model = "not_known"),
                 "Unknown model")

    expect_equal(mse(goldStandard = c(0,2), estimates = c(0,0)), 2)
})

# @file ModelFit.R
#
# Copyright 2016 Observational Health Data Sciences and Informatics
#
# This file is part of cyclops
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#' @title Fit a Cyclops model
#'
#' @description
#' \code{fitCyclopsModel} fits a Cyclops model data object
#'
#' @details
#' This function performs numerical optimization to fit a Cyclops model data object.
#'
#' @param cyclopsData			A Cyclops data object
#' @template prior
#' @param control  A \code{"cyclopsControl"} object constructed by \code{\link{createControl}}
#' @param weights Vector of 0/1 weights for each data row
#' @param forceNewObject Logical, forces the construction of a new Cyclops model fit object
#' @param returnEstimates Logical, return regression coefficient estimates in Cyclops model fit object
#' @param startingCoefficients Vector of starting values for optimization
#' @param fixedCoefficients Vector of booleans indicating if coefficient should be fix
#' @param warnings Logical, report regularization warnings
#' @param computeDevice String: Name of compute device to employ; defaults to \code{"native"} C++ on CPU
#'
#' @return
#' A list that contains a Cyclops model fit object pointer and an operation duration
#'
#' @references
#' Suchard MA, Simpson SE, Zorych I, Ryan P, Madigan D.
#' Massive parallelization of serial inference algorithms for complex generalized linear models.
#' ACM Transactions on Modeling and Computer Simulation, 23, 10, 2013.
#'
#' Simpson SE, Madigan D, Zorych I, Schuemie M, Ryan PB, Suchard MA.
#' Multiple self-controlled case series for large-scale longitudinal observational databases.
#' Biometrics, 69, 893-902, 2013.
#'
#' Mittal S, Madigan D, Burd RS, Suchard MA.
#' High-dimensional, massive sample-size Cox proportional hazards regression for survival analysis.
#' Biostatistics, 15, 207-221, 2014.
#'
#' @examples
#' ## Dobson (1990) Page 93: Randomized Controlled Trial :
#' counts <- c(18,17,15,20,10,20,25,13,12)
#' outcome <- gl(3,1,9)
#' treatment <- gl(3,3)
#' cyclopsData <- createCyclopsData(counts ~ outcome + treatment, modelType = "pr")
#' cyclopsFit <- fitCyclopsModel(cyclopsData, prior = createPrior("none"))
#' coef(cyclopsFit)
#' confint(cyclopsFit, c("outcome2","treatment3"))
#' predict(cyclopsFit)
#'
#' @export
fitCyclopsModel <- function(cyclopsData,
                            prior = createPrior("none"),
                            control = createControl(),
                            weights = NULL,
                            forceNewObject = FALSE,
                            returnEstimates = TRUE,
                            startingCoefficients = NULL,
                            fixedCoefficients = NULL,
                            warnings = TRUE,
							computeDevice = "native") {

    # Delegate to control$setHook if exists
    if (!is.null(control$setHook)) {
        return(control$setHook(cyclopsData, prior, control,
                               weights, forceNewObject, returnEstimates,
                               startingCoefficients, fixedCoefficients))
    }

    # Delegate to prior$fitHook if exists
    if (!is.null(prior$fitHook)) {
        return(prior$fitHook(cyclopsData, prior, control,
                             weights, forceNewObject, returnEstimates,
                             startingCoefficients, fixedCoefficients))
    }

    cl <- match.call()

    # Check conditions
    .checkData(cyclopsData)

    if (getNumberOfRows(cyclopsData) < 1 ||
            getNumberOfStrata(cyclopsData) < 1 ||
            getNumberOfCovariates(cyclopsData) < 1) {
        stop("Data are incompletely loaded")
    }

    .checkInterface(cyclopsData, computeDevice =  computeDevice, forceNewObject = forceNewObject)

    # Set up prior
    stopifnot(inherits(prior, "cyclopsPrior"))

    if (!is.null(prior$setHook)) {

        prior$setHook(cyclopsData, warnings) # Call-back

    } else {
        prior$exclude <- .checkCovariates(cyclopsData, prior$exclude)

        if (!is.null(prior$neighborhood)) {
            prior$neighborhood <- lapply(prior$neighborhood,
                                         function(element) {
                                             list(.checkCovariates(cyclopsData, element[[1]]),
                                                  .checkCovariates(cyclopsData, element[[2]]))
                                         })
        }

        if (prior$priorType[1] != "none" &&
            is.null(prior$graph) && # TODO Ignore hierarchical models for now
            .cyclopsGetHasIntercept(cyclopsData) &&
            !prior$forceIntercept) {
            interceptId <- bit64::as.integer64(.cyclopsGetInterceptLabel(cyclopsData))
            warn <- FALSE
            if (is.null(prior$exclude)) {
                prior$exclude <- c(interceptId)
                warn <- TRUE
            } else {
                if (!(interceptId %in% prior$exclude)) {
                    prior$exclude <- c(interceptId, prior$exclude)
                    warn <- TRUE
                }
            }
            if (warn && warnings) {
                warning("Excluding intercept from regularization")
            }
        }

        if (is.null(prior$graph)) {
            graph <- NULL
        } else {
            graph <- .makeHierarchyGraph(cyclopsData, prior$graph)
            if (length(prior$priorType) != length(prior$variance)) {
                stop("Prior types and variances have a dimensionality mismatch")
            }
            if (any(prior$priorType != "normal")) {
                stop("Only normal-normal hierarchies are currently supported")
            }
        }

        if (is.null(prior$neighborhood)) {
            neighborhood <- NULL
        } else {
            neighborhood <- prior$neighborhood
            if (length(prior$priorType) != length(prior$variance)) {
                stop("Prior types and variances have a dimensionality mismatch");
            }
            if (any(prior$priorType  != "laplace")) {
                stop("Only Laplace-Laplace fused neighborhoods are currently supported")
            }
        }

        if (is.null(graph) && is.null(neighborhood) && length(prior$priorType) > 1) {
            if (length(prior$priorType) != getNumberOfCovariates(cyclopsData)) {
                stop("Length of priors must equal the number of covariates")
            }
        }

        if (any(prior$priorType == "jeffreys")) {
            if (Cyclops::getNumberOfCovariates(cyclopsData) > 1) {
                stop("Jeffreys prior is currently only implemented for 1 covariate")
            }

            covariate <- Cyclops::getCovariateIds(cyclopsData)
            if (Cyclops::getCovariateTypes(cyclopsData, covariate) != "indicator") {
                count <- reduce(cyclopsData, covariate, power = 0)
                sum <- reduce(cyclopsData, covariate, power = 1)
                mean <- sum / count
                if (!(mean == 0.0 || mean == 1.0)) {
                    stop("Jeffreys prior is currently only implemented for indicator covariates")
                }
            }
        }

        .cyclopsSetPrior(cyclopsData$cyclopsInterfacePtr, prior$priorType, prior$variance,
                         prior$exclude, graph, neighborhood)
    }

    if (control$selectorType == "auto") {
        if (cyclopsData$modelType %in% c("pr", "lr")) {
            control$selectorType <- "byRow"
        } else {
            rowsPerStratum <- (getNumberOfRows(cyclopsData) / getNumberOfStrata(cyclopsData))
            if (rowsPerStratum < getNumberOfStrata(cyclopsData)) {
                control$selectorType <- "byPid"
            } else {
                control$selectorType <- "byRow"
            }
        }
        if (prior$useCrossValidation && control$noiseLevel != "silent") {
            writeLines(paste("Using cross-validation selector type", control$selectorType))
        }
    }

    if (control$cvRepetitions == "auto") {
        control$cvRepetitions <- .getNumberOfRepetitions(getNumberOfRows(cyclopsData))
    }

    control <- .setControl(cyclopsData$cyclopsInterfacePtr, control)
    threads <- control$threads

    if (!is.null(startingCoefficients)) {

        if (length(startingCoefficients) != getNumberOfCovariates(cyclopsData)) {
            stop("Must provide a value for each coefficient")
        }

        if (.cyclopsGetHasOffset(cyclopsData)) {
            startingCoefficients <- c(1.0, startingCoefficients)
        }

        .cyclopsSetBeta(cyclopsData$cyclopsInterfacePtr, startingCoefficients)
        .cyclopsSetStartingBeta(cyclopsData$cyclopsInterfacePtr, startingCoefficients)
    }

    if (!is.null(fixedCoefficients)) {
        if (length(fixedCoefficients) != getNumberOfCovariates(cyclopsData)) {
            stop("Must provide a boolean for each coefficient")
        }

        offset <- ifelse(.cyclopsGetHasOffset(cyclopsData), 1, 0)
        for (i in 1:length(fixedCoefficients)) {
            .cyclopsSetFixedBeta(cyclopsData$cyclopsInterfacePtr, offset + i, fixedCoefficients[i] == TRUE)
        }
    }

    if (.cyclopsGetHasIntercept(cyclopsData) &&
        is.null(prior$fitHook) &&
        is.null(startingCoefficients)) {
        yMean <- mean(getYVector(cyclopsData))
        if (cyclopsData$modelType == "lr") {
            interceptValue <- log(yMean / (1 - yMean))
        } else if (cyclopsData$modelType == "pr") {
            interceptValue <- log(yMean)
        } else if (cyclopsData$modelType == "ls") {
            interceptValue <- yMean
        }
        betas <- c(interceptValue, rep(0.0, getNumberOfCovariates(cyclopsData) - 1))
        if (.cyclopsGetHasOffset(cyclopsData)) {
            betas <- c(1.0, betas)
        }
        .cyclopsSetBeta(cyclopsData$cyclopsInterfacePtr, betas)
        .cyclopsSetStartingBeta(cyclopsData$cyclopsInterfacePtr, betas)
    }

    # Handle weights

    weightsUnsorted <- TRUE
    if (!is.null(cyclopsData$weights)) {
        if (!is.null(weights)) {
            warning("Using weights passed to fitCyclopsModel()")
        } else {
            weights <- cyclopsData$weights
            weightsUnsorted <- FALSE
        }
    }
    if (!is.null(weights)) {
        if (length(weights) != getNumberOfRows(cyclopsData)) {
            stop("Must provide a weight for each data row")
        }
        if (any(weights < 0)) {
            stop("Only non-negative weights are allowed")
        }

        if (weightsUnsorted) {
            if (!is.null(cyclopsData$sortOrder)) {
                weights <- weights[cyclopsData$sortOrder]
            }
        }

        .cyclopsSetWeights(cyclopsData$cyclopsInterfacePtr, weights)
    }

    # censorWeight check for the Fine-Gray model
    if (cyclopsData$modelType == "fgr" & is.null(cyclopsData$censorWeights)) {
        stop("Subject-specific censoring weights must be specified for modelType = 'fgr'.")
    }

    if (!is.null(cyclopsData$censorWeights)) {
        if (cyclopsData$modelType != 'fgr' && warnings) {
            warning(paste0("modelType = '", cyclopsData$modelType, "' does not use censorWeights. These weights will not be passed further."))
        }
        if (length(cyclopsData$censorWeights) != getNumberOfRows(cyclopsData)) {
            stop("Must provide a censorWeight for each data row")
        }
        if (any(cyclopsData$censorWeights < 0) || any(cyclopsData$censorWeights > 1)) {
            stop("Only weights between 0 and 1 are allowed for censorWeights")
        }
        .cyclopsSetCensorWeights(cyclopsData$cyclopsInterfacePtr, cyclopsData$censorWeights)
    }

    if (prior$useCrossValidation) {
        minCVData <- control$minCVData
        if (control$selectorType == "byRow" && minCVData > getNumberOfRows(cyclopsData)) {
            stop("Insufficient data count for cross validation")
        }
        if (control$selectorType == "byPid" && minCVData > getNumberOfStrata(cyclopsData)) {
            stop("Insufficient data count for cross validation")
        }

        fit <- .cyclopsRunCrossValidation(cyclopsData$cyclopsInterfacePtr)
    } else {
        fit <- .cyclopsFitModel(cyclopsData$cyclopsInterfacePtr)
    }

    if (fit$return_flag == "POOR_BLR_STEP" && control$convergenceType == "gradient") {

        if (warnings) {
            warning("BLR convergence criterion failed; coefficient may be infinite")
        }

        control$convergenceType <- "lange"
        return(fitCyclopsModel(cyclopsData = cyclopsData,
                               prior = prior,
                               control = control,
                               weights = weights,
                               forceNewObject = forceNewObject,
                               returnEstimates = returnEstimates,
                               startingCoefficients = startingCoefficients,
                               fixedCoefficients = fixedCoefficients,
                               computeDevice = computeDevice))
    }

    if (returnEstimates) {
        estimates <- .cyclopsLogModel(cyclopsData$cyclopsInterfacePtr)
        fit <- c(fit, estimates)
        fit$estimation <- as.data.frame(fit$estimation)
    }
    fit$call <- cl
    fit$cyclopsData <- cyclopsData
    fit$coefficientNames <- cyclopsData$coefficientNames
    if (!is.null(fixedCoefficients)) {
        fit$fixedCoefficients <- fixedCoefficients
    }
    fit$rowNames <- cyclopsData$rowNames
    fit$scale <- cyclopsData$scale
    fit$threads <- threads
    fit$seed <- control$seed

    if (prior$useCrossValidation) {
        fit$cvRepetitions <- control$cvRepetitions
    }

    class(fit) <- "cyclopsFit"
    return(fit)
}

#' @title Get full path to installed Cyclops library
#'
#' @description
#' \code{cyclopsLibraryFileName} returns the full path to the installed Cyclops .so/.dll file
#'
#' @return String
cyclopsLibraryFileName <- function() {
    normalizePath(system.file("libs", .Platform$r_arch,
                              paste0("Cyclops", .Platform$dynlib.ext),
                              package = "Cyclops"))
}

#' @title Place Cyclops model fit object in a persistent cache
#'
#' @description
#' \code{cacheCyclopsModelForJava} places a Cyclops model fit object into a persistent cache,
#' so that the object may be used across language-barriers.
#'
#' @param object    Cyclops model fit object
#'
#' @return Integer index of object in persistent cache
#'
#' @export
cacheCyclopsModelForJava <- function(object) {
    .checkInterface(object$cyclopsData, testOnly = TRUE)
    instance <- attr(object, "jniCache")

    if (is.null(instance)) {
        instance <- .cyclopsCacheForJava(object$cyclopsData$cyclopsInterfacePtr)
        bit::setattr(object, "jniCache", instance)
    }

    return(instance)
}

#' @title Clear the Cyclops persistent cache
#'
#' @description
#' \code{clearCyclopsModelCache} clears the persistent cache holding
#' Cyclops model fit objects that may be used across language-barriers.
#'
#' @export
clearCyclopsModelCache <- function() {
    stop("Not yet implemented")
}

.checkCovariates <- function(cyclopsData, covariates) {
    if (!is.null(covariates)) {
        saved <- covariates

        indices <- NULL

        if (inherits(covariates, "character")) {
            # Try to match names
            indices <- match(covariates, cyclopsData$coefficientNames)
            covariates <- getCovariateIds(cyclopsData)[indices]
        }

        if (!bit64::is.integer64(covariates)) {
            covariates <- bit64::as.integer64(covariates)
        }

        if (any(is.na(covariates))) {
            stop("Unable to match all covariates: ", paste(saved, collapse = ", "))
        }

        attr(covariates, "indices") <- indices
    }
    covariates
}

.checkData <- function(x) {
    # Check conditions
    if (missing(x) || is.null(x$cyclopsDataPtr) || !inherits(x$cyclopsDataPtr, "externalptr")) {
        stop("Improperly constructed cyclopsData object")
    }
    if (.isRcppPtrNull(x$cyclopsDataPtr)) {
        stop("Data object is no longer initialized")
    }
}

.checkInterface <- function(x, computeDevice = "native", forceNewObject = FALSE, testOnly = FALSE) {
    if (forceNewObject
        || is.null(x$cyclopsInterfacePtr)
        || !inherits(x$cyclopsInterfacePtr, "externalptr")
        || .isRcppPtrNull(x$cyclopsInterfacePtr)
        #|| .cyclopsGetComputeDevice(x$cyclopsInterfacePtr) != computeDevice TODO is this necessary?
    ) {

        if (testOnly == TRUE) {
            stop("Interface object is not initialized")
        }

        if (computeDevice != "native") {
            stopifnot(computeDevice %in% listGPUDevices())
        }

        # Build interface
        interface <- .cyclopsInitializeModel(x$cyclopsDataPtr, modelType = x$modelType, computeDevice, computeMLE = TRUE)
        # TODO Check for errors
        assign("cyclopsInterfacePtr", interface$interface, x)
    }
}



#' @title Extract model coefficients
#'
#' @description
#' \code{coef.cyclopsFit} extracts model coefficients from an Cyclops model fit object
#'
#' @param object    Cyclops model fit object
#' @param rescale   Boolean: rescale coefficients for unnormalized covariate values
#' @param ignoreConvergence Boolean: return coefficients even if fit object did not converge
#' @param ...       Other arguments
#'
#' @return Named numeric vector of model coefficients.
#'
#' @export
coef.cyclopsFit <- function(object, rescale = FALSE, ignoreConvergence = FALSE, ...) {

    if (object$return_flag != "SUCCESS" && !ignoreConvergence) {
        stop("Cyclops estimation is null; suspect that estimation did not converge.")
    }

    result <- object$estimation$estimate
    if (is.null(object$coefficientNames)) {
        names(result) <- object$estimation$column_label
        if ("0" %in% names(result)) {
            names(result)[which(names(result) == "0")] <- "(Intercept)"
        }
    } else {
        names(result) <- object$coefficientNames
    }

    if (!is.null(object$scale) && rescale) {
        result <- result * object$scale
    }
    result
}

#' @title Get hyperparameter
#'
#' @description
#' \code{getHyperParameter} returns the current hyper parameter in a Cyclops model fit object
#'
#' @param object    A Cyclops model fit object
#'
#' @template elaborateExample
#'
#' @export
getHyperParameter <- function(object) {
    if (inherits(object, "cyclopsFit")) {
        object$variance
    } else {
        NULL
    }
}

#' @title Extract log-likelihood
#'
#' @description
#' \code{logLik} returns the current log-likelihood of the fit in a Cyclops model fit object
#'
#' @param object    A Cyclops model fit object
#' @param ...       Additional arguments
#'
#' @template elaborateExample
#'
#' @export
logLik.cyclopsFit <- function(object, ...) {
    out <- object$log_likelihood
    attr(out, 'df') <- sum(!is.na(coefficients(object)))
    attr(out, 'nobs') <- getNumberOfRows(object$cyclopsData)
    class(out) <- 'logLik'
    out
}

#' @title Extract gradient
#'
#' @description
#' \code{gradient} returns the current gradient wrt the regression parameters of
#' the log-likelihood of the fit in a Cyclops model fit object
#'
#' @param object    A Cyclops model fit object
#'
#' @export
gradient <- function(object) {

    .checkInterface(object$cyclopsData, testOnly = TRUE)
    gradient <- .cyclopsGetLogLikelihoodGradient(object$interface)
    names(gradient) <- names(coef(object))

    return(gradient)
}


#' @method print cyclopsFit
#' @title Print a Cyclops model fit object
#'
#' @description
#' \code{print.cyclopsFit} displays information about a Cyclops model fit object
#'
#' @param x    A Cyclops model fit object
#' @param show.call Logical: display last call to update the Cyclops model fit object
#' @param ...   Additional arguments
#'
#' @export
print.cyclopsFit <- function(x, show.call=TRUE ,...) {
    cat("Cyclops model fit object\n\n")

    if (show.call && !is.null(x$call)) {
        cat("Call: ",paste(deparse(x$call),sep="\n",collapse="\n"),"\n\n",sep="")
    }
    cat("           Model: ", x$cyclopsData$modelType, "\n", sep="")
    cat("           Prior: ", x$prior_info, "\n", sep="")
    cat("  Hyperparameter: ", paste(x$variance, collapse=" "), "\n", sep="")
    cat("     Return flag: ", x$return_flag, "\n", sep="")
    if (x$return_flag == "SUCCESS") {
        cat("Log likelikehood: ", x$log_likelihood, "\n", sep="")
        cat("       Log prior: ", x$log_prior, "\n", sep="")
    }
    invisible(x)
}

#' @title Create a Cyclops control object
#'
#' @description
#' \code{createControl} creates a Cyclops control object for use with \code{\link{fitCyclopsModel}}.
#'
#' @param maxIterations			Integer: maximum iterations of Cyclops to attempt before returning a failed-to-converge error
#' @param tolerance					Numeric: maximum relative change in convergence criterion from successive iterations to achieve convergence
#' @param convergenceType		String: name of convergence criterion to employ (described in more detail below)
#' @param cvType						String: name of cross validation search.
#' 													Option \code{"auto"} selects an auto-search following BBR.
#' 													Option \code{"grid"} selects a grid-search cross validation
#' @param fold							Numeric: Number of random folds to employ in cross validation
#' @param lowerLimit				Numeric: Lower prior variance limit for grid-search
#' @param upperLimit				Numeric: Upper prior variance limit for grid-search
#' @param gridSteps					Numeric: Number of steps in grid-search
#' @param cvRepetitions			Numeric: Number of repetitions of X-fold cross validation
#' @param minCVData					Numeric: Minimum number of data for cross validation
#' @param noiseLevel				String: level of Cyclops screen output (\code{"silent"}, \code{"quiet"}, \code{"noisy"})
#' @param threads               Numeric: Specify number of CPU threads to employ in cross-validation; default = 1 (auto = -1)
#' @param seed                  Numeric: Specify random number generator seed. A null value sets seed via \code{\link{Sys.time}}.
#' @param resetCoefficients     Logical: Reset all coefficients to 0 between model fits under cross-validation
#' @param startingVariance      Numeric: Starting variance for auto-search cross-validation; default = -1 (use estimate based on data)
#' @param useKKTSwindle Logical: Use the Karush-Kuhn-Tucker conditions to limit search
#' @param tuneSwindle    Numeric: Size multiplier for active set
#' @param selectorType  String: name of exchangeable sampling unit.
#'                              Option \code{"byPid"} selects entire strata.
#'                              Option \code{"byRow"} selects single rows.
#'                              If set to \code{"auto"}, \code{"byRow"} will be used for all models except conditional models where
#'                              the average number of rows per stratum is smaller than the number of strata.
#' @param initialBound          Numeric: Starting trust-region size
#' @param maxBoundCount         Numeric: Maximum number of tries to decrease initial trust-region size
#' @param algorithm             String: name of fitting algorithm to employ; default is `ccd`
#' @param doItAll               Currently unused
#' @param syncCV                Currently unused
#'
#' Todo: Describe convegence types
#'
#' @return
#' A Cyclops control object of class inheriting from \code{"cyclopsControl"} for use with \code{\link{fitCyclopsModel}}.
#'
#' @template elaborateExample
#'
#' @export
createControl <- function(maxIterations = 1000,
                          tolerance = 1E-6,
                          convergenceType = "gradient",
                          cvType = "auto",
                          fold = 10,
                          lowerLimit = 0.01,
                          upperLimit = 20.0,
                          gridSteps = 10,
                          cvRepetitions = 1,
                          minCVData = 100,
                          noiseLevel = "silent",
                          threads = 1,
                          seed = NULL,
                          resetCoefficients = FALSE,
                          startingVariance = -1,
                          useKKTSwindle = FALSE,
                          tuneSwindle = 10,
                          selectorType = "auto",
                          initialBound = 2.0,
                          maxBoundCount = 5,
                          algorithm = "ccd",
                          doItAll = TRUE,
                          syncCV = FALSE) {
    validCVNames = c("grid", "auto")
    stopifnot(cvType %in% validCVNames)

    validNLNames = c("silent", "quiet", "noisy")
    stopifnot(noiseLevel %in% validNLNames)
    stopifnot(threads == -1 || threads >= 1)
    stopifnot(startingVariance == -1 || startingVariance > 0)
    stopifnot(selectorType %in% c("auto","byPid", "byRow"))

    validAlgorithmNames = c("ccd", "mm")
    stopifnot(algorithm %in% validAlgorithmNames)

    structure(list(maxIterations = maxIterations,
                   tolerance = tolerance,
                   convergenceType = convergenceType,
                   autoSearch = (cvType == "auto"),
                   fold = fold,
                   lowerLimit = lowerLimit,
                   upperLimit = upperLimit,
                   gridSteps = gridSteps,
                   minCVData = minCVData,
                   cvRepetitions = cvRepetitions,
                   noiseLevel = noiseLevel,
                   threads = threads,
                   seed = seed,
                   resetCoefficients = resetCoefficients,
                   startingVariance = startingVariance,
                   useKKTSwindle = useKKTSwindle,
                   tuneSwindle = tuneSwindle,
                   selectorType = selectorType,
                   initialBound = initialBound,
                   maxBoundCount = maxBoundCount,
                   algorithm = algorithm,
                   doItAll = doItAll,
                   syncCV = syncCV),
              class = "cyclopsControl")
}

#' @title Create a Cyclops prior object
#'
#' @description
#' \code{createPrior} creates a Cyclops prior object for use with \code{\link{fitCyclopsModel}}.
#'
#' @param priorType     Character: specifies prior distribution.  See below for options
#' @param variance      Numeric: prior distribution variance
#' @param exclude       A vector of numbers or covariateId names to exclude from prior
#' @param graph         Child-to-parent mapping for a hierarchical prior
#' @param neighborhood  A list of first-order neighborhoods for a partially fused prior
#' @param useCrossValidation    Logical: Perform cross-validation to determine prior \code{variance}.
#' @param forceIntercept  Logical: Force intercept coefficient into prior
#'
#' @section Prior types:
#'
#' We specify all priors in terms of their variance parameters.
#' Similar fitting tools for regularized regression often parameterize the Laplace distribution
#' in terms of a rate \code{"lambda"} per observation.
#' See \code{"glmnet"}, for example.
#'
#' variance = 2 * / (nobs * lambda)^2 or lambda = sqrt(2 / variance) / nobs
#'
#' @template elaborateExample
#'
#' @return
#' A Cyclops prior object of class inheriting from \code{"cyclopsPrior"} for use with \code{fitCyclopsModel}.
#'
#' @export
createPrior <- function(priorType,
                        variance = 1,
                        exclude = c(),
                        graph = NULL,
                        neighborhood = NULL,
                        useCrossValidation = FALSE,
                        forceIntercept = FALSE) {
    validNames = c("none", "laplace","normal", "barupdate", "hierarchical", "jeffreys")
    stopifnot(priorType %in% validNames)
    if (!is.null(exclude)) {
        if (!inherits(exclude, "character") &&
                !inherits(exclude, "numeric") &&
                !inherits(exclude, "integer")
        ) {
            stop(cat("Unable to parse excluded covariates:"), exclude)
        }
    }

    if (length(priorType) != length(variance)) {
        stop("Prior types and variances have a dimensionality mismatch")
    }

    if (all(priorType == "none") && useCrossValidation) {
        stop("Cannot perform cross validation with a flat prior")
    }
    if (any(priorType == "barupdate") && useCrossValidation) {
        stop("Cannot perform cross valudation with BAR updates")
    }
    if (any(priorType == "hierarchical") && missing(graph)) {
        stop("Must provide a graph for a hierarchical prior")
    }
    if (!is.null(neighborhood)) {
        allNames <- unlist(neighborhood)
        if (!inherits(allNames, "character") &&
            !inherits(allNames, "numeric") &&
            !inherits(allNames, "integer")) {
            stop(cat("Unable to parse neighborhood covariates:"), allNames)
        }
    }
    structure(list(priorType = priorType, variance = variance, exclude = exclude,
                   graph = graph,
                   neighborhood = neighborhood,
                   useCrossValidation = useCrossValidation, forceIntercept = forceIntercept),
              class = "cyclopsPrior")
}

# .cyclopsSetCoefficients <- function(object, coefficients) {
#     .checkInterface(object$cyclopsData, testOnly = TRUE)
#
#     if (length(coefficients) != getNumberOfCovariates(object$cyclopsData)) {
#         stop("Must provide a value for each coefficient")
#     }
#
#     if (.cyclopsGetHasOffset(object$cyclopsData)) {
#         coefficients <- c(1.0, coefficients)
#     }
#
#     .cyclopsSetBeta(object$cyclopsData$cyclopsInterfacePtr, coefficients)
# }

#' @title Compute predictive log-likelihood from a Cyclops model fit
#'
#' @description
#' \code{getCyclopsPredictiveLogLikelihood} returns the log-likelihood of a subset of the data in a Cyclops model fit object.
#'
#' @param object    A Cyclops model fit object
#' @param weights   Numeric vector: vector of 0/1 identifying subset (=1) of rows from \code{object} to use in computing the log-likelihood
#' @return The predictive log-likelihood
#'
#' @keywords internal
getCyclopsPredictiveLogLikelihood <- function(object, weights) {
    .checkInterface(object$cyclopsData, testOnly = TRUE)

    if (length(weights) != getNumberOfRows(object$cyclopsData)) {
        stop("Must provide a weight for each data row")
    }
    if (any(weights < 0)) {
        stop("Only non-negative weights are allowed")
    }

    if(!is.null(object$cyclopsData$sortOrder)) {
        weights <- weights[object$cyclopsData$sortOrder]
    }
    # TODO Remove code duplication with weights section of fitCyclopsModel

    .cyclopsGetNewPredictiveLogLikelihood(object$cyclopsData$cyclopsInterfacePtr, weights)
}

#' @title Get cross-validation information from a Cyclops model fit
#'
#' @description {getCrossValidationInfo} returns the predicted optimal cross-validation point and ordinate
#'
#' @param object A Cyclops model fit object
#'
#' @keywords internal
getCrossValidationInfo <- function(object) {
    info <- object$cross_validation

    if (is.na(info) || info == "") {
        stop("No cross-validation information is available")
    }

    values <- as.numeric(unlist(strsplit(info, " ")))
    list(ordinate = values[1],
         point = values[-1])
}

.setControl <- function(cyclopsInterfacePtr, control) {
    if (!missing(control)) {
        stopifnot(inherits(control, "cyclopsControl"))

        if (is.null(control$seed)) {
            control$seed <- as.integer(Sys.time())
        }

        if (is.null(control$algorithm) || is.na(control$algorithm)) { # Provide backwards compatibility
            control$algorithm <- "ccd"
        }

        .cyclopsSetControl(cyclopsInterfacePtr, control$maxIterations, control$tolerance,
                           control$convergenceType, control$autoSearch, control$fold,
                           (control$fold * control$cvRepetitions),
                           control$lowerLimit, control$upperLimit, control$gridSteps,
                           control$noiseLevel, control$threads, control$seed, control$resetCoefficients,
                           control$startingVariance, control$useKKTSwindle, control$tuneSwindle,
                           control$selectorType, control$initialBound, control$maxBoundCount,
                           control$algorithm, control$doItAll, control$syncCV
                          )
        return(control)
    }

    return(NULL)
}

#' @title Extract standard errors
#'
#' @description
#' \code{getSEs} extracts asymptotic standard errors for specific covariates from a Cyclops model fit object.
#'
#' @details This function first computes the (partial) Fisher information matrix for
#' just the requested covariates and then returns the square root of the diagonal elements of
#' the inverse of the Fisher information matrix.  These are the asymptotic standard errors
#' when all possible covariates are included.
#' When the requested covariates do not equate to all coefficients in the model,
#' then interpretation is more challenging.
#'
#' @param object    A Cyclops model fit object
#' @param covariates    Integer or string vector: list of covariates for which asymptotic standard errors are wanted
#'
#' @return Vector of standard error estimates
#'
#' @keywords internal
getSEs <- function(object, covariates) {
    .checkInterface(object$cyclopsData, testOnly = TRUE)
    covariates <- .checkCovariates(object$cyclopsData, covariates)
    if (getNumberOfCovariates(object$cyclopsData) != length(covariates)) {
        warning("Asymptotic standard errors are only valid if computed for all covariates simultaneously")
    }

    fisherInformation <- .cyclopsGetFisherInformation(object$cyclopsData$cyclopsInterfacePtr, covariates)
    ses <- sqrt(diag(solve(fisherInformation)))
    names(ses) <- object$coefficientNames[as.integer(covariates)]
    ses
}

#' @title Run bootstrap for Cyclops model parameters
#'
#' @param object          A fitted Cyclops model object
#' @param replicates      Numeric: number of bootstrap samples
#' @param ciCoverage      Numeric (0-1): coverage proportion for bootstrap confidence intervals
#' @param na.rm           Logical: strip out bootstrap replicates with NA values (most likely without a finite estimate)
#'
#' @export
runBootstrap <- function(object, replicates, ciCoverage = 0.95, na.rm = FALSE) {
    .checkInterface(object$cyclopsData, testOnly = TRUE)
    bs <- .cyclopsRunBootstrap(object$cyclopsData$cyclopsInterfacePtr, "", "1", replicates) # TODO Remove unneeded arguments

    if (.cyclopsGetHasOffset(object$cyclopsData)) {
        bs$samples <- bs$samples[(replicates + 1):length(bs$samples)]
    }

    bs$samples <- matrix(bs$samples, nrow = replicates)

    colnames(bs$samples) <- names(coef(object))
    bs$samples <- as_tibble(bs$samples)

    if (na.rm) {
        bs$samples <- bs$samples %>% tidyr::drop_na()
    }

    epsilon <- 1 - ciCoverage

    bs$summary <- data.frame(
        original = coef(object),
        bias = sapply(bs$samples, mean) - coef(object),
        std_err = sapply(bs$samples, function(x) { sqrt(stats::var(x)) }),
        bpi_lower = sapply(bs$samples, stats::quantile, probs = epsilon / 2),
        bpi_upper = sapply(bs$samples, stats::quantile, probs = 1 - epsilon / 2)
    )

    return(bs)
}

#' @title Confidence intervals for Cyclops model parameters
#'
#' @description
#' \code{confinit.cyclopsFit} profiles the data likelihood to construct confidence intervals of
#' arbitrary level. Usually it only makes sense to do this for variables that have not been regularized.
#'
#' @param object    A fitted Cyclops model object
#' @param parm      A specification of which parameters require confidence intervals,
#'                  either a vector of numbers of covariateId names
#' @param level     Numeric: confidence level required
## @param control   A Cyclops \code{\link{control}} object
#' @param overrideNoRegularization   Logical: Enable confidence interval estimation for regularized parameters
#' @param includePenalty    Logical: Include regularized covariate penalty in profile
#' @param rescale   Boolean: rescale coefficients for unnormalized covariate values
#' @param maximumCurvature Numeric: maximum curvature of log-likelihood profile required to compute CI
#' @param ... Additional argument(s) for methods
#'
#' @return
#' A matrix with columns reporting lower and upper confidence limits for each parameter.
#' These columns are labelled as (1-level) / 2 and 1 - (1 - level) / 2 in percent
#' (by default 2.5 percent and 97.5 percent)
#'
#' @template elaborateExample
#'
#' @export
confint.cyclopsFit <- function(object, parm, level = 0.95, #control,
                               overrideNoRegularization = FALSE,
                               includePenalty = TRUE,
                               rescale = FALSE,
                               maximumCurvature = -1E-3,
                               ...) {
    .checkInterface(object$cyclopsData, testOnly = TRUE)
    #.setControl(object$cyclopsData$cyclopsInterfacePtr, control)

    savedParm <- parm
    parm <- .checkCovariates(object$cyclopsData, parm)
    if (level < 0.01 || level > 0.99) {
        stop("level must be between 0 and 1")
    }
    threshold <- qchisq(level, df = 1) / 2
    threads <- object$threads

    if (!is.null(object$fixedCoefficients)) {
        if (any(object$fixedCoefficients[as.integer(parm)])) {
            stop("Cannot estimate confidence interval for a fixed coefficient")
        }
    }

    hessianDiagonal <- .cyclopsGetLogLikelihoodHessianDiagonal(object$cyclopsData$cyclopsInterfacePtr,
                                                               parm)

    convergenceType <- .cyclopsGetConvergenceType(object$cyclopsData$cyclopsInterfacePtr)
    prof <- NULL
    if (convergenceType != 1 && any(hessianDiagonal >= maximumCurvature)) {

        .cyclopsSetConvergenceType(object$interface, "lange")
        newFit <- .cyclopsFitModel(object$cyclopsData$cyclopsInterfacePtr)
        newHessianDiagonal <- .cyclopsGetLogLikelihoodHessianDiagonal(object$cyclopsData$cyclopsInterfacePtr,
                                                                      parm)

        if (isTRUE(any(newHessianDiagonal > 0))) {
            warning("Cannot estimate confidence interval for positive-curvature log-likelihood function")

            prof <- data.frame(covariate = parm, # was 'savedParm' but unsure why
                               lower = rep(NA, length(parm)),
                               upper = rep(NA, length(parm)),
                               evaluations = rep(NA, length(parm)))
        }
    }

    if (is.null(prof)) {

        prof <- .cyclopsProfileModel(object$cyclopsData$cyclopsInterfacePtr, parm,
                                     threads, threshold,
                                     overrideNoRegularization,
                                     includePenalty)
    }

    indices <- match(parm, getCovariateIds(object$cyclopsData))

    if (!is.null(object$scale) && rescale) {
        prof$lower <- prof$lower * object$scale[indices]
        prof$upper <- prof$upper * object$scale[indices]
    }
    prof <- as.matrix(as.data.frame(prof))
    rownames(prof) <- object$coefficientNames[indices]
    qs <- c((1 - level) / 2, 1 - (1 - level) / 2) * 100
    colnames(prof)[2:3] <- paste(sprintf("%.1f", qs), "%")

    # Change NaN to NA
    prof[which(is.nan(prof[, 2])), 2] <- NA
    prof[which(is.nan(prof[, 3])), 3] <- NA

    prof
}

.initAdaptiveProfile <- function(object, parm, bounds, includePenalty, optimalWarmStart, returnDerivatives) {
    # If an MLE was found, let's not throw that bit of important information away:
    if (object$return_flag == "SUCCESS" &&
        coef(object)[as.character(parm)] > bounds[1] &&
        coef(object)[as.character(parm)] < bounds[2]) {
        eval <- fixedGridProfileLogLikelihood(object, parm, coef(object)[as.character(parm)], includePenalty,
        									  optimalWarmStart, returnDerivatives)
        point <- coef(object)[as.character(parm)]
        if (returnDerivatives) {
            profile <- tibble(point = point,
                              value = eval$value,
                              derivative = eval$derivative)
        } else {
            profile <- tibble(point = point,
                              value = eval$value)
        }
    } else {
        profile <- tibble()
    }
}

#' @title Profile likelihood for Cyclops model parameters
#'
#' @description
#' \code{getCyclopsProfileLogLikelihood} evaluates the profile likelihood at a grid of parameter values.
#'
#' @param object    Fitted Cyclops model object
#' @param parm      Specification of which parameter requires profiling,
#'                  either a vector of numbers or covariateId names
#' @param x         Vector of values of the parameter
#' @param bounds    Pair of values to bound adaptive profiling
#' @param tolerance Absolute tolerance allowed for adaptive profiling
#' @param initialGridSize Initial grid size for adaptive profiling
#' @param maxResets Maximum allowed number of recomputing the likelihood when coefficient drift is
#'                  detected.
#' @param includePenalty    Logical: Include regularized covariate penalty in profile
#' @param optimalWarmStart Logical: Use optimal warm-starting when parallelizing evaluations
#' @param returnDerivatives Return derivative of log-likelihood wrt the parameter `param` at each evaluation point
#'
#' @return
#' A data frame containing the profile log likelihood. Returns NULL when the adaptive profiling fails
#' to converge.
#'
#' @export
getCyclopsProfileLogLikelihood <- function(object,
                                           parm,
                                           x = NULL,
                                           bounds = NULL,
                                           tolerance = 1E-3,
                                           initialGridSize = 10,
                                           maxResets = 10,
                                           includePenalty = TRUE,
                                           optimalWarmStart = TRUE,
                                           returnDerivatives = FALSE) {

    if (!xor(is.null(x), is.null(bounds))) {
        stop("Must provide either `x` or `bounds`, but not both.")
    }

    logLik <- .cyclopsGetLogLikelihood(object$interface)
    if (is.na(logLik)) {
        object <- fitCyclopsModel(
            object$cyclopsData,
            fixedCoefficients = rep(TRUE, getNumberOfCovariates(object$cyclopsData)))
    }

    if (!is.null(bounds)) { # Adaptive profiling using recursive calls

        if (length(bounds) != 2 || bounds[1] >= bounds[2]) {
            stop("Must provide bounds[1] < bounds[2]")
        }
        profile <- .initAdaptiveProfile(object, parm, bounds, includePenalty, optimalWarmStart, returnDerivatives)

        # Start with sparse grid:
        grid <- seq(bounds[1], bounds[2], length.out = initialGridSize)

        # Iterate until stopping criteria met:
        priorMaxMaxError <- Inf
        resetsPerformed <- 0
        while (length(grid) != 0) {
            ll <- fixedGridProfileLogLikelihood(object, parm, grid, includePenalty, optimalWarmStart, returnDerivatives)
            profile <- bind_rows(profile, ll) %>% arrange(.data$point)
            invalid <- is.nan(profile$value) | is.infinite(profile$value)
            if (any(invalid)) {
                if (all(invalid)) {
                    warning("Failing to compute likelihood at entire initial grid.")
                    return(NULL)
                }

                start <- min(which(!invalid))
                end <- max(which(!invalid))
                if (start == end) {
                    warning("Failing to compute likelihood at entire grid except one. Giving up")
                    return(NULL)
                }
                profile <- profile[start:end, ]
                invalid <- invalid[start:end]
                if (any(invalid)) {
                    warning("Failing to compute likelihood in non-extreme regions. Giving up.")
                    return(NULL)
                }
                warning("Failing to compute likelihood at extremes. Truncating bounds.")
            }

            deltaX <- profile$point[2:nrow(profile)] - profile$point[1:(nrow(profile) - 1)]
            deltaY <- profile$value[2:nrow(profile)] - profile$value[1:(nrow(profile) - 1)]
            slopes <- deltaY / deltaX

            if (resetsPerformed < maxResets && !all(slopes[2:length(slopes)] < slopes[1:(length(slopes)-1)])) {
                warning("Coefficient drift detected. Resetting Cyclops object and recomputing all likelihood values computed so far.")
                grid <- profile$point
                profile <- tibble()
                interface <- .cyclopsInitializeModel(object$cyclopsData$cyclopsDataPtr, modelType = object$cyclopsData$modelType, "native", computeMLE = TRUE)
                assign("cyclopsInterfacePtr", interface$interface, object)
                resetsPerformed <- resetsPerformed + 1
                next
            }

            # Compute where prior and posterior slopes intersect
            slopes <- c(slopes[1] + (slopes[2] - slopes[3]),
                        slopes,
                        slopes[length(slopes)] - (slopes[length(slopes) - 1] - slopes[length(slopes)]))

            interceptX <- (profile$value[2:nrow(profile)] -
                               profile$point[2:nrow(profile)] * slopes[3:length(slopes)] -
                               profile$value[1:(nrow(profile) - 1)] +
                               profile$point[1:(nrow(profile) - 1)] * slopes[1:(length(slopes) - 2)]) /
                (slopes[1:(length(slopes) - 2)] - slopes[3:length(slopes)])

            # Compute absolute difference between linear interpolation and worst case scenario (which is at the intercept):
            maxError <- abs((profile$value[1:(nrow(profile) - 1)] + (interceptX - profile$point[1:(nrow(profile) - 1)]) * slopes[1:(length(slopes) - 2)]) -
                                (profile$value[1:(nrow(profile) - 1)] + (interceptX - profile$point[1:(nrow(profile) - 1)]) * slopes[2:(length(slopes) - 1)]))

            maxMaxError <- max(maxError, na.rm = TRUE)
            if (is.na(maxMaxError) || maxMaxError > priorMaxMaxError) {
                warning("Failing to converge when using adaptive profiling.")
                return(NULL)
            }
            priorMaxMaxError <- maxMaxError

            exceed <- which(maxError > tolerance)
            grid <- (profile$point[exceed] + profile$point[exceed + 1]) / 2
        }
    } else { # Use x
        profile <- fixedGridProfileLogLikelihood(object, parm, x, includePenalty, optimalWarmStart, returnDerivatives)
    }

    return(profile)
}

fixedGridProfileLogLikelihood <- function(object, parm, x, includePenalty,
                                          optimalWarmStart = TRUE, returnDerivatives = FALSE) {

    .checkInterface(object$cyclopsData, testOnly = TRUE)
    parm <- .checkCovariates(object$cyclopsData, parm)
    threads <- object$threads

    if (getNumberOfCovariates(object$cyclopsData) == 1 || length(x) == 1) {
        grid <- .cyclopsGetProfileLikelihood(object$cyclopsData$cyclopsInterfacePtr, parm, x,
                                             threads, includePenalty, returnDerivatives)
    } else if (!optimalWarmStart && length(x) / 2 >= threads) {
        grid <- .cyclopsGetProfileLikelihood(object$cyclopsData$cyclopsInterfacePtr, parm, x,
                                             threads, includePenalty, returnDerivatives)
    } else {
        # Partition sequence
        y <- sort(x)
        midPt <- floor(length(x) / 2)
        lower <- y[midPt:1]
        upper <- y[(midPt + 1):length(x)]

        # Execute: TODO chunk and repeat until ill-conditioned
        gridLower <- .cyclopsGetProfileLikelihood(object$cyclopsData$cyclopsInterfacePtr, parm, lower,
                                                  threads, includePenalty, returnDerivatives)
        gridUpper <- .cyclopsGetProfileLikelihood(object$cyclopsData$cyclopsInterfacePtr, parm, upper,
                                                  threads, includePenalty, returnDerivatives)
        # Merge
        grid <- rbind(gridLower, gridUpper)
        grid <- grid[order(grid$point),]
        rownames(grid) <- NULL
    }

    return(grid)
}

#' @title Asymptotic confidence intervals for a fitted Cyclops model object
#'
#' @description
#' \code{aconfinit} constructs confidence intervals of
#' arbitrary level using asymptotic standard error estimates.
#'
#' @param object    A fitted Cyclops model object
#' @param parm      A specification of which parameters require confidence intervals,
#'                  either a vector of numbers of covariateId names
#' @param level     Numeric: confidence level required
#' @param control   A \code{"cyclopsControl"} object constructed by \code{\link{createControl}}
#' @param overrideNoRegularization   Logical: Enable confidence interval estimation for regularized parameters
#' @param ... Additional argument(s) for methods
#'
#' @return
#' A matrix with columns reporting lower and upper confidence limits for each parameter.
#' These columns are labelled as (1-level) / 2 and 1 - (1 - level) / 2 in %
#' (by default 2.5% and 97.5%)
#'
#' @keywords internal
#' @export
aconfint <- function(object, parm, level = 0.95, control,
                     overrideNoRegularization = FALSE, ...) {
    .checkInterface(object$cyclopsData, testOnly = TRUE)
    .setControl(object$cyclopsData$cyclopsInterfacePtr, control)
    cf <- coef(object)
    if (missing(parm)) {
        parm <- names(cf)
    }
    #parm <- .checkCovariates(object$cyclopsData, parm)
    if (level < 0.01 || level > 0.99) {
        stop("level must be between 0 and 1")
    }
    a <- (1 - level) / 2
    a <- c(a, 1 - a)
    pct <- paste(sprintf("%.1f", a * 100), "%")
    fac <- qnorm(a)
    ci <- array(NA, dim = c(length(parm), 2L), dimnames = list(parm, pct))
    ses <- sqrt(diag(vcov(object)))[parm]
    ci[] <- cf[parm] + ses %o% fac
    ci
}

#' @title Calculate variance-covariance matrix for a fitted Cyclops model object
#'
#' @description
#' \code{vcov.cyclopsFit} returns the variance-covariance matrix for all covariates of a Cyclops model object
#'
#' @param object    A fitted Cyclops model object
#' @param control    A \code{"cyclopsControl"} object constructed by \code{\link{createControl}}
#' @param overrideNoRegularization   Logical: Enable variance-covariance estimation for regularized parameters
#' @param ... Additional argument(s) for methods
#'
#' @return
#' A matrix of the estimates covariances between all covariate estimates.
#'
#' @export
vcov.cyclopsFit <- function(object, control, overrideNoRegularization = FALSE, ...) {
    .checkInterface(object$cyclopsData, testOnly = TRUE)
    .setControl(object$cyclopsData$cyclopsInterfacePtr, control)
    fisherInformation <- .cyclopsGetFisherInformation(object$cyclopsData$cyclopsInterfacePtr, NULL)
    if (.cyclopsGetHasOffset(object$cyclopsData)) {
        fisherInformation <- fisherInformation[2:nrow(fisherInformation), 2:ncol(fisherInformation)]
    }
    vcov <- solve(fisherInformation)
    if (!is.null(object$coefficientNames)) {
        rownames(vcov) <- object$coefficientNames
        colnames(vcov) <- object$coefficientNames
    }
    vcov
}

#' @title Convert to Cyclops Prior Variance
#'
#' @description
#' \code{convertToCyclopsVariance} converts the regularization parameter \code{lambda}
#' from \code{glmnet} into a prior variance.
#'
#' @param lambda    Regularization parameter from \code{glmnet}
#' @param nobs      Number of observation rows in dataset
#'
#' @return Prior variance under a Laplace() prior
#'
#' @keywords internal
convertToCyclopsVariance <- function(lambda, nobs) {
    2 / (nobs * lambda)^2
}

#' @title Convert to glmnet regularization parameter
#'
#' @description
#' \code{convertToGlmnetLambda} converts a prior variance
#' from \code{Cyclops} into the regularization parameter \code{lambda}.
#'
#' @param variance  Prior variance
#' @param nobs      Number of observation rows in dataset
#'
#' @return \code{lambda}
#'
#' @keywords internal
convertToGlmnetLambda <- function(variance, nobs) {
    sqrt(2 / variance) / nobs
}

.makeHierarchyGraph <- function(cyclopsData, graph) {
    nTypes <- getNumberOfTypes(cyclopsData)
    if (nTypes < 2 || graph != "type") stop("Only multitype hierarchies are currently supported")
    hasOffset <- .cyclopsGetHasOffset(cyclopsData)
    if (hasOffset) stop("Hierarchies with offset covariates are currently not supported")

    # Build Multitype hierarchy
    nChild <- getNumberOfCovariates(cyclopsData)
    nParents <- nChild / nTypes

    graph <-  lapply(0:(nParents - 1), function(n, types) { 0:(nTypes - 1) + n * nTypes }, types = nTypes)
    graph
}

.getNumberOfRepetitions <- function(nrows) {
    top <- 1E2
    factor <- 0.5
    pmax(pmin(
        ceiling(top / nrows^(factor))
        , 10), 1)
}

library("testthat")

#
# Small Poisson MLE regression
#

test_that("Small Poisson dense regression", {
    dobson <- data.frame(
        counts = c(18,17,15,20,10,20,25,13,12),
        outcome = gl(3,1,9),
        treatment = gl(3,3)
    )
    tolerance <- 1E-4

    glmFit <- glm(counts ~ outcome + treatment, data = dobson, family = poisson()) # gold standard

    dataPtrD <- createCyclopsData(counts ~ outcome + treatment, data = dobson,
                                   modelType = "pr")
    cyclopsFitD <- fitCyclopsModel(dataPtrD,
                           prior = createPrior("none"),
                           control = createControl(noiseLevel = "silent"))
    expect_equal(coef(cyclopsFitD), coef(glmFit), tolerance = tolerance)
    expect_equal(cyclopsFitD$log_likelihood, logLik(glmFit)[[1]], tolerance = tolerance)
    expect_equal(confint(cyclopsFitD, c(1:3))[,2:3], confint(glmFit, c(1:3)), tolerance = tolerance)
    expect_equal(predict(cyclopsFitD), predict(glmFit, type = "response"), tolerance = tolerance)
    expect_equal(confint(cyclopsFitD, c("(Intercept)","outcome3")), confint(cyclopsFitD, c(1,3)))
})

test_that("Small Poisson dense regression in 32bit", {
    dobson <- data.frame(
        counts = c(18,17,15,20,10,20,25,13,12),
        outcome = gl(3,1,9),
        treatment = gl(3,3)
    )
    tolerance <- 1E-4

    gold <- glm(counts ~ outcome + treatment, data = dobson, family = poisson()) # gold standard

    data <- createCyclopsData(counts ~ outcome + treatment, data = dobson,
                                  modelType = "pr", floatingPoint = 32)
    expect_equal(getFloatingPointSize(data), 32)

    fit <- fitCyclopsModel(data,
                           prior = createPrior("none"),
                           control = createControl(noiseLevel = "silent"))

    expect_equal(coef(fit), coef(gold), tolerance = tolerance)

    expect_equal(Cyclops:::.cyclopsGetMeanOffset(data), 0)
})

test_that("Errors for different criteria", {
    dobson <- data.frame(
        counts = c(18,17,15,20,10,20,25,13,12),
        outcome = gl(3,1,9),
        treatment = gl(3,3)
    )
    tolerance <- 1E-3

    gold <- glm(counts ~ outcome + treatment, data = dobson, family = poisson()) # gold standard

    data <- createCyclopsData(counts ~ outcome + treatment, data = dobson,
                              modelType = "pr")

    fit <- fitCyclopsModel(data,
                           prior = createPrior("none"),
                           control = createControl(convergenceType = "mittal",
                                                   noiseLevel = "silent"))

    expect_equal(coef(fit), coef(gold), tolerance = tolerance)
})

test_that("Small Poisson dense regression with offset", {
    dobson <- data.frame(
        counts = c(18,17,15,20,10,20,25,13,12),
        outcome = as.numeric(gl(3,1,9)),
        treatment = gl(3,3)
    )
    tolerance <- 1E-4

    gold <- glm(counts ~  treatment, offset = outcome, data = dobson, family = poisson()) # gold standard

    data <- createCyclopsData(counts ~  treatment, offset = as.numeric(outcome),
                                    data = dobson, modelType = "pr")

    fit <- fitCyclopsModel(data,
                           prior = createPrior("none"),
                           control = createControl(noiseLevel = "silent"))

    expect_equal(coef(fit), coef(gold), tolerance = tolerance)

    expect_error(Cyclops:::.cyclopsGetMeanOffset(fit),
                 "Input must be a cyclopsData object")

    expect_equal(Cyclops:::.cyclopsGetMeanOffset(data), 2)
})

test_that("Small Poisson fixed beta", {
    dobson <- data.frame(
        counts = c(18,17,15,20,10,20,25,13,12),
        outcome = gl(3,1,9),
        treatment = gl(3,3)
    )

    dataPtrD <- createCyclopsData(counts ~ outcome + treatment, data = dobson,
                                  modelType = "pr")

    cyclopsFitD <- fitCyclopsModel(dataPtrD,
                                   startingCoefficients = rep(0, 5),
                                   fixedCoefficients = c(FALSE, TRUE, FALSE, TRUE, FALSE),
                                   prior = createPrior("none"),
                                   control = createControl(noiseLevel = "silent"))
    expect_equivalent(coef(cyclopsFitD)[2], 0)
    expect_equivalent(coef(cyclopsFitD)[4], 0)
})

# test_that("Parallel confint", {
#     dobson <- data.frame(
#         counts = c(18,17,15,20,10,20,25,13,12),
#         outcome = gl(3,1,9),
#         treatment = gl(3,3)
#     )
#     tolerance <- 1E-4
#
#     glmFit <- glm(counts ~ outcome + treatment, data = dobson, family = poisson()) # gold standard
#
#     dataPtrD <- createCyclopsData(counts ~ outcome + treatment, data = dobson,
#                                   modelType = "pr")
#     cyclopsFit1 <- fitCyclopsModel(dataPtrD,
#                                    prior = createPrior("none"),
#                                    control = createControl(noiseLevel = "silent",
#                                                            threads = 1))
#     cyclopsFit2 <- fitCyclopsModel(dataPtrD,
#                                    prior = createPrior("none"),
#                                    control = createControl(noiseLevel = "silent",
#                                                            threads = 2))
#
#     expect_equal(confint(cyclopsFit1, c(1:3)), confint(cyclopsFit2, c(1:3)))
#     ## TODO Check output of confint for "Using 2 thread(s)"
# })

test_that("Specify CI level", {
###function(object, parm, level, ...)
    counts <- c(18,17,15,20,10,20,25,13,12)
    outcome <- gl(3,1,9)
    treatment <- gl(3,3)
    tolerance <- 1E-4

    glmFit <- glm(counts ~ outcome + treatment, family = poisson()) # gold standard

    dataPtr <- createCyclopsData(counts ~ outcome + treatment,
                                   modelType = "pr")
    cyclopsFit <- fitCyclopsModel(dataPtr,
                           prior = createPrior("none"),
                           control = createControl(noiseLevel = "silent"))

    expect_equal(
        confint(cyclopsFit, c(1:3), level = 0.99)[,2:3],
        confint(glmFit, c(1:3), level = 0.99),
        tolerance = tolerance)


})

test_that("Small Poisson indicator regression", {
    counts <- c(18,17,15,20,10,20,25,13,12)
    outcome <- gl(3,1,9)
    treatment <- gl(3,3)
    tolerance <- 1E-4

    glmFit <- glm(counts ~ outcome + treatment, family = poisson()) # gold standard

    dataPtrI <- createCyclopsData(counts ~ outcome, indicatorFormula =  ~ treatment,
                                   modelType = "pr")

    cyclopsFitI <- fitCyclopsModel(dataPtrI,
                           prior = createPrior("none"),
                           control = createControl(noiseLevel = "silent"))
    expect_equal(coef(cyclopsFitI), coef(glmFit), tolerance = tolerance)
    expect_equal(cyclopsFitI$log_likelihood, logLik(glmFit)[[1]], tolerance = tolerance)

    expect_equal(logLik(cyclopsFitI), logLik(glmFit))

    expect_equal(confint(cyclopsFitI, c(1:3))[,2:3], confint(glmFit, c(1:3)), tolerance = tolerance)
    expect_equal(predict(cyclopsFitI), predict(glmFit, type = "response"), tolerance = tolerance)
})

test_that("Small Poisson sparse regression", {
    counts <- c(18,17,15,20,10,20,25,13,12)
    outcome <- gl(3,1,9)
    treatment <- gl(3,3)
    tolerance <- 1E-4

    glmFit <- glm(counts ~ outcome + treatment, family = poisson()) # gold standard

    dataPtrS <- createCyclopsData(counts ~ outcome, sparseFormula =  ~ treatment,
                                   modelType = "pr")
    cyclopsFitS <- fitCyclopsModel(dataPtrS,
                           prior = createPrior("none"),
                           control = createControl(noiseLevel = "silent"))
    expect_equal(coef(cyclopsFitS), coef(glmFit), tolerance = tolerance)
    expect_equal(cyclopsFitS$log_likelihood, logLik(glmFit)[[1]], tolerance = tolerance)
    expect_equal(confint(cyclopsFitS, c(1:3))[,2:3], confint(glmFit, c(1:3)), tolerance = tolerance)
    expect_equal(predict(cyclopsFitS), predict(glmFit, type = "response"), tolerance = tolerance)
})


test_that("Get SEs in small Poisson model", {
    counts <- c(18,17,15,20,10,20,25,13,12)
    outcome <- gl(3,1,9)
    treatment <- gl(3,3)
    tolerance <- 1E-4

    gold <- glm(counts ~ outcome + treatment, family = poisson()) # gold standard
    goldSE <- summary(gold)$coefficients[,2]

    dataPtr <- createCyclopsData(counts ~ outcome + treatment,
                                  modelType = "pr")
    cyclopsFit <- fitCyclopsModel(dataPtr,
                          prior = createPrior("none"))

    cyclopsSE <- getSEs(cyclopsFit, c(1:5))

    expect_equal(goldSE, cyclopsSE, tolerance = tolerance)
})

# test_that("Playing with standardization", {
#     counts <- c(18,17,15,20,10,20,25,13,12)
#     outcome <- gl(3,1,9)
#     treatment <- gl(3,3)
#     tolerance <- 1E-4
#
#     dataPtr <- createCyclopsData(counts ~ outcome + treatment,
#                                       modelType = "pr")
#     cyclopsFit <- fitCyclopsModel(dataPtr,
#                                   prior = createPrior("none"))
#
#     dataPtrS <- createCyclopsData(counts ~ outcome + treatment,
#                                        modelType = "pr")
#     cyclopsFitS <- fitCyclopsModel(dataPtrS,
#                                    prior = createPrior("none"))
#
#     coef(cyclopsFit)
#     coef(cyclopsFitS)
# })

library("testthat")

context("test-parameterizedPriors.R")
suppressWarnings(RNGversion("3.5.0"))

test_that("lazy parameterization evaluation", {

    # initially invalid
    test <- Cyclops:::.cyclopsTestParameterizedPrior(priorFunction = function(x) { list(c(x[1],x[2])) },
                                             startingParameters = c(0, 0),
                                             indices = c(0), # no set
                                             values = c(1))
    expect_equal(test$valid, 0)

    # invalid after one-update and then returns correct evaluation
    test <- Cyclops:::.cyclopsTestParameterizedPrior(priorFunction = function(x) { list(c(x[1],x[2])) },
                                                     startingParameters = c(0, 0),
                                                     indices = c(1),
                                                     values = c(1))
    expect_equal(test$valid, 0)
    expect_equal(test$evaluation, list(c(1,0)))

    # multiple hits
    test <- Cyclops:::.cyclopsTestParameterizedPrior(priorFunction = function(x) { list(c(x[1],x[2])) },
                                                     startingParameters = c(0, 0),
                                                     indices = c(1,2,0),
                                                     values = c(1,2,0))
    expect_equal(test$valid, c(0,0,1)) # already valid at last index
    expect_equal(test$evaluation[[3]], c(1,2))
})

test_that("Specify parameterized L1 regularization", {
    counts <- c(18,17,15,20,10,20,25,13,12)
    outcome <- gl(3,1,9)
    treatment <- gl(3,3)

    data <- createCyclopsData(counts ~ outcome + treatment,
                              modelType = "pr")

    expect_error(
        prior <- createParameterizedPrior(priorType = "laplace",
                                          parameterize =  function(x) { list(c(0,x)) })
        , "provide parameter values")

    expect_error(
        prior <- createParameterizedPrior(priorType = "laplace",
                                          values = c(1),
                                          parameterize =  function(x) { c(0,x) })
        , "return a list")

    expect_error(
        prior <- createParameterizedPrior(priorType = c("laplace", "laplace"),
                                          values = c(1),
                                          parameterize =  function(x) { list(c(0,x)) })
        , "dimensionality mismatch")

    prior <- createParameterizedPrior(priorType = c(rep("laplace", 5)),
                                      values = c(2),
                                      parameterize = function(x) {
                                          lapply(1:5, function(i) {
                                              c(0,x)
                                          })
                                      })

    expect_warning(
        fit <- fitCyclopsModel(data, prior, forceNewObject = TRUE),
        "Excluding intercept")

    comp <- fitCyclopsModel(data, prior = createPrior("laplace", variance = 2),
                            warnings = FALSE, forceNewObject = TRUE)

    expect_equal(coef(fit), coef(comp))

    prior <- createParameterizedPrior(priorType = c(rep("laplace", 5)),
                                      values = c(0.2, 0.1),
                                      parameterize = function(x) {
                                          lapply(1:5, function(i) {
                                              c(x[1],x[2])
                                          })
                                      })

    fit <- fitCyclopsModel(data, prior, warnings = FALSE, forceNewObject = TRUE)

    expect_equivalent(coef(fit)[4:5], c(0.2, 0.2))
})

test_that("Using parameterized cross-validation", {
    skip_on_cran() # Do not run on CRAN

    ntest <- 1000
    ntrain <- 1000
    ncovars <- 2000

    set.seed(666)

    data <- simulateCyclopsData(nstrata=1,nrows=ntest+ntrain,ncovars=ncovars,model="logistic")
    test <- list(outcomes = data$outcomes[1:ntest,], covariates = data$covariates[data$covariates$rowId %in% data$outcomes$rowId[1:ntest],])
    train <- list(outcomes = data$outcomes[(ntest+1):(ntest+ntrain),], covariates = data$covariates[data$covariates$rowId %in% data$outcomes$rowId[(ntest+1):(ntest+ntrain)],])
    cyclopsData <- convertToCyclopsData(train$outcomes,train$covariates,modelType = "lr",addIntercept = TRUE)

    prior1 <- createParameterizedPrior(c("none", rep("laplace", ncovars)),
                                      parameterize = function(x) {
                                          lapply(1:(ncovars + 1), function(i) { c(0,x) })
                                      },
                                      values = c(1),
                                      useCrossValidation = TRUE)

    prior2 <- createPrior("laplace", exclude = c(0), useCrossValidation = TRUE)

    control <- createControl(noiseLevel = "silent", cvType = "auto", cvRepetitions = 1, seed = 666, threads = 1, resetCoefficients = TRUE)

    time1 <- system.time(fit1 <- fitCyclopsModel(cyclopsData,
                                                 prior=prior1,
                                                 control=control,
                                                 forceNewObject = TRUE))

    time2 <- system.time(fit2 <- fitCyclopsModel(cyclopsData,
                                                 prior=prior2,
                                                 control=control,
                                                 forceNewObject = TRUE))
    expect_equal(fit1$variance, fit2$variance)

    prior3 <- createParameterizedPrior(c("none", rep("laplace", ncovars)),
                                       parameterize = function(x) {
                                           lapply(1:(ncovars + 1),
                                                  function(i) {
                                                      if (i < (ncovars / 2)) {
                                                          return(c(0,x[1]))
                                                      } else {
                                                          return(c(0,x[2]))
                                                      }
                                                  }
                                           )
                                       },
                                       values = c(1, 0.5),
                                       useCrossValidation = TRUE)

    control3 <- createAutoGridCrossValidationControl(outerGrid = c(0.1, 0.1, 0.2))

    time3 <- system.time(fit3 <- fitCyclopsModel(cyclopsData,
                                                 prior=prior3,
                                                 control=control3,
                                                 forceNewObject = TRUE))

    expect_equal(length(unlist(strsplit(fit3$cross_validation, split = " "))), 3)

    expect_error(createAutoGridCrossValidationControl(outerGrid = c(1,2,3),
                                                      autoPosition = 0),
                 "Auto-position is invalid")

    # The following should execute w/o error
    createAutoGridCrossValidationControl(outerGrid = matrix(c(1,2,3,4,5,6), nrow = 2),
                                         autoPosition = 2)
    createAutoGridCrossValidationControl(outerGrid = matrix(c(1,2,3,4,5,6), nrow = 2),
                                         autoPosition = 3)
})

library("testthat")
library("survival")
library("gnm")

context("test-conditionalPoisson.R")

test_that("Check simple SCCS as conditional logistic regression", {
#     source("helper-conditionalPoisson.R")
    tolerance <- 1E-6
    gold.clogit <- clogit(event ~ exgr + agegr + strata(indiv) + offset(loginterval),
                          data = Cyclops::oxford)

    dataPtr <- createCyclopsData(event ~ exgr + agegr + strata(indiv) + offset(loginterval),
                                  data = Cyclops::oxford,
                                  modelType = "clr")
    cyclopsFit <- fitCyclopsModel(dataPtr,
                          prior = createPrior("none"))
    expect_equivalent(logLik(cyclopsFit), logLik(gold.clogit))
    expect_equal(coef(cyclopsFit), coef(gold.clogit), tolerance = tolerance)
})

test_that("Check simple SCCS as indicator conditional logistic regression", {
	#     source("helper-conditionalPoisson.R")
	tolerance <- 1E-6
	gold.clogit <- clogit(event ~ exgr + agegr + strata(indiv) + offset(loginterval),
												data = Cyclops::oxford)

	dataPtr <- createCyclopsData(event ~ strata(indiv) + offset(loginterval),
															 indicatorFormula = ~ exgr + agegr,
															 data = Cyclops::oxford,
															 modelType = "clr")
	cyclopsFit <- fitCyclopsModel(dataPtr,
																prior = createPrior("none"))
	expect_equivalent(logLik(cyclopsFit), logLik(gold.clogit))
	expect_equal(coef(cyclopsFit), coef(gold.clogit), tolerance = tolerance)
})

test_that("Check simple SCCS as sparse conditional logistic regression", {
	#     source("helper-conditionalPoisson.R")
	tolerance <- 1E-6
	gold.clogit <- clogit(event ~ exgr + agegr + strata(indiv) + offset(loginterval),
												data = Cyclops::oxford)

	dataPtr <- createCyclopsData(event ~ strata(indiv) + offset(loginterval),
															 sparseFormula = ~ exgr + agegr,
															 data = Cyclops::oxford,
															 modelType = "clr")
	cyclopsFit <- fitCyclopsModel(dataPtr,
																prior = createPrior("none"))
	expect_equivalent(logLik(cyclopsFit), logLik(gold.clogit))
	expect_equal(coef(cyclopsFit), coef(gold.clogit), tolerance = tolerance)
})

test_that("Check simple SCCS as conditional Poisson regression", {
#     source("helper-conditionalPoisson.R")
    tolerance <- 1E-3
    gold.cp <- gnm(event ~ exgr + agegr + offset(loginterval),
                   family = poisson, eliminate = indiv,
                   data = Cyclops::oxford)

    dataPtr <- createCyclopsData(event ~ exgr + agegr + strata(indiv) + offset(loginterval),
                                  data = Cyclops::oxford,
                                  modelType = "cpr")
    cyclopsFit <- fitCyclopsModel(dataPtr,
                          prior = createPrior("none"))

    expect_equal(coef(cyclopsFit)[1:2], coef(gold.cp)[1:2], tolerance = tolerance)
    expect_equal(confint(cyclopsFit, c("exgr1","agegr2"))[,2:3],
                 confint(gold.cp), tolerance = tolerance)
})

test_that("Check simple SCCS as SCCS", {
#     source("helper-conditionalPoisson.R")
    tolerance <- 1E-6
    gold.clogit <- clogit(event ~ exgr + agegr + strata(indiv) + offset(loginterval),
                          data = Cyclops::oxford)

    dataPtr <- createCyclopsData(event ~ exgr + agegr + strata(indiv), time = Cyclops::oxford$interval,
                                  data = Cyclops::oxford,
                                  modelType = "sccs")
    cyclopsFit <- fitCyclopsModel(dataPtr,
                          prior = createPrior("none"))
    expect_equivalent(logLik(cyclopsFit), logLik(gold.clogit))
    expect_equal(coef(cyclopsFit), coef(gold.clogit), tolerance = tolerance)
})

library("testthat")
library("survival")

suppressWarnings(RNGversion("3.5.0"))

test_that("Check very small Cox example with no ties, but with/without strata", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3.5,1,2,0
0, 3,  0,0,1
0, 2.5,1,0,1
0, 2,  1,1,1
0, 1.5,0,1,0
0, 1,  1,1,0
")

    goldCounting <-  coxph( Surv(start, length, event) ~ x1 + x2, test)
    summary(goldCounting)

    goldRight <- coxph(Surv(length, event) ~ x1 + x2, test)
    summary(goldRight)

    dataPtrRight <- createCyclopsData(Surv(length, event) ~ x1 + x2, data = test,
                                       modelType = "cox")
    cyclopsFitRight <- fitCyclopsModel(dataPtrRight)

    dataPtrCounting <- createCyclopsData(Surv(start, length, event) ~ x1 + x2, data = test,
                                          modelType = "cox")
    cyclopsFitCounting <- fitCyclopsModel(dataPtrCounting)

    expect_equal(coef(cyclopsFitRight), coef(cyclopsFitCounting))

    tolerance <- 1E-4
    expect_equal(coef(cyclopsFitRight), coef(goldRight), tolerance = tolerance)

    # Test residuals

    # goldRes <- residuals(goldCounting, "schoenfeld")[,1]
    # cyclopsRes <- residuals(cyclopsFitRight, "schoenfeld")
#
    # expect_equal(cyclopsRes, goldRes, tolerance = tolerance)

    # cox2.zph(goldCounting, transform = "identity")
    # residuals2.coxph(goldCounting, type = "scaledsch")
    #
    # # rr <- rr * vv * ndead + coef
    #
    # hand <- rev(navg) * goldCounting$var[1,1] * sum(event)  + coef(goldCounting)
    #
    # expect_equivalent(res2, hand) # scaled

    goldStrat <- coxph(Surv(length, event) ~ x1 + strata(x2), test)

    dataPtrStrat <- createCyclopsData(Surv(length, event) ~ x1 + strata(x2),
                                       data = test,
                                       modelType = "cox")
    cyclopsFitStrat <- fitCyclopsModel(dataPtrStrat)
    expect_equal(coef(cyclopsFitStrat), coef(goldStrat), tolerance = tolerance)

    goldStratRes <- residuals(goldStrat, "schoenfeld")
    cyclopsStratRes <- residuals(cyclopsFitStrat, "schoenfeld")
    # TODO Not yet the same order!!!
})



test_that("Check very small Cox example with time-ties, but no failure ties", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3,  1,2,0
0, 3,  0,0,1
0, 2,  1,0,1
0, 2,  0,1,1
0, 1,  0,1,0
0, 1,  1,1,0
")

    goldRight <- coxph(Surv(length, event) ~ x1 + x2, test)
    summary(goldRight)

    dataPtrRight <- createCyclopsData(Surv(length, event) ~ x1 + x2, data = test,
                                       modelType = "cox")
    cyclopsFitRight <- fitCyclopsModel(dataPtrRight)

    tolerance <- 1E-4
    expect_equal(coef(cyclopsFitRight), coef(goldRight), tolerance = tolerance)
})

test_that("Check very small Cox example with failure ties, no risk-set contribution after tie", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3,  1,2,0
0, 3,  0,0,1
0, 2,  1,0,1
0, 2,  1,1,1
0, 2,  1,0,0
")
    goldRight <- coxph(Surv(length, event) ~ x1 + x2, test, ties = "breslow")
    coef(goldRight)

    dataPtrRight <- createCyclopsData(Surv(length, event) ~ x1 + x2, data = test,
                                       modelType = "cox")
    cyclopsFitRight <- fitCyclopsModel(dataPtrRight)

    tolerance <- 1E-4
    expect_equal(coef(cyclopsFitRight), coef(goldRight), tolerance = tolerance)
})

test_that("Check very small Cox example with failure ties, with risk-set contribution after tie", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3,  1,2,0
0, 3,  0,0,1
0, 2,  1,0,1
0, 2,  1,1,1
0, 1,  0,1,0
0, 1,  1,1,0
")

        # We get the correct answer when last entry is censored
    goldRight <- coxph(Surv(length, event) ~ x1 + x2, test, ties = "breslow")

    dataPtrRight <- createCyclopsData(Surv(length, event) ~ x1 + x2, data = test,
                                           modelType = "cox")
    cyclopsFitRight <- fitCyclopsModel(dataPtrRight)

    tolerance <- 1E-4
    expect_equal(coef(cyclopsFitRight), coef(goldRight), tolerance = tolerance)
})

test_that("Check sparse Cox example with failure ties and strata", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3,  1,2,0
0, 3,  0,0,1
0, 2,  1,0,1
0, 2,  1,1,1
0, 1,  0,1,0
0, 1,  1,1,0
")

    # We get the correct answer when last entry is censored
    goldRight <- coxph(Surv(length, event) ~ x1 + strata(x2), test, ties = "breslow")

    dataPtrRight <- createCyclopsData(Surv(length, event) ~ x1 + strata(x2), data = test,
                                           modelType = "cox")
    cyclopsFitRight <- fitCyclopsModel(dataPtrRight)

    tolerance <- 1E-4
    expect_equal(coef(cyclopsFitRight), coef(goldRight), tolerance = tolerance)

    # Attempt sparse
    dataSparse <- createCyclopsData(Surv(length, event) ~ strata(x2),
                                         sparseFormula = ~ x1,
                                         data = test, modelType = "cox")

    cyclopsSparse <- fitCyclopsModel(dataSparse)
    expect_equal(coef(cyclopsSparse), coef(goldRight), tolerance = tolerance)
})

test_that("Check sparse Cox example with failure ties, strata and data weights", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3,  1,2,0
0, 3,  0,0,1
0, 2,  1,0,1
0, 2,  1,1,1
0, 1,  0,1,0
0, 1,  1,1,0
")

    goldRight <- coxph(Surv(length, event) ~ x1 + strata(x2), test, ties = "breslow")

    test2 <- rbind(data.frame(test, index = 1:7), data.frame(test, index = 1:7))
    test2 <- test2[order(test2$index),]
    test2[1,"x1"] <- 5
    test2[5,"x1"] <- 6

    dataPtrRight <- createCyclopsData(Surv(length, event) ~ x1 + strata(x2), data = test2,
                                      modelType = "cox")

    cyclopsFitRight <- fitCyclopsModel(dataPtrRight,
                                       weights = rep(c(0,1), 7))

    tolerance <- 1E-4
    expect_equal(coef(cyclopsFitRight), coef(goldRight), tolerance = tolerance)

    test2clean <- rbind(data.frame(test, index = 1:7), data.frame(test, index = 1:7))
    test2clean <- test2clean[order(test2clean$index),]

    dataPtrClean <- createCyclopsData(Surv(length, event) ~ x1 + strata(x2), data = test2clean,
                                      modelType = "cox")

    cyclopsFitClean <- fitCyclopsModel(dataPtrClean,
                                       weights = rep(c(0,1), 7))

    weights <- rep(c(1,0), 7)
    predictiveLogLik <- Cyclops:::getCyclopsPredictiveLogLikelihood(cyclopsFitClean, weights)
    expect_equal(predictiveLogLik, logLik(goldRight)[[1]])

    # Test error cases in predictive likelihood
    expect_error(Cyclops:::getCyclopsPredictiveLogLikelihood(cyclopsFitClean, weights = c(1,2)),
                 "Must provide")

    expect_error(Cyclops:::getCyclopsPredictiveLogLikelihood(cyclopsFitClean, weights = rep(-1, 14)),
                 "Only non-negative weights")

    cap <- capture.output(print(cyclopsFitClean))
    expect_false(is.null(cap))
})

test_that("Check SQL interface for a very small Cox example with failure ties and strata", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3,  1,2,0
0, 3,  0,0,1
0, 2,  1,0,1
0, 2,  1,1,1
0, 1,  0,1,0
0, 1,  1,1,0
")

    # We get the correct answer when last entry is censored
    goldRight <- coxph(Surv(length, event) ~ x1 + strata(x2), test, ties = "breslow")

    data <- test
    data$row_id <- 1:nrow(data)
    data$covariate_id = 1

    data <- data[order(data$x2, -data$length, data$event),] # Must sort by: strata, into risk set (with events before censorsed)

    dataPtr <- createSqlCyclopsData(modelType = "cox")

    count <- appendSqlCyclopsData(dataPtr,
                                  data$x2,
                                  data$row_id,
                                  data$event,
                                  data$length,
                                  data$row_id,
                                  data$covariate_id,
                                  data$x1)

    cyclopsFitStrat <- fitCyclopsModel(dataPtr)

    tolerance <- 1E-4
    # SQL interface provides different names ('1' instead of 'x1')
    t1 <- coef(cyclopsFitStrat)
    t2 <- coef(goldRight)
    names(t1) <- NULL
    names(t2) <- NULL
    expect_equal(t1, t2, tolerance = tolerance)
})

test_that("Check confidence intervals with fixed coefficients",{
    set.seed(666)
    simulant <- simulateCyclopsData(nstrata = 1,
                                    ncovars = 5,
                                    model = "survival")

    # Fix covariate 2:
    data <- convertToCyclopsData(simulant$outcomes, simulant$covariates, modelType = "cox")
    fixed <- c(FALSE, TRUE, FALSE, FALSE, FALSE)
    names(fixed) <- c("1","2","3","4","5")
    fit <- fitCyclopsModel(data, prior = createPrior("none"),
                           forceNewObject = TRUE,
                           fixedCoefficients = fixed)
    expect_error(confint(fit, parm = c("1","2","3","4","5")))
    expect_equal(nrow(confint(fit, parm = c("1","3","4","5"))), 4)
    expect_error(confint(fit, parm = c(2)))
})

#
# test_that("More SQL checks for stratified cox models", {
#     data(lung)
#     lung$status = lung$status -1
#     lung <- lung[!is.na(lung$ph.ecog),]
#
#     goldRight <- coxph(Surv(time, status) ~ age + ph.ecog + strata(sex), lung, ties = "breslow")
#
#     dataPtrRight <- createCyclopsData(Surv(time, status) ~ age + ph.ecog + strata(sex),
#                                            method = "debug",
#                                            data = lung, modelType = "cox")
#     #This crashed R:
#     cyclopsFitRight <- fitCyclopsModel(dataPtrRight,
#                                        control = createControl(noiseLevel = "silent"))
#
#     lung$row_id <- 1:nrow(lung)
#     out <- data.frame(row_id = lung$row_id, stratum_id = lung$sex, time = lung$time, y = lung$status)
#     covAge <- data.frame(row_id = lung$row_id, stratum_id = lung$sex, time = lung$time, y = lung$status, covariate_value = lung$age)
#     covAge$covariate_id = 1
#     covPhEcog <- data.frame(row_id = lung$row_id, stratum_id = lung$sex, time = lung$time, y = lung$status, covariate_value = lung$ph.ecog)
#     covPhEcog$covariate_id = 2
#     cov <- rbind(covAge,covPhEcog)
#
#     # Sometimes, rows match on all (stratum, time, outcome): in which case sort may differ between out/cov
#     out <- out[order(out$stratum_id, -out$time, out$y, out$row_id),] # Must sort by: strata, into risk set (with events before censorsed)
#     cov <- cov[order(cov$stratum_id, -cov$time, cov$y, cov$row_id),] # Must sort by: strata, into risk set (with events before censorsed)
#
#     dataPtr <- createSqlCyclopsData(modelType = "cox")
#
#     count <- appendSqlCyclopsData(dataPtr,
#                                   out$stratum_id,
#                                   out$row_id,
#                                   out$y,
#                                   out$time,
#                                   cov$row_id,
#                                   cov$covariate_id,
#                                   cov$covariate_value)
#     #This crashes R
#     cyclopsFitStrat <- fitCyclopsModel(dataPtr)
# })

library("testthat")
library("survival")

context("test-infiniteMLE.R")

simulate <- function() {
    set.seed(123)
    n <- 1000
    timeToCensor <- rexp(n, 0.01)
    exposure <- runif(n) < 0.5
    timeToEvent <- rep(Inf, n)
    timeToEvent[!exposure] <- rexp(sum(!exposure), 0.01)
    time <- pmin(timeToCensor, timeToEvent)
    outcome <- timeToEvent < timeToCensor
    list(time = time, outcome = outcome, exposure = exposure)
}

test_that("Check for infinite MLE in Cox example with no outcomes in one treatment arm", {

    data <- simulate()

    expect_warning(gold <- coxph(Surv(time, outcome) ~ exposure, data = data),
                   regexp = ".*coefficient may be infinite.*")

    cyclopsData <- createCyclopsData(Surv(time, outcome) ~ exposure, data = data, modelType = "cox")
    expect_warning(fit <- fitCyclopsModel(cyclopsData), regexp =".*coefficient may be infinite.*")

    ci <- confint(fit, parm = "exposureTRUE", level = 0.9)
    expect_true(is.na(ci[2]))
})

test_that("Check multivariable Jeffreys prior", {

    data <- simulate()
    data$x2 <- runif(length(data$time), 0.5) < 0.5

    cyclopsData <- createCyclopsData(Surv(time, outcome) ~ exposure + x2, data = data, modelType = "cox")
    expect_error(fit <- fitCyclopsModel(cyclopsData,
                                        prior = createPrior(priorType = "jeffreys")),
                 regexp = ".*1 covariate.*")
})

test_that("Check Jeffreys prior with correct control", {

    data <- simulate()

    cyclopsData <- createCyclopsData(Surv(time, outcome) ~ exposure, data = data, modelType = "cox")
    expect_warning(fit <- fitCyclopsModel(cyclopsData,
                                        prior = createPrior(priorType = "jeffreys")),
                 regexp = "BLR convergence")

    fit <- fitCyclopsModel(cyclopsData = cyclopsData, forceNewObject = TRUE,
                           prior = createPrior(priorType = "jeffreys"),
                           control = createControl(convergenceType = "lange"))
    expect_true(fit$return_flag == "SUCCESS")
})

test_that("Check Jeffreys prior with indicator only", {

    data <- data.frame(time = c(1,2),
                       outcome = c(1,0),
                       exposure = c(1,2))

    cyclopsData <- createCyclopsData(Surv(time, outcome) ~ exposure, data = data, modelType = "cox")
    expect_error(fitCyclopsModel(cyclopsData, prior = createPrior(priorType = "jeffreys")),
                 regexp = "*.for indicator covariates")

})

test_that("Check univariable Jeffreys prior", {

    data <- simulate()

    cyclops1 <- createCyclopsData(Surv(time, outcome) ~ exposure, data = data, modelType = "cox")
    cyclops2 <- createCyclopsData(Surv(time, outcome) ~ exposure, data = data, modelType = "cox")

    expect_warning(uniform <- fitCyclopsModel(cyclops1))

    jeffreys <- fitCyclopsModel(cyclops2,
                                prior = createPrior(priorType = "jeffreys"),
                                control = createControl(convergenceType = "lange",
                                                        tolerance = 1E-7),
                                forceNewObject = TRUE)

    jci <- confint(jeffreys, parm = "exposureTRUE", level = 0.95, overrideNoRegularization = TRUE)
    expect_false(is.na(jci[2]))

    # grid <- seq(from=-8, to = -2, length.out = 1000)
    # uniformProfile <- getCyclopsProfileLogLikelihood(uniform, parm = "exposureTRUE", x = grid)
    # jeffreysProfile <- getCyclopsProfileLogLikelihood(jeffreys, parm = "exposureTRUE", x = grid)
    #
    # plot(uniformProfile$point, uniformProfile$value - max(uniformProfile$value), type = "l")
    # lines(jeffreysProfile$point, jeffreysProfile$value - max(uniformProfile$value, lwd = 2))
    #
    # plot(jeffreysProfile$point, jeffreysProfile$value - max(jeffreysProfile$value), type = "l")
    #
    # maxProfile <- function(profile) {
    #     idx <- which(profile$value == max(profile$value))
    #     list(point = profile$point[idx], value = profile$value[idx])
    # }
    #
    # maxProfile(jeffreysProfile)
})

library("testthat")
library("survival")

suppressWarnings(RNGversion("3.5.0"))

test_that("gradient", {

    data <- Cyclops::createCyclopsData(Surv(stop, event) ~ (rx - 1) + size, data = bladder, modelType = "cox")

    fit <- Cyclops::fitCyclopsModel(data)

    gradientAtMode <- gradient(fit)

    expect_equivalent(gradientAtMode, rep(0, length(coef(fit))))

    mode <- coef(fit)
    eps <- 1E-2
    Cyclops:::.cyclopsSetBeta(fit$interface,
                              mode + c(-eps, 0))
    grad <- gradient(fit)

    Cyclops:::.cyclopsSetBeta(fit$interface,
                              mode + c(-2*eps, 0))
    h0 <- Cyclops:::.cyclopsGetLogLikelihood(fit$interface)
    central <- (fit$log_likelihood - h0) / (2 * eps)

    expect_equivalent(grad[1], central, tolerance = 1E-3)
})

library("testthat")

context("test-outOfSample.R")
suppressWarnings(RNGversion("3.5.0"))

test_that("Out-of-sample optimization", {
    skip_on_cran()
    set.seed(666)

    nrows <- 10000
    outOfSampleProp <- 0.2

    simulant <- simulateCyclopsData(nstrata = 1,
                                    nrows = nrows,
                                ncovars = 5,
                                model = "logistic")

    nOutSample <- nrows * outOfSampleProp
    nInSample <- nrows - nOutSample

    weights <- c(rep(1, nInSample), rep(0, nOutSample))

    data <- convertToCyclopsData(simulant$outcomes,
                                 simulant$covariates,
                                 modelType = "lr",
                                        addIntercept = TRUE)

    prior <- createPrior("laplace", variance = 1, exclude = "(Intercept)",
                         useCrossValidation = TRUE)

    control <- createWeightBasedSearchControl(initialValue = 1,
                                              noiseLevel = "noisy")

    optimal <- fitCyclopsModel(data, prior, control, weights = weights)

})

library("testthat")

context("test-smallNormal.R")

#
# Small Normal regression
#

test_that("Small Normal dense regression with CCD algorithm", {

    x <- log(c(1,5,10,20,30,40,50,75,100,150,200))
    y <- c(0,3,6,7,9,13,17,12,11,14,13)

    tolerance <- 1E-4

    gold <- lm(y ~ x) # gold standard

    data <- createCyclopsData(y ~ x, modelType = "ls")
    fit <- fitCyclopsModel(data, prior = createPrior("none"),
                           control = createControl(tolerance = 1E-8))
    expect_equal(coef(fit), coef(gold), tolerance = tolerance)
})

test_that("Small Normal sparse regression with CCD algorithm", {

    x <- log(c(1,5,10,20,30,40,50,75,100,150,200))
    y <- c(0,3,6,7,9,13,17,12,11,14,13)

    tolerance <- 1E-4

    gold <- lm(y ~ x) # gold standard

    data <- createCyclopsData(y ~ 1, sparseFormula = ~ x, modelType = "ls")
    fit <- fitCyclopsModel(data, prior = createPrior("none"),
                           control = createControl(tolerance = 1E-8))
    expect_equal(coef(fit), coef(gold), tolerance = tolerance)
})

library("testthat")
library("survival")
library("gnm")

context("test-smallCLR.R")

#
# Small conditional logistic regression
#

test_that("Small conditional logistic regression", {
    gold <- clogit(case ~ spontaneous + induced + strata(stratum), data=infert)

    dataPtr <- createCyclopsData(case ~ spontaneous + induced + strata(stratum),
                                  data = infert,
                                  modelType = "clr")

    cyclopsFit <- fitCyclopsModel(dataPtr, prior = createPrior("none"))

	tolerance <- 1E-4

	expect_equal(coef(cyclopsFit), coef(gold), tolerance = tolerance)
	expect_equal(cyclopsFit$log_likelihood, logLik(gold)[[1]], tolerance = tolerance)

    expect_equal(vcov(cyclopsFit), vcov(gold), tolerance = tolerance)

    expect_equal(aconfint(cyclopsFit), confint(gold), tolerance = tolerance)

    expect_equal(confint(cyclopsFit, c(1:2), includePenalty = TRUE),
                 confint(cyclopsFit, c(1:2), includePenalty = FALSE))

    # dataPtrR <- createCyclopsData(case ~ spontaneous + induced + strata(stratum),
    #                                    data = infert,
    #                                    modelType = "clr")
    #
    # cyclopsFitR <- fitCyclopsModel(dataPtrR,
    #                                prior = createPrior("laplace", 1, exclude = 1))

#     expect_not_equal(confint(cyclopsFitR, c(1), includePenalty = TRUE),
#                  confint(cyclopsFitR, c(1), includePenalty = FALSE))
#       # How to test for inequality?
# This is broken:
# 	expect_equal(predict(cyclopsFit), predict(gold, type = "risk"), tolerance = tolerance)
})

test_that("Small conditional poisson regression with an offset", {

    gold.cp <- gnm(event ~ exgr + agegr + offset(loginterval),
                   family = poisson, eliminate = indiv,
                   data = Cyclops::oxford)
    vcov(gold.cp)

    dataPtr <- createCyclopsData(event ~ exgr + agegr + strata(indiv) + offset(loginterval),
                                 data = Cyclops::oxford,
                                 modelType = "cpr")
    cyclopsFit <- fitCyclopsModel(dataPtr,
                                  prior = createPrior("none"))

    expect_equal(as.numeric(coef(cyclopsFit)), as.numeric(coef(gold.cp)),
                 tolerance = 1E-6)
    expect_equal(as.numeric(vcov(cyclopsFit)), as.numeric(vcov(gold.cp)),
                 tolerance = 1E-6)
})

library(testthat)
suppressWarnings(RNGversion("3.5.0"))

test_that("Logistic regression with cross-validation", {
    skip_on_cran()
    set.seed(123)

    sim <- simulateCyclopsData(nstrata=1,
                               ncovars=100,
                               nrows=10000,
                               effectSizeSd=0.5,
                               eCovarsPerRow=2,
                               model="logistic")

    cyclopsData <- convertToCyclopsData(outcomes = sim$outcomes, covariates = sim$covariates, modelType = "lr")
    fit <- fitCyclopsModel(cyclopsData = cyclopsData, warnings = FALSE,
                           prior = createPrior("laplace", useCrossValidation = TRUE),
                           control = createControl(fold = 10, cvRepetitions = 1,
                                                   startingVariance = 0.1, noiseLevel = "quiet", seed = 123))
    coef(fit)[1:10]

    # (Intercept)            1            2            3            4            5
    # -0.99228229  0.00000000  0.13961894  0.01131845  0.00000000  0.26817152
    # 6           7           8           9
    # -0.02982060  0.00000000  0.09693092  0.00000000

    expect_equal(coef(fit)[1], -0.99, tolerance = 0.01, check.attributes = FALSE)
})

library("testthat")

context("test-floatingPoint.R")
suppressWarnings(RNGversion("3.5.0"))

#
# FP precision
#

test_that("Double precision", {
	binomial_bid <- c(1,5,10,20,30,40,50,75,100,150,200)
	binomial_n <- c(31,29,27,25,23,21,19,17,15,15,15)
	binomial_y <- c(0,3,6,7,9,13,17,12,11,14,13)

	log_bid <- log(c(rep(rep(binomial_bid, binomial_n - binomial_y)), rep(binomial_bid, binomial_y)))
	y <- c(rep(0, sum(binomial_n - binomial_y)), rep(1, sum(binomial_y)))

	tolerance <- 1E-4

	glmFit <- glm(y ~ log_bid, family = binomial()) # gold standard

	dataPtrD <- createCyclopsData(y ~ log_bid, modelType = "lr",
	                              floatingPoint = 64)

	expect_equal(getFloatingPointSize(dataPtrD), 64)

	cyclopsFitD <- fitCyclopsModel(dataPtrD, prior = createPrior("none"),
	                       control = createControl(noiseLevel = "silent"))

	expect_equal(coef(cyclopsFitD), coef(glmFit), tolerance = tolerance)
})

test_that("Single precision", {
    binomial_bid <- c(1,5,10,20,30,40,50,75,100,150,200)
    binomial_n <- c(31,29,27,25,23,21,19,17,15,15,15)
    binomial_y <- c(0,3,6,7,9,13,17,12,11,14,13)

    log_bid <- log(c(rep(rep(binomial_bid, binomial_n - binomial_y)), rep(binomial_bid, binomial_y)))
    y <- c(rep(0, sum(binomial_n - binomial_y)), rep(1, sum(binomial_y)))

    tolerance <- 1E-4

    glmFit <- glm(y ~ log_bid, family = binomial()) # gold standard

    dataPtrS <- createCyclopsData(y ~ log_bid, modelType = "lr",
                                  floatingPoint = 32)

    expect_equal(getFloatingPointSize(dataPtrS), 32)

    cyclopsFitS <- fitCyclopsModel(dataPtrS, prior = createPrior("none"),
                                   control = createControl(noiseLevel = "silent"))

    expect_equal(coef(cyclopsFitS), coef(glmFit), tolerance = tolerance * 10)
})

test_that("Speed difference", {
    skip("for debugging purposes; simulation takes too long")

    set.seed(123)
    sim <- simulateCyclopsData(nstrata=1,
                               ncovars=1000,
                               nrows=100000,
                               effectSizeSd=0.5,
                               eCovarsPerRow=2,
                               model="logistic")

    dataD <- convertToCyclopsData(outcomes = sim$outcomes,
                                  covariates = sim$covariates, modelType = "lr",
                                  floatingPoint = 64)

    dataS <- convertToCyclopsData(outcomes = sim$outcomes,
                                  covariates = sim$covariates, modelType = "lr",
                                  floatingPoint = 32)

    system.time(
        fitD <- fitCyclopsModel(cyclopsData = dataD,
                                prior = createPrior("laplace", variance = 1),
                                control = createControl(fold = 10, cvRepetitions = 1,
                                                        startingVariance = 0.1, noiseLevel = "silent",
                                                        seed = 123),
                                forceNewObject = TRUE))

    system.time(
        fitS <- fitCyclopsModel(cyclopsData = dataS,
                                prior = createPrior("laplace", variance = 1),
                                control = createControl(fold = 10, cvRepetitions = 1,
                                                        startingVariance = 0.1, noiseLevel = "silent",
                                                        seed = 123),
                                forceNewObject = TRUE))

    expect_equal(coef(fitD)[1:10], coef(fitS)[1:10], tolerance = 1E-3)
})

library("testthat")
library("survival")
library("Andromeda")
library("gnm")

context("test-dataConversionUnstratified.R")

test_that("Test data.frame to data for lr", {

  gold <- glm(case ~ spontaneous + induced, data=infert, family="binomial")

  #Convert infert dataset to Cyclops format:
  covariates <- data.frame(rowId = rep(1:nrow(infert),2),
                           covariateId = rep(1:2,each=nrow(infert)),
                           covariateValue = c(infert$spontaneous,infert$induced))
  outcomes <- data.frame(rowId = 1:nrow(infert),
                         y = infert$case)
  #Make sparse:
  covariates <- covariates[covariates$covariateValue != 0,]

  andr <- andromeda(outcomes = outcomes, covariates = covariates)
  cyclopsDataAndr <- convertToCyclopsData(andr$outcomes,andr$covariates,modelType = "lr",addIntercept = TRUE)
  fitAndr <- fitCyclopsModel(cyclopsDataAndr,prior = createPrior("none"))

  cyclopsData <- convertToCyclopsData(outcomes,covariates,modelType = "lr",addIntercept = TRUE)
  fit <- fitCyclopsModel(cyclopsData,prior = createPrior("none"))


  tolerance <- 1E-4
  expect_equal(as.vector(coef(fit)), as.vector(coef(gold)), tolerance = tolerance)
  expect_equal(as.vector(coef(fitAndr)), as.vector(coef(gold)), tolerance = tolerance)
})

test_that("Test unstratified cox using lung dataset ", {

  test <- lung
  test[is.na(test)] <- 0 # Don't want to bother with missing values

  gold <- coxph(Surv(time, status) ~ age + ph.ecog + ph.karno + pat.karno + meal.cal + wt.loss, test, method="breslow")

  cyclopsDataFormula <- createCyclopsData(Surv(time, status) ~ age + ph.ecog + ph.karno + pat.karno + meal.cal + wt.loss, data=test,modelType="cox")
  fitFormula <- fitCyclopsModel(cyclopsDataFormula)

  #Convert to data.frames for Cyclops:
  nCovars = 6
  covariates <- data.frame(stratumId = 0,
                           rowId = rep(1:nrow(test),nCovars),
                           covariateId = rep(1:nCovars,each = nrow(test)),
                           covariateValue = c(test$age,test$ph.ecog,test$ph.karno,test$pat.karno,test$meal.cal,test$wt.loss))
  outcomes <- data.frame(stratumId = 0,
                         rowId = 1:nrow(test),
                         y = test$status-1,
                         time = test$time)

  #Make sparse:
  covariates <- covariates[covariates$covariateValue != 0,]

  andr <- andromeda(outcomes = outcomes, covariates = covariates)
  cyclopsDataAndr <- convertToCyclopsData(andr$outcomes,andr$covariates,modelType = "cox")
  fitAndr <- fitCyclopsModel(cyclopsDataAndr,prior = createPrior("none"))

  cyclopsData <- convertToCyclopsData(outcomes,covariates,modelType = "cox")
  fit <- fitCyclopsModel(cyclopsData,prior = createPrior("none"))

  tolerance <- 1E-4
  expect_equal(as.vector(coef(fit)), as.vector(coef(fitFormula)), tolerance = tolerance)
  expect_equal(as.vector(coef(fit)), as.vector(coef(gold)), tolerance = tolerance)
  expect_equal(as.vector(coef(fitAndr)), as.vector(coef(gold)), tolerance = tolerance)
})


test_that("Test poisson regression", {

  sim <- simulateCyclopsData(nstrata = 1, nrows = 10000, ncovars = 2, eCovarsPerRow = 0.5, effectSizeSd = 1,model = "poisson")
  covariates <- sim$covariates
  outcomes <- sim$outcomes

  #Convert to data format for gnm:
  ncovars <- max(covariates$covariateId)
  nrows <- nrow(outcomes)
  m <- matrix(0,nrows,ncovars)
  for (i in 1:nrow(covariates)){
    m[covariates$rowId[i],covariates$covariateId[i]] <- 1
  }
  data <- as.data.frame(m)

  data$rowId <- 1:nrow(data)
  data <- merge(data,outcomes)
  data <- data[order(data$stratumId,data$rowId),]
  formula <- as.formula(paste(c("y ~ V1",paste("V",2:ncovars,sep="")),collapse=" + "))

  gold <- gnm(formula, family=poisson, offset=log(time), data = data)

  cyclopsDataFormula <- createCyclopsData(formula, offset=log(time), data=data,modelType="pr")
  fitFormula <- fitCyclopsModel(cyclopsDataFormula)

  andr <- andromeda(outcomes = outcomes, covariates = covariates)
  cyclopsDataAndr <- convertToCyclopsData(andr$outcomes,andr$covariates,modelType = "pr",addIntercept = TRUE)
  fitAndr <- fitCyclopsModel(cyclopsDataAndr,prior = createPrior("none"))

  cyclopsData <- convertToCyclopsData(outcomes,covariates,modelType = "pr",addIntercept = TRUE)
  fit <- fitCyclopsModel(cyclopsData,prior = createPrior("none"))

  tolerance <- 1E-4
  expect_equal(as.vector(sort(coef(fit))), as.vector(sort(coef(fitFormula))), tolerance = tolerance)
  expect_equal(as.vector(sort(coef(fit))), as.vector(sort(coef(gold))), tolerance = tolerance)
  expect_equal(as.vector(sort(coef(fitAndr))), as.vector(sort(coef(gold))), tolerance = tolerance)
})

library("testthat")
library("boot")
library("rsample")
library("Cyclops")
library("survival")

## helper functions

convertToDf <- function(sim, ncovars) {

    if (ncovars < max(sim$covariates$covariateId)) {
        stop("Too few covariates")
    }

    mat <- matrix(0, nrow = nrow(sim$outcomes), ncol = ncovars)

    invisible(mapply(function(rowId, covariateId, covariateValue) {
        mat[rowId, covariateId] <<- covariateValue
    },
    sim$covariates$rowId, sim$covariates$covariateId, sim$covariates$covariateValue))

    df <- as.data.frame(mat)
    outcomes <- sim$outcomes[order(sim$outcomes$rowId),]
    df$y <- outcomes$y
    df$stratumId <- outcomes$stratumId

    if (!is.null(outcomes$weights)) {
        df$weights <- outcomes$weights
    }

    if (!is.null(outcomes$time)) {
        df$time <- outcomes$time
    }

    return(df)
}

getAllPercentileIntervals <- function(bb) {
    result <- t(apply(bb, 2L, quantile, c(0.025, 0.975)))
    colnames(result) <- c("bpi_lower", "bpi_upper")
    as.data.frame(result)
}

bootstrap <- function(df, func, R) {

    replicates <- rsample::bootstraps(df, times = R)

    betas <- lapply(replicates$splits, func) %>%
        bind_rows()

    return(betas)
}

## End helper functions

test_that("Small Poisson bootstrap examples with and without weights", {

    # skip_on_cran()

    set.seed(123)

    dobson <- data.frame(
        counts = c(18,17,15,20,10,20,25,13,12),
        outcome2 = c(0,1,0,0,1,0,0,1,0),
        outcome3 = c(0,0,1,0,0,1,0,0,1),
        treatment2 = c(0,0,0,1,1,1,0,0,0),
        treatment3 = c(0,0,0,0,0,0,1,1,1),
        weights = c(0.17,0.03,0.93,0.64,0.94,0.19,0.12,0.58,0.01)
    )

    execute <- function(d, f) {
        cd <- createCyclopsData(counts ~ outcome2 + outcome3 + treatment2 + treatment3, data = d[f,], modelType = "pr")
        fit <- fitCyclopsModel(cd, prior = createPrior("normal", exclude = "(Intercept)"))
        coef(fit)
    }

    bb <- boot(dobson, execute, R = 4999)
    bBias <- apply(bb$t, 2L, mean) - bb$t0

    cd <- createCyclopsData(counts ~ outcome2 + outcome3 + treatment2 + treatment3, data = dobson, modelType = "pr")
    cf <- fitCyclopsModel(cd, prior = createPrior("normal", exclude = "(Intercept)"),
                          control = createControl(seed = 123))
    cb <- runBootstrap(cf, replicates = 4999)

    expect_equivalent(bBias, cb$summary$bias, tolerance = 1E-2) # Check unweighted bias

    wexecute <- function(d, f) {
        cd <- createCyclopsData(counts ~ outcome2 + outcome3 + treatment2 + treatment3, data = d[f,], modelType = "pr")
        fit <- fitCyclopsModel(cd, weights = d[f,]$weights,
                               prior = createPrior("normal", exclude = "(Intercept)"))
        coef(fit)
    }

    wbb <- boot(dobson, wexecute, R = 4999)
    wbbBias <- apply(wbb$t, 2L, mean) - wbb$t0

    wcd <- createCyclopsData(counts ~ outcome2 + outcome3 + treatment2 + treatment3, data = dobson, modelType = "pr")
    wcf <- fitCyclopsModel(wcd, weights = dobson$weights,
                          prior = createPrior("normal", exclude = "(Intercept)"),
                          control = createControl(seed = 123))
    wcb <- runBootstrap(wcf, replicates = 4999)

    expect_equivalent(wbbBias, wcb$summary$bias, tolerance = 1E-2) # Check weighted bias

    expect_gt(mean(abs(cb$summary$bias - wcb$summary$bias)), 0.01) # Difference estimates with and without weights
})

test_that("Small Poisson bootstrap examples with an offset", {

    # skip_on_cran()

    dobson <- data.frame(
        counts = c(18,17,15,20,10,20,25,13,12),
        outcome2 = c(0,1,0,0,1,0,0,1,0),
        outcome3 = c(0,0,1,0,0,1,0,0,1),
        treatment2 = c(0,0,0,1,1,1,0,0,0),
        treatment3 = c(0,0,0,0,0,0,1,1,1),
        offset = c(1,1,1,1,1,1,1,1,1)
    )

    set.seed(123)
    cdo <- createCyclopsData(counts ~ outcome2 + outcome3 + treatment2 + treatment3 + offset(offset),
                            data = dobson, modelType = "pr")
    cfo <- fitCyclopsModel(cdo, prior = createPrior("normal", exclude = "(Intercept)"),
                          control = createControl(seed = 123))
    cbo <- runBootstrap(cfo, replicates = 4999)

    set.seed(123)
    cdn <- createCyclopsData(counts ~ outcome2 + outcome3 + treatment2 + treatment3,
                             data = dobson, modelType = "pr")
    cfn <- fitCyclopsModel(cdn, prior = createPrior("normal", exclude = "(Intercept)"),
                           control = createControl(seed = 123))
    cbn <- runBootstrap(cfn, replicates = 4999)

    expect_equal(cbo$summary$original + c(1,0,0,0,0), cbn$summary$original, tolerance = 1E-4)

    expect_equal(cbo$summary$bpi_lower + c(1,0,0,0,0), cbn$summary$bpi_lower, tolerance = 1E-4)
})


# boot.out <- list(
#     t = as.matrix(cb$samples),
#     sim = "ordinary",
#     stype = "i",
#     R = nrow(cb$samples),
#     t0 = coef(cf),
#     data = matrix(nrow = Cyclops::getNumberOfRows(cd)), # bb$data
#     strata = rep(1, Cyclops::getNumberOfRows(cd)) # bb$strata
# )
#
# boot.ci(boot.out, index = 1L, type = "bca")

test_that("bootstrap option for na.rm", {

    # skip_on_cran() # ASAN error https://github.com/OHDSI/Cyclops/actions/runs/16373844416

    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3,  1,2,0
0, 3,  0,0,1
0, 2,  1,0,1
0, 2,  1,1,1
0, 1,  0,1,0
0, 1,  1,1,0
")
    test <- rbind(data.frame(test, index = 1:7), data.frame(test, index = 1:7))

    set.seed(123)

    cyclopsData <- createCyclopsData(Surv(length, event) ~ x1 + strata(x2),
                                     data = test,
                                     modelType = "cox")

    fit <- fitCyclopsModel(cyclopsData, weights = rep(c(0,1), 7))
    expect_error(bootstrap <- runBootstrap(fit, replicates = 4999), "missing values and NaN")

    fit <- fitCyclopsModel(cyclopsData, weights = rep(c(0,1), 7))
    bootstrap <- runBootstrap(fit, replicates = 4999, na.rm = TRUE)
    expect_lt(nrow(bootstrap$samples), 4999)
})

test_that("Large logistic bootstrap with and without weights", {

    # skip_on_cran()

    set.seed(123)
    sim <- simulateCyclopsData(nstrata=100,
                               ncovars=4,
                               nrows=1000,
                               effectSizeSd=0.5,
                               eCovarsPerRow=2,
                               model="logistic")
    cyclopsData <- convertToCyclopsData(outcomes = sim$outcomes,
                                        covariates = sim$covariates,
                                        modelType = "lr")
    fitCyclopsNoWeights <- fitCyclopsModel(cyclopsData = cyclopsData)
    bsNoWeights <- runBootstrap(fitCyclopsNoWeights, replicates = 1999)

    bbNoWeights <- bootstrap(
        convertToDf(sim, 4),
        function(split) {
            data <- rsample::analysis(split)
            fit <- glm(y ~ V1 + V2 + V3 + V4,
                       family = "binomial", data = data)
            coef(fit)
        }, R = 1999)

    bbNoStdError <- sqrt(apply(bbNoWeights, 2L, var))
    expect_equivalent(bsNoWeights$summary$std_err, bbNoStdError, tolerance = 0.02)

    expect_equivalent(bsNoWeights$summary[,c("bpi_lower", "bpi_upper")],
                      getAllPercentileIntervals(bbNoWeights),
                      tolerance = 0.1)

    sim$outcomes$weights <- rep(c(0.1,0.9), nrow(sim$outcomes) / 2)

    cyclopsData <- convertToCyclopsData(outcomes = sim$outcomes,
                                        covariates = sim$covariates,
                                        modelType = "lr")
    fitCyclopsYesWeights <- fitCyclopsModel(cyclopsData = cyclopsData)
    bsYesWeights <- runBootstrap(fitCyclopsYesWeights, replicates = 1999)

    bbYesWeights <- bootstrap(convertToDf(sim, 4),
        function(split) {
            data <- rsample::analysis(split)
            suppressWarnings(
                fit <- glm(y ~ V1 + V2 + V3 + V4, weights = data$weights,
                           family = "binomial", data = data))
            coef(fit)
        }, R = 1999)

    bbYesStdError <- sqrt(apply(bbYesWeights, 2L, var))
    expect_equivalent(bsYesWeights$summary$std_err, bbYesStdError, tolerance = 0.02)

    expect_equivalent(bsYesWeights$summary[,c("bpi_lower", "bpi_upper")],
                      getAllPercentileIntervals(bbYesWeights),
                      tolerance = 0.1)
})

test_that("Large Poisson bootstrap with and without weights", {

    # skip_on_cran()

    set.seed(123)
    sim <- simulateCyclopsData(nstrata=100,
                               ncovars=4,
                               nrows=1000,
                               effectSizeSd=0.5,
                               eCovarsPerRow=2,
                               model="poisson")
    cyclopsData <- convertToCyclopsData(outcomes = sim$outcomes,
                                        covariates = sim$covariates,
                                        modelType = "pr")
    fitCyclopsNoWeights <- fitCyclopsModel(cyclopsData = cyclopsData)
    bsNoWeights <- runBootstrap(fitCyclopsNoWeights, replicates = 1999)

    bbNoWeights <- bootstrap(
        convertToDf(sim, 4),
        function(split) {
            data <- rsample::analysis(split)
            fit <- glm(y ~ V1 + V2 + V3 + V4, offset = log(time),
                       family = "poisson", data = data)
            coef(fit)
        }, R = 1999)

    bbNoStdError <- sqrt(apply(bbNoWeights, 2L, var))
    expect_equivalent(bsNoWeights$summary$std_err, bbNoStdError, tolerance = 0.02)

    expect_equivalent(bsNoWeights$summary[,c("bpi_lower", "bpi_upper")],
                      getAllPercentileIntervals(bbNoWeights),
                      tolerance = 0.075)

    sim$outcomes$weights <- rep(c(0.1,0.9), nrow(sim$outcomes) / 2)

    cyclopsData <- convertToCyclopsData(outcomes = sim$outcomes,
                                        covariates = sim$covariates,
                                        modelType = "pr")
    fitCyclopsYesWeights <- fitCyclopsModel(cyclopsData = cyclopsData)
    bsYesWeights <- runBootstrap(fitCyclopsYesWeights, replicates = 1999)

    bbYesWeights <- bootstrap(
        convertToDf(sim, 4),
        function(split) {
            data <- rsample::analysis(split)
            fit <- glm(y ~ V1 + V2 + V3 + V4, offset = log(time),
                       family = "poisson", weights = data$weights, data = data)
            coef(fit)
        }, R = 1999)

    bbYesStdError <- sqrt(apply(bbYesWeights, 2L, var))
    expect_equivalent(bsYesWeights$summary$std_err, bbYesStdError, tolerance = 0.02)

    expect_equivalent(bsYesWeights$summary[,c("bpi_lower", "bpi_upper")],
                      getAllPercentileIntervals(bbYesWeights),
                      tolerance = 0.075)
})

test_that("Large Cox bootstrap with and without weights", {

    # skip_on_cran()

    set.seed(123)
    sim <- simulateCyclopsData(nstrata=100,
                               ncovars=4,
                               nrows=1000,
                               effectSizeSd=0.5,
                               eCovarsPerRow=2,
                               model="survival")
    sim$outcomes$stratumId <- NULL
    sim$covariates$stratumId <- NULL
    cyclopsData <- convertToCyclopsData(outcomes = sim$outcomes,
                                        covariates = sim$covariates,
                                        modelType = "cox")
    fitCyclopsNoWeights <- fitCyclopsModel(cyclopsData = cyclopsData)
    bsNoWeights <- runBootstrap(fitCyclopsNoWeights, replicates = 3999)

    bbNoWeights <- bootstrap(
        convertToDf(sim, 4),
        function(split) {
            data <- rsample::analysis(split)
            fit <- coxph(Surv(time, y) ~ V1 + V2 + V3 + V4,
                         data = data)
            coef(fit)
        }, R = 3999)

    bbNoStdError <- sqrt(apply(bbNoWeights, 2L, var))
    expect_equivalent(bsNoWeights$summary$std_err, bbNoStdError, tolerance = 0.02)

    expect_equivalent(bsNoWeights$summary[,c("bpi_lower", "bpi_upper")],
                      getAllPercentileIntervals(bbNoWeights),
                      tolerance = 0.075)

    sim$outcomes$weights <- rep(c(0.1,0.9), nrow(sim$outcomes) / 2)

    cyclopsData <- convertToCyclopsData(outcomes = sim$outcomes,
                                        covariates = sim$covariates,
                                        modelType = "cox")
    fitCyclops <- fitCyclopsModel(cyclopsData = cyclopsData)
    bsYesWeights <- runBootstrap(fitCyclops, replicates = 3999)

    bbYesWeights <- bootstrap(
        convertToDf(sim, 4),
        function(split) {
            data <- rsample::analysis(split)
            fit <- coxph(Surv(time, y) ~ V1 + V2 + V3 + V4,
                         weights = data$weights, data = data)
            coef(fit)
        }, R = 3999)

    bbYesStdError <- sqrt(apply(bbYesWeights, 2L, var))
    expect_equivalent(bsYesWeights$summary$std_err, bbYesStdError, tolerance = 0.02)

    expect_equivalent(bsYesWeights$summary[,c("bpi_lower", "bpi_upper")],
                      getAllPercentileIntervals(bbYesWeights),
                      tolerance = 0.075)
})

library("testthat")
library("survival")

suppressWarnings(RNGversion("3.5.0"))

test_that("Check evaluate profile likelihood with two variables", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3.5,1,2,0
0, 3,  0,0,1
0, 2.5,1,0,1
0, 2,  1,1,1
0, 1.5,0,1,0
0, 1,  1,1,0
")
    data <- createCyclopsData(Surv(length, event) ~ x1 + x2, data = test,
                              modelType = "cox")
    fit <- fitCyclopsModel(data)

    x <- seq(from = 0.8, to = 0.9, length = 100)
    out <- getCyclopsProfileLogLikelihood(fit, "x1", x)

    argMax <- out$point[which(out$value == max(out$value))]
    expect_equivalent(coef(fit)["x1"], argMax, tolerance = 0.01)
})

test_that("Check evaluate profile likelihood with multiple threads", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3.5,1,2,0
0, 3,  0,0,1
0, 2.5,1,0,1
0, 2,  1,1,1
0, 1.5,0,1,0
0, 1,  1,1,0
")
    data <- createCyclopsData(Surv(length, event) ~ x1 + x2, data = test,
                              modelType = "cox")
    fit <- fitCyclopsModel(data, control = createControl(threads = 2))

    x <- seq(from = 0.8, to = 0.9, length = 100)
    out <- getCyclopsProfileLogLikelihood(fit, "x1", x)

    argMax <- out$point[which(out$value == max(out$value))]
    expect_equivalent(coef(fit)["x1"], argMax, tolerance = 0.01)
})

test_that("Check evaluate profile likelihood with one variable; cold-start", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3.5,1,2,0
0, 3,  0,0,1
0, 2.5,1,0,1
0, 2,  1,1,1
0, 1.5,0,1,0
0, 1,  1,1,0
")
    data <- createCyclopsData(Surv(length, event) ~ x1, data = test,
                              modelType = "cox")
    fit <- fitCyclopsModel(data, fixedCoefficients = c(TRUE))

    x <- seq(from = 0.3, to = 0.4, length = 100)
    out <- getCyclopsProfileLogLikelihood(fit, "x1", x)
    argMax <- out$point[which(out$value == max(out$value))]

    gold <- fitCyclopsModel(data, forceNewObject = TRUE)
    expect_equivalent(coef(gold)["x1"], argMax, tolerance = 0.01)
    expect_equivalent(coef(fit)["x1"], 0)
})

test_that("Check adapative profiling likelihood", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3.5,1,2,0
0, 3,  0,0,1
0, 2.5,1,0,1
0, 2,  1,1,1
0, 1.5,0,1,0
0, 1,  1,1,0
")
    data <- createCyclopsData(Surv(length, event) ~ x1 + x2, data = test,
                              modelType = "cox")
    fit <- fitCyclopsModel(data)

    expect_error(getCyclopsProfileLogLikelihood(fit, "x1"))
    expect_error(getCyclopsProfileLogLikelihood(fit, "x1", x = c(0,1), bounds = c(0.8, 0.9)))
    expect_error(getCyclopsProfileLogLikelihood(fit, "x1", bounds = c(0.8, 0.8)))

    out <- getCyclopsProfileLogLikelihood(fit, "x1", bounds = c(0, 2), initialGridSize = 10)
    expect_gt(nrow(out), 10)

    argMax <- out$point[which(out$value == max(out$value))]
    expect_equal(coef(fit)["x1"], argMax)
})

test_that("Check adapative profiling likelihood with derivative information", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3.5,1,2,0
0, 3,  0,0,1
0, 2.5,1,0,1
0, 2,  1,1,1
0, 1.5,0,1,0
0, 1,  1,1,0
")
    data <- createCyclopsData(Surv(length, event) ~ x1 + x2, data = test,
                              modelType = "cox")
    fit <- fitCyclopsModel(data)

    out <- getCyclopsProfileLogLikelihood(fit, "x1", bounds = c(0, 2), initialGridSize = 10,
                                          returnDerivatives = TRUE)

    pt <- which(out$value == max(out$value))
    argMax <- out[pt, ]
    expect_equal(0.0, argMax$derivative, tolerance = 1E-4)
    increasing <- out[pt - 1, ]
    expect_gt(increasing$derivative, 0.0)
})

test_that("Check adapative profiling likelihood, other covariate is perfect predictor", {
	skip("Needs debugging") # DEVELOP
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,1
0, 4,1,2,1
0, 4,  0,0,0
0, 4,1,0,1
0, 4,  1,1,1
0, 4,0,1,0
0, 4,  1,1,1
")
    data <- createCyclopsData(event ~ x1 + x2, data = test,
                              modelType = "cpr")

    # Recreating combination of circumstances that lead to infinite log likelihood
    # Note: infinite variance would have been selected by cross-validation if we had enough data:
    fit <- fitCyclopsModel(data,
                           prior = createPrior("laplace", variance = Inf, exclude = "x1"))
    expect_warning(getCyclopsProfileLogLikelihood(fit, "x1", bounds = c(0, 2), initialGridSize = 10),
                   "Failing to compute likelihood at entire initial grid")

    # This just produces NaN likelihood:
    fit <- fitCyclopsModel(data)
    expect_warning(getCyclopsProfileLogLikelihood(fit, "x1", bounds = c(0, 2), initialGridSize = 10),
                   "Failing to compute likelihood at entire initial grid")
})

library("testthat")
library("survival")

#
# Small exact, conditional logistic regression
#

test_that("Small exact, conditional logistic regression with no ties", {
    skip_on_cran()

    gold <- clogit(case ~ spontaneous + induced + strata(stratum), data=infert)

    dataPtrNoTies <- createCyclopsData(case ~ spontaneous + induced + strata(stratum),
                                            data = infert,
                                            modelType = "clr")

    cyclopsFitNoTies <- fitCyclopsModel(dataPtrNoTies, prior = createPrior("none"))

    dataPtrExact <- createCyclopsData(case ~ spontaneous + induced + strata(stratum),
                                           data = infert,
                                           modelType = "clr_exact")

    cyclopsFitExact <- fitCyclopsModel(dataPtrExact, prior = createPrior("none"))

    tolerance <- 1E-4

    expect_equal(coef(cyclopsFitNoTies), coef(gold), tolerance = tolerance)
    expect_equal(coef(cyclopsFitExact), coef(gold), tolerance = tolerance)
})

 test_that("Small, exact conditinal logistic regression with ties" , {
    skip_on_cran()

    withTies <- read.table(system.file("extdata/test1-clr.txt", package="Cyclops"), sep=",")
    names(withTies) <- c("stratum", "y",paste("x", 1:10, sep=""))

    goldWithTies <- clogit(y ~ x1 + x2 + x3 + x4 + x5 + x6 + x7 + x8 + x9 + x10 + strata(stratum),
                           data = withTies, method="exact")

    goldWithTiesBreslow <- clogit(y ~ x1 + x2 + x3 + x4 + x5 + x6 + x7 + x8 + x9 + x10 + strata(stratum),
                                  data = withTies, method="breslow")

    dataPtrWithTies <-createCyclopsData(y ~ x1 + x2 + x3 + x4 + x5 + x6 + x7 + x8 + x9 + x10 + strata(stratum),
                                             data = withTies,
                                             modelType = "clr_exact")

    cyclopsFitWithTies <- fitCyclopsModel(dataPtrWithTies, prior = createPrior("none"))

    dataPtrWithTiesBreslow <-createCyclopsData(y ~ x1 + x2 + x3 + x4 + x5 + x6 + x7 + x8 + x9 + x10 + strata(stratum),
                                                    data = withTies,
                                                    modelType = "clr")

    cyclopsFitWithTiesBreslow <- fitCyclopsModel(dataPtrWithTiesBreslow, prior = createPrior("none"))

    tolerance <- 1E-4

    expect_equal(coef(cyclopsFitWithTies), coef(goldWithTies), tolerance = tolerance)
    expect_equal(coef(cyclopsFitWithTiesBreslow), coef(goldWithTiesBreslow), tolerance = tolerance)
})

# test_that("Evaluate speed of exact method without ties (should be same as Breslow)", {
#     gold <- clogit(case ~ spontaneous + induced + strata(stratum), data=infert)
#
#     dataPtrBreslow <- createCyclopsData(case ~ spontaneous + induced + strata(stratum),
#                                              data = infert,
#                                              modelType = "clr")
#
#     cyclopsFitBreslow <- fitCyclopsModel(dataPtrBreslow, prior = createPrior("none"), forceNewObject = TRUE)
#
#     dataPtrExact <- createCyclopsData(case ~ strata(stratum), sparseFormula = ~spontaneous + induced,
#                                            data = infert,
#                                            modelType = "clr_exact")
#
#     cyclopsFitExact <- fitCyclopsModel(dataPtrExact, prior = createPrior("none"), forceNewObject = TRUE)
#
#     library(microbenchmark)
#     microbenchmark(
#         clogit(case ~ spontaneous + induced + strata(stratum), data=infert),
#         {
#             dataPtrBreslow <- createCyclopsData(case ~ spontaneous + induced + strata(stratum),
#                                                      data = infert,
#                                                      modelType = "clr")
#             fitCyclopsModel(dataPtrBreslow, prior = createPrior("none"), forceNewObject = TRUE)
#         },
#         fitCyclopsModel(dataPtrExact, prior = createPrior("none"), forceNewObject = TRUE),
#         times = 100L
#     )
# })

library("testthat")

context("test-multitypePoisson.R")

#
# Small Poisson MLE regression
#

test_that("Small multi-type Poisson dense regression", {
    dobson1 <- data.frame(
        counts = c(18,17,15,20,10,20,25,13,12),
        outcome = gl(3,1,9),
        treatment = gl(3,3)
    )
    dobson <- rbind(dobson1, dobson1)
    dobson$type = as.factor(c(rep("A",9),rep("B",9)))
    tolerance <- 1E-4

    goldFit <- glm(counts ~ outcome + treatment, data = dobson1, family = poisson())

    # glmFit <- glm(counts ~ outcome + treatment, data = dobson, contrasts = dobson$type,
    #               family = poisson()) # gold standard

    dataPtrD <- createCyclopsData(counts ~ outcome + treatment, data = dobson,
                                       type = dobson$type,
                                       modelType = "pr", method = "debug")

    cyclopsFitD <- fitCyclopsModel(dataPtrD,
                           prior = createPrior("none"),
                           control = createControl(noiseLevel = "silent"))

    dataPtrE <- createCyclopsData(Multitype(counts, type) ~ outcome + treatment, data = dobson,
                                       modelType = "pr", method = "debug")

    cyclopsFitE <- fitCyclopsModel(dataPtrE,
                                   prior = createPrior("none"),
                                   control = createControl(noiseLevel = "silent"))

    expect_equal(coef(cyclopsFitD), coef(cyclopsFitE))

    dataPtrI <- createCyclopsData(Multitype(counts, type) ~ 1, indicatorFormula = ~ outcome + treatment, data = dobson,
                                       modelType = "pr", method = "debug")

    cyclopsFitI <- fitCyclopsModel(dataPtrI,
                                   prior = createPrior("none"),
                                   control = createControl(noiseLevel = "silent"))

    expect_equal(coef(cyclopsFitI), coef(cyclopsFitD))

    dataPtrS <- createCyclopsData(Multitype(counts, type) ~ 1, sparseFormula = ~ outcome + treatment, data = dobson,
                                       modelType = "pr", method = "debug")

    cyclopsFitS <- fitCyclopsModel(dataPtrS,
                                   prior = createPrior("none"),
                                   control = createControl(noiseLevel = "silent"))

    expect_equal(coef(cyclopsFitS), coef(cyclopsFitD))
})

test_that("coef throws error when not converged", {
    dobson1 <- data.frame(
        counts = c(18,17,15,20,10,20,25,13,12),
        outcome = gl(3,1,9),
        treatment = gl(3,3)
    )
    dobson2 <- data.frame(
        counts = c(18,17,15,20,10,20,25,13,12)-10,
        outcome = gl(3,1,9),
        treatment = gl(3,3)
    )
    dobson <- rbind(dobson1, dobson2)
    dobson$type = as.factor(c(rep("A",9),rep("B",9)))
    tolerance <- 1E-4

    dataPtrD <- createCyclopsData(Multitype(counts, type) ~ outcome + treatment, data = dobson,
                                       modelType = "pr")

    cyclopsFitD <- fitCyclopsModel(dataPtrD,
                                   prior = createPrior(c("normal","normal"), c(0.0001,10), graph = "type"),
                                   control = createControl(noiseLevel = "silent"))
    expect_error(coef(cyclopsFitD), "did not converge")
})


test_that("confirm dimension check", {
    dobson1 <- data.frame(
        counts = c(18,17,15,20,10,20,25,13,12),
        outcome = gl(3,1,9),
        treatment = gl(3,3)
    )
    dobson2 <- data.frame(
        counts = c(18,17,15,20,10,20,25,13,12)-10,
        outcome = gl(3,1,9),
        treatment = gl(3,3)
    )
    dobson <- rbind(dobson1, dobson2)
    dobson$type = as.factor(c(rep("A",9),rep("B",9)))
    tolerance <- 1E-4

    dataPtrD <- createCyclopsData(Multitype(counts, type) ~ outcome + treatment, data = dobson,
                                       modelType = "pr")


    expect_error(fitCyclopsModel(dataPtrD,
                                 prior = createPrior(c("normal"), c(0.0001,10), graph = "type"),
                                 control = createControl(noiseLevel = "silent")), "dimensionality mismatch")
    expect_error(fitCyclopsModel(dataPtrD,
                                 prior = createPrior(c("normal", "normal"), c(0.0001), graph = "type"),
                                 control = createControl(noiseLevel = "silent")), "dimensionality mismatch")
})


test_that("Small multi-type Poisson with hierarchical prior", {
    dobson1 <- data.frame(
        counts = c(18,17,15,20,10,20,25,13,12),
        outcome = gl(3,1,9),
        treatment = gl(3,3)
    )
    dobson2 <- data.frame(
        counts = c(18,17,15,20,10,20,25,13,12)-10,
        outcome = gl(3,1,9),
        treatment = gl(3,3)
    )
    dobson <- rbind(dobson1, dobson2)
    dobson$type = as.factor(c(rep("A",9),rep("B",9)))
    tolerance <- 1E-4

    # glmFit <- glm(counts ~ outcome + treatment, data = dobson, contrasts = dobson$type,
    #               family = poisson()) # gold standard

    dataPtrD <- createCyclopsData(Multitype(counts, type) ~ outcome + treatment, data = dobson,
                                       modelType = "pr")

    cyclopsFitD <- fitCyclopsModel(dataPtrD,
                                   prior = createPrior(c("normal","normal"), c(0.0001,10), graph = "type"),
                                   control = createControl(noiseLevel = "silent", maxIterations = 2000))

    cyclopsFitE <- fitCyclopsModel(dataPtrD,
                                   prior = createPrior(c("normal","normal"), c(0.0001,0.0001), graph = "type"),
                                   control = createControl(noiseLevel = "silent"))


})


test_that("Check multitype SCCS", {
#
#   stratumID must be unique by type and all patient entries must be replicated by type
#     (how to specify this in input) ???
#
#     source("helper-conditionalPoisson.R")
#     tolerance <- 1E-6
#     gold.clogit <- clogit(event ~ exgr + agegr + strata(indiv) + offset(loginterval),
#                           data = oxford)
#
#     dataPtr <- createCyclopsData(event ~ exgr + agegr + strata(indiv), time = oxford$interval,
#                                       data = oxford,
#                                       modelType = "sccs")
#     cyclopsFit <- fitCyclopsModel(dataPtr,
#                                   prior = createPrior("none"))
#     expect_equal(logLik(cyclopsFit), logLik(gold.clogit)[1])
#     expect_equal(coef(cyclopsFit), coef(gold.clogit), tolerance = tolerance)
})

library("testthat")

context("test-smallMM.R")

#
# Small MM regression
#


test_that("Small Bernoulli dense regression with MM algorithm", {
    binomial_bid <- c(1,5,10,20,30,40,50,75,100,150,200)
    binomial_n <- c(31,29,27,25,23,21,19,17,15,15,15)
    binomial_y <- c(0,3,6,7,9,13,17,12,11,14,13)

    log_bid <- log(c(rep(rep(binomial_bid, binomial_n - binomial_y)), rep(binomial_bid, binomial_y)))
    y <- c(rep(0, sum(binomial_n - binomial_y)), rep(1, sum(binomial_y)))

    tolerance <- 1E-3

    gold <- glm(y ~ log_bid, family = binomial()) # gold standard

    data <- createCyclopsData(y ~ log_bid, modelType = "lr")
    fit <- fitCyclopsModel(data, prior = createPrior("none"),
                           control = createControl(algorithm = "mm"))
    expect_equal(coef(fit), coef(gold), tolerance = tolerance)
})

test_that("Small Bernoulli sparse regression with MM algorithm", {
    binomial_bid <- c(1,5,10,20,30,40,50,75,100,150,200)
    binomial_n <- c(31,29,27,25,23,21,19,17,15,15,15)
    binomial_y <- c(0,3,6,7,9,13,17,12,11,14,13)

    log_bid <- log(c(rep(rep(binomial_bid, binomial_n - binomial_y)), rep(binomial_bid, binomial_y)))
    y <- c(rep(0, sum(binomial_n - binomial_y)), rep(1, sum(binomial_y)))

    tolerance <- 1E-3

    gold <- glm(y ~ log_bid, family = binomial()) # gold standard

    data <- createCyclopsData(y ~ 1, sparseFormula = ~ log_bid, modelType = "lr")
    fit <- fitCyclopsModel(data, prior = createPrior("none"),
                           control = createControl(algorithm = "mm"))
    expect_equal(coef(fit), coef(gold), tolerance = tolerance)
})

test_that("Small Bernoulli indicator regression with MM algorithm", {
    binomial_bid <- c(0,0,0,0,0,0,1,1,1,1,1)
    binomial_n <- c(31,29,27,25,23,21,19,17,15,15,15)
    binomial_y <- c(0,3,6,7,9,13,17,12,11,14,13)

    log_bid <- c(rep(rep(binomial_bid, binomial_n - binomial_y)), rep(binomial_bid, binomial_y))
    y <- c(rep(0, sum(binomial_n - binomial_y)), rep(1, sum(binomial_y)))

    tolerance <- 1E-3

    gold <- glm(y ~ log_bid, family = binomial()) # gold standard

    data <- createCyclopsData(y ~ 1, indicatorFormula = ~ log_bid, modelType = "lr")
    fit <- fitCyclopsModel(data, prior = createPrior("none"),
                           control = createControl(algorithm = "mm"))
    expect_equal(coef(fit), coef(gold), tolerance = tolerance)
})

library("testthat")
library("survival")
library("gnm")
library("Andromeda")

context("test-dataConversionStratified.R")
suppressWarnings(RNGversion("3.5.0"))

test_that("Test clr", {

  gold <- clogit(case ~ spontaneous + induced + strata(stratum), data=infert)

  #Convert infert dataset to Cyclops format:
  covariates <- data.frame(stratumId = rep(infert$stratum,2),
                           rowId = rep(1:nrow(infert),2),
                           covariateId = rep(1:2,each=nrow(infert)),
                           covariateValue = c(infert$spontaneous,infert$induced))
  outcomes <- data.frame(stratumId = infert$stratum,
                         rowId = 1:nrow(infert),
                         y = infert$case)
  #Make sparse:
  covariates <- covariates[covariates$covariateValue != 0,]

  andr <- andromeda(outcomes = outcomes, covariates = covariates)
  cyclopsDataAndr <- convertToCyclopsData(andr$outcomes, andr$covariates, modelType = "clr", addIntercept = FALSE)
  fitAndr <- fitCyclopsModel(cyclopsDataAndr,prior = createPrior("none"))

  cyclopsData <- convertToCyclopsData(outcomes,covariates,modelType = "clr",addIntercept = FALSE)
  fit <- fitCyclopsModel(cyclopsData,prior = createPrior("none"))

  tolerance <- 1E-4
  expect_equal(as.vector(coef(fit)), as.vector(coef(gold)), tolerance = tolerance)
  expect_equal(as.vector(coef(fitAndr)), as.vector(coef(gold)), tolerance = tolerance)
})

test_that("Test stratified cox", {

  test1 <- data.frame(time=c(4,3,1,1,2,2,3),
                      status=c(1,1,1,0,1,1,0),
                      x=c(0,2,1,1,1,0,0),
                      sex=c(0,0,0,0,1,1,1))

  gold <- coxph(Surv(time, status) ~ x + strata(sex), test1, method="breslow")

  cyclopsDataFormula <- createCyclopsData(Surv(time, status) ~ x + strata(sex), data=test1,modelType="cox")
  fitFormula <- fitCyclopsModel(cyclopsDataFormula)

  #Convert to data.frames for Cyclops:
  covariates <- data.frame(stratumId = test1$sex,
                           rowId = 1:nrow(test1),
                           covariateId = rep(1,nrow(test1)),
                           covariateValue = test1$x)
  outcomes <- data.frame(stratumId = test1$sex,
                         rowId = 1:nrow(test1),
                         y = test1$status,
                         time = test1$time)

  #Make sparse:
  covariates <- covariates[covariates$covariateValue != 0,]

  andr <- andromeda(outcomes = outcomes, covariates = covariates)
  cyclopsDataAndr <- convertToCyclopsData(andr$outcomes, andr$covariates ,modelType = "cox")
  fitAndr <- fitCyclopsModel(cyclopsDataAndr,prior = createPrior("none"))

  cyclopsData <- convertToCyclopsData(outcomes,covariates,modelType = "cox")
  fit <- fitCyclopsModel(cyclopsData,prior = createPrior("none"))

  tolerance <- 1E-4
  expect_equal(as.vector(coef(fit)), as.vector(coef(fitFormula)), tolerance = tolerance)
  expect_equal(as.vector(coef(fit)), as.vector(coef(gold)), tolerance = tolerance)
  expect_equal(as.vector(coef(fitAndr)), as.vector(coef(gold)), tolerance = tolerance)
})

test_that("Test stratified cox using lung dataset ", {

  test <- lung
  test[is.na(test)] <- 0 # Don't want to bother with missing values

  gold <- coxph(Surv(time, status) ~ age + ph.ecog + ph.karno + pat.karno + meal.cal + wt.loss + strata(sex), test, method="breslow")

  cyclopsDataFormula <- createCyclopsData(Surv(time, status) ~ age + ph.ecog + ph.karno + pat.karno + meal.cal + wt.loss + strata(sex), data=test,modelType="cox")
  fitFormula <- fitCyclopsModel(cyclopsDataFormula)

  #Convert to data.frames for Cyclops:
  nCovars = 6
  covariates <- data.frame(stratumId = rep(test$sex,nCovars),
                           rowId = rep(1:nrow(test),nCovars),
                           covariateId = rep(1:nCovars,each = nrow(test)),
                           covariateValue = c(test$age,test$ph.ecog,test$ph.karno,test$pat.karno,test$meal.cal,test$wt.loss))
  outcomes <- data.frame(stratumId = test$sex,
                         rowId = 1:nrow(test),
                         y = test$status-1,
                         time = test$time)

  #Make sparse:
  covariates <- covariates[covariates$covariateValue != 0,]

  andr <- andromeda(outcomes = outcomes, covariates = covariates)
  cyclopsDataAndr <- convertToCyclopsData(andr$outcomes,andr$covariates,modelType = "cox")
  fitAndr <- fitCyclopsModel(cyclopsDataAndr,prior = createPrior("none"))

  cyclopsData <- convertToCyclopsData(outcomes,covariates,modelType = "cox")
  fit <- fitCyclopsModel(cyclopsData,prior = createPrior("none"))

  tolerance <- 1E-4
  expect_equal(as.vector(coef(fit)), as.vector(coef(fitFormula)), tolerance = tolerance)
  expect_equal(as.vector(coef(fit)), as.vector(coef(gold)), tolerance = tolerance)
  expect_equal(as.vector(coef(fitAndr)), as.vector(coef(gold)), tolerance = tolerance)
})


test_that("Test conditional poisson regression", {
  skip_on_cran() # Do not run on CRAN
  #skip("Do not run")
  set.seed(123)
  sim <- simulateCyclopsData(nstrata = 2, nrows = 10000, ncovars = 2, eCovarsPerRow = 0.5, effectSizeSd = 1,model = "poisson")
  covariates <- sim$covariates
  outcomes <- sim$outcomes

  #Convert to data format for gnm:
  ncovars <- max(covariates$covariateId)
  nrows <- nrow(outcomes)
  m <- matrix(0,nrows,ncovars)
  for (i in 1:nrow(covariates)){
    m[covariates$rowId[i],covariates$covariateId[i]] <- 1
  }
  data <- as.data.frame(m)

  data$rowId <- 1:nrow(data)
  data <- merge(data,outcomes)
  data <- data[order(data$stratumId,data$rowId),]
  formula <- as.formula(paste(c("y ~ V1",paste("V",2:ncovars,sep="")),collapse=" + "))

  gold <- gnm(formula, family=poisson, offset=log(time), eliminate=as.factor(stratumId), data = data)

  formula <- as.formula(paste(c("y ~ V1",paste("V",2:ncovars,sep=""),"strata(stratumId)"),collapse=" + "))
  cyclopsDataFormula <- createCyclopsData(formula, offset=log(time), data=data,modelType="cpr")
  fitFormula <- fitCyclopsModel(cyclopsDataFormula)

  andr <- andromeda(outcomes = outcomes, covariates = covariates)
  cyclopsDataAndr <- convertToCyclopsData(andr$outcomes,andr$covariates,modelType = "cpr",addIntercept = FALSE)
  fitAndr <- fitCyclopsModel(cyclopsDataAndr,prior = createPrior("none"))

  cyclopsData <- convertToCyclopsData(outcomes,covariates,modelType = "cpr",addIntercept = FALSE)
  fit <- fitCyclopsModel(cyclopsData,prior = createPrior("none"))

  tolerance <- 1E-4
  expect_equal(as.vector(sort(coef(fit))), as.vector(sort(coef(fitFormula))), tolerance = tolerance)
  expect_equal(as.vector(sort(coef(fit))), as.vector(sort(coef(gold))), tolerance = tolerance)
  expect_equal(as.vector(sort(coef(fitAndr))), as.vector(sort(coef(gold))), tolerance = tolerance)
})

library("testthat")

context("test-cv.R")
suppressWarnings(RNGversion("3.5.0"))

# ### COX ERROR
#
# library(Cyclops)
# set.seed(100)
# data <- simulateCyclopsData(nstrata=1,nrows=1000,ncovars=200,model="survival")
# cyclopsData <- convertToCyclopsData(data$outcomes,data$covariates,modelType = "cox")
#
# prior <- createPrior("laplace", useCrossValidation = TRUE)
# control <- createControl(noiseLevel = "quiet",lowerLimit = 0.000001,upperLimit = 100,
#                          cvType = "auto", startingVariance = 0.1,
#                          seed = 100,
#                          selectorType = "default")
# fit <- fitCyclopsModel(cyclopsData,prior=prior,control=control)
#
#
# fit <- fitCyclopsModel(cyclopsData,
#                        prior = createPrior("laplace", variance = 0.0464159),
#                        control=control)
#
# # This generates nan on first eval
# library(Cyclops)
# set.seed(100)
# data <- simulateCyclopsData(nstrata=1,nrows=1000,ncovars=200,model="survival")
# cyclopsData <- convertToCyclopsData(data$outcomes,data$covariates,modelType = "cox")
# prior <- createPrior("laplace", useCrossValidation = TRUE)
# control <- createControl(noiseLevel = "quiet",lowerLimit = 0.0464159,upperLimit = 0.0464159,gridSteps=1,
#                          seed = 100,
#                          selectorType = "default")
# fit <- fitCyclopsModel(cyclopsData,prior=prior,control=control)

test_that("Grid in R and auto-search in C++", {
    skip_on_cran() # Do not run on CRAN
    skip("Do not run")

    seed <- 666
    set.seed(seed)

    nrows <- 1000
    ncovars <- 200

    data <- simulateCyclopsData(nstrata = 1,
                                nrows = nrows,
                                ncovars = ncovars,
                                model = "logistic")

    cyclopsData <- convertToCyclopsData(data$outcomes,
                                        data$covariates,
                                        modelType = "lr",
                                        addIntercept = TRUE)

    prior <- createPrior("laplace", exclude = c(0), useCrossValidation = TRUE)

    control <- createControl(noiseLevel = "quiet",
                              cvType = "auto",
                              cvRepetitions = 1,
                              seed = seed)

    out <- capture.output(fit <- fitCyclopsModel(cyclopsData, prior, control))

    cv <- Cyclops:::getCrossValidationInfo(fit)

    printValue <- as.numeric(gsub(".* log likelihood \\((.*)\\) estimated.*", "\\1",
                                  capture.output(cat(out))))

    expect_equal(cv$ordinate, printValue)

    outerGrid <- c(0.5, 1, 2)
    expect_error(createAutoGridCrossValidationControl(outerGrid, autoPosition = 0), "Auto-position")
    expect_error(createAutoGridCrossValidationControl(outerGrid, autoPosition = 3), "Auto-position")

    expect_error(
        fitCyclopsModel(cyclopsData,
                        prior = prior,
                        control = createAutoGridCrossValidationControl(outerGrid)),
        "Auto-grid"
    )

    prior <-
        createParameterizedPrior(
            c("none", rep("laplace", ncovars)),
            parameterize = function(x) {
                lapply(1:(ncovars + 1),
                       function(i) {
                           if (i < (ncovars / 2)) {
                               return(c(0, x[1]))
                           } else {
                               return(c(0, x[1] * x[2]))
                           }
                       })
            },
            values = c(1, 0.5),
            useCrossValidation = TRUE
        )

    fit <- fitCyclopsModel(cyclopsData,
                    prior = prior,
                    control = createAutoGridCrossValidationControl(outerGrid,
                                                                   autoPosition = 1))
    expect_equal(length(fit$searchResults), 3)
    expect_equal(length(Cyclops:::getCrossValidationInfo(fit)$point), 2)

})

test_that("Specify starting variance with auto-search", {
    skip_on_cran() # Do not run on CRAN
    skip("Do not run")
    seed <- 666
    set.seed(seed)
    ntrain <- 100
    data <- simulateCyclopsData(nstrata = 1,
                         nrows = ntrain,
                         ncovars = 2000,
                         model = "logistic")
    cyclopsData <- convertToCyclopsData(data$outcomes,
                                        data$covariates,
                                        modelType = "lr",
                                        addIntercept = TRUE)
    prior <- createPrior("laplace", exclude = c(0), useCrossValidation = TRUE)
    control1 <- createControl(noiseLevel = "quiet",
                              cvType = "auto",
                              cvRepetitions = 1,
                              seed = seed)
    out1 <- capture.output(fit1 <- fitCyclopsModel(cyclopsData,
                            prior = prior,
                            control = control1))

    control2 <- createControl(noiseLevel = "quiet",
                              cvType = "auto",
                              cvRepetitions = 1,
                              seed = seed,
                              startingVariance = 0.1)

    out2 <- capture.output(fit2 <- fitCyclopsModel(cyclopsData,
                                                   prior = prior,
                                                   control = control2))

    expect_true(grep("default", out1[3]) == 1)
    expect_true(length(grep("default", out2[3])) == 0)
    expect_false(out1[3] == out2[3])
})

test_that("Using multi-core CV", {
    skip_on_cran() # Do not run on CRAN
    skip("Do not run")
    ntest <- 1000
    ntrain <- 1000

    set.seed(666)

    data <- simulateCyclopsData(nstrata=1,nrows=ntest+ntrain,ncovars=2000,model="logistic")
    test <- list(outcomes = data$outcomes[1:ntest,], covariates = data$covariates[data$covariates$rowId %in% data$outcomes$rowId[1:ntest],])
    train <- list(outcomes = data$outcomes[(ntest+1):(ntest+ntrain),], covariates = data$covariates[data$covariates$rowId %in% data$outcomes$rowId[(ntest+1):(ntest+ntrain)],])
    cyclopsData <- convertToCyclopsData(train$outcomes,train$covariates,modelType = "lr",addIntercept = TRUE)
    prior <- createPrior("laplace", exclude = c(0), useCrossValidation = TRUE)

    # 1 thread, cold start
    control <- createControl(noiseLevel = "silent", cvType = "auto", cvRepetitions = 1, seed = 666, threads = 1, resetCoefficients = TRUE)
    time1 <- system.time(fit1 <- fitCyclopsModel(cyclopsData,prior=prior,control=control, forceNewObject = TRUE))

    # 4 thread, cold start
    control <- createControl(noiseLevel = "silent", cvType = "auto", cvRepetitions = 1, seed = 666, threads = 4, resetCoefficients = TRUE)
    time2 <- system.time(fit2 <- fitCyclopsModel(cyclopsData,prior=prior,control=control, forceNewObject = TRUE))

    # 1 thread, warm start
    control <- createControl(noiseLevel = "silent", cvType = "auto", cvRepetitions = 1, seed = 666, threads = 1)
    time3 <- system.time(fit3 <- fitCyclopsModel(cyclopsData,prior=prior,control=control, forceNewObject = TRUE))

    # 4 thread, warm start
    control <- createControl(noiseLevel = "silent", cvType = "auto", cvRepetitions = 1, seed = 666, threads = 4)
    time4 <- system.time(fit4 <- fitCyclopsModel(cyclopsData,prior=prior,control=control,  forceNewObject = TRUE))

    cat("Variance estimates\n")
    cat(fit1$variance,"\n")
    cat(fit2$variance,"\n")
    cat(fit3$variance,"\n")
    cat(fit4$variance,"\n")

    # Cold starting each model fit will return equal single- to multi-core estimates
    expect_equal(fit1$variance, fit2$variance)

    cat("Times\n")
    cat(time1,"\n")
    cat(time2,"\n")
    cat(time3,"\n")
    cat(time4,"\n")

    # Multi-core should be faster
    expect_less_than(time2[3], time1[3])
    # Warm starting should be faster
    expect_less_than(time3[3], time1[3])
})

test_that("Seed gets returned", {
    y <- 0
    x <- 1
    data <- createCyclopsData(y ~ x, modelType = "lr")
    fit <- fitCyclopsModel(data, control = createControl(seed = 123), warnings = FALSE)
    expect_equal(fit$seed, 123)

    fit <- fitCyclopsModel(data, forceNewObject = TRUE,
                           control = createControl(seed = NULL), warnings = FALSE)
    expect_true(!is.null(fit$seed))
})

test_that("Auto cvRepetitions", {
    data <- simulateCyclopsData(nstrata = 1, nrows = 1000, ncovars = 10, model = "logistic")
    cyclopsData <- convertToCyclopsData(data$outcomes,
                                        data$covariates,
                                        modelType = "lr",
                                        addIntercept = TRUE)

    prior <- createPrior("laplace", exclude = c(0), useCrossValidation = TRUE)

    control <- createControl(noiseLevel = "quiet",
                             cvType = "auto",
                             cvRepetitions = "auto")

    fit <- fitCyclopsModel(cyclopsData,
                           prior = prior,
                           control = control)

    expect_equal(fit$cvRepetitions, 4)
})

library("testthat")
library("Andromeda")

context("test-predict.R")

test_that("Test predict for Poisson regression", {

    sim <- simulateCyclopsData(nstrata = 1, nrows = 10000, ncovars = 2, eCovarsPerRow = 0.5, effectSizeSd = 1,model = "poisson")
    covariates <- sim$covariates
    outcomes <- sim$outcomes

    cyclopsData <- convertToCyclopsData(outcomes, covariates, modelType = "pr", addIntercept = TRUE)
    fit <- fitCyclopsModel(cyclopsData, prior = createPrior("none"))
    predictOriginal <- predict(fit)

    # Test using data frames
    predictNew <- predict(fit, outcomes, covariates)
    expect_equal(predictOriginal, predictNew)

    # Test using Andromeda
    andr <- andromeda(outcomes = outcomes, covariates = covariates)
    predictNew <- predict(fit, andr$outcomes, andr$covariates)
    expect_equal(predictOriginal, predictNew)
})

test_that("Test predict for logistic regression", {

    sim <- simulateCyclopsData(nstrata = 1, nrows = 10000, ncovars = 2, eCovarsPerRow = 0.5, effectSizeSd = 1,model = "logistic")
    covariates <- sim$covariates
    outcomes <- sim$outcomes

    cyclopsData <- convertToCyclopsData(outcomes, covariates, modelType = "lr", addIntercept = TRUE)
    fit <- fitCyclopsModel(cyclopsData, prior = createPrior("none"))
    predictOriginal <- predict(fit)

    # Test using data frames
    predictNew <- predict(fit, outcomes, covariates)
    expect_equal(predictOriginal, predictNew)

    # Test using Andromeda
    andr <- andromeda(outcomes = outcomes, covariates = covariates)
    predictNew <- predict(fit, andr$outcomes, andr$covariates)
    expect_equal(predictOriginal, predictNew)
})

test_that("Test predict for pr with all-zero betas", {

    sim <- simulateCyclopsData(nstrata = 1, nrows = 1000, ncovars = 2, eCovarsPerRow = 0.5, effectSizeSd = 1,model = "poisson")
    covariates <- sim$covariates
    outcomes <- sim$outcomes

    cyclopsData <- convertToCyclopsData(outcomes, covariates, modelType = "pr", addIntercept = TRUE)
    fit <- fitCyclopsModel(cyclopsData, prior = createPrior("laplace", variance = 0.0001, exclude = 0))
    predictOriginal <- predict(fit)

    # Test using data frames
    predictNew <- predict(fit, outcomes, covariates)
    expect_equal(predictOriginal, predictNew)

    # Test using Andromeda
    andr <- andromeda(outcomes = outcomes, covariates = covariates)
    predictNew <- predict(fit, andr$outcomes, andr$covariates)
    expect_equal(predictOriginal, predictNew)

    # newCovariates <- as.ffdf(covariates)
    # newOutcomes <- as.ffdf(outcomes)
    # object <- fit
})

test_that("Test predict for lr with all-zero betas", {

    sim <- simulateCyclopsData(nstrata = 1, nrows = 1000, ncovars = 2, eCovarsPerRow = 0.5, effectSizeSd = 1,model = "logistic")
    covariates <- sim$covariates
    outcomes <- sim$outcomes

    cyclopsData <- convertToCyclopsData(outcomes, covariates, modelType = "lr", addIntercept = TRUE)
    fit <- fitCyclopsModel(cyclopsData, prior = createPrior("laplace", variance = 0.0001, exclude = 0))
    predictOriginal <- predict(fit)

    # Test using data frames
    predictNew <- predict(fit, outcomes, covariates)
    expect_equal(predictOriginal, predictNew)

    # Test using Andromeda
    andr <- andromeda(outcomes = outcomes, covariates = covariates)
    predictNew <- predict(fit, andr$outcomes, andr$covariates)
    expect_equal(predictOriginal, predictNew)
})

library("testthat")

context("test-largeBernoulli.R")
suppressWarnings(RNGversion("3.5.0"))

#
# Large Bernoulli regression
#

# test_that("Large Bernoulli CCD data file read", {
# 	tolerance <- 1E-4
# 	dataPtr <- readCyclopsData(system.file("extdata/CCD_LOGISTIC_TEST_17var.txt",
# 																		 package="Cyclops"), "lr")
# 	expect_equal(getNumberOfRows(dataPtr), 22296) # Reads rows
# 	expect_equal(getNumberOfStrata(dataPtr), 22296) # Generates unique ids
# 	expect_equal(getNumberOfCovariates(dataPtr), 18) # Adds intercept
#
# 	cyclopsFit <- fitCyclopsModel(dataPtr, prior = createPrior("none"),
# 												control = createControl(noiseLevel = "silent"))
# 	expect_equal(cyclopsFit$log_likelihood * 2, -1578.046, tolerance = tolerance) # SAS fit
# 	expect_named(coef(cyclopsFit)) # Reads covariate names from file
# 	expect_named(predict(cyclopsFit)) # Reads row names from file
# })

test_that("Separable covariates in logistic regression", {
    skip_on_cran()
    set.seed(666)

    simulant <- simulateCyclopsData(nstrata = 1,
                                ncovars = 5,
                                model = "logistic")

    successes <- simulant$outcomes$rowId[simulant$outcomes$y == 1]
    separable <- simulant$covariates[simulant$covariates$covariateId == 3 &
                                         simulant$covariates$rowId %in% successes,]
    separable$covariateId <- 6

    simulant$covariates <- rbind(simulant$covariates, separable)

    data <- convertToCyclopsData(simulant$outcomes, simulant$covariates, modelType = "lr",
                                        addIntercept = TRUE)

    fit <- fitCyclopsModel(data, prior = createPrior("none"))

    expect_error(coef(fit), "did not converge")

    # Use separability condition
    separability <- getUnivariableSeparability(data)
    expect_equal(sum(separability), 1.0)

    scFit <- fitCyclopsModel(data, prior = createPrior("none"),
                             forceNewObject = TRUE,
                             fixedCoefficients = separability)

    expect_equivalent(coef(scFit)[separability], 0.0)

    # Use estimate existence condition
    separability <- is.nan(coef(fit, ignoreConvergence = TRUE))

    ecFit <- fitCyclopsModel(data, prior = createPrior("none"),
                             forceNewObject = TRUE,
                             fixedCoefficients = separability)

    expect_equivalent(coef(ecFit)[separability], 0.0)

    # Use non-separable MLE prior
    nsFit <- fitCyclopsModel(data, prior = createNonSeparablePrior(),
                             forceNewObject = TRUE)
    expect_true(is.na(coef(nsFit)[7]))
})

test_that("Separable covariates in cox regression", {
    skip_on_cran()
    set.seed(666)

    simulant <- simulateCyclopsData(nstrata = 1,
                                    ncovars = 5,
                                    model = "survival")

    successes <- simulant$outcomes$rowId[simulant$outcomes$y == 1]
    separable <- simulant$covariates[simulant$covariates$covariateId == 3 &
                                         simulant$covariates$rowId %in% successes,]
    separable$covariateId <- 6

    simulant$covariates <- rbind(simulant$covariates, separable)

    data <- convertToCyclopsData(simulant$outcomes, simulant$covariates, modelType = "cox")

    fit <- fitCyclopsModel(data, prior = createPrior("none"))

    #expect_error(coef(fit), "did not converge") TODO

    separability <- getUnivariableSeparability(data)
    expect_equal(sum(separability), 1.0)

    fit <- fitCyclopsModel(data, prior = createPrior("none"),
                           forceNewObject = TRUE,
                           fixedCoefficients = separability)

    expect_equivalent(coef(fit)[separability], 0.0)
})

library("testthat")
library("survival")
library("gnm")

tolerance <- 1E-4

test_that("Check very small Cox example with ties, but without weights",{
    test <- read.table(header=T, sep = ",", text = "
                       start, length, status, x1, x2
                       0, 5, 0, 0, 1
                       0, 4, 1, 1, 2
                       0, 3, 0, 0, 1
                       0, 2, 0, 1, 0
                       0, 2, 1, 0, 0
                       0, 2, 1, 0, 1
                       0, 1, 0, 0, 0
                       0, 1, 1, 2, 1 ")
    tolerance <- 1E-4

    gold <- coxph(Surv(length, status) ~ x1 + x2, test, ties = "breslow")

    cyclopsData <- createCyclopsData(Surv(length, status) ~ x1 + x2, data = test, modelType = "cox")
    cyclopsFit <- fitCyclopsModel(cyclopsData)

    expect_equal(coef(gold), coef(cyclopsFit), tolerance = tolerance)
    expect_equivalent(logLik(gold),logLik(cyclopsFit))

})

test_that("Check very small Cox example without ties, but with weights",{
    test <- read.table(header=T, sep = ",", text = "
                   start, length, status, x1, x2
                       0, 5,  0,0,1
                       0, 4,  1,1,2
                       0, 3,  0,0,1
                       0, 2,  0,1,0
                       0, 2,  1,0,0
                       0, 1,  0,1,0
                       0, 1,  1,2,1 ")
    weights <- c(1,2,1,2,3,2,1)

    gold <- coxph(Surv(length, status) ~ x1 + x2, test, weights = weights)

    cyclopsData <- createCyclopsData(Surv(length, status) ~ x1 + x2, data = test, modelType = "cox")
    cyclopsFit <- fitCyclopsModel(cyclopsData, weights = weights)

    expect_equal(coef(gold), coef(cyclopsFit))
    expect_equivalent(logLik(gold),logLik(cyclopsFit))
})

test_that("Check very small Cox example with ties, with weights",{
    test <- read.table(header=T, sep = ",", text = "
                       start, length, status, x1, x2
                       0, 5, 0, 0, 1
                       0, 4, 1, 1, 2
                       0, 3, 0, 0, 1
                       0, 2, 0, 1, 0
                       0, 2, 1, 0, 0
                       0, 2, 1, 0, 1
                       0, 1, 0, 0, 0
                       0, 1, 1, 2, 1 ")
    weights <- c(1,2,1,2,4,3,2,1)

    gold <- coxph(Surv(length, status) ~ x1 + x2, test, weights, ties = "breslow")

    cyclopsData <- createCyclopsData(Surv(length, status) ~ x1 + x2, data = test, modelType = "cox")
    cyclopsFit <- fitCyclopsModel(cyclopsData, weights=weights)

    expect_equal(coef(gold), coef(cyclopsFit))
    expect_equivalent(logLik(gold),logLik(cyclopsFit))

})

test_that("Check predictive log likelihood",{
    test <- read.table(header=T, sep = ",", text = "
                       start, length, event, x1, x2
                       0, 4,  1,0,0
                       0, 3,  1,2,0
                       0, 3,  0,0,1
                       0, 2,  1,0,1
                       0, 2,  1,1,1
                       0, 1,  0,1,0
                       0, 1,  1,1,0")

    gold <- coxph(Surv(length, event) ~ x1 + strata(x2), test, ties = "breslow")

    # Double the data
    test2 <- rbind(data.frame(test, index = 1:7), data.frame(test, index = 1:7))
    test2 <- test2[order(test2$index),]

    data <- createCyclopsData(Surv(length, event) ~ x1 + strata(x2), data = test2,
                              modelType = "cox")

    # Fit with the second set
    weights = rep(c(0,1), 7)
    fit <- fitCyclopsModel(data, weights = weights)

    tolerance <- 1E-4
    expect_equal(coef(fit), coef(gold), tolerance = tolerance)
    expect_equivalent(logLik(fit), logLik(gold))

    # Get predictive log likelihood of first set
    pred <- Cyclops:::.cyclopsGetNewPredictiveLogLikelihood(fit$interface, weights = 1 - weights)
    expect_equal(pred, as.numeric(logLik(gold)), tolerance)
})

test_that("Check very small Cox example with weighting", {
    test <- read.table(header=T, sep = ",", text = "
                       start, length, event, x1, x2
                       0, 4,  1,0,0
                       0, 3.5,1,2,0
                       0, 3,  0,0,1
                       0, 2.5,1,0,1
                       0, 2,  1,1,1
                       0, 1.5,0,1,0
                       0, 1,  1,1,0")

    gold <- coxph(Surv(length, event) ~ x1 + x2, test, ties = "breslow")

    # Duplicate some rows, so we can reweigh later:
    testDup <- rbind(test, test[c(1,3,4),])
    weights <- c(0.5, 1, 0.5, 0.5, 1, 1, 1, 0.5, 0.5, 0.5)

    goldDup <- coxph(Surv(length, event) ~ x1 + x2, testDup, weights = weights, ties = "breslow")

    expect_equal(coef(gold), coef(goldDup))

    cyclopsData <- createCyclopsData(Surv(length, event) ~ x1 + x2, data = testDup, modelType = "cox")
    cyclopsFit <- fitCyclopsModel(cyclopsData, weights = weights)

    expect_equal(coef(gold), coef(cyclopsFit), tolerance = 1E-6)
    expect_equal(coef(goldDup), coef(cyclopsFit), tolerance = 1E-6)

    # Weights with sparse and indicator values
    cyclopsData4 <- createCyclopsData(Surv(length, event) ~ 1, sparseFormula = ~ x1,
                                      indicatorFormula = ~ x2,
                                      data = testDup, modelType = "cox",
                                      weights = weights)
    cyclopsFit4 <- fitCyclopsModel(cyclopsData4)
    expect_equal(coef(goldDup), coef(cyclopsFit4), tolerance = 0.000001)

})

test_that("Multi-core weights", {
    test <- read.table(header=T, sep = ",", text = "
                       start, length, status, x1, x2
                       0, 5,  0,0,1
                       0, 4,  1,1,2
                       0, 3,  0,0,1
                       0, 2,  0,1,0
                       0, 2,  1,0,0
                       0, 1,  0,1,0
                       0, 1,  1,2,1 ")
    weights <- c(1,2,1,2,3,2,1)

    cyclopsData <- createCyclopsData(Surv(length, status) ~ x1 + x2, data = test, modelType = "cox")
    cyclopsFit <- fitCyclopsModel(cyclopsData, weights = weights,
                                  control = createControl(noiseLevel = "silent"))
    ci1 <- confint(cyclopsFit, "x1")
    cyclopsFit <- fitCyclopsModel(cyclopsData, weights = weights,
                                  control = createControl(threads = 2,
                                                          noiseLevel = "silent"))
    ci2 <- confint(cyclopsFit, "x1")

    expect_equal(ci1, ci2)
})

test_that("Small Bernoulli dense regression with weighting", {
    binomial_bid <- c(1,5,10,20,30,40,50,75,100,150,200)
    binomial_n <- c(31,29,27,25,23,21,19,17,15,15,15)
    binomial_y <- c(0,3,6,7,9,13,17,12,11,14,13)

    log_bid <- log(c(rep(rep(binomial_bid, binomial_n - binomial_y)), rep(binomial_bid, binomial_y)))
    y <- c(rep(0, sum(binomial_n - binomial_y)), rep(1, sum(binomial_y)))

    weights <- as.numeric(gl(5,1,length(y)))

    tolerance <- 1E-4

    glmFit <- glm(y ~ log_bid, family = binomial(), weights = weights) # gold standard

    dataPtrD <- createCyclopsData(y ~ log_bid, modelType = "lr")
    cyclopsFitD <- fitCyclopsModel(dataPtrD, prior = createPrior("none"),
                                   control = createControl(noiseLevel = "silent"), weights = weights)

    expect_equal(coef(cyclopsFitD), coef(glmFit), tolerance = tolerance)
    expect_equal(cyclopsFitD$log_likelihood, logLik(glmFit)[[1]], tolerance = tolerance)
    expect_equal(confint(cyclopsFitD, c(1:2))[,2:3], confint(glmFit, c(1:2)), tolerance = tolerance)
    expect_equal(predict(cyclopsFitD), predict(glmFit, type = "response"), tolerance = tolerance)
})

test_that("Small Bernoulli dense regression with zero-weights", {
    binomial_bid <- c(1,5,10,20,30,40,50,75,100,150,200)
    binomial_n <- c(31,29,27,25,23,21,19,17,15,15,15)
    binomial_y <- c(0,3,6,7,9,13,17,12,11,14,13)

    log_bid <- log(c(rep(rep(binomial_bid, binomial_n - binomial_y)), rep(binomial_bid, binomial_y)))
    y <- c(rep(0, sum(binomial_n - binomial_y)), rep(1, sum(binomial_y)))

    weights <- as.numeric(gl(2,1,length(y))) - 1

    y_sub <- y[which(weights == 1)]
    log_bid_sub <- log_bid[which(weights == 1)]

    glmFit <- glm(y ~ log_bid, family = binomial(), weights = weights) # gold standard

    tolerance <- 1E-4

    data <- createCyclopsData(y ~ log_bid, modelType = "lr")
    data_sub <- createCyclopsData(y_sub ~ log_bid_sub, modelType = "lr")

    fit <- fitCyclopsModel(data, prior = createPrior("none"),
                           control = createControl(noiseLevel = "silent"),
                           weights = weights)

    fit_sub <- fitCyclopsModel(data_sub, prior = createPrior("none"),
                               control = createControl(noiseLevel = "silent"))

    expect_equal(coef(fit), coef(glmFit), tolerance = tolerance)
    expect_equal(fit$log_likelihood, logLik(glmFit)[[1]], tolerance = tolerance)
    expect_equal(predict(glmFit, type = "response"), predict(fit), tolerance = tolerance)

    expect_equal(coef(fit), coef(fit_sub), check.attributes = FALSE)
    expect_equal(fit$log_likelihood, fit_sub$log_likelihood)

    expect_equal(predict(fit)[which(weights == 1)], predict(fit_sub), check.attributes = FALSE)
})

test_that("Small Poisson dense regression with weighting", {
    dobson <- data.frame(
        counts = c(18,17,15,20,10,20,25,13,12),
        outcome = gl(3,1,9),
        treatment = gl(3,3)
    )
    tolerance <- 1E-4

    weights <- c(1,2,1,2,4,3,2,1,1)

    glmFit <- glm(counts ~ outcome + treatment, data = dobson, weights, family = poisson()) # gold standard

    dataPtrD <- createCyclopsData(counts ~ outcome + treatment, data = dobson,
                                  modelType = "pr")
    cyclopsFitD <- fitCyclopsModel(dataPtrD,
                                   prior = createPrior("none"),
                                   control = createControl(noiseLevel = "silent"), weights = weights)

    expect_equal(coef(cyclopsFitD), coef(glmFit), tolerance = tolerance)
    expect_equal(cyclopsFitD$log_likelihood, logLik(glmFit)[[1]], tolerance = tolerance)
    expect_equal(confint(cyclopsFitD, c(1:3))[,2:3], confint(glmFit, c(1:3)), tolerance = tolerance)
    expect_equal(predict(cyclopsFitD), predict(glmFit, type = "response"), tolerance = tolerance)
    expect_equal(confint(cyclopsFitD, c("(Intercept)","outcome3")), confint(cyclopsFitD, c(1,3)))

    fit <- fitCyclopsModel(dataPtrD,
                           prior = createPrior("laplace", useCrossValidation = TRUE),
                           weights = weights, warnings = FALSE,
                           control = createControl(minCVData = 1, noiseLevel = "quiet"))
})

test_that("Check conditional Poisson with cross-validation",{
    set.seed(123)
    sim <- simulateCyclopsData(nstrata=1000,
                               ncovars=10,
                               nrows=10000,
                               effectSizeSd=0.5,
                               eCovarsPerRow=2,
                               model="poisson")
    suppressWarnings(
        cyclopsData <- convertToCyclopsData(outcomes = sim$outcomes,
                                            covariates = sim$covariates,
                                            modelType = "cpr")
    )
    fit <- fitCyclopsModel(cyclopsData = cyclopsData,
                           prior = createPrior("laplace",
                                               useCrossValidation = TRUE,
                                               exclude = 1),
                           control = createControl(fold = 10,
                                                   cvRepetitions = 1,
                                                   startingVariance = 0.1,
                                                   noiseLevel = "quiet",
                                                   seed = 123))
})


test_that("Large Cox regression with weighting",{
    set.seed(123)
    sim <- simulateCyclopsData(nstrata=1000,
                               ncovars=10,
                               nrows=10000,
                               effectSizeSd=0.5,
                               eCovarsPerRow=2,
                               model="survival")
    sim$outcomes$weights <- 1/sim$outcomes$rr

    # Gold standard
    covariates <- sim$covariates
    ncovars <- max(covariates$covariateId)
    nrows <- nrow(sim$outcomes)
    m <- matrix(0,nrows,ncovars)
    for (i in 1:nrow(covariates)){
        m[covariates$rowId[i],covariates$covariateId[i]] <- 1
    }
    data <- as.data.frame(m)

    data$rowId <- 1:nrow(data)
    data <- merge(data,sim$outcomes)
    data <- data[order(data$stratumId,data$rowId),]
    formula <- as.formula(paste(c("Surv(time,y) ~ strata(stratumId)",paste("V",1:ncovars,sep="")),collapse=" + "))
    fitCoxph <- survival::coxph(formula, data = data, weights = data$weights, ties = "breslow")

    # Cyclops
    cyclopsData <- convertToCyclopsData(outcomes = sim$outcomes,
                                        covariates = sim$covariates,
                                        modelType = "cox")
    fitCyclops <- fitCyclopsModel(cyclopsData = cyclopsData)

    expect_equivalent(coef(fitCyclops), coef(fitCoxph), tolerance = tolerance)
})

# # currently broken
# test_that("Small conditional logistic regression with weighting", {
#
#     weights <- as.numeric(gl(5,1,length(infert$case)))
#
#     gold <- clogit(case ~ spontaneous + induced + strata(stratum), data=infert, weights = weights, method=c("approximate"))
#
#     dataPtr <- createCyclopsData(case ~ spontaneous + induced + strata(stratum),
#                                  data = infert,
#                                  modelType = "clr")
#
#     cyclopsFit <- fitCyclopsModel(dataPtr, prior = createPrior("none"), weights = weights)
#
#     tolerance <- 1E-4
#
#     expect_equal(coef(cyclopsFit), coef(gold), tolerance = tolerance)
#     expect_equal(cyclopsFit$log_likelihood, logLik(gold)[[1]], tolerance = tolerance)
#
#     # expect_equal(vcov(cyclopsFit), vcov(gold), tolerance = tolerance)
#     # expect_equal(aconfint(cyclopsFit), confint(gold), tolerance = tolerance)
#
#     expect_equal(confint(cyclopsFit, c(1:2), includePenalty = TRUE),
#                  confint(cyclopsFit, c(1:2), includePenalty = FALSE))
#
# })
#
# test_that("Check simple SCCS as conditional Poisson regression with weighting", {
#     #     source("helper-conditionalPoisson.R")
#     tolerance <- 1E-3
#     weights <- as.numeric(gl(5,1,38))
#     gold.cp <- gnm(event ~ exgr + agegr + offset(loginterval),
#                    family = poisson, eliminate = indiv, weights = weights,
#                    data = Cyclops::oxford)
#
#     dataPtr <- createCyclopsData(event ~ exgr + agegr + strata(indiv) + offset(loginterval),
#                                  data = Cyclops::oxford,
#                                  modelType = "cpr")
#     cyclopsFit <- fitCyclopsModel(dataPtr, weights = weights,
#                                   prior = createPrior("none"))
#
#     expect_equal(coef(cyclopsFit)[1:2], coef(gold.cp)[1:2], tolerance = tolerance)
#     # expect_equal(confint(cyclopsFit, c("exgr1","agegr2"))[,2:3],
#     # confint(gold.cp), tolerance = tolerance)
# })
#
# test_that("Check simple SCCS as SCCS with weighting", {
#     #     source("helper-conditionalPoisson.R")
#     tolerance <- 1E-6
#     weights <- as.numeric(gl(5,1,38))
#
#     gold.clogit <- clogit(event ~ exgr + agegr + strata(indiv) + offset(loginterval),weights = weights,
#                           data = Cyclops::oxford)
#
#     dataPtr <- createCyclopsData(event ~ exgr + agegr + strata(indiv), time = Cyclops::oxford$interval,
#                                  data = Cyclops::oxford,
#                                  modelType = "sccs")
#     cyclopsFit <- fitCyclopsModel(dataPtr,weights = weights,
#                                   prior = createPrior("none"))
#     expect_equivalent(logLik(cyclopsFit), logLik(gold.clogit))
#     expect_equal(coef(cyclopsFit), coef(gold.clogit), tolerance = tolerance)
# })

library("testthat")
library("survival")

context("test-believedBroken.R")

#
# These tests are believed to be broken; they need confirmation and fixes
#

# test_that("Returns correct counts" ,{
#   Number of rows
#   Number of unique denominators
#   Number of strata
# })

# test_that("Returns cross-validated variance estimate" ,{})

# test_that("Data validity when loaded with (y,dx,sx,ix)" ,{})

# test_that("Dimension checking on objects in createCyclopsModelData" ,{})

# test_that("Approximations for ties in CLR" ,{})

# test_that("getSEs() throws error when all covariates are not included", {})

# test_that("Extract Y from data object", {})

# test_that("Extract X from data object", {})

# test_that("Predict CLR model", {})

# test_that("Predict SCCS model", {})

# test_that("Predict Cox model", {})

# test_that("Throw error with more than one case in CLR" ,{})

# test_that("Check SCCS model via SQL", {})

# test_that("Check default regularization variance", {})

# test_that("Check starting regularization with cross validation", {})

# test_that("Standardize covariates", {})

# test_that("Check correct dimensions in matrices in createCyclopsData", {})

# test_that("Fail to convergence", {})

# test_that("Make intercept dense in SQL input", {})

# test_that("SCCS as conditional Poisson regression likelihoods" ,{
#     expect_equal(logLik(cyclopsFit), logLik(gold.cp)) # TODO Why are these different?
#})

# test_that("SCCS as SCCS likelihoods" ,{
#     expect_equal(logLik(cyclopsFit), MJS values) # TODO Why are these different?
#})

# test_that("Reuse data object", {
#
#     dataPtr <- createCyclopsData(case ~ spontaneous + induced + strata(stratum),
#                                       data = infert,
#                                       modelType = "clr")
#
#     cyclopsFit <- fitCyclopsModel(dataPtr, prior = createPrior("none"))
#
#     cyclopsFitR <- fitCyclopsModel(dataPtr,
#                                    prior = createPrior("laplace", 1, exclude = 1))
#
#     # Error: both cyclopsFit and cyclopsFitR share the same interface ptr
#     confint(cyclopsFit, c(1:2), includePenalty = TRUE) # Should not throw error
# })

test_that("Check asymptotic variance in Cox example with failure ties and strata", {
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
0, 4,  1,0,0
0, 3,  1,2,0
0, 3,  0,0,1
0, 2,  1,0,1
0, 2,  1,1,1
0, 1,  0,1,0
0, 1,  1,1,0
")

    # We get the correct answer when last entry is censored
    gold <- coxph(Surv(length, event) ~ x1 + strata(x2), test, ties = "breslow")

    dataPtr <- createCyclopsData(Surv(length, event) ~ x1 + strata(x2), data = test,
                                      modelType = "cox")

    cyclopsFit <- fitCyclopsModel(dataPtr)

    tolerance <- 1E-4
    #     expect_equal(vcov(cyclopsFit), vcov(gold), tolerance = tolerance)

})

library("testthat")

context("test-covariateRegularization.R")

test_that("Find covariate by name and number", {
    counts <- c(18,17,15,20,10,20,25,13,12)
    outcome <- gl(3,1,9)
    treatment <- gl(3,3)
    tolerance <- 1E-4

    dataPtr <- createCyclopsData(counts ~ outcome + treatment,
                                  modelType = "pr")

    cyclopsFit <- fitCyclopsModel(dataPtr,
                          prior = createPrior("laplace",
                                        exclude = c("(Intercept)", "outcome2", "outcome3")),
                          control = createControl(noiseLevel = "silent"))

    # Shrinkage on treatment-effects
    expect_equivalent(coef(cyclopsFit)[4], 0.0)
    expect_equivalent(coef(cyclopsFit)[5], 0.0)

    dataPtr2 <- createCyclopsData(counts ~ outcome + treatment,
                                   modelType = "pr")

    cyclopsFit2 <- fitCyclopsModel(dataPtr2,
                           prior = createPrior("laplace",
                                         exclude = c(1:3)),
                           control = createControl(noiseLevel = "silent"))
    # Check c(i:j) notation
    expect_equal(coef(cyclopsFit), coef(cyclopsFit2))
})

test_that("Error when covariate not found", {
    counts <- c(18,17,15,20,10,20,25,13,12)
    outcome <- gl(3,1,9)
    treatment <- gl(3,3)
    tolerance <- 1E-4

    dataPtr <- createCyclopsData(counts ~ outcome + treatment,
                                  modelType = "pr")

    expect_error(
        fitCyclopsModel(dataPtr,
                    prior = createPrior("laplace",
                                  exclude = c("BAD", "outcome2", "outcome3")),
                    control = createControl(noiseLevel = "silent")))

    dataPtr2 <- createCyclopsData(counts ~ outcome + treatment,
                                   modelType = "pr")

    expect_error(
        fitCyclopsModel(dataPtr2,
                    prior = createPrior("laplace",
                                  exclude = c(10,1:3)),
                    control = createControl(noiseLevel = "silent")))
})

test_that("Preclude profiling regularized coefficients", {
    counts <- c(18,17,15,20,10,20,25,13,12)
    outcome <- gl(3,1,9)
    treatment <- gl(3,3)
    tolerance <- 1E-4

    dataPtr <- createCyclopsData(counts ~ outcome + treatment,
                                  modelType = "pr")

    cyclopsFit <- fitCyclopsModel(dataPtr,
                          prior = createPrior("laplace", exclude = "(Intercept)"),
                          control = createControl(noiseLevel = "silent"))

    expect_true(
        !is.null(confint(cyclopsFit, "(Intercept)")) # not regularized
    )
    expect_error(
        confint(cyclopsFit, "outcome2") # regularized
    )
})

test_that("Preclude intercept regularization by default", {
    counts <- c(18,17,15,20,10,20,25,13,12)
    outcome <- gl(3,1,9)
    treatment <- gl(3,3)
    tolerance <- 1E-4

    dataPtr <- createCyclopsData(counts ~ outcome + treatment,
                                  modelType = "pr")

#     expect_error(fitCyclopsModel(dataPtr,
#                 prior = createPrior("laplace", 0.1)))

    c1 <- fitCyclopsModel(dataPtr,
                      forceNewObject = TRUE,
                      prior = createPrior("laplace", 0.1, forceIntercept = TRUE))

    c2 <- fitCyclopsModel(dataPtr,
                      forceNewObject = TRUE,
                      prior = createPrior("laplace", 0.1, exclude = "(Intercept)"))

    c3 <- fitCyclopsModel(dataPtr,
                      forceNewObject = TRUE,
                      prior = createPrior("laplace", 0.1, exclude = 1))

    expect_equal(coef(c2),
                 coef(c3))

    expect_lt(coef(c1)[1],  # Intercept is regularized
                     coef(c2)[1])
})

test_that("Specify each prior independently", {
    counts <- c(18,17,15,20,10,20,25,13,12)
    outcome <- gl(3,1,9)
    treatment <- gl(3,3)

    dataPtr <- createCyclopsData(counts ~ outcome + treatment,
                                 modelType = "pr")

    prior <- createPrior(priorType = c("none", "laplace", "none", "laplace", "none"),
                         variance = c(0, 1, 0, 1, 1))

    cyclopsFit <- fitCyclopsModel(dataPtr, prior = prior)
    expect_equal(length(strsplit(cyclopsFit$prior_info, ' ')[[1]]), 5) # 5 different covariates

    expect_true(coef(cyclopsFit)[4] == 0)
    expect_true(coef(cyclopsFit)[5] != 0)

    expect_equal(getHyperParameter(cyclopsFit), c(1,1))

    expect_false(Cyclops:::.cyclopsGetIsRegularized(cyclopsFit$interface, 0))
    expect_true(Cyclops:::.cyclopsGetIsRegularized(cyclopsFit$interface, 1))

    expect_equal(Cyclops:::.cyclopsGetLogLikelihood(cyclopsFit$interface),
                 cyclopsFit$log_likelihood)
})


test_that("Mixture report should show full details of components", {
    counts <- c(18,17,15,20,10,20,25,13,12)
    outcome <- gl(3,1,9)
    treatment <- gl(3,3)

    dataPtr <- createCyclopsData(counts ~ outcome + treatment,
                                  modelType = "pr")
    cyclopsFit <- fitCyclopsModel(dataPtr,
                          prior = createPrior("laplace",
                                        exclude = c("(Intercept)", "outcome2", "outcome3")))
    expect_equal(length(strsplit(cyclopsFit$prior_info, ' ')[[1]]),
                 4) # 4 different prior assignments
})

test_that("Random-walk fusion prior", {
    counts <- c(18,17,15,20,10,20,25,13,12)
    outcome <- gl(3,1,9)
    treatment <- gl(3,3)

    dataPtr <- createCyclopsData(counts ~ outcome + treatment,
                                 modelType = "pr")

    cyclopsFit0 <- fitCyclopsModel(dataPtr,
                                   prior = createPrior("laplace",
                                                       1.0,
                                                       exclude = c("(Intercept)")),
                                   startingCoefficients = c(1,2,3,4,5))

    cyclopsFit1 <- fitCyclopsModel(dataPtr,
                                   prior = createPrior(c("laplace", "laplace"),
                                                       c(1.0, 1E20), # Effectively no fused regularization
                                                       exclude = c("(Intercept)"),
                                                       neighborhood = list(list("outcome2", c("outcome3")),
                                                                           list("outcome3", c("outcome2")))
                                   ),
                                   startingCoefficients = c(1,2,3,4,5))

    expect_equal(coef(cyclopsFit0), coef(cyclopsFit1))

    cyclopsFit2 <- fitCyclopsModel(dataPtr,
                                   prior = createPrior(c("laplace", "laplace"),
                                                       c(1.0, 0.1), # Strong fused regularization
                                                       exclude = c("(Intercept)"),
                                                       neighborhood = list(list("outcome2", c("outcome3")),
                                                                           list("outcome3", c("outcome2")))
                                   ),
                                   startingCoefficients = c(1,2,3,4,5))

    expect_equivalent(coef(cyclopsFit2)[2], coef(cyclopsFit2)[3]) # Have different names
})

library("testthat")

context("test-fastBarUpdate.R")

test_that("Small Bernoulli dense regression", {
	binomial_bid <- c(1,5,10,20,30,40,50,75,100,150,200)
	binomial_n <- c(31,29,27,25,23,21,19,17,15,15,15)
	binomial_y <- c(0,3,6,7,9,13,17,12,11,14,13)

	log_bid <- log(c(rep(rep(binomial_bid, binomial_n - binomial_y)), rep(binomial_bid, binomial_y)))
	y <- c(rep(0, sum(binomial_n - binomial_y)), rep(1, sum(binomial_y)))

	tolerance <- 1E-4

	data <- createCyclopsData(y ~ log_bid, modelType = "lr")

	expect_error(prior <- createPrior(priorType = "barupdate",
	                                  useCrossValidation = TRUE),
	             "Cannot perform")

	prior <- createPrior(priorType = "barupdate", variance = 1)
	control <- createControl(convergenceType = "onestep")

	expect_warning(fit <- fitCyclopsModel(data, prior, control),
	               "Excluding intercept")

	expect_equal(fit$iterations, 1)
	expect_equal(fit$return_flag, "SUCCESS")
})

test_that("Small Poisson dense regression; unpenalized regression", {
    y <- c(1, 6, 16, 23, 27, 39, 31, 30, 43, 51, 63, 70, 88, 97, 91, 104, 110, 113, 149, 159)
    x <- 1:20

    data <- createCyclopsData(y ~ x, modelType = "pr")
    #Run "unpenalized regression"
    tolerance <- 1E-8
    delta <- 1
    b.old <- c(0, 0)
    while(delta > tolerance) {
        prior <- createPrior(priorType = "barupdate", variance = Inf)
        control <- createControl(convergenceType = "onestep")
        fit <- fitCyclopsModel(data, prior = prior, control = control,
                               startingCoefficients = b.old, warnings = FALSE)
        delta <- max(abs((coef(fit) - b.old)))
        b.old <- coef(fit)
    }
    glm.fit <- glm(y ~ x, family = "poisson")
    expect_equal(b.old, glm.fit$coefficients, tolerance = 1E-6)
})

library(Cyclops)
library("testthat")
library("survival")


GpuDevice <- listGPUDevices()[1]
tolerance <- 1E-4


# small cox
test_that("Check small Cox on GPU", {
    skip_if(length(listGPUDevices()) == 0, "GPU not available")
    test <- read.table(header=T, sep = ",", text = "
                   start, length, event, x1, x2
                       0, 4,  1,0,0
                       0, 3.5,1,2,0
                       0, 3,  0,0,1
                       0, 2.5,1,0,1
                       0, 2,  1,1,1
                       0, 1.5,0,1,0
                       0, 1,  1,1,0")

    goldRight <- coxph(Surv(length, event) ~ x1 + x2, test)
    dataPtrRight_CPU <- createCyclopsData(Surv(length, event) ~ x1 + x2, data = test,
                                          modelType = "cox", floatingPoint = 32)
    cyclopsFitRight_CPU <- fitCyclopsModel(dataPtrRight_CPU)

    dataPtrRight_GPU <- createCyclopsData(Surv(length, event) ~ x1 + x2, data = test,
                                          modelType = "cox", floatingPoint = 32)
    cyclopsFitRight_GPU <- fitCyclopsModel(dataPtrRight_GPU, computeDevice = GpuDevice)

    expect_equal(coef(cyclopsFitRight_CPU), coef(goldRight), tolerance = tolerance)
    expect_equal(coef(cyclopsFitRight_GPU), coef(cyclopsFitRight_CPU), tolerance = tolerance)
})

test_that("Check very small Cox example with time-ties", {
    skip_if(length(listGPUDevices()) == 0, "GPU not available")
    test <- read.table(header=T, sep = ",", text = "
start, length, event, x1, x2
                       0, 4,  1,0,0
                       0, 3,  1,2,0
                       0, 3,  0,0,1
                       0, 2,  1,0,1
                       0, 2,  0,1,1
                       0, 1,  0,1,0
                       0, 1,  1,1,0")
    # cpu
    dataPtrRight_CPU <- createCyclopsData(Surv(length, event) ~ x1 + x2, data = test,
                                          modelType = "cox", floatingPoint = 32)
    cyclopsFitRight_CPU <- fitCyclopsModel(dataPtrRight_CPU)

    # gpu
    dataPtrRight_GPU <- createCyclopsData(Surv(length, event) ~ x1 + x2, data = test,
                                          modelType = "cox", floatingPoint = 32)
    cyclopsFitRight_GPU <- fitCyclopsModel(dataPtrRight_GPU, computeDevice = GpuDevice)

    expect_equal(coef(cyclopsFitRight_GPU), coef(cyclopsFitRight_CPU), tolerance = tolerance)
})


# large cox
test_that("Check Cox on GPU", {
    skip_if(length(listGPUDevices()) == 0, "GPU not available")
    set.seed(123)
    sim <- simulateCyclopsData(nstrata = 1,
                               nrows = 100000,
                               ncovars = 100,
                               effectSizeSd = 1,
                               zeroEffectSizeProp = 0.8,
                               eCovarsPerRow = 0.2,
                               model = "survival")
    set.seed(12)
    sim$outcomes$time <- sim$outcomes$time + rnorm(100000, mean = 0, sd = 0.00001)

    cyclopsData_CPU <- convertToCyclopsData(sim$outcomes, sim$covariates,
                                            modelType = "cox",floatingPoint = 64,
                                            addIntercept = TRUE)
    fit_CPU <- fitCyclopsModel(cyclopsData_CPU) # iterations: 6 lastObjFunc: 733.17
    cyclopsData_GPU <- convertToCyclopsData(sim$outcomes, sim$covariates,
                                            modelType = "cox",floatingPoint = 64,
                                            addIntercept = TRUE)
    fit_GPU <- fitCyclopsModel(cyclopsData_GPU, computeDevice = GpuDevice) # iterations: 6 lastObjFunc: 733.17
    expect_equal(coef(fit_GPU), coef(fit_CPU), tolerance = tolerance)
})

# lasso cv
test_that("Check cross-validation for lasso Cox on GPU", {
    skip_if(length(listGPUDevices()) == 0, "GPU not available")
    set.seed(123)
    sim <- simulateCyclopsData(nstrata = 1,
                               nrows = 900,
                               ncovars = 35,
                               effectSizeSd = 1,
                               zeroEffectSizeProp = 0.8,
                               eCovarsPerRow = 1,
                               model = "survival")
    set.seed(123)
    sim$outcomes$time <- sim$outcomes$time + rnorm(900, mean = 0, sd = 0.00001)

    prior <- createPrior("laplace", useCrossValidation = TRUE)
    control <- createControl(noiseLevel = "quiet", lowerLimit = 0.000001, upperLimit = 100,
                                cvType = "auto", fold = 10, cvRepetitions = 1, startingVariance = 0.01, threads = 1,
                                seed = 123)

    cyclopsData_CPU <- convertToCyclopsData(sim$outcomes, sim$covariates,modelType = "cox",floatingPoint = 64,addIntercept = TRUE)
    fit_CPU <- fitCyclopsModel(cyclopsData_CPU, prior = prior, control = control)

    cyclopsData_GPU <- convertToCyclopsData(sim$outcomes, sim$covariates,modelType = "cox",floatingPoint = 64,addIntercept = TRUE)
    fit_GPU <- fitCyclopsModel(cyclopsData_GPU, prior = prior, control = control, computeDevice = GpuDevice)

    expect_equal(getHyperParameter(fit_GPU), getHyperParameter(fit_CPU), tolerance = tolerance)
    expect_equal(coef(fit_GPU), coef(fit_CPU), tolerance = tolerance)

})

# multi-core
test_that("Check multi-core cross-validation for lasso Cox on GPU", {
    skip_if(length(listGPUDevices()) == 0, "GPU not available")
    set.seed(123)
    sim <- simulateCyclopsData(nstrata = 1,
                               nrows = 900,
                               ncovars = 35,
                               effectSizeSd = 1,
                               zeroEffectSizeProp = 0.8,
                               eCovarsPerRow = 1,
                               model = "survival")
    set.seed(123)
    sim$outcomes$time <- sim$outcomes$time + rnorm(900, mean = 0, sd = 0.00001)

    prior <- createPrior("laplace", useCrossValidation = TRUE)
    control <- createControl(noiseLevel = "quiet", lowerLimit = 0.000001, upperLimit = 100,
                                cvType = "auto", fold = 10, cvRepetitions = 1, startingVariance = 0.01, threads = 2,
                                seed = 123)

    cyclopsData_CPU <- convertToCyclopsData(sim$outcomes, sim$covariates,modelType = "cox",floatingPoint = 64,addIntercept = TRUE)
    fit_CPU <- fitCyclopsModel(cyclopsData_CPU, prior = prior, control = control)

    cyclopsData_GPU <- convertToCyclopsData(sim$outcomes, sim$covariates,modelType = "cox",floatingPoint = 64,addIntercept = TRUE)
    fit_GPU <- fitCyclopsModel(cyclopsData_GPU, prior = prior, control = control, computeDevice = GpuDevice)

    expect_equal(getHyperParameter(fit_GPU), getHyperParameter(fit_CPU), tolerance = tolerance)
    expect_equal(coef(fit_GPU), coef(fit_CPU), tolerance = tolerance)

})



# test_that("Check small Cox example with failure ties and strata on GPU", {
#     test <- read.table(header=T, sep = ",", text = "
#                        start, length, event, x1, x2
#                        0, 4,  1,0,0
#                        0, 3,  1,2,0
#                        0, 3,  0,0,1
#                        0, 2,  1,0,1
#                        0, 2,  1,1,1
#                        0, 1,  0,1,0
#                        0, 1,  1,1,0")
#
#     # We get the correct answer when last entry is censored
#     goldRight <- coxph(Surv(length, event) ~ x1 + strata(x2), test, ties = "breslow")
#
#     dataPtrRight_CPU <- createCyclopsData(Surv(length, event) ~ x1 + strata(x2), data = test,
#                                            modelType = "cox", floatingPoint = 32)
#     cyclopsFitRight_CPU <- fitCyclopsModel(dataPtrRight_CPU)
#
#     dataPtrRight_GPU <- createCyclopsData(Surv(length, event) ~ x1 + strata(x2), data = test,
#                                           modelType = "cox", floatingPoint = 32)
#     cyclopsFitRight_GPU <- fitCyclopsModel(dataPtrRight_GPU, computeDevice = GpuDevice)
#
#     tolerance <- 1E-4
#     expect_equal(coef(cyclopsFitRight_CPU), coef(goldRight), tolerance = tolerance)
#     expect_equal(coef(cyclopsFitRight_CPU), coef(cyclopsFitRight_GPU), tolerance = tolerance)
# })

# make sure logistic regression still works
# test_that("Small Bernoulli dense regression using GPU", {
#     binomial_bid <- c(1,5,10,20,30,40,50,75,100,150,200)
#     binomial_n <- c(31,29,27,25,23,21,19,17,15,15,15)
#     binomial_y <- c(0,3,6,7,9,13,17,12,11,14,13)
# 
#     log_bid <- log(c(rep(rep(binomial_bid, binomial_n - binomial_y)), rep(binomial_bid, binomial_y)))
#     y <- c(rep(0, sum(binomial_n - binomial_y)), rep(1, sum(binomial_y)))
#     tolerance <- 1E-4

#     # gold standard
#     glmFit <- glm(y ~ log_bid, family = binomial())
# 
#     # # cpu
#     # dataPtrD_c <- createCyclopsData(y ~ log_bid, modelType = "lr", floatingPoint = 32)
#     # cyclopsFitD_c <- fitCyclopsModel(dataPtrD_c, prior = createPrior("none"),
#     #                                  control = createControl(noiseLevel = "silent"))
# 
#     # gpu
#     dataPtrD <- createCyclopsData(y ~ log_bid, modelType = "lr", floatingPoint = 32)
#     cyclopsFitD <- fitCyclopsModel(dataPtrD, prior = createPrior("none"),
#                                    control = createControl(noiseLevel = "silent"),
#                                    computeDevice = GpuDevice)
# 
#     expect_equal(coef(cyclopsFitD), coef(glmFit), tolerance = tolerance)
# })

library("testthat")
library(MASS)

context("test-offsetPoisson.R")

test_that("Check offset in model formula", {

    tolerance <- 1E-4
    Insurance$logHolders <- log(Insurance$Holders)

    glmFit <- glm(Claims ~ District + Group + Age + offset(logHolders),
                  data = Insurance,
                  family = poisson()) # gold standard

    dataPtr <- createCyclopsData(Claims ~ District + Group + Age + logHolders,
                                  data = Insurance,
                                  modelType = "pr")
    finalizeSqlCyclopsData(dataPtr, useOffsetCovariate = "logHolders", offsetAlreadyOnLogScale = TRUE)

    ## Test new number of covariates

    cyclopsFit <- fitCyclopsModel(dataPtr,
                          prior = createPrior("none"),
                          control = createControl(noiseLevel = "silent"))
    expect_equal(coef(cyclopsFit), coef(glmFit), tolerance = tolerance)

    dataPtr2 <- createCyclopsData(Claims ~ District + Group + Age + offset(logHolders),
                                   data = Insurance,
                                   modelType = "pr")

    cyclopsFit2 <- fitCyclopsModel(dataPtr2,
                                   startingCoefficients = rep(0.5,10),
                                   prior = createPrior("none"))

    expect_equal(coef(cyclopsFit2), coef(glmFit), tolerance = tolerance)

    # Need to test now using finalize to (1) add intercept and (2) log-transform
})

test_that("Check active set", {
    skip("Current not working")
    tolerance <- 1E-4
    Insurance$logHolders <- log(Insurance$Holders)

    dataPtr2 <- createCyclopsData(Claims ~ District + Group + Age + offset(logHolders),
                                  data = Insurance,
                                  modelType = "pr")

    out <- capture.output(cyclopsFit2 <- fitCyclopsModel(dataPtr2,
                                   control = createControl(useKKTSwindle = TRUE,
                                                           noiseLevel = "quiet",
                                                            tuneSwindle = 4),
                                   prior = createPrior("laplace", exclude=c(1,3))))

    expect_equal(length(out), 8) # Should have 3 (+ prior line) swindle sets
})

library("testthat")

#
# Small Poisson MLE regression
#

test_that("JNI with small Poisson regression", {
    dobson <- data.frame(
        counts = c(18,17,15,20,10,20,25,13,12),
        outcome = gl(3,1,9),
        treatment = gl(3,3)
    )
    tolerance <- 1E-4

    dataPtrD <- createCyclopsData(counts ~ outcome + treatment, data = dobson,
                                   modelType = "pr")
    cyclopsFitD <- fitCyclopsModel(dataPtrD,
                           prior = createPrior("none"),
                           control = createControl(noiseLevel = "silent"))

    instance <- cacheCyclopsModelForJava(cyclopsFitD)

    instance2 <- cacheCyclopsModelForJava(cyclopsFitD)
    expect_equal(instance, instance2)

   })

# Gets run when compiling

makevars_in <- file.path("src", "Makevars.in")
makevars_win_in <- file.path("src", "Makevars.win.in")

makevars_out <- file.path("src", "Makevars")
makevars_win_out <- file.path("src", "Makevars.win")

txt <- readLines(makevars_in)
txt_win <- readLines(makevars_win_in)
if (getRversion() < "4.3") { # macOS / linux
    if (!any(grepl("^CXX_STD", txt))) {
        txt <- c("CXX_STD = CXX11", txt)
    }
}

if (getRversion() < "4.2") { # Windoz
    if (!any(grepl("^CXX_STD", txt_win))) {
        txt_win <- c("CXX_STD = CXX11", txt_win)
    }
}

##### can find JNI #####

doJni <- TRUE
if (doJni) {

if (.Platform$OS.type == "unix") {
    java_home <- Sys.getenv("JAVA_HOME")
} else {
    system("make -C tools/jvm-w32")
    java_home <- system("tools/jvm-w32/findjava -s -f", intern = TRUE)
    system("make -C tools/jvm-w32 clean")
}

if (is.null(java_home) || nchar(java_home) == 0) {
    message("No JAVA_HOME defined; ignoring JNI compilation")
} else {
    message("Using JAVA_HOME=", java_home)
    jni_path = file.path(java_home, "include")

    # modify Makevars.in if Java is available
    txt[grep("^PKG_CPPFLAGS", txt)] <- paste(txt[grep("^PKG_CPPFLAGS", txt)],
                                             paste0("-I", jni_path),
                                             paste0("-I", jni_path, "/darwin"), # TODO Make OS-dependent
                                             paste0("-I", jni_path, "/linux"))
    engine_idx <- grep("^OBJECTS.engine", txt)
    txt[engine_idx+1] <- paste(txt[engine_idx+1],
                               "cyclops/jni/dr_inference_regression_RegressionJNIWrapper.o",
                               "cyclops/jni/dr_inference_regression_NewRegressionJNIWrapper.o")
    # Handle Windows
    txt_win[grep("^PKG_CPPFLAGS", txt_win)] <- paste(txt_win[grep("^PKG_CPPFLAGS", txt_win)],
                                             paste0("-I", jni_path),
                                             paste0("-I", jni_path, "/win32"))

    sources_idx <- grep("^SOURCES", txt_win)
    txt_win[sources_idx + 1] <- paste0(txt_win[sources_idx + 1],
                                      "\n\t\t\t\t\tcyclops/jni/*.cpp \\")
}

} # doJNI

#################### CUDA Toolkit ####################

cuda_home <- system2(command = "find", args = c("/usr/local/", "-maxdepth", "1" ,"-name", "cuda"), stdout  = TRUE)
if (TRUE || nchar(cuda_home)==0) { # By default, no CUDA build
    message("no CUDA installation found; only compile host code")
} else {
    message(paste0("using CUDA_HOME=", cuda_home))
    nvcc <- c(paste0(cuda_home, "/bin/nvcc -arch=sm_70")) # TODO remove hardcoding of ARCH
    cub_path <- paste0(cuda_home ,"/include") # CUB is included since CUDA Toolkit 11.0

    # whether this is the 64 bit linux version of CUDA
    cu_libdir <- system2(command = "find", args = c(paste0(cuda_home ,"/lib64")), stdout  = TRUE)
    if (nchar(cu_libdir) == 0) {
        cu_libdir <- paste0(cuda_home ,"/lib")
    }

    cuda_libs <- paste0("-L", cu_libdir, " -lcudart")
    cuda_cppflags <-paste0("-DHAVE_CUDA -I", cuda_home, "/include -I", cu_libdir, " -pthread -rdynamic")

    # modify Makevars.in if CUDA is available
    txt[grep("^PKG_LIBS", txt)] <- paste(txt[grep("^PKG_LIBS", txt)], cuda_libs)
    txt[grep("^PKG_CPPFLAGS", txt)] <- paste(txt[grep("^PKG_CPPFLAGS", txt)], cuda_cppflags)
    engine_idx <- grep("^OBJECTS.engine", txt)
    txt[engine_idx+1] <- paste(txt[engine_idx+1],
                               "cyclops/engine/CudaKernel.o",
                               "cyclops/engine/CudaDetail.o")
    txt <- c(txt,
             'all: $(OBJECTS)',
             '%.o: %.cu',
             paste0('\t', nvcc,' --default-stream per-thread -c -Xcompiler "-fPIC $(CPPFLAGS) -c" -I', cub_path,' $(R_PATH_LINKER) $^ -o $@'),
             'clean:',
             '\trm -rf *o',
             '.PHONY: all clean')
}

######################################################

if (.Platform$OS.type == "unix") {
	cat(txt, file = makevars_out, sep = "\n")
} else {
	cat(txt_win, file = makevars_win_out, sep = "\n")
}

# @file getFineGrayWeights.R
#
# Copyright 2020 Observational Health Data Sciences and Informatics
#
# This file is part of cyclops
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#' @title Creates a \code{Surv} object that forces in competing risks and the IPCW needed for Fine-Gray estimation.
#'
#' @description
#' \code{getFineGrayWeights} creates a list \code{Surv} object and vector of weights required for estimation.
#'
#' @param ftime    Numeric: Observed event (failure) times
#' @param fstatus  Numeric: Observed event (failure) types
#' @param cvweights Numeric: Vector of 0/1 (cross-validation) weights for each data row
#' @param cencode Numeric: Code to denote censored observations (Default is 0)
#' @param failcode Numeric: Code to denote event of interest (Default is 1)
#'
#' @return A list that returns both an object of class \code{Surv} that forces in the competing risks indicators and a vector of weights needed for parameter estimation.
#' @examples
#' ftime <- c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
#' fstatus <- c(1, 2, 0, 1, 2, 0, 1, 2, 0, 1)
#' getFineGrayWeights(ftime, fstatus, cencode = 0, failcode = 1)
#' @importFrom survival survfit Surv
#' @importFrom stats approx
#' @export
getFineGrayWeights <- function(ftime, fstatus, cvweights = NULL,
                               cencode = 0, failcode = 1) {

    # Check for cross-validation weights
    if(is.null(cvweights)) cvweights <- rep(1, length(ftime))
    if (!all(cvweights %in% c(0,1))) {
        stop("Only 0/1 weights are currently supported")
    }

    # Check for errors
    if(!cencode %in% unique(fstatus)) stop("cencode must be a valid value from fstatus")
    if(!failcode %in% unique(fstatus)) stop("failcode must be a valid value from fstatus")
    if(any(ftime < 0)) stop("all values of ftime must be positive valued")

    obj <- suppressWarnings(survival::Surv(ftime, fstatus)) # Suppress warning given by Surv function
    obj[, 2] <- fstatus # Changes NA's to competing risks indicators
    cenind   <- ifelse(fstatus == cencode, 1, 0)
    obj[, 2] <- ifelse(obj[, 2] == failcode, 1, 2 * (1 - cenind)) # Changes competing risks to 2

    # Create IPCW here (see original F&G code)
    u <- do.call('survfit', list(formula = Surv(ftime, cenind) ~ 1, weights = cvweights,
                                 data = data.frame(ftime, cenind)))
    u <- approx(c(0, u$time, max(u$time) * (1 + 10 * .Machine$double.eps)), c(1, u$surv, 0),
                xout = ftime * (1 - 100 * .Machine$double.eps), method = 'constant',
                f = 0, rule = 2)
    uuu <- cvweights * u$y
    return(list(surv = obj,
                weights = uuu))
}

# @file SpecialPriors.R
#
# Copyright 2018 Observational Health Data Sciences and Informatics
#
# This file is part of Cyclops
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @author Marc A. Suchard

#' @title Create a Cyclops prior object that returns the MLE of non-separable coefficients
#'
#' @description
#' \code{createNonSeparablePrior} creates a Cyclops prior object for use with \code{\link{fitCyclopsModel}}.
#'
#' @param maxIterations Numeric: maximum iterations to achieve convergence
#' @param ... Additional argument(s) for \code{\link{fitCyclopsModel}}
#'
#' @examples
#' prior <- createNonSeparablePrior()
#'
#' @return
#' A Cyclops prior object of class inheriting from
#' \code{"cyclopsPrior"} for use with \code{fitCyclopsModel}.
#'
#' @export
createNonSeparablePrior <- function(maxIterations = 10,
                                    ...) {

    fitHook <- function(...) {
      # closure to capture fitCyclopsModel control
      nonSeparableHook(maxIterations, ...)
    }

    prior <- createPrior(priorType = "none")
    prior$fitHook <- fitHook

    return(prior)
}

# Below are package-private functions

nonSeparableHook <- function(
                    maxIterations,
                    cyclopsData,
                    prior,
                    control,
                    weights,
                    forceNewObject,
                    returnEstimates,
                    startingCoefficients,
                    fixedCoefficients) {

    prior <- createPrior(priorType = "none")

    dim <- length(getCovariateIds(cyclopsData))
    separable <- rep(FALSE, dim)

    continue <- TRUE
    count <- 0

    while (continue) {
        count <- count + 1

        fit <- fitCyclopsModel(cyclopsData,
                                        prior = prior,
                                        control, weights, forceNewObject, returnEstimates,
                                        startingCoefficients,
                                        fixedCoefficients = separable)

        if (fit$return_flag == "ILLCONDITIONED") {
            new_separable <- is.nan(coef(fit, ignoreConvergence = TRUE)) | separable
        } else {
            new_separable <- separable
        }

        if (all(new_separable == separable) || count >= maxIterations) {
            continue <- FALSE
        } else {
            separable <- new_separable
        }
    }

    fit$estimation[separable, "estimate"] <- NaN
    fit$non_separable_iterations <- count

    return(fit)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

.cyclopsGetModelTypeNames <- function() {
    .Call(`_Cyclops_cyclopsGetModelTypeNames`)
}

.cyclopsGetRemoveInterceptNames <- function() {
    .Call(`_Cyclops_cyclopsGetRemoveInterceptNames`)
}

.cyclopsGetIsSurvivalNames <- function() {
    .Call(`_Cyclops_cyclopsGetIsSurvivalNames`)
}

.cyclopsGetUseOffsetNames <- function() {
    .Call(`_Cyclops_cyclopsGetUseOffsetNames`)
}

.cyclopsGetComputeDevice <- function(inRcppCcdInterface) {
    .Call(`_Cyclops_cyclopsGetComputeDevice`, inRcppCcdInterface)
}

.cyclopsSetBeta <- function(inRcppCcdInterface, beta) {
    invisible(.Call(`_Cyclops_cyclopsSetBeta`, inRcppCcdInterface, beta))
}

.cyclopsGetBeta <- function(inRcppCcdInterface, index) {
    .Call(`_Cyclops_cyclopsGetBeta`, inRcppCcdInterface, index)
}

.cyclopsSetStartingBeta <- function(inRcppCcdInterface, inStartingBeta) {
    invisible(.Call(`_Cyclops_cyclopsSetStartingBeta`, inRcppCcdInterface, inStartingBeta))
}

.cyclopsSetFixedBeta <- function(inRcppCcdInterface, beta, fixed) {
    invisible(.Call(`_Cyclops_cyclopsSetFixedBeta`, inRcppCcdInterface, beta, fixed))
}

.cyclopsGetFixedBeta <- function(inRcppCcdInterface, index) {
    .Call(`_Cyclops_cyclopsGetFixedBeta`, inRcppCcdInterface, index)
}

.cyclopsGetIsRegularized <- function(inRcppCcdInterface, index) {
    .Call(`_Cyclops_cyclopsGetIsRegularized`, inRcppCcdInterface, index)
}

.cyclopsSetWeights <- function(inRcppCcdInterface, weights) {
    invisible(.Call(`_Cyclops_cyclopsSetWeights`, inRcppCcdInterface, weights))
}

.cyclopsSetCensorWeights <- function(inRcppCcdInterface, weights) {
    invisible(.Call(`_Cyclops_cyclopsSetCensorWeights`, inRcppCcdInterface, weights))
}

.cyclopsGetNewPredictiveLogLikelihood <- function(inRcppCcdInterface, weights) {
    .Call(`_Cyclops_cyclopsGetNewPredictiveLogLikelihood`, inRcppCcdInterface, weights)
}

.cyclopsGetLogLikelihood <- function(inRcppCcdInterface) {
    .Call(`_Cyclops_cyclopsGetLogLikelihood`, inRcppCcdInterface)
}

.cyclopsMakeDirty <- function(inRcppCcdInterface) {
    invisible(.Call(`_Cyclops_cyclopsMakeDirty`, inRcppCcdInterface))
}

.cyclopsLogResults <- function(inRcppCcdInterface, fileName, withASE) {
    invisible(.Call(`_Cyclops_cyclopsLogResult`, inRcppCcdInterface, fileName, withASE))
}

.cyclopsTestProportionality <- function(inRcppCcdInterface, sexpBitCovariates, covariate) {
    .Call(`_Cyclops_cyclopsTestProportionality`, inRcppCcdInterface, sexpBitCovariates, covariate)
}

.cyclopsGetSchoenfeldResiduals <- function(inRcppCcdInterface, sexpBitCovariates) {
    .Call(`_Cyclops_cyclopsGetSchoenfeldResiduals`, inRcppCcdInterface, sexpBitCovariates)
}

.cyclopsGetFisherInformation <- function(inRcppCcdInterface, sexpBitCovariates) {
    .Call(`_Cyclops_cyclopsGetFisherInformation`, inRcppCcdInterface, sexpBitCovariates)
}

.cyclopsSetPrior <- function(inRcppCcdInterface, priorTypeName, variance, excludeNumeric, sexpGraph, sexpNeighborhood) {
    invisible(.Call(`_Cyclops_cyclopsSetPrior`, inRcppCcdInterface, priorTypeName, variance, excludeNumeric, sexpGraph, sexpNeighborhood))
}

.cyclopsTestParameterizedPrior <- function(priorFunction, startingParameters, indices, values) {
    .Call(`_Cyclops_cyclopsTestParameterizedPrior`, priorFunction, startingParameters, indices, values)
}

.cyclopsSetParameterizedPrior <- function(inRcppCcdInterface, priorTypeName, priorFunction, startingParameters, excludeNumeric) {
    invisible(.Call(`_Cyclops_cyclopsSetParameterizedPrior`, inRcppCcdInterface, priorTypeName, priorFunction, startingParameters, excludeNumeric))
}

.cyclopsGetProfileLikelihood <- function(inRcppCcdInterface, inCovariate, points, threads, includePenalty, returnDerivatives) {
    .Call(`_Cyclops_cyclopsGetProfileLikelihood`, inRcppCcdInterface, inCovariate, points, threads, includePenalty, returnDerivatives)
}

.cyclopsProfileModel <- function(inRcppCcdInterface, sexpCovariates, threads, threshold, override, includePenalty) {
    .Call(`_Cyclops_cyclopsProfileModel`, inRcppCcdInterface, sexpCovariates, threads, threshold, override, includePenalty)
}

.cyclopsPredictModel <- function(inRcppCcdInterface) {
    .Call(`_Cyclops_cyclopsPredictModel`, inRcppCcdInterface)
}

.cyclopsSetConvergenceType <- function(inRcppCcdInterface, convergenceType) {
    invisible(.Call(`_Cyclops_cyclopsSetConvergenceType`, inRcppCcdInterface, convergenceType))
}

.cyclopsGetConvergenceType <- function(inRcppCcdInterface) {
    .Call(`_Cyclops_cyclopsGetConvergenceType`, inRcppCcdInterface)
}

.cyclopsSetControl <- function(inRcppCcdInterface, maxIterations, tolerance, convergenceType, useAutoSearch, fold, foldToCompute, lowerLimit, upperLimit, gridSteps, noiseLevel, threads, seed, resetCoefficients, startingVariance, useKKTSwindle, swindleMultipler, selectorType, initialBound, maxBoundCount, algorithm, doItAll, syncCV) {
    invisible(.Call(`_Cyclops_cyclopsSetControl`, inRcppCcdInterface, maxIterations, tolerance, convergenceType, useAutoSearch, fold, foldToCompute, lowerLimit, upperLimit, gridSteps, noiseLevel, threads, seed, resetCoefficients, startingVariance, useKKTSwindle, swindleMultipler, selectorType, initialBound, maxBoundCount, algorithm, doItAll, syncCV))
}

.cyclopsRunCrossValidation <- function(inRcppCcdInterface) {
    .Call(`_Cyclops_cyclopsRunCrossValidationl`, inRcppCcdInterface)
}

.cyclopsCacheForJava <- function(inRcppCcdInterface) {
    .Call(`_Cyclops_cyclopsCacheForJava`, inRcppCcdInterface)
}

.cyclopsClearCacheForJava <- function() {
    invisible(.Call(`_Cyclops_cyclopsClearCacheForJava`))
}

.cyclopsFitModel <- function(inRcppCcdInterface) {
    .Call(`_Cyclops_cyclopsFitModel`, inRcppCcdInterface)
}

.cyclopsRunBootstrap <- function(inRcppCcdInterface, outFileName, treatmentId, replicates) {
    .Call(`_Cyclops_cyclopsRunBootstrap`, inRcppCcdInterface, outFileName, treatmentId, replicates)
}

.cyclopsGetLogLikelihoodGradient <- function(inRcppCcdInterface) {
    .Call(`_Cyclops_cyclopsGetLogLikelihoodGradient`, inRcppCcdInterface)
}

.cyclopsLogModel <- function(inRcppCcdInterface) {
    .Call(`_Cyclops_cyclopsLogModel`, inRcppCcdInterface)
}

.cyclopsInitializeModel <- function(inModelData, modelType, computeDevice, computeMLE = FALSE) {
    .Call(`_Cyclops_cyclopsInitializeModel`, inModelData, modelType, computeDevice, computeMLE)
}

.cyclopsGetLogLikelihoodHessianDiagonal <- function(inRcppCcdInterface, sexpCovariates) {
    .Call(`_Cyclops_cyclopsGetLogLikelihoodHessianDiagonal`, inRcppCcdInterface, sexpCovariates)
}

#' @title List available GPU devices
#'
#' @description
#' \code{listGPUDevices} list available GPU devices
#'
#' @export
listGPUDevices <- function() {
    .Call(`_Cyclops_listGPUDevices`)
}

.getDefaultGPUDevice <- function() {
    .Call(`_Cyclops_getDefaultGPUDevice`)
}

.isSorted <- function(dataFrame, indexes, ascending) {
    .Call(`_Cyclops_isSorted`, dataFrame, indexes, ascending)
}

.isSortedVectorList <- function(vectorList, ascending) {
    .Call(`_Cyclops_isSortedVectorList`, vectorList, ascending)
}

.isRcppPtrNull <- function(x) {
    .Call(`_Cyclops_isRcppPtrNull`, x)
}

#' @title Get number of strata
#'
#' @description
#' \code{getNumberOfStrata} return the number of unique strata in a Cyclops data object
#'
#' @param object    A Cyclops data object
#'
#' @export
getNumberOfStrata <- function(object) {
    .Call(`_Cyclops_cyclopsGetNumberOfStrata`, object)
}

#' @title Get covariate identifiers
#'
#' @description
#' \code{getCovariateIds} returns a vector of integer64 covariate identifiers in a Cyclops data object
#'
#' @param object    A Cyclops data object
#'
#' @export
getCovariateIds <- function(object) {
    .Call(`_Cyclops_cyclopsGetCovariateIds`, object)
}

.getCovariateTypes <- function(object, bitCovariateLabel) {
    .Call(`_Cyclops_cyclopsGetCovariateType`, object, bitCovariateLabel)
}

#' @title Get floating point size
#'
#' @description
#' \code{getFloatingPointSize} returns the floating-point representation size in a Cyclops data object
#'
#' @param object   A Cyclops data object
#'
#' @export
getFloatingPointSize <- function(object) {
    .Call(`_Cyclops_cyclopsGetFloatingPointSize`, object)
}

#' @title Get total number of covariates
#'
#' @description
#' \code{getNumberOfCovariates} returns the total number of covariates in a Cyclops data object
#'
#' @param object    A Cyclops data object
#'
#' @export
getNumberOfCovariates <- function(object) {
    .Call(`_Cyclops_cyclopsGetNumberOfColumns`, object)
}

#' @title Print Cyclops data matrix to file
#'
#' @description
#' \code{printMatrixMarket} prints the data matrix to a file
#'
#' @param object      A Cyclops data object
#' @param file        Filename
#'
#' @keywords internal
printMatrixMarket <- function(object, file) {
    invisible(.Call(`_Cyclops_cyclopsPrintMatrixMarket`, object, file))
}

#' @title Get total number of rows
#'
#' @description
#' \code{getNumberOfRows} returns the total number of outcome rows in a Cyclops data object
#'
#' @param object    A Cyclops data object
#'
#' @export
getNumberOfRows <- function(object) {
    .Call(`_Cyclops_cyclopsGetNumberOfRows`, object)
}

#' @title Get total number of outcome types
#'
#' @description
#' \code{getNumberOfTypes} returns the total number of outcome types in a Cyclops data object
#'
#' @param object    A Cyclops data object
#'
#' @keywords internal
getNumberOfTypes <- function(object) {
    .Call(`_Cyclops_cyclopsGetNumberOfTypes`, object)
}

.cyclopsUnivariableCorrelation <- function(x, bitCovariateLabel) {
    .Call(`_Cyclops_cyclopsUnivariableCorrelation`, x, bitCovariateLabel)
}

.cyclopsUnivariableSeparability <- function(x, bitCovariateLabel) {
    .Call(`_Cyclops_cyclopsUnivariableSeparability`, x, bitCovariateLabel)
}

.cyclopsSumByGroup <- function(x, bitCovariateLabel, bitGroupByLabel, power) {
    .Call(`_Cyclops_cyclopsSumByGroup`, x, bitCovariateLabel, bitGroupByLabel, power)
}

.cyclopsSumByStratum <- function(x, bitCovariateLabel, power) {
    .Call(`_Cyclops_cyclopsSumByStratum`, x, bitCovariateLabel, power)
}

.cyclopsSum <- function(x, bitCovariateLabel, power) {
    .Call(`_Cyclops_cyclopsSum`, x, bitCovariateLabel, power)
}

.cyclopsNewSqlData <- function(modelTypeName, noiseLevel, floatingPoint) {
    .Call(`_Cyclops_cyclopsNewSqlData`, modelTypeName, noiseLevel, floatingPoint)
}

.cyclopsMedian <- function(vector) {
    .Call(`_Cyclops_cyclopsMedian`, vector)
}

.cyclopsQuantile <- function(vector, q) {
    .Call(`_Cyclops_cyclopsQuantile`, vector, q)
}

.cyclopsNormalizeCovariates <- function(x, normalizationName) {
    .Call(`_Cyclops_cyclopsNormalizeCovariates`, x, normalizationName)
}

.cyclopsSetHasIntercept <- function(x, hasIntercept) {
    invisible(.Call(`_Cyclops_cyclopsSetHasIntercept`, x, hasIntercept))
}

.cyclopsGetHasIntercept <- function(x) {
    .Call(`_Cyclops_cyclopsGetHasIntercept`, x)
}

.cyclopsGetHasOffset <- function(x) {
    .Call(`_Cyclops_cyclopsGetHasOffset`, x)
}

.cyclopsGetMeanOffset <- function(x) {
    .Call(`_Cyclops_cyclopsGetMeanOffset`, x)
}

getYVector <- function(object) {
    .Call(`_Cyclops_cyclopsGetYVector`, object)
}

getTimeVector <- function(object) {
    .Call(`_Cyclops_cyclopsGetTimeVector`, object)
}

.cyclopsFinalizeData <- function(x, addIntercept, sexpOffsetCovariate, offsetAlreadyOnLogScale, sortCovariates, sexpCovariatesDense, magicFlag = FALSE) {
    invisible(.Call(`_Cyclops_cyclopsFinalizeData`, x, addIntercept, sexpOffsetCovariate, offsetAlreadyOnLogScale, sortCovariates, sexpCovariatesDense, magicFlag))
}

.loadCyclopsDataY <- function(x, stratumId, rowId, y, time) {
    invisible(.Call(`_Cyclops_cyclopsLoadDataY`, x, stratumId, rowId, y, time))
}

.loadCyclopsDataMultipleX <- function(x, covariateId, rowId, covariateValue, checkCovariateIds, checkCovariateBounds, append, forceSparse) {
    .Call(`_Cyclops_cyclopsLoadDataMultipleX`, x, covariateId, rowId, covariateValue, checkCovariateIds, checkCovariateBounds, append, forceSparse)
}

.loadCyclopsDataX <- function(x, bitCovariateId, rowId, covariateValue, replace, append, forceSparse) {
    .Call(`_Cyclops_cyclopsLoadDataX`, x, bitCovariateId, rowId, covariateValue, replace, append, forceSparse)
}

.loadCyclopsDataStratTimeEffects <- function(x, stratumId, rowId, subjectId, timeEffectCovariateId) {
    .Call(`_Cyclops_cyclopsLoadDataStratTimeEffects`, x, stratumId, rowId, subjectId, timeEffectCovariateId)
}

.appendSqlCyclopsData <- function(x, oStratumId, oRowId, oY, oTime, cRowId, cCovariateId, cCovariateValue) {
    .Call(`_Cyclops_cyclopsAppendSqlData`, x, oStratumId, oRowId, oY, oTime, cRowId, cCovariateId, cCovariateValue)
}

.cyclopsGetInterceptLabel <- function(x) {
    .Call(`_Cyclops_cyclopsGetInterceptLabel`, x)
}

.cyclopsReadData <- function(fileName, modelTypeName) {
    .Call(`_Cyclops_cyclopsReadFileData`, fileName, modelTypeName)
}

.cyclopsModelData <- function(pid, y, z, offs, dx, sx, ix, modelTypeName, useTimeAsOffset = FALSE, numTypes = 1L, floatingPoint = 64L) {
    .Call(`_Cyclops_cyclopsModelData`, pid, y, z, offs, dx, sx, ix, modelTypeName, useTimeAsOffset, numTypes, floatingPoint)
}

library(testthat)

test_that("quantile", {
    set.seed(123)
    vector <- runif(100)

    expect_equal(
        Cyclops:::.cyclopsQuantile(vector, q = 0.5),
        median(vector))
    expect_equal(
        Cyclops:::.cyclopsMedian(vector),
        median(vector))

    expect_error(
        Cyclops:::.cyclopsQuantile(vector, q = -0.1),
        "Invalid quantile")
})

test_that("null ptr", {
    expect_error(
        Cyclops:::.isRcppPtrNull(1),
        "Input must be an Rcpp externalptr")
})

test_that("print MatrixMarket format", {
    binomial_bid <- c(1,5,10,20,30,40,50,75,100,150,200)
    binomial_n <- c(31,29,27,25,23,21,19,17,15,15,15)
    binomial_y <- c(0,3,6,7,9,13,17,12,11,14,13)

    log_bid <- log(c(rep(rep(binomial_bid, binomial_n - binomial_y)), rep(binomial_bid, binomial_y)))
    y <- c(rep(0, sum(binomial_n - binomial_y)), rep(1, sum(binomial_y)))
    data <- createCyclopsData(y ~ log_bid, modelType = "lr")

    tempFile <- tempfile()
    Cyclops:::printMatrixMarket(data, file = tempFile)
    text <- readr::read_delim(file = tempFile, delim = " ", skip = 3, col_names = FALSE, col_types = readr::cols())
    expect_equal(nrow(text), getNumberOfCovariates(data) * getNumberOfRows(data))
    expect_equal(ncol(text), 1 + getNumberOfCovariates(data))

    unlink(tempFile)
})

library("testthat")

context("test-sorting.R")

#
# Sort functions
#

test_that("Check sorting functions", {
    set.seed(123)
    vector <- c(1:100)

    expect_true(Cyclops:::.isSortedVectorList(vectorList = list(vector[1]), ascending = FALSE))
    expect_true(Cyclops:::.isSortedVectorList(vectorList = list(vector), ascending = c(TRUE)))
    expect_false(Cyclops:::.isSortedVectorList(vectorList = list(vector), ascending = c(FALSE)))
})

#' @title Create a Cyclops control object that supports in- / out-of-sample hyperparameter search using weights
#'
#' @description
#' \code{createWeightBasedSearchControl} creates a Cyclops control object for use with \code{\link{fitCyclopsModel}}
#' that supports hyperparameter optimization through an auto-search where weight = 1 identifies in-sample observations
#' and weight = 0 identifies out-of-sample observations.
#'
#' @param cvType        Must equal "auto"
#' @param initialValue  Initial value for auto-search parameter
#' @param ...           Additional parameters passed through to \code{\link{createControl}}
#'
#' @return
#' A Cyclops prior object of class inheriting from \code{"cyclopsControl"}
#' for use with \code{fitCyclopsModel}.
#'
#' @export
createWeightBasedSearchControl <- function(cvType = "auto",
                                           initialValue = 1,
                                           ...) {
    if (cvType != "auto") {
        stop("Only auto-search is currently implemented")
    }
    control <- createControl(cvType = cvType, fold = -1, cvRepetitions = 1, minCVData = 2, ...)

    control$setHook <- function(cyclopsData, prior, control, weights, ...) {
        # closure to capture arguments

        if (!prior$useCrossValidation) {
            stop("Prior specification must require cross-validation")
        }

        if (is.null(weights)) {
            stop("Must specify weights")
        }

        control$setHook <- NULL # Do not re-enter call-back

        fit <- fitCyclopsModel(cyclopsData,
                               prior = prior,
                               control = control,
                               weights = weights, ...)

        return(fit)
    }

    return(control)
}

# @file cyclops.R
#
# Copyright 2014 Observational Health Data Sciences and Informatics
#
# This file is part of Cyclops
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @author Observational Health Data Sciences and Informatics
# @author Marc Suchard
# @author Trevor Shaddox


#' Cyclops: Cyclic coordinate descent for logistic, Poisson and survival analysis
#'
#' The Cyclops package incorporates cyclic coordinate descent and
#' majorization-minimization approaches to fit a variety of regression models
#' found in large-scale observational healthcare data.  Implementations focus
#' on computational optimization and fine-scale parallelization to yield
#' efficient inference in massive datasets.
#'
#' @name cyclops
#' @import Rcpp Matrix dplyr
#'
#' @importFrom methods as
#' @importFrom stats aggregate as.formula coef coefficients confint contrasts deviance model.matrix model.offset model.response pchisq poisson qchisq qnorm rbinom rexp rnorm rpois runif terms time vcov
#' @importFrom rlang .data
#'
#' @useDynLib Cyclops, .registration = TRUE
"_PACKAGE"

.onUnload <- function (libpath) {
  library.dynam.unload("Cyclops", libpath)
}

#' @title Set GPU device
#'
#' @description
#' \code{setOpenCLDevice} set GPU device
#'
#' @param name String: Name of GPU device
#'
#' @export
setOpenCLDevice <- function(name) {
    devices <- listGPUDevices()

    if (!(name %in% devices)) {
        stop("Unable to find device.")
    }

    Sys.setenv(BOOST_COMPUTE_DEFAULT_DEVICE = name)
    # TODO set CUDA device
}

#' @title Split the analysis time into several intervals for time-varying coefficients.
#'
#' @description
#' \code{splitTime} split the analysis time into several intervals for time-varying coefficients
#'
#' @param shortOut     A data frame containing the outcomes with predefined columns (see below).
#' @param cut          Numeric: Time points to cut at
#'
#' These columns are expected in the shortOut object:
#' \tabular{lll}{
#'   \verb{rowId}   \tab(integer) \tab Row ID is used to link multiple covariates (x) to a single outcome (y) \cr
#'   \verb{y}       \tab(real) \tab Observed event status \cr
#'   \verb{time}    \tab(real) \tab Observed event time \cr
#' }
#'
#' @return A long outcome table for time-varying coefficients.
#' @importFrom survival survfit survSplit
#' @export
splitTime <- function(shortOut, cut) {

    if (!"time" %in% colnames(shortOut)) stop("Must provide observed event time.")
    if (!"y" %in% colnames(shortOut)) stop("Must provide observed event status.")
    if ("rowId" %in% colnames(shortOut)) {
        shortOut <- shortOut %>%
            rename(subjectId = .data$rowId) %>%
            arrange(.data$subjectId)
    } else {
        shortOut <- shortOut %>%
            mutate(subjectId = row_number())
    }

    shortOut <- collect(shortOut)
    longOut <- do.call('survSplit', list(formula = Surv(shortOut$time, shortOut$y)~.,
                                         data = shortOut,
                                         cut = cut,
                                         episode = "stratumId",
                                         id = "newSubjectId"))
    longOut <- longOut %>%
        rename(y = .data$event) %>%
        mutate(time = .data$tstop - .data$tstart) %>%
        select(-c(.data$newSubjectId, .data$tstart, .data$tstop)) %>%
        arrange(.data$stratumId, .data$subjectId)

    # Restore rowIds
    SubjectIds <- shortOut$subjectId
    newSubjectId <- max(SubjectIds)+1
    longOut$rowId <-c(SubjectIds, # rowId = subjectId at 1st stratum
                      newSubjectId:(newSubjectId+(nrow(longOut)-length(SubjectIds))-1)) # create new distinct rowIds for other strata

    # Reorder columns
    longOut <- longOut %>%
        select(.data$stratumId, .data$subjectId, .data$rowId, everything())
        # select(.data$rowId, everything()) %>%
        # select(.data$subjectId, everything()) %>%
        # select(.data$stratumId, everything())

    return(longOut)
}

#' @title Convert short sparse covariate table to long sparse covariate table for time-varying coefficients.
#'
#' @description
#' \code{convertToTimeVaryingCoef} convert short sparse covariate table to long sparse covariate table for time-varying coefficients.
#'
#' @param shortCov       A data frame containing the covariate with predefined columns (see below).
#' @param longOut        A data frame containing the outcomes with predefined columns (see below), output of \code{splitTime}.
#' @param timeVaryCoefId   Integer: A numeric identifier of a time-varying coefficient
#'
#' @details
#' These columns are expected in the shortCov object:
#' \tabular{lll}{
#'   \verb{rowId}  	       \tab(integer) \tab Row ID is used to link multiple covariates (x) to a single outcome (y) \cr
#'   \verb{covariateId}    \tab(integer) \tab A numeric identifier of a covariate  \cr
#'   \verb{covariateValue} \tab(real) \tab The value of the specified covariate \cr
#' }
#'
#' These columns are expected in the longOut object:
#' \tabular{lll}{
#'   \verb{stratumId}   \tab(integer) \tab Stratum ID for time-varying models \cr
#'   \verb{subjectId}  	\tab(integer) \tab Subject ID is used to link multiple covariates (x) at different time intervals to a single subject \cr
#'   \verb{rowId}  	    \tab(integer) \tab Row ID is used to link multiple covariates (x) to a single outcome (y) \cr
#'   \verb{y}           \tab(real) \tab The outcome variable \cr
#'   \verb{time}        \tab(real) \tab For models that use time (e.g. Poisson or Cox regression) this contains time \cr
#'                      \tab       \tab(e.g. number of days) \cr
#' }
#' @return A long sparse covariate table for time-varying coefficients.
#' @export
convertToTimeVaryingCoef <- function(shortCov, longOut, timeVaryCoefId) {

    # Process time-varying coefficients
    timeVaryCoefId <- sort(unique(timeVaryCoefId))
    numTime <- length(timeVaryCoefId) # number of time-varying covariates
    maxCovId <- max(shortCov$covariateId)

    # First stratum
    longCov <- shortCov
    longCov$stratumId <- 1
    colnames(longCov)[which(names(longCov) == "rowId")] <- "subjectId"
    colnames(shortCov)[which(names(shortCov) == "rowId")] <- "subjectId"

    # Rest of strata
    maxStrata <- max(longOut$stratumId)
    for (st in 2:maxStrata) {

        # get valid subjects in current stratum
        subId <- longOut[longOut$stratumId == st, ]$subjectId

        # get valid sparse covariates information in current stratum
        curStrata <- shortCov[shortCov$subjectId %in% subId, ]

        if (any(curStrata$covariateId %in% timeVaryCoefId)) { # skip when valid subjects only have non-zero time-indep covariates
            curStrata$stratumId <- st # assign current stratumId

            # recode covariateId for time-varying coefficients
            # TODO update label
            for (i in 1:numTime) {
                curStrata[curStrata$covariateId == timeVaryCoefId[i], "covariateId"] <- maxCovId + numTime * (st - 2) + i
            }

            # bind current stratum to longCov
            longCov <- rbind(longCov, curStrata)
        }
    }

    # match rowId in longCov
    longCov$rowId <- NA
    for (i in 1:nrow(longCov)) {
        longCov$rowId[i] <- longOut[with(longOut, subjectId == longCov$subjectId[i] & stratumId == longCov$stratumId[i]), "rowId"]
    }

    return(longCov)
}

library(survival)
library(testthat)

GpuDevice <- listGPUDevices()[1]
tolerance <- 1E-4

test_that("Check very small Cox example with time-varying coefficient as stratified model", {
    skip_if(length(listGPUDevices()) == 0, "GPU not available")
    test <- read.table(header=T, sep = ",", text = "
            start, length, event, x1, x2
            0, 4,  1,0,0
            0, 3.5,1,2,0
            0, 3,  0,0,1
            0, 2.5,1,0,1
            0, 2,  1,1,1
            0, 1.5,0,1,0
            0, 1,  1,1,0")
    test_split <- survSplit(Surv(length, event)~., data = test,
                            cut = c(2),
                            episode = "stratumId")

    # assume x1 has time-varying effect on the outcome
    test_cyclops <- test_split
    test_cyclops$time <- test_cyclops$length - test_cyclops$tstart
    test_cyclops$x1.1 <- ifelse(test_cyclops$stratumId == 1, test_cyclops$x1, 0)
    test_cyclops$x1.2 <- ifelse(test_cyclops$stratumId == 2, test_cyclops$x1, 0)
    test_cyclops <- test_cyclops[, c("stratumId", "time", "event",
                                     "x1.1", "x1.2", "x2")]

    goldStrat <- coxph(Surv(tstart, length, event) ~ x2 + x1:strata(stratumId), test_split)

    dataPtrStrat_CPU <- createCyclopsData(Surv(time, event) ~ x2 + x1.1 + x1.2 + strata(stratumId),
                                      data = test_cyclops,
                                      modelType = "cox_time")
    cyclopsFitStrat_CPU <- fitCyclopsModel(dataPtrStrat_CPU)

    dataPtrStrat_GPU <- createCyclopsData(Surv(time, event) ~ x2 + x1.1 + x1.2 + strata(stratumId),
                                      data = test_cyclops,
                                      modelType = "cox_time")
    cyclopsFitStrat_GPU <- fitCyclopsModel(dataPtrStrat_GPU, computeDevice = GpuDevice)

    expect_equal(unname(coef(cyclopsFitStrat_CPU)), unname(coef(goldStrat)), tolerance = tolerance)
    expect_equal(coef(cyclopsFitStrat_CPU), coef(cyclopsFitStrat_GPU), tolerance = tolerance)


    # short sparse cov
    sparseShortCov <- data.frame(stratumId = 1,
                                 rowId = rep(1:nrow(test), 2),
                                 covariateId = rep(1:2, each = nrow(test)),
                                 covariateValue = c(test$x2, test$x1))
    sparseShortCov <- sparseShortCov[sparseShortCov$covariateValue != 0,]

    # long out
    longOut <- splitTime(data.frame(time = test$length,
                                    y = test$event),
                         cut = c(2))

    # long sparse cov
    sparseLongCov <- convertToTimeVaryingCoef(sparseShortCov, longOut, timeVaryCoefId = c(2))

    # CPU: long out and long cov
    dataPtrSparse_CPU <- convertToCyclopsData(outcomes = longOut,
                                              covariates = sparseLongCov,
                                              modelType = "cox_time")
    cyclopsFitSparse_CPU <- fitCyclopsModel(dataPtrSparse_CPU)

    # GPU: long out and long cov
    dataPtrSparse_GPU <- convertToCyclopsData(outcomes = longOut,
                                              covariates = sparseLongCov,
                                              modelType = "cox_time")
    cyclopsFitSparse_GPU <- fitCyclopsModel(dataPtrSparse_GPU, computeDevice = GpuDevice)

    expect_equal(unname(coef(cyclopsFitSparse_CPU)), unname(coef(goldStrat)), tolerance = tolerance)
    expect_equal(coef(cyclopsFitSparse_CPU), coef(cyclopsFitSparse_GPU), tolerance = tolerance)

    # GPU: long out and short cov
    shortCovPtrSparse_GPU <- convertToCyclopsData(outcomes = longOut,
                                                  covariates = sparseShortCov,
                                                  timeEffectMap = data.frame(covariateId = c(2)),
                                                  modelType = "cox_time")
    shortCovFitSparse_GPU <- fitCyclopsModel(shortCovPtrSparse_GPU, computeDevice = GpuDevice)
    expect_equal(unname(coef(cyclopsFitSparse_GPU)), unname(coef(shortCovFitSparse_GPU)), tolerance = tolerance)

})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{convertToGlmnetLambda}
\alias{convertToGlmnetLambda}
\title{Convert to glmnet regularization parameter}
\usage{
convertToGlmnetLambda(variance, nobs)
}
\arguments{
\item{variance}{Prior variance}

\item{nobs}{Number of observation rows in dataset}
}
\value{
\code{lambda}
}
\description{
\code{convertToGlmnetLambda} converts a prior variance
from \code{Cyclops} into the regularization parameter \code{lambda}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{getCovariateIds}
\alias{getCovariateIds}
\title{Get covariate identifiers}
\usage{
getCovariateIds(object)
}
\arguments{
\item{object}{A Cyclops data object}
}
\description{
\code{getCovariateIds} returns a vector of integer64 covariate identifiers in a Cyclops data object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{convertToCyclopsVariance}
\alias{convertToCyclopsVariance}
\title{Convert to Cyclops Prior Variance}
\usage{
convertToCyclopsVariance(lambda, nobs)
}
\arguments{
\item{lambda}{Regularization parameter from \code{glmnet}}

\item{nobs}{Number of observation rows in dataset}
}
\value{
Prior variance under a Laplace() prior
}
\description{
\code{convertToCyclopsVariance} converts the regularization parameter \code{lambda}
from \code{glmnet} into a prior variance.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataManagement.R
\name{appendSqlCyclopsData}
\alias{appendSqlCyclopsData}
\title{appendSqlCyclopsData}
\usage{
appendSqlCyclopsData(
  object,
  oStratumId,
  oRowId,
  oY,
  oTime,
  cRowId,
  cCovariateId,
  cCovariateValue
)
}
\arguments{
\item{object}{OHDSI Cyclops data object to append entries}

\item{oStratumId}{Integer vector (optional): non-unique stratum identifier for each row in outcomes table}

\item{oRowId}{Integer vector: unique row identifier for each row in outcomes table}

\item{oY}{Numeric vector: model outcome variable for each row in outcomes table}

\item{oTime}{Numeric vector (optional): exposure interval or censoring time for each row in outcomes table}

\item{cRowId}{Integer vector: non-unique row identifier for each row in covariates table that matches a single outcomes table entry}

\item{cCovariateId}{Integer vector: covariate identifier}

\item{cCovariateValue}{Numeric vector: covariate value}
}
\description{
\code{appendSqlCyclopsData} appends data to an OHDSI data object.
}
\details{
Append data using two tables.  The outcomes table is dense and contains ...  The covariates table is sparse and contains ...
All entries in the outcome table must be sorted in increasing order by (oStratumId, oRowId).  All entries in the covariate table
must be sorted in increasing order by (cRowId). Each cRowId value must match exactly one oRowId value.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{cyclopsLibraryFileName}
\alias{cyclopsLibraryFileName}
\title{Get full path to installed Cyclops library}
\usage{
cyclopsLibraryFileName()
}
\value{
String
}
\description{
\code{cyclopsLibraryFileName} returns the full path to the installed Cyclops .so/.dll file
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{confint.cyclopsFit}
\alias{confint.cyclopsFit}
\title{Confidence intervals for Cyclops model parameters}
\usage{
\method{confint}{cyclopsFit}(
  object,
  parm,
  level = 0.95,
  overrideNoRegularization = FALSE,
  includePenalty = TRUE,
  rescale = FALSE,
  maximumCurvature = -0.001,
  ...
)
}
\arguments{
\item{object}{A fitted Cyclops model object}

\item{parm}{A specification of which parameters require confidence intervals,
either a vector of numbers of covariateId names}

\item{level}{Numeric: confidence level required}

\item{overrideNoRegularization}{Logical: Enable confidence interval estimation for regularized parameters}

\item{includePenalty}{Logical: Include regularized covariate penalty in profile}

\item{rescale}{Boolean: rescale coefficients for unnormalized covariate values}

\item{maximumCurvature}{Numeric: maximum curvature of log-likelihood profile required to compute CI}

\item{...}{Additional argument(s) for methods}
}
\value{
A matrix with columns reporting lower and upper confidence limits for each parameter.
These columns are labelled as (1-level) / 2 and 1 - (1 - level) / 2 in percent
(by default 2.5 percent and 97.5 percent)
}
\description{
\code{confinit.cyclopsFit} profiles the data likelihood to construct confidence intervals of
arbitrary level. Usually it only makes sense to do this for variables that have not been regularized.
}
\examples{
#Generate some simulated data:
sim <- simulateCyclopsData(nstrata = 1, nrows = 1000, ncovars = 2, eCovarsPerRow = 0.5, 
                           model = "poisson")
cyclopsData <- convertToCyclopsData(sim$outcomes, sim$covariates, modelType = "pr", 
                                    addIntercept = TRUE)

#Define the prior and control objects to use cross-validation for finding the 
#optimal hyperparameter:
prior <- createPrior("laplace", exclude = 0, useCrossValidation = TRUE)
control <- createControl(cvType = "auto", noiseLevel = "quiet")

#Fit the model
fit <- fitCyclopsModel(cyclopsData,prior = prior, control = control)  

#Find out what the optimal hyperparameter was:
getHyperParameter(fit)

#Extract the current log-likelihood, and coefficients
logLik(fit)
coef(fit)

#We can only retrieve the confidence interval for unregularized coefficients:
confint(fit, c(0))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataManagement.R
\name{isInitialized}
\alias{isInitialized}
\title{Check if a Cyclops data object is initialized}
\usage{
isInitialized(object)
}
\arguments{
\item{object}{Cyclops data object to test}
}
\description{
\code{isInitialized} determines if an Cyclops data object is properly
initialized and remains in memory.  Cyclops data objects do not
serialized/deserialize their back-end memory across R sessions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{vcov.cyclopsFit}
\alias{vcov.cyclopsFit}
\title{Calculate variance-covariance matrix for a fitted Cyclops model object}
\usage{
\method{vcov}{cyclopsFit}(object, control, overrideNoRegularization = FALSE, ...)
}
\arguments{
\item{object}{A fitted Cyclops model object}

\item{control}{A \code{"cyclopsControl"} object constructed by \code{\link{createControl}}}

\item{overrideNoRegularization}{Logical: Enable variance-covariance estimation for regularized parameters}

\item{...}{Additional argument(s) for methods}
}
\value{
A matrix of the estimates covariances between all covariate estimates.
}
\description{
\code{vcov.cyclopsFit} returns the variance-covariance matrix for all covariates of a Cyclops model object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataManagement.R
\name{reduce}
\alias{reduce}
\title{Apply simple data reductions}
\usage{
reduce(object, covariates, groupBy, power = 1)
}
\arguments{
\item{object}{A Cyclops data object}

\item{covariates}{Integer or string vector: list of covariates to report}

\item{groupBy}{Integer or string (optional): generates a segmented reduction stratified by this covariate.  Setting \code{groupBy = "stratum"} segments reduction for strataID}

\item{power}{Integer: 0 = non-zero count, 1 = sum, 2 = sum-of-squares}
}
\value{
Specified reduction as number or \code{data.frame} if segmented.
}
\description{
\code{reduce} reports the count of non-zero elements, sum and sum-of-squares for specified covariates in a Cyclops data object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Multitype.R
\name{Multitype}
\alias{Multitype}
\title{Create a multitype outcome object}
\usage{
Multitype(y, type)
}
\arguments{
\item{y}{Numeric: Response count(s)}

\item{type}{Numeric or factor: Response type}
}
\value{
An object of class \code{Multitype} with length equal to the length of \code{y} and \code{type}.
}
\description{
\code{Multitype} creates a multitype outcome object, usually used as a response variable in a
hierarchical Cyclops model fit.
}
\examples{
Multitype(c(0,1,0), as.factor(c("A","A","B")))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataManagement.R
\name{getCovariateTypes}
\alias{getCovariateTypes}
\title{Get covariate types}
\usage{
getCovariateTypes(object, covariateLabel)
}
\arguments{
\item{object}{A Cyclops data object}

\item{covariateLabel}{Integer vector: covariate identifiers to return}
}
\description{
\code{getCovariateTypes} returns a vector covariate types in a Cyclops data object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Simulation.R
\name{coverage}
\alias{coverage}
\title{Coverage}
\usage{
coverage(goldStandard, lowerBounds, upperBounds)
}
\arguments{
\item{goldStandard}{Numeric vector}

\item{lowerBounds}{Numeric vector. Lower bound of the confidence intervals}

\item{upperBounds}{Numeric vector. Upper bound of the confidence intervals}
}
\value{
The proportion of times \code{goldStandard} falls between \code{lowerBound} and \code{upperBound}
}
\description{
\code{coverage} computes the coverage on confidence intervals
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{getSEs}
\alias{getSEs}
\title{Extract standard errors}
\usage{
getSEs(object, covariates)
}
\arguments{
\item{object}{A Cyclops model fit object}

\item{covariates}{Integer or string vector: list of covariates for which asymptotic standard errors are wanted}
}
\value{
Vector of standard error estimates
}
\description{
\code{getSEs} extracts asymptotic standard errors for specific covariates from a Cyclops model fit object.
}
\details{
This function first computes the (partial) Fisher information matrix for
just the requested covariates and then returns the square root of the diagonal elements of
the inverse of the Fisher information matrix.  These are the asymptotic standard errors
when all possible covariates are included.
When the requested covariates do not equate to all coefficients in the model,
then interpretation is more challenging.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{createControl}
\alias{createControl}
\title{Create a Cyclops control object}
\usage{
createControl(
  maxIterations = 1000,
  tolerance = 1e-06,
  convergenceType = "gradient",
  cvType = "auto",
  fold = 10,
  lowerLimit = 0.01,
  upperLimit = 20,
  gridSteps = 10,
  cvRepetitions = 1,
  minCVData = 100,
  noiseLevel = "silent",
  threads = 1,
  seed = NULL,
  resetCoefficients = FALSE,
  startingVariance = -1,
  useKKTSwindle = FALSE,
  tuneSwindle = 10,
  selectorType = "auto",
  initialBound = 2,
  maxBoundCount = 5,
  algorithm = "ccd",
  doItAll = TRUE,
  syncCV = FALSE
)
}
\arguments{
\item{maxIterations}{Integer: maximum iterations of Cyclops to attempt before returning a failed-to-converge error}

\item{tolerance}{Numeric: maximum relative change in convergence criterion from successive iterations to achieve convergence}

\item{convergenceType}{String: name of convergence criterion to employ (described in more detail below)}

\item{cvType}{String: name of cross validation search.
Option \code{"auto"} selects an auto-search following BBR.
Option \code{"grid"} selects a grid-search cross validation}

\item{fold}{Numeric: Number of random folds to employ in cross validation}

\item{lowerLimit}{Numeric: Lower prior variance limit for grid-search}

\item{upperLimit}{Numeric: Upper prior variance limit for grid-search}

\item{gridSteps}{Numeric: Number of steps in grid-search}

\item{cvRepetitions}{Numeric: Number of repetitions of X-fold cross validation}

\item{minCVData}{Numeric: Minimum number of data for cross validation}

\item{noiseLevel}{String: level of Cyclops screen output (\code{"silent"}, \code{"quiet"}, \code{"noisy"})}

\item{threads}{Numeric: Specify number of CPU threads to employ in cross-validation; default = 1 (auto = -1)}

\item{seed}{Numeric: Specify random number generator seed. A null value sets seed via \code{\link{Sys.time}}.}

\item{resetCoefficients}{Logical: Reset all coefficients to 0 between model fits under cross-validation}

\item{startingVariance}{Numeric: Starting variance for auto-search cross-validation; default = -1 (use estimate based on data)}

\item{useKKTSwindle}{Logical: Use the Karush-Kuhn-Tucker conditions to limit search}

\item{tuneSwindle}{Numeric: Size multiplier for active set}

\item{selectorType}{String: name of exchangeable sampling unit.
Option \code{"byPid"} selects entire strata.
Option \code{"byRow"} selects single rows.
If set to \code{"auto"}, \code{"byRow"} will be used for all models except conditional models where
the average number of rows per stratum is smaller than the number of strata.}

\item{initialBound}{Numeric: Starting trust-region size}

\item{maxBoundCount}{Numeric: Maximum number of tries to decrease initial trust-region size}

\item{algorithm}{String: name of fitting algorithm to employ; default is \code{ccd}}

\item{doItAll}{Currently unused}

\item{syncCV}{Currently unused

Todo: Describe convegence types}
}
\value{
A Cyclops control object of class inheriting from \code{"cyclopsControl"} for use with \code{\link{fitCyclopsModel}}.
}
\description{
\code{createControl} creates a Cyclops control object for use with \code{\link{fitCyclopsModel}}.
}
\examples{
#Generate some simulated data:
sim <- simulateCyclopsData(nstrata = 1, nrows = 1000, ncovars = 2, eCovarsPerRow = 0.5, 
                           model = "poisson")
cyclopsData <- convertToCyclopsData(sim$outcomes, sim$covariates, modelType = "pr", 
                                    addIntercept = TRUE)

#Define the prior and control objects to use cross-validation for finding the 
#optimal hyperparameter:
prior <- createPrior("laplace", exclude = 0, useCrossValidation = TRUE)
control <- createControl(cvType = "auto", noiseLevel = "quiet")

#Fit the model
fit <- fitCyclopsModel(cyclopsData,prior = prior, control = control)  

#Find out what the optimal hyperparameter was:
getHyperParameter(fit)

#Extract the current log-likelihood, and coefficients
logLik(fit)
coef(fit)

#We can only retrieve the confidence interval for unregularized coefficients:
confint(fit, c(0))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Residuals.R
\name{testProportionality}
\alias{testProportionality}
\title{Test hazard ratio proportionality assumption}
\usage{
testProportionality(object, parm = NULL, transform = "km")
}
\arguments{
\item{object}{A Cyclops model fit object}

\item{parm}{A specification of which parameters require a proportionality test,
either a vector of numbers or covariateId names}

\item{transform}{Same as \code{transform} in \code{cox.zph}}
}
\description{
\code{testProportionality} tests the hazard ratio proportionality assumption
of a Cyclops model fit object using a score test
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{printMatrixMarket}
\alias{printMatrixMarket}
\title{Print Cyclops data matrix to file}
\usage{
printMatrixMarket(object, file)
}
\arguments{
\item{object}{A Cyclops data object}

\item{file}{Filename}
}
\description{
\code{printMatrixMarket} prints the data matrix to a file
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Simulation.R
\name{simulateCyclopsData}
\alias{simulateCyclopsData}
\title{Simulation Cyclops dataset}
\usage{
simulateCyclopsData(
  nstrata = 200,
  nrows = 10000,
  ncovars = 20,
  effectSizeSd = 1,
  zeroEffectSizeProp = 0.9,
  eCovarsPerRow = ncovars/100,
  model = "survival"
)
}
\arguments{
\item{nstrata}{Numeric: Number of strata}

\item{nrows}{Numeric: Number of observation rows}

\item{ncovars}{Numeric: Number of covariates}

\item{effectSizeSd}{Numeric: Standard derivation of the non-zero simulated regression coefficients}

\item{zeroEffectSizeProp}{Numeric: Expected proportion of zero effect size}

\item{eCovarsPerRow}{Number: Effective number of non-zero covariates per data row}

\item{model}{String: Simulation model. Choices are: \code{logistic}, \code{poisson} or \code{survival}}
}
\value{
A simulated data set
}
\description{
\code{simulateCyclopsData} generates a simulated large, sparse data set for use by \code{fitCyclopsSimulation}.
}
\examples{
#Generate some simulated data:
sim <- simulateCyclopsData(nstrata = 1, nrows = 1000, ncovars = 2, eCovarsPerRow = 0.5, 
                           model = "poisson")
cyclopsData <- convertToCyclopsData(sim$outcomes, sim$covariates, modelType = "pr", 
                                    addIntercept = TRUE)

#Define the prior and control objects to use cross-validation for finding the 
#optimal hyperparameter:
prior <- createPrior("laplace", exclude = 0, useCrossValidation = TRUE)
control <- createControl(cvType = "auto", noiseLevel = "quiet")

#Fit the model
fit <- fitCyclopsModel(cyclopsData,prior = prior, control = control)  

#Find out what the optimal hyperparameter was:
getHyperParameter(fit)

#Extract the current log-likelihood, and coefficients
logLik(fit)
coef(fit)

#We can only retrieve the confidence interval for unregularized coefficients:
confint(fit, c(0))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{runBootstrap}
\alias{runBootstrap}
\title{Run bootstrap for Cyclops model parameters}
\usage{
runBootstrap(object, replicates, ciCoverage = 0.95, na.rm = FALSE)
}
\arguments{
\item{object}{A fitted Cyclops model object}

\item{replicates}{Numeric: number of bootstrap samples}

\item{ciCoverage}{Numeric (0-1): coverage proportion for bootstrap confidence intervals}

\item{na.rm}{Logical: strip out bootstrap replicates with NA values (most likely without a finite estimate)}
}
\description{
Run bootstrap for Cyclops model parameters
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ParameterizedPrior.R
\name{createParameterizedPrior}
\alias{createParameterizedPrior}
\title{Create a Cyclops parameterized prior object}
\usage{
createParameterizedPrior(
  priorType,
  parameterize,
  values,
  useCrossValidation = FALSE,
  forceIntercept = FALSE
)
}
\arguments{
\item{priorType}{Character vector: specifies prior distribution.  See below for options}

\item{parameterize}{Function list: parameterizes location and variance}

\item{values}{Numeric vector: initial parameter values}

\item{useCrossValidation}{Logical: Perform cross-validation to determine \code{parameters}.}

\item{forceIntercept}{Logical: Force intercept coefficient into prior}
}
\value{
A Cyclops prior object of class inheriting from \code{"cyclopsPrior"} and \code{"cyclopsFunctionalPrior"}
for use with \code{fitCyclopsModel}.
}
\description{
\code{createParameterizedPrior} creates a Cyclops prior object for use with \code{\link{fitCyclopsModel}}
in which arbitrary \code{R} functions parameterize the prior location and variance.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{logLik.cyclopsFit}
\alias{logLik.cyclopsFit}
\title{Extract log-likelihood}
\usage{
\method{logLik}{cyclopsFit}(object, ...)
}
\arguments{
\item{object}{A Cyclops model fit object}

\item{...}{Additional arguments}
}
\description{
\code{logLik} returns the current log-likelihood of the fit in a Cyclops model fit object
}
\examples{
#Generate some simulated data:
sim <- simulateCyclopsData(nstrata = 1, nrows = 1000, ncovars = 2, eCovarsPerRow = 0.5, 
                           model = "poisson")
cyclopsData <- convertToCyclopsData(sim$outcomes, sim$covariates, modelType = "pr", 
                                    addIntercept = TRUE)

#Define the prior and control objects to use cross-validation for finding the 
#optimal hyperparameter:
prior <- createPrior("laplace", exclude = 0, useCrossValidation = TRUE)
control <- createControl(cvType = "auto", noiseLevel = "quiet")

#Fit the model
fit <- fitCyclopsModel(cyclopsData,prior = prior, control = control)  

#Find out what the optimal hyperparameter was:
getHyperParameter(fit)

#Extract the current log-likelihood, and coefficients
logLik(fit)
coef(fit)

#We can only retrieve the confidence interval for unregularized coefficients:
confint(fit, c(0))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataManagement.R
\name{finalizeSqlCyclopsData}
\alias{finalizeSqlCyclopsData}
\title{finalizeSqlCyclopsData}
\usage{
finalizeSqlCyclopsData(
  object,
  addIntercept = FALSE,
  useOffsetCovariate = NULL,
  offsetAlreadyOnLogScale = FALSE,
  sortCovariates = FALSE,
  makeCovariatesDense = NULL
)
}
\arguments{
\item{object}{Cyclops data object}

\item{addIntercept}{Add an intercept covariate if one was not imported through SQL}

\item{useOffsetCovariate}{Specify is a covariate should be used as an offset (fixed coefficient = 1).
Set option to \code{-1} to specify the time-to-event column,
otherwise include a single numeric or character covariate name.}

\item{offsetAlreadyOnLogScale}{Set to \code{TRUE} to indicate that offsets were log-transformed before importing into Cyclops data object.}

\item{sortCovariates}{Sort covariates in numeric-order with intercept first if it exists.}

\item{makeCovariatesDense}{List of numeric or character covariates names to densely represent in Cyclops data object.
For efficiency, we suggest making at least the intercept dense.}
}
\description{
\code{finalizeSqlCyclopsData} finalizes a Cyclops data object
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataManagement.R
\name{createCyclopsData}
\alias{createCyclopsData}
\title{Create a Cyclops data object}
\usage{
createCyclopsData(
  formula,
  sparseFormula,
  indicatorFormula,
  modelType,
  data,
  subset = NULL,
  weights = NULL,
  censorWeights = NULL,
  offset = NULL,
  time = NULL,
  pid = NULL,
  y = NULL,
  type = NULL,
  dx = NULL,
  sx = NULL,
  ix = NULL,
  model = FALSE,
  normalize = NULL,
  floatingPoint = 64,
  method = "cyclops.fit"
)
}
\arguments{
\item{formula}{An object of class \code{"\link{formula}"} that provides a symbolic description of the numerically dense model response and terms.}

\item{sparseFormula}{An object of class \code{"\link{formula}"} that provides a symbolic description of numerically sparse model terms.}

\item{indicatorFormula}{An object of class \code{"\link{formula}"} that provides a symbolic description of \{0,1\} model terms.}

\item{modelType}{character string: Valid types are listed below.}

\item{data}{An optional data frame, list or environment containing the variables in the model.}

\item{subset}{Currently unused}

\item{weights}{Currently unused}

\item{censorWeights}{Vector of subject-specific censoring weights (between 0 and 1). Currently only supported in \code{modelType = "fgr"}.}

\item{offset}{Currently unused}

\item{time}{Currently undocumented}

\item{pid}{Optional vector of integer stratum identifiers. If supplied, all rows must be sorted by increasing identifiers}

\item{y}{Currently undocumented}

\item{type}{Currently undocumented}

\item{dx}{Optional dense \code{"\link[Matrix]{Matrix}"} of covariates}

\item{sx}{Optional sparse \code{"\link[Matrix]{Matrix}"} of covariates}

\item{ix}{Optional \{0,1\} \code{"\link[Matrix]{Matrix}"} of covariates}

\item{model}{Currently undocumented}

\item{normalize}{String: Name of normalization for all non-indicator covariates (possible values: stdev, max, median)}

\item{floatingPoint}{Integer: Floating-point representation size (32 or 64)}

\item{method}{Currently undocumented}
}
\value{
A list that contains a Cyclops model data object pointer and an operation duration
}
\description{
\code{createCyclopsData} creates a Cyclops data object from an R formula or data matrices.
}
\details{
This function creates a Cyclops model data object from R \code{"\link{formula}"} or directly from
numeric vectors and matrices to define the model response and covariates.
If specifying a model using a \code{"\link{formula}"}, then the left-hand side define the model response and the
right-hand side defines dense covariate terms.
Objects provided with \code{"sparseFormula"} and \code{"indicatorFormula"} must be include left-hand side responses and terms are
coersed into sparse and indicator representations for computational efficiency.

Items to discuss:
\itemize{
\item Only use formula or (y,dx,...)
\item stratum() in formula
\item offset() in formula
\item when \code{"stratum"} (renamed from pid) are necessary
\item when \code{"time"} are necessary
}
}
\section{Models}{

Currently supported model types are:
\tabular{ll}{
\verb{	"ls"} \tab Least squares \cr
\verb{	"pr"} \tab Poisson regression \cr
\verb{	"lr"} \tab Logistic regression \cr
\verb{	"clr"} \tab Conditional logistic regression \cr
\verb{	"cpr"} \tab Conditional Poisson regression \cr
\verb{	"sccs"} \tab Self-controlled case series \cr
\verb{	"cox"} \tab Cox proportional hazards regression \cr
\verb{	"fgr"} \tab Fine-Gray proportional subdistribution hazards regression \cr
}
}

\examples{
## Dobson (1990) Page 93: Randomized Controlled Trial :
counts <- c(18, 17, 15, 20, 10, 20, 25, 13, 12)
outcome <- gl(3, 1, 9)
treatment <- gl(3, 3)
cyclopsData <- createCyclopsData(
     counts ~ outcome + treatment,
     modelType = "pr")
cyclopsFit <- fitCyclopsModel(cyclopsData)

cyclopsData2 <- createCyclopsData(
     counts ~ outcome,
     indicatorFormula = ~ treatment,
     modelType = "pr")
summary(cyclopsData2)
cyclopsFit2 <- fitCyclopsModel(cyclopsData2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{coef.cyclopsFit}
\alias{coef.cyclopsFit}
\title{Extract model coefficients}
\usage{
\method{coef}{cyclopsFit}(object, rescale = FALSE, ignoreConvergence = FALSE, ...)
}
\arguments{
\item{object}{Cyclops model fit object}

\item{rescale}{Boolean: rescale coefficients for unnormalized covariate values}

\item{ignoreConvergence}{Boolean: return coefficients even if fit object did not converge}

\item{...}{Other arguments}
}
\value{
Named numeric vector of model coefficients.
}
\description{
\code{coef.cyclopsFit} extracts model coefficients from an Cyclops model fit object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Predict.R
\name{meanLinearPredictor}
\alias{meanLinearPredictor}
\title{Calculates xbar*beta}
\usage{
meanLinearPredictor(cyclopsFit)
}
\arguments{
\item{cyclopsFit}{A Cyclops model fit object}
}
\description{
\code{meanLinearPredictor} computes xbar*beta for model fit
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Residuals.R
\name{residuals.cyclopsFit}
\alias{residuals.cyclopsFit}
\title{Model residuals}
\usage{
\method{residuals}{cyclopsFit}(object, parm = NULL, type = "schoenfeld", ...)
}
\arguments{
\item{object}{A Cyclops model fit object}

\item{parm}{A specification of which parameters require residuals,
either a vector of numbers or covariateId names}

\item{type}{Character string indicating the type of residual desires. Possible
values are "schoenfeld".}

\item{...}{Additional parameters for compatibility with S3 parent function}
}
\description{
\code{residuals.cyclopsFit} computes model residuals for Cox model-based Cyclops objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{getNumberOfCovariates}
\alias{getNumberOfCovariates}
\title{Get total number of covariates}
\usage{
getNumberOfCovariates(object)
}
\arguments{
\item{object}{A Cyclops data object}
}
\description{
\code{getNumberOfCovariates} returns the total number of covariates in a Cyclops data object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Simulation.R
\name{plotCyclopsSimulationFit}
\alias{plotCyclopsSimulationFit}
\title{Plot Cyclops simulation model fit}
\usage{
plotCyclopsSimulationFit(fit, goldStandard, label)
}
\arguments{
\item{fit}{A Cyclops simulation fit generated by \code{fitCyclopsSimulation}}

\item{goldStandard}{Numeric vector.  True relative risks.}

\item{label}{String. Name of estimate type.}
}
\description{
\code{plotCyclopsSimulationFit} generates a plot that compares \code{goldStandard} coefficients to their
Cyclops model \code{fit}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Simulation.R
\name{fitCyclopsSimulation}
\alias{fitCyclopsSimulation}
\title{Fit simulated data}
\usage{
fitCyclopsSimulation(
  sim,
  useCyclops = TRUE,
  model = "logistic",
  coverage = TRUE,
  includePenalty = FALSE,
  computeDevice = "native"
)
}
\arguments{
\item{sim}{A simulated Cyclops dataset generated via \code{simulateCyclopsData}}

\item{useCyclops}{Logical: use Cyclops or a standard routine}

\item{model}{String: Fitted regression model type}

\item{coverage}{Logical: report coverage statistics}

\item{includePenalty}{Logical: include regularized regression penalty in computing profile likelihood based confidence intervals}

\item{computeDevice}{String: Name of compute device to employ; defaults to \code{"native"} C++ on CPU}
}
\description{
\code{fitCyclopsSimulation} fits simulated Cyclops data using Cyclops or a standard routine.
This function is useful for simulation studies comparing the performance of Cyclops when considering
large, sparse datasets.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{getCyclopsPredictiveLogLikelihood}
\alias{getCyclopsPredictiveLogLikelihood}
\title{Compute predictive log-likelihood from a Cyclops model fit}
\usage{
getCyclopsPredictiveLogLikelihood(object, weights)
}
\arguments{
\item{object}{A Cyclops model fit object}

\item{weights}{Numeric vector: vector of 0/1 identifying subset (=1) of rows from \code{object} to use in computing the log-likelihood}
}
\value{
The predictive log-likelihood
}
\description{
\code{getCyclopsPredictiveLogLikelihood} returns the log-likelihood of a subset of the data in a Cyclops model fit object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataManagement.R
\name{getUnivariableCorrelation}
\alias{getUnivariableCorrelation}
\title{Get univariable correlation}
\usage{
getUnivariableCorrelation(cyclopsData, covariates = NULL, threshold = 0)
}
\arguments{
\item{cyclopsData}{A Cyclops data object}

\item{covariates}{Integer or string vector: list of covariates to report; default (NULL) implies all covariates}

\item{threshold}{Correlation threshold for reporting}
}
\value{
A list of covariates whose absolute correlation with the outcome is greater than or equal to the threshold
}
\description{
\code{getUnivariableCorrelation} reports covariates that have high correlation with the outcome
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Gpu.R
\name{setOpenCLDevice}
\alias{setOpenCLDevice}
\title{Set GPU device}
\usage{
setOpenCLDevice(name)
}
\arguments{
\item{name}{String: Name of GPU device}
}
\description{
\code{setOpenCLDevice} set GPU device
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TimeEffects.R
\name{convertToTimeVaryingCoef}
\alias{convertToTimeVaryingCoef}
\title{Convert short sparse covariate table to long sparse covariate table for time-varying coefficients.}
\usage{
convertToTimeVaryingCoef(shortCov, longOut, timeVaryCoefId)
}
\arguments{
\item{shortCov}{A data frame containing the covariate with predefined columns (see below).}

\item{longOut}{A data frame containing the outcomes with predefined columns (see below), output of \code{splitTime}.}

\item{timeVaryCoefId}{Integer: A numeric identifier of a time-varying coefficient}
}
\value{
A long sparse covariate table for time-varying coefficients.
}
\description{
\code{convertToTimeVaryingCoef} convert short sparse covariate table to long sparse covariate table for time-varying coefficients.
}
\details{
These columns are expected in the shortCov object:
\tabular{lll}{
\verb{rowId}  	       \tab(integer) \tab Row ID is used to link multiple covariates (x) to a single outcome (y) \cr
\verb{covariateId}    \tab(integer) \tab A numeric identifier of a covariate  \cr
\verb{covariateValue} \tab(real) \tab The value of the specified covariate \cr
}

These columns are expected in the longOut object:
\tabular{lll}{
\verb{stratumId}   \tab(integer) \tab Stratum ID for time-varying models \cr
\verb{subjectId}  	\tab(integer) \tab Subject ID is used to link multiple covariates (x) at different time intervals to a single subject \cr
\verb{rowId}  	    \tab(integer) \tab Row ID is used to link multiple covariates (x) to a single outcome (y) \cr
\verb{y}           \tab(real) \tab The outcome variable \cr
\verb{time}        \tab(real) \tab For models that use time (e.g. Poisson or Cox regression) this contains time \cr
\tab       \tab(e.g. number of days) \cr
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/NewDataConversion.R
\name{convertToCyclopsData}
\alias{convertToCyclopsData}
\alias{convertToCyclopsData.data.frame}
\alias{convertToCyclopsData.tbl_dbi}
\title{Convert data from two data frames or ffdf objects into a CyclopsData object}
\usage{
convertToCyclopsData(
  outcomes,
  covariates,
  modelType = "lr",
  timeEffectMap = NULL,
  addIntercept = TRUE,
  checkSorting = NULL,
  checkRowIds = TRUE,
  normalize = NULL,
  quiet = FALSE,
  floatingPoint = 64
)

\method{convertToCyclopsData}{data.frame}(
  outcomes,
  covariates,
  modelType = "lr",
  timeEffectMap = NULL,
  addIntercept = TRUE,
  checkSorting = NULL,
  checkRowIds = TRUE,
  normalize = NULL,
  quiet = FALSE,
  floatingPoint = 64
)

\method{convertToCyclopsData}{tbl_dbi}(
  outcomes,
  covariates,
  modelType = "lr",
  timeEffectMap = NULL,
  addIntercept = TRUE,
  checkSorting = NULL,
  checkRowIds = TRUE,
  normalize = NULL,
  quiet = FALSE,
  floatingPoint = 64
)
}
\arguments{
\item{outcomes}{A data frame or ffdf object containing the outcomes with predefined columns (see below).}

\item{covariates}{A data frame or ffdf object containing the covariates with predefined columns (see below).}

\item{modelType}{Cyclops model type. Current supported types are "pr", "cpr", lr", "clr", or "cox"}

\item{timeEffectMap}{A data frame or ffdf object containing the convariates that have time-varying effects on the outcome}

\item{addIntercept}{Add an intercept to the model?}

\item{checkSorting}{(DEPRECATED) Check if the data are sorted appropriately, and if not, sort.}

\item{checkRowIds}{Check if all rowIds in the covariates appear in the outcomes.}

\item{normalize}{String: Name of normalization for all non-indicator covariates (possible values: stdev, max, median)}

\item{quiet}{If true, (warning) messages are suppressed.}

\item{floatingPoint}{Specified floating-point representation size (32 or 64)}
}
\value{
An object of type cyclopsData
}
\description{
\code{convertToCyclopsData} loads data from two data frames or ffdf objects, and inserts it into a Cyclops data object.
}
\details{
These columns are expected in the outcome object:
\tabular{lll}{
\verb{stratumId}    \tab(integer) \tab (optional) Stratum ID for conditional regression models \cr
\verb{rowId}  	\tab(integer) \tab Row ID is used to link multiple covariates (x) to a single outcome (y) \cr
\verb{y}    \tab(real) \tab The outcome variable \cr
\verb{time}    \tab(real) \tab For models that use time (e.g. Poisson or Cox regression) this contains time \cr
\tab        \tab(e.g. number of days) \cr
\verb{weights} \tab(real) \tab (optional) Non-negative weights to apply to outcome \cr
\verb{censorWeights} \tab(real) \tab (optional) Non-negative censoring weights for competing risk model; will be computed if not provided.
}

These columns are expected in the covariates object:
\tabular{lll}{
\verb{stratumId}    \tab(integer) \tab (optional) Stratum ID for conditional regression models \cr
\verb{rowId}  	\tab(integer) \tab Row ID is used to link multiple covariates (x) to a single outcome (y) \cr
\verb{covariateId}    \tab(integer) \tab A numeric identifier of a covariate  \cr
\verb{covariateValue}    \tab(real) \tab The value of the specified covariate \cr
}

These columns are expected in the timeEffectMap object:
\tabular{lll}{
\verb{covariateId}    \tab(integer) \tab A numeric identifier of the covariates that have time-varying effects on the outcome \cr
}
}
\section{Methods (by class)}{
\itemize{
\item \code{convertToCyclopsData(data.frame)}: Convert data from two \code{data.frame}

\item \code{convertToCyclopsData(tbl_dbi)}: Convert data from two \code{Andromeda} tables

}}
\examples{
#Convert infert dataset to Cyclops format:
covariates <- data.frame(stratumId = rep(infert$stratum, 2),
                         rowId = rep(1:nrow(infert), 2),
                         covariateId = rep(1:2, each = nrow(infert)),
                         covariateValue = c(infert$spontaneous, infert$induced))
outcomes <- data.frame(stratumId = infert$stratum,
                       rowId = 1:nrow(infert),
                       y = infert$case)
#Make sparse:
covariates <- covariates[covariates$covariateValue != 0, ]

#Create Cyclops data object:
cyclopsData <- convertToCyclopsData(outcomes, covariates, modelType = "clr",
                                    addIntercept = FALSE)

#Fit model:
fit <- fitCyclopsModel(cyclopsData, prior = createPrior("none"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataManagement.R
\name{summary.cyclopsData}
\alias{summary.cyclopsData}
\title{Cyclops data object summary}
\usage{
\method{summary}{cyclopsData}(object, ...)
}
\arguments{
\item{object}{A Cyclops data object}

\item{...}{Additional arguments}
}
\value{
Returns a \code{data.frame} that reports simply summarize statistics for each covariate in a Cyclops data object.
}
\description{
\code{summary.cyclopsData} summarizes the data held in an Cyclops data object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataManagement.R
\name{print.cyclopsData}
\alias{print.cyclopsData}
\title{Print a Cyclops data object}
\usage{
\method{print}{cyclopsData}(x, show.call = TRUE, ...)
}
\arguments{
\item{x}{A Cyclops data model object}

\item{show.call}{Logical: display last call to construct the Cyclops data model object}

\item{...}{Additional arguments}
}
\description{
\code{print.cyclopsData} displays information about a Cyclops data model object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{getFloatingPointSize}
\alias{getFloatingPointSize}
\title{Get floating point size}
\usage{
getFloatingPointSize(object)
}
\arguments{
\item{object}{A Cyclops data object}
}
\description{
\code{getFloatingPointSize} returns the floating-point representation size in a Cyclops data object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{createPrior}
\alias{createPrior}
\title{Create a Cyclops prior object}
\usage{
createPrior(
  priorType,
  variance = 1,
  exclude = c(),
  graph = NULL,
  neighborhood = NULL,
  useCrossValidation = FALSE,
  forceIntercept = FALSE
)
}
\arguments{
\item{priorType}{Character: specifies prior distribution.  See below for options}

\item{variance}{Numeric: prior distribution variance}

\item{exclude}{A vector of numbers or covariateId names to exclude from prior}

\item{graph}{Child-to-parent mapping for a hierarchical prior}

\item{neighborhood}{A list of first-order neighborhoods for a partially fused prior}

\item{useCrossValidation}{Logical: Perform cross-validation to determine prior \code{variance}.}

\item{forceIntercept}{Logical: Force intercept coefficient into prior}
}
\value{
A Cyclops prior object of class inheriting from \code{"cyclopsPrior"} for use with \code{fitCyclopsModel}.
}
\description{
\code{createPrior} creates a Cyclops prior object for use with \code{\link{fitCyclopsModel}}.
}
\section{Prior types}{


We specify all priors in terms of their variance parameters.
Similar fitting tools for regularized regression often parameterize the Laplace distribution
in terms of a rate \code{"lambda"} per observation.
See \code{"glmnet"}, for example.

variance = 2 * / (nobs * lambda)^2 or lambda = sqrt(2 / variance) / nobs
}

\examples{
#Generate some simulated data:
sim <- simulateCyclopsData(nstrata = 1, nrows = 1000, ncovars = 2, eCovarsPerRow = 0.5, 
                           model = "poisson")
cyclopsData <- convertToCyclopsData(sim$outcomes, sim$covariates, modelType = "pr", 
                                    addIntercept = TRUE)

#Define the prior and control objects to use cross-validation for finding the 
#optimal hyperparameter:
prior <- createPrior("laplace", exclude = 0, useCrossValidation = TRUE)
control <- createControl(cvType = "auto", noiseLevel = "quiet")

#Fit the model
fit <- fitCyclopsModel(cyclopsData,prior = prior, control = control)  

#Find out what the optimal hyperparameter was:
getHyperParameter(fit)

#Extract the current log-likelihood, and coefficients
logLik(fit)
coef(fit)

#We can only retrieve the confidence interval for unregularized coefficients:
confint(fit, c(0))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{aconfint}
\alias{aconfint}
\title{Asymptotic confidence intervals for a fitted Cyclops model object}
\usage{
aconfint(
  object,
  parm,
  level = 0.95,
  control,
  overrideNoRegularization = FALSE,
  ...
)
}
\arguments{
\item{object}{A fitted Cyclops model object}

\item{parm}{A specification of which parameters require confidence intervals,
either a vector of numbers of covariateId names}

\item{level}{Numeric: confidence level required}

\item{control}{A \code{"cyclopsControl"} object constructed by \code{\link{createControl}}}

\item{overrideNoRegularization}{Logical: Enable confidence interval estimation for regularized parameters}

\item{...}{Additional argument(s) for methods}
}
\value{
A matrix with columns reporting lower and upper confidence limits for each parameter.
These columns are labelled as (1-level) / 2 and 1 - (1 - level) / 2 in \%
(by default 2.5\% and 97.5\%)
}
\description{
\code{aconfinit} constructs confidence intervals of
arbitrary level using asymptotic standard error estimates.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{gradient}
\alias{gradient}
\title{Extract gradient}
\usage{
gradient(object)
}
\arguments{
\item{object}{A Cyclops model fit object}
}
\description{
\code{gradient} returns the current gradient wrt the regression parameters of
the log-likelihood of the fit in a Cyclops model fit object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{cacheCyclopsModelForJava}
\alias{cacheCyclopsModelForJava}
\title{Place Cyclops model fit object in a persistent cache}
\usage{
cacheCyclopsModelForJava(object)
}
\arguments{
\item{object}{Cyclops model fit object}
}
\value{
Integer index of object in persistent cache
}
\description{
\code{cacheCyclopsModelForJava} places a Cyclops model fit object into a persistent cache,
so that the object may be used across language-barriers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ExampleData.R
\docType{data}
\name{oxford}
\alias{oxford}
\title{Oxford self-controlled case series data}
\format{
A data frame with 38 rows and 6 variables:
\describe{
\item{indiv}{patient identifier}
\item{event}{number of events in interval}
\item{interval}{interval length in days}
\item{agegr}{age group}
\item{exgr}{exposure group}
\item{loginterval}{log interval length}
...
}
}
\usage{
data(oxford)
}
\description{
A dataset containing the MMR vaccination / meningitis in Oxford example from
Farrington and Whitaker.  There are 10 patients comprising 38 unique exposure intervals.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Survfit.R
\name{survfit.cyclopsFit}
\alias{survfit.cyclopsFit}
\title{Calculate baseline hazard function}
\usage{
\method{survfit}{cyclopsFit}(formula, type = "aalen", ...)
}
\arguments{
\item{formula}{A Cyclops survival model fit object}

\item{type}{type of baseline survival, choices are: "aalen" (Breslow)}

\item{...}{for future methods}
}
\value{
Baseline survival function for mean covariates
}
\description{
\code{survfit.cyclopsFit} computes baseline hazard function
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{getNumberOfStrata}
\alias{getNumberOfStrata}
\title{Get number of strata}
\usage{
getNumberOfStrata(object)
}
\arguments{
\item{object}{A Cyclops data object}
}
\description{
\code{getNumberOfStrata} return the number of unique strata in a Cyclops data object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{getCrossValidationInfo}
\alias{getCrossValidationInfo}
\title{Get cross-validation information from a Cyclops model fit}
\usage{
getCrossValidationInfo(object)
}
\arguments{
\item{object}{A Cyclops model fit object}
}
\description{
{getCrossValidationInfo} returns the predicted optimal cross-validation point and ordinate
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{fitCyclopsModel}
\alias{fitCyclopsModel}
\title{Fit a Cyclops model}
\usage{
fitCyclopsModel(
  cyclopsData,
  prior = createPrior("none"),
  control = createControl(),
  weights = NULL,
  forceNewObject = FALSE,
  returnEstimates = TRUE,
  startingCoefficients = NULL,
  fixedCoefficients = NULL,
  warnings = TRUE,
  computeDevice = "native"
)
}
\arguments{
\item{cyclopsData}{A Cyclops data object}

\item{prior}{A prior object. More details are given below.}

\item{control}{A \code{"cyclopsControl"} object constructed by \code{\link{createControl}}}

\item{weights}{Vector of 0/1 weights for each data row}

\item{forceNewObject}{Logical, forces the construction of a new Cyclops model fit object}

\item{returnEstimates}{Logical, return regression coefficient estimates in Cyclops model fit object}

\item{startingCoefficients}{Vector of starting values for optimization}

\item{fixedCoefficients}{Vector of booleans indicating if coefficient should be fix}

\item{warnings}{Logical, report regularization warnings}

\item{computeDevice}{String: Name of compute device to employ; defaults to \code{"native"} C++ on CPU}
}
\value{
A list that contains a Cyclops model fit object pointer and an operation duration
}
\description{
\code{fitCyclopsModel} fits a Cyclops model data object
}
\details{
This function performs numerical optimization to fit a Cyclops model data object.
}
\section{Prior}{

Currently supported prior types are:
\tabular{ll}{
\verb{	"none"} \tab Useful for finding MLE \cr
\verb{	"laplace"} \tab L_1 regularization \cr
\verb{  "normal"} \tab L_2 regularization \cr
}
}

\examples{
## Dobson (1990) Page 93: Randomized Controlled Trial :
counts <- c(18,17,15,20,10,20,25,13,12)
outcome <- gl(3,1,9)
treatment <- gl(3,3)
cyclopsData <- createCyclopsData(counts ~ outcome + treatment, modelType = "pr")
cyclopsFit <- fitCyclopsModel(cyclopsData, prior = createPrior("none"))
coef(cyclopsFit)
confint(cyclopsFit, c("outcome2","treatment3"))
predict(cyclopsFit)

}
\references{
Suchard MA, Simpson SE, Zorych I, Ryan P, Madigan D.
Massive parallelization of serial inference algorithms for complex generalized linear models.
ACM Transactions on Modeling and Computer Simulation, 23, 10, 2013.

Simpson SE, Madigan D, Zorych I, Schuemie M, Ryan PB, Suchard MA.
Multiple self-controlled case series for large-scale longitudinal observational databases.
Biometrics, 69, 893-902, 2013.

Mittal S, Madigan D, Burd RS, Suchard MA.
High-dimensional, massive sample-size Cox proportional hazards regression for survival analysis.
Biostatistics, 15, 207-221, 2014.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SpecialPriors.R
\name{createNonSeparablePrior}
\alias{createNonSeparablePrior}
\title{Create a Cyclops prior object that returns the MLE of non-separable coefficients}
\usage{
createNonSeparablePrior(maxIterations = 10, ...)
}
\arguments{
\item{maxIterations}{Numeric: maximum iterations to achieve convergence}

\item{...}{Additional argument(s) for \code{\link{fitCyclopsModel}}}
}
\value{
A Cyclops prior object of class inheriting from
\code{"cyclopsPrior"} for use with \code{fitCyclopsModel}.
}
\description{
\code{createNonSeparablePrior} creates a Cyclops prior object for use with \code{\link{fitCyclopsModel}}.
}
\examples{
prior <- createNonSeparablePrior()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{getCyclopsProfileLogLikelihood}
\alias{getCyclopsProfileLogLikelihood}
\title{Profile likelihood for Cyclops model parameters}
\usage{
getCyclopsProfileLogLikelihood(
  object,
  parm,
  x = NULL,
  bounds = NULL,
  tolerance = 0.001,
  initialGridSize = 10,
  maxResets = 10,
  includePenalty = TRUE,
  optimalWarmStart = TRUE,
  returnDerivatives = FALSE
)
}
\arguments{
\item{object}{Fitted Cyclops model object}

\item{parm}{Specification of which parameter requires profiling,
either a vector of numbers or covariateId names}

\item{x}{Vector of values of the parameter}

\item{bounds}{Pair of values to bound adaptive profiling}

\item{tolerance}{Absolute tolerance allowed for adaptive profiling}

\item{initialGridSize}{Initial grid size for adaptive profiling}

\item{maxResets}{Maximum allowed number of recomputing the likelihood when coefficient drift is
detected.}

\item{includePenalty}{Logical: Include regularized covariate penalty in profile}

\item{optimalWarmStart}{Logical: Use optimal warm-starting when parallelizing evaluations}

\item{returnDerivatives}{Return derivative of log-likelihood wrt the parameter \code{param} at each evaluation point}
}
\value{
A data frame containing the profile log likelihood. Returns NULL when the adaptive profiling fails
to converge.
}
\description{
\code{getCyclopsProfileLogLikelihood} evaluates the profile likelihood at a grid of parameter values.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{listGPUDevices}
\alias{listGPUDevices}
\title{List available GPU devices}
\usage{
listGPUDevices()
}
\description{
\code{listGPUDevices} list available GPU devices
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cyclops.R
\docType{package}
\name{cyclops}
\alias{Cyclops}
\alias{Cyclops-package}
\alias{cyclops}
\title{Cyclops: Cyclic coordinate descent for logistic, Poisson and survival analysis}
\description{
The Cyclops package incorporates cyclic coordinate descent and
majorization-minimization approaches to fit a variety of regression models
found in large-scale observational healthcare data.  Implementations focus
on computational optimization and fine-scale parallelization to yield
efficient inference in massive datasets.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/ohdsi/cyclops}
  \item Report bugs at \url{https://github.com/ohdsi/cyclops/issues}
}

}
\author{
\strong{Maintainer}: Marc A. Suchard \email{msuchard@ucla.edu}

Authors:
\itemize{
  \item Martijn J. Schuemie
  \item Trevor R. Shaddox
  \item Yuxi Tian
  \item Jianxiao Yang
  \item Eric Kawaguchi
}

Other contributors:
\itemize{
  \item Sushil Mittal [contributor]
  \item Observational Health Data Sciences and Informatics [copyright holder]
  \item Marcus Geelnard (provided the TinyThread library) [copyright holder, contributor]
  \item Rutgers University (provided the HParSearch routine) [copyright holder, contributor]
  \item R Development Core Team (provided the ZeroIn routine) [copyright holder, contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{print.cyclopsFit}
\alias{print.cyclopsFit}
\title{Print a Cyclops model fit object}
\usage{
\method{print}{cyclopsFit}(x, show.call = TRUE, ...)
}
\arguments{
\item{x}{A Cyclops model fit object}

\item{show.call}{Logical: display last call to update the Cyclops model fit object}

\item{...}{Additional arguments}
}
\description{
\code{print.cyclopsFit} displays information about a Cyclops model fit object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GetFrineGrayWeights.R
\name{getFineGrayWeights}
\alias{getFineGrayWeights}
\title{Creates a \code{Surv} object that forces in competing risks and the IPCW needed for Fine-Gray estimation.}
\usage{
getFineGrayWeights(ftime, fstatus, cvweights = NULL, cencode = 0, failcode = 1)
}
\arguments{
\item{ftime}{Numeric: Observed event (failure) times}

\item{fstatus}{Numeric: Observed event (failure) types}

\item{cvweights}{Numeric: Vector of 0/1 (cross-validation) weights for each data row}

\item{cencode}{Numeric: Code to denote censored observations (Default is 0)}

\item{failcode}{Numeric: Code to denote event of interest (Default is 1)}
}
\value{
A list that returns both an object of class \code{Surv} that forces in the competing risks indicators and a vector of weights needed for parameter estimation.
}
\description{
\code{getFineGrayWeights} creates a list \code{Surv} object and vector of weights required for estimation.
}
\examples{
ftime <- c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
fstatus <- c(1, 2, 0, 1, 2, 0, 1, 2, 0, 1)
getFineGrayWeights(ftime, fstatus, cencode = 0, failcode = 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataManagement.R
\name{getUnivariableSeparability}
\alias{getUnivariableSeparability}
\title{Get univariable linear separability}
\usage{
getUnivariableSeparability(cyclopsData, covariates = NULL)
}
\arguments{
\item{cyclopsData}{A Cyclops data object}

\item{covariates}{Integer or string vector: list of covariates to report; default (NULL) implies all covariates}
}
\value{
A list of covariates that are univariably separable with the outcome
}
\description{
\code{getUnivariableSeparability} reports covariates that are univariably separable with the outcome
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Simulation.R
\name{mse}
\alias{mse}
\title{Mean squared error}
\usage{
mse(goldStandard, estimates)
}
\arguments{
\item{goldStandard}{Numeric vector}

\item{estimates}{Numeric vector}
}
\value{
MSE(\code{goldStandard}, \code{estimates})
}
\description{
\code{mse} computes the mean squared error between two numeric vectors
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{getNumberOfRows}
\alias{getNumberOfRows}
\title{Get total number of rows}
\usage{
getNumberOfRows(object)
}
\arguments{
\item{object}{A Cyclops data object}
}
\description{
\code{getNumberOfRows} returns the total number of outcome rows in a Cyclops data object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CrossValidationControl.R
\name{createAutoGridCrossValidationControl}
\alias{createAutoGridCrossValidationControl}
\title{Create a Cyclops control object that supports multiple hyperparameters}
\usage{
createAutoGridCrossValidationControl(
  outerGrid,
  autoPosition = 1,
  refitAtMaximum = TRUE,
  cvType = "auto",
  initialValue = 1,
  ...
)
}
\arguments{
\item{outerGrid}{List or data.frame of grid parameters to explore}

\item{autoPosition}{Vector position for auto-search parameter (concatenated into outerGrid)}

\item{refitAtMaximum}{Logical: re-fit Cyclops object at maximal cross-validation parameters}

\item{cvType}{Must equal "auto"}

\item{initialValue}{Initial value for auto-search parameter}

\item{...}{Additional parameters passed through to \code{\link{createControl}}}
}
\value{
A Cyclops prior object of class inheriting from \code{"cyclopsPrior"} and \code{"cyclopsFunctionalPrior"}
for use with \code{fitCyclopsModel}.
}
\description{
\code{createCrossValidationControl} creates a Cyclops control object for use with \code{\link{fitCyclopsModel}}
that supports multiple hyperparameters through an auto-search in one dimension and a grid-search over the remaining
dimensions
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/WeightBasedSearchControl.R
\name{createWeightBasedSearchControl}
\alias{createWeightBasedSearchControl}
\title{Create a Cyclops control object that supports in- / out-of-sample hyperparameter search using weights}
\usage{
createWeightBasedSearchControl(cvType = "auto", initialValue = 1, ...)
}
\arguments{
\item{cvType}{Must equal "auto"}

\item{initialValue}{Initial value for auto-search parameter}

\item{...}{Additional parameters passed through to \code{\link{createControl}}}
}
\value{
A Cyclops prior object of class inheriting from \code{"cyclopsControl"}
for use with \code{fitCyclopsModel}.
}
\description{
\code{createWeightBasedSearchControl} creates a Cyclops control object for use with \code{\link{fitCyclopsModel}}
that supports hyperparameter optimization through an auto-search where weight = 1 identifies in-sample observations
and weight = 0 identifies out-of-sample observations.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Predict.R
\name{predict.cyclopsFit}
\alias{predict.cyclopsFit}
\title{Model predictions}
\usage{
\method{predict}{cyclopsFit}(object, newOutcomes, newCovariates, ...)
}
\arguments{
\item{object}{A Cyclops model fit object}

\item{newOutcomes}{An optional data frame or Andromeda table object, similar to the object used in \code{\link{convertToCyclopsData}}.}

\item{newCovariates}{An optional data frame or Andromeda table object, similar to the object used in \code{\link{convertToCyclopsData}}.}

\item{...}{Additional arguments}
}
\description{
\code{predict.cyclopsFit} computes model response-scale predictive values for all data rows
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TimeEffects.R
\name{splitTime}
\alias{splitTime}
\title{Split the analysis time into several intervals for time-varying coefficients.}
\usage{
splitTime(shortOut, cut)
}
\arguments{
\item{shortOut}{A data frame containing the outcomes with predefined columns (see below).}

\item{cut}{Numeric: Time points to cut at

These columns are expected in the shortOut object:
\tabular{lll}{
\verb{rowId}   \tab(integer) \tab Row ID is used to link multiple covariates (x) to a single outcome (y) \cr
\verb{y}       \tab(real) \tab Observed event status \cr
\verb{time}    \tab(real) \tab Observed event time \cr
}}
}
\value{
A long outcome table for time-varying coefficients.
}
\description{
\code{splitTime} split the analysis time into several intervals for time-varying coefficients
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{getNumberOfTypes}
\alias{getNumberOfTypes}
\title{Get total number of outcome types}
\usage{
getNumberOfTypes(object)
}
\arguments{
\item{object}{A Cyclops data object}
}
\description{
\code{getNumberOfTypes} returns the total number of outcome types in a Cyclops data object
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFit.R
\name{clearCyclopsModelCache}
\alias{clearCyclopsModelCache}
\title{Clear the Cyclops persistent cache}
\usage{
clearCyclopsModelCache()
}
\description{
\code{clearCyclopsModelCache} clears the persistent cache holding
Cyclops model fit objects that may be used across language-barriers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataManagement.R
\name{readCyclopsData}
\alias{readCyclopsData}
\title{Read Cyclops data from file}
\usage{
readCyclopsData(fileName, modelType)
}
\arguments{
\item{fileName}{Name of text file to be read. If fileName does not contain an absolute path,}

\item{modelType}{character string: Valid types are listed below.}
}
\value{
A list that contains a Cyclops model data object pointer and an operation duration
}
\description{
\code{readCyclopsData} reads a Cyclops-formatted text file.
}
\details{
This function reads a Cyclops-formatted text file and returns a Cyclops data object. The first line of the
file may start with '\samp{#}', indicating that it contains header options.  Valid header options are:

\tabular{ll}{
\verb{	row_label}		\tab (assume file contains a numeric column of unique row identifiers) \cr
\verb{	stratum_label}\tab (assume file contains a numeric column of stratum identifiers) \cr
\verb{	weight}				\tab (assume file contains a column of row-specific model weights, currently unused) \cr
\verb{	offset}				\tab (assume file contains a dense column of linear predictor offsets) \cr
\verb{	bbr_outcome}	\tab (assume logistic outcomes are encoded -1/+1 following BBR) \cr
\verb{	log_offset}		\tab (assume file contains a dense column of values x_i for which log(x_i) is the offset) \cr
\verb{	add_intercept}\tab (automatically include an intercept column of all 1s for each entry) \cr
\verb{	indicator_only}\tab(assume all covariates 0/1-valued and only covariate name is given) \cr
\verb{	sparse}				\tab (force all BBR formatted covariates to be represented as sparse, instead of \cr
\tab sparse-indicator, columns .. really only for debugging) \cr
\verb{	dense}				\tab (force all BBR formatted covariates to be represented as dense columns.. really\cr
\tab only for debugging) \cr
}

Successive lines of the file are white-space delimited and follow the format:

\preformatted{	[Row ID] {Stratum ID} [Weight] <Outcome> {Censored} {Offset} <BBR covariates>}

\itemize{
\item \verb{[optional]}
\item \verb{<required>}
\item \verb{{required or optional depending on model}}
}

Bayesian binary regression (BBR) covariates are white-space delimited and generally in a sparse
\samp{<name>:<value>} format, where \samp{name} must (currently) be numeric and \samp{value} is non-zero.
If option \samp{indicator_only} is specified, then format is simply \samp{<name>}.
\samp{Row ID} and \samp{Stratum ID} must be numeric, and rows must be sorted such that equal \samp{Stratum ID}
are consecutive.
\samp{Stratum ID} is required for \samp{clr} and \samp{sccs} models.
\samp{Censored} is required for a \samp{cox} model.
\samp{Offset} is (currently) required for a \samp{sccs} model.
}
\section{Models}{

Currently supported model types are:
\tabular{ll}{
\verb{	"ls"} \tab Least squares \cr
\verb{	"pr"} \tab Poisson regression \cr
\verb{	"lr"} \tab Logistic regression \cr
\verb{	"clr"} \tab Conditional logistic regression \cr
\verb{	"cpr"} \tab Conditional Poisson regression \cr
\verb{	"sccs"} \tab Self-controlled case series \cr
\verb{	"cox"} \tab Cox proportional hazards regression \cr
\verb{	"fgr"} \tab Fine-Gray proportional subdistribution hazards regression \cr
}
}

\examples{
\dontrun{
dataPtr = readCyclopsData(system.file("extdata/infert_ccd.txt", package="Cyclops"), "clr")
}
}
